Package: reactable
Type: Package
Title: Interactive Data Tables for R
Version: 0.4.5
Authors@R: c(
    person("Greg", "Lin", email = "glin@glin.io", role = c("aut", "cre")),
    person("Tanner", "Linsley", role = c("ctb", "cph"), comment = "React Table library"),
    person(family = "Emotion team and other contributors", role = c("ctb", "cph"),
           comment = "Emotion library"),
    person("Kent", "Russell", role = c("ctb", "cph"), comment = "reactR package"),
    person("Ramnath", "Vaidyanathan", role = c("ctb", "cph"), comment = "htmlwidgets package"),
    person("Joe", "Cheng", role = c("ctb", "cph"), comment = "htmlwidgets package"),
    person("JJ", "Allaire", role = c("ctb", "cph"), comment = "htmlwidgets package"),
    person("Yihui", "Xie", role = c("ctb", "cph"), comment = "htmlwidgets package"),
    person("Kenton", "Russell", role = c("ctb", "cph"), comment = "htmlwidgets package"),
    person(family = "Facebook, Inc. and its affiliates", role = c("ctb", "cph"), comment = "React library"),
    person(family = "FormatJS", role = c("ctb", "cph"), comment = "FormatJS libraries"),
    person(family = "Feross Aboukhadijeh, and other contributors", role = c("ctb", "cph"),
           comment = "buffer library"),
    person("Roman", "Shtylman", role = c("ctb", "cph"), comment = "process library"),
    person("James", "Halliday", role = c("ctb", "cph"), comment = "stream-browserify library"),
    person(family = "Posit Software, PBC", role = c("fnd", "cph"))
    )
Description: Interactive data tables for R, based on the 'React Table'
    JavaScript library. Provides an HTML widget that can be used in 'R Markdown'
    or 'Quarto' documents, 'Shiny' applications, or viewed from an R console.
License: MIT + file LICENSE
URL: https://glin.github.io/reactable/,
        https://github.com/glin/reactable
BugReports: https://github.com/glin/reactable/issues
Depends: R (>= 3.1)
Imports: digest, htmltools (>= 0.5.2), htmlwidgets (>= 1.5.3),
        jsonlite, reactR
Suggests: covr, crosstalk, dplyr, fontawesome, knitr, leaflet, MASS,
        rmarkdown, shiny, sparkline, testthat, tippy, V8
Encoding: UTF-8
RoxygenNote: 7.2.1
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2025-12-01 15:17:46 UTC; greg
Author: Greg Lin [aut, cre],
  Tanner Linsley [ctb, cph] (React Table library),
  Emotion team and other contributors [ctb, cph] (Emotion library),
  Kent Russell [ctb, cph] (reactR package),
  Ramnath Vaidyanathan [ctb, cph] (htmlwidgets package),
  Joe Cheng [ctb, cph] (htmlwidgets package),
  JJ Allaire [ctb, cph] (htmlwidgets package),
  Yihui Xie [ctb, cph] (htmlwidgets package),
  Kenton Russell [ctb, cph] (htmlwidgets package),
  Facebook, Inc. and its affiliates [ctb, cph] (React library),
  FormatJS [ctb, cph] (FormatJS libraries),
  Feross Aboukhadijeh, and other contributors [ctb, cph] (buffer library),
  Roman Shtylman [ctb, cph] (process library),
  James Halliday [ctb, cph] (stream-browserify library),
  Posit Software, PBC [fnd, cph]
Maintainer: Greg Lin <glin@glin.io>
Repository: CRAN
Date/Publication: 2025-12-01 16:00:02 UTC
Built: R 4.6.0; ; 2026-04-26 15:51:26 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as.tags,reactable)
export(JS)
export(colDef)
export(colFormat)
export(colGroup)
export(getReactableState)
export(knit_print.reactable)
export(reactable)
export(reactableLang)
export(reactableOutput)
export(reactableTheme)
export(renderReactable)
export(updateReactable)
import(htmltools)
import(htmlwidgets)
importFrom(htmlwidgets,JS)

# reactable

[![CRAN Status](https://www.r-pkg.org/badges/version/reactable)](https://cran.r-project.org/package=reactable)
[![Build Status](https://github.com/glin/reactable/workflows/build/badge.svg)](https://github.com/glin/reactable/actions)
[![codecov](https://codecov.io/gh/glin/reactable/branch/master/graph/badge.svg)](https://app.codecov.io/gh/glin/reactable)
[![lifecycle](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)

Interactive data tables for R, based on the
[React Table](https://github.com/tanstack/table/tree/v7) library and made with
[reactR](https://github.com/react-R/reactR).

## Features
- Sorting, filtering, pagination
- Grouping and aggregation
- Built-in column formatting
- Custom rendering via R or JavaScript — use Shiny HTML tags and HTML widgets in tables
- Expandable rows and nested tables
- Conditional styling
- Works seamlessly within R Markdown documents and Shiny apps

## Demos
- [Twitter Followers](https://glin.github.io/reactable/articles/twitter-followers/twitter-followers.html)
- [Women's World Cup Predictions](https://glin.github.io/reactable/articles/womens-world-cup/womens-world-cup.html)
- [NBA Box Score](https://glin.github.io/reactable/articles/nba-box-score/nba-box-score.html)
- [CRAN Packages](https://glin.github.io/reactable/articles/cran-packages/cran-packages.html)
- [Spotify Charts](https://glin.github.io/reactable/articles/spotify-charts/spotify-charts.html)
- [Popular Movies](https://glin.github.io/reactable/articles/popular-movies/popular-movies.html)
- [Demo Cookbook](https://glin.github.io/reactable/articles/cookbook/cookbook.html)
- [100k Rows](https://glin.github.io/reactable/articles/100k-rows.html)
- [Shiny Demo](https://glin.github.io/reactable/articles/shiny-demo.html)

## Installation
You can install reactable from CRAN with:
```r
install.packages("reactable")
```

Or install the development version from GitHub with:
```r
# install.packages("devtools")
devtools::install_github("glin/reactable")
```

## Usage
To create a table, use `reactable()` on a data frame or matrix:
```r
library(reactable)

reactable(iris)
```

<!-- pkgdown:example -->
[![reactable example using the iris dataset](man/figures/iris.png)](https://glin.github.io/reactable/articles/examples.html)

You can embed tables in **R Markdown** or **Quarto** documents:
````
```{r}
library(reactable)

reactable(iris)
```
````

Or use them in **Shiny** applications:
```r
library(shiny)
library(reactable)

ui <- fluidPage(
  reactableOutput("table")
)

server <- function(input, output) {
  output$table <- renderReactable({
    reactable(iris)
  })
}

shinyApp(ui, server)
```

To learn more about using reactable, check out the examples below.

## Examples
- [Basic Usage](https://glin.github.io/reactable/articles/examples.html#basic-usage)
- [Sorting](https://glin.github.io/reactable/articles/examples.html#sorting)
- [Filtering](https://glin.github.io/reactable/articles/examples.html#filtering)
- [Searching](https://glin.github.io/reactable/articles/examples.html#searching)
- [Pagination](https://glin.github.io/reactable/articles/examples.html#pagination)
- [Grouping and Aggregation](https://glin.github.io/reactable/articles/examples.html#grouping-and-aggregation)
- [Column Formatting](https://glin.github.io/reactable/articles/examples.html#column-formatting)
- [Custom Rendering](https://glin.github.io/reactable/articles/examples.html#custom-rendering)
- [Embedding HTML Widgets](https://glin.github.io/reactable/articles/examples.html#embedding-html-widgets)
- [Footers and Total Rows](https://glin.github.io/reactable/articles/examples.html#footers)
- [Expandable Rows and Nested Tables](https://glin.github.io/reactable/articles/examples.html#expandable-row-details)
- [Conditional Styling](https://glin.github.io/reactable/articles/examples.html#conditional-styling)
- [Table Styling](https://glin.github.io/reactable/articles/examples.html#table-styling)
- [Theming](https://glin.github.io/reactable/articles/examples.html#theming)
- [Column Groups](https://glin.github.io/reactable/articles/examples.html#column-groups)
- [Column Resizing](https://glin.github.io/reactable/articles/examples.html#column-resizing)
- [Sticky Columns](https://glin.github.io/reactable/articles/examples.html#sticky-columns)
- [Row Names and Row Headers](https://glin.github.io/reactable/articles/examples.html#row-names-and-row-headers)
- [Cell Click Actions](https://glin.github.io/reactable/articles/examples.html#cell-click-actions)
- [Language Options](https://glin.github.io/reactable/articles/examples.html#language-options)
- [Shiny](https://glin.github.io/reactable/articles/examples.html#shiny)
- [Cross-Widget Interactions with Crosstalk](https://glin.github.io/reactable/articles/examples.html#cross-widget-interactions)
- [JavaScript API](https://glin.github.io/reactable/articles/examples.html#javascript-api)

## Browser Support
| [<img src="https://raw.githubusercontent.com/alrra/browser-logos/master/src/edge/edge_48x48.png" alt="IE / Edge" width="24px" height="24px" />](https://godban.github.io/browsers-support-badges/)<br>IE / Edge | [<img src="https://raw.githubusercontent.com/alrra/browser-logos/master/src/firefox/firefox_48x48.png" alt="Firefox" width="24px" height="24px" />](https://godban.github.io/browsers-support-badges/)<br>Firefox | [<img src="https://raw.githubusercontent.com/alrra/browser-logos/master/src/chrome/chrome_48x48.png" alt="Chrome" width="24px" height="24px" />](https://godban.github.io/browsers-support-badges/)<br>Chrome | [<img src="https://raw.githubusercontent.com/alrra/browser-logos/master/src/safari/safari_48x48.png" alt="Safari" width="24px" height="24px" />](https://godban.github.io/browsers-support-badges/)<br>Safari | [<img src="https://raw.githubusercontent.com/alrra/browser-logos/master/src/opera/opera_48x48.png" alt="Opera" width="24px" height="24px" />](https://godban.github.io/browsers-support-badges/)<br>Opera |
| --------- | --------- | --------- | --------- | --------- |
| IE 11*, Edge | last 2 versions | last 2 versions | last 2 versions | last 2 versions |

\* Support for Internet Explorer 11 was deprecated in reactable v0.4.0.

## License
MIT

library(shiny)
library(reactable)

db <- tools::CRAN_package_db()

ui <- fluidPage(
  reactableOutput("table")
)

server <- function(input, output) {
  pkgs <- c("rlang", "magrittr", "Rcpp", "dplyr", "tibble", "ggplot2", "glue", "pillar", "data.table")
  pkgs <- db[db$Package %in% pkgs, c("Package", "Version", "Description")]
  pkgs$Downloads <- c(70, 105, 38, 51, 49, 64, 53, 43, 22)

  output$table <- renderReactable({
    reactable(
      pkgs,
      searchable = TRUE,
      wrap = FALSE,
      resizable = TRUE,
      fullWidth = FALSE,
      defaultSorted = "Downloads",
      defaultColDef = colDef(minWidth = 200),
      columns = list(
        Downloads = colDef(defaultSortOrder = "desc")
      )
    )
  })

  observe({
    # Update download counts
    invalidateLater(100)
    pkgs$Downloads <<- pkgs$Downloads + floor(runif(9, 0, 10))
    updateReactable("table", data = pkgs)
  })
}

shinyApp(ui, server)

library(shiny)
library(reactable)

prismCodeBlock <- function(code, language = "r") {
  # Generate a pre-code tag without indentation
  html <- as.character(tags$code(class = paste0("language-", language), code))
  html <- sprintf("<pre>%s</pre>", html)
  tagList(
    HTML(html),
    tags$script("Prism.highlightAll()")
  )
}

prismDependencies <- tags$head(
  tags$script(src = "https://cdnjs.cloudflare.com/ajax/libs/prism/1.16.0/prism.min.js"),
  tags$script(src = "https://cdnjs.cloudflare.com/ajax/libs/prism/1.16.0/components/prism-r.min.js"),
  tags$link(rel = "stylesheet", type = "text/css",
            href = "https://cdnjs.cloudflare.com/ajax/libs/prism/1.16.0/themes/prism.min.css")
)

themes <- list(
  "Light 1" = quote(reactableTheme(
    borderColor = "#dfe2e5",
    stripedColor = "#f6f8fa",
    highlightColor = "#f0f5f9",
    cellPadding = "8px 12px"
  )),
  "Dark 1" = quote(reactableTheme(
    color = "hsl(0, 0%, 87%)",
    backgroundColor = "hsl(220, 13%, 18%)",
    borderColor = "hsl(0, 0%, 22%)",
    stripedColor = "rgba(255, 255, 255, 0.04)",
    highlightColor = "rgba(255, 255, 255, 0.06)",
    inputStyle = list(backgroundColor = "hsl(0, 0%, 24%)"),
    selectStyle = list(backgroundColor = "hsl(0, 0%, 24%)"),
    pageButtonHoverStyle = list(backgroundColor = "hsl(0, 0%, 24%)"),
    pageButtonActiveStyle = list(backgroundColor = "hsl(0, 0%, 28%)")
  )),
  "Dark 2" = quote(reactableTheme(
    color = "hsl(233, 9%, 87%)",
    backgroundColor = "hsl(233, 9%, 19%)",
    borderColor = "hsl(233, 9%, 22%)",
    stripedColor = "hsl(233, 12%, 22%)",
    highlightColor = "hsl(233, 12%, 24%)",
    inputStyle = list(backgroundColor = "hsl(233, 9%, 25%)"),
    selectStyle = list(backgroundColor = "hsl(233, 9%, 25%)"),
    pageButtonHoverStyle = list(backgroundColor = "hsl(233, 9%, 25%)"),
    pageButtonActiveStyle = list(backgroundColor = "hsl(233, 9%, 28%)")
  )),
  "Dark 3" = quote(reactableTheme(
    color = "hsl(0, 0%, 90%)",
    backgroundColor = "hsl(0, 0%, 10%)",
    borderColor = "hsl(0, 0%, 18%)",
    stripedColor = "hsl(0, 0%, 13%)",
    headerStyle = list(
      "&:hover[aria-sort]" = list(backgroundColor = "hsl(0, 0%, 14%)")
    ),
    tableBodyStyle = list(color = "hsl(0, 0%, 75%)"),
    rowHighlightStyle = list(color = "hsl(0, 0%, 90%)", backgroundColor = "hsl(0, 0%, 14%)"),
    selectStyle = list(backgroundColor = "hsl(0, 0%, 20%)"),
    inputStyle = list(
      backgroundColor = "hsl(0, 0%, 10%)",
      borderColor = "hsl(0, 0%, 21%)",
      "&:hover, &:focus" = list(borderColor = "hsl(0, 0%, 30%)")
    ),
    pageButtonHoverStyle = list(backgroundColor = "hsl(0, 0%, 20%)"),
    pageButtonActiveStyle = list(backgroundColor = "hsl(0, 0%, 24%)")
  ))
)

ui <- fluidPage(
  tags$head(
    tags$title("reactable demo"),
    includeCSS("assets/styles.css"),
    prismDependencies
  ),

  div(class = "main",
      div(class = "sidebar",
          h1("reactable demo", class = "title"),

          checkboxGroupInput(
            "options",
            "Table Options",
            choices = c(
              "Column groups" = "columnGroups",
              "Footer" = "footer",
              Filterable = "filterable",
              Searchable = "searchable",
              Sortable = "sortable",
              Resizable = "resizable",
              "Default sorted" = "defaultSorted",
              "Show sortable" = "showSortable",
              Outlined = "outlined",
              Bordered = "bordered",
              Borderless = "borderless",
              Striped = "striped",
              Highlight = "highlight",
              Compact = "compact",
              "Full width" = "fullWidth"
            ),
            selected = c("sortable", "resizable", "highlight", "fullWidth")
          ),

          selectInput(
            "theme",
            "Theme",
            c("Default", names(themes)),
            selectize = FALSE,
            width = 170
          ),

          checkboxGroupInput(
            "pagination",
            "Pagination",
            choices = c(
              "Enable pagination" = "pagination",
              "Show page size options" = "showPageSizeOptions",
              "Show page info" = "showPageInfo"
            ),
            selected = c("pagination", "showPageSizeOptions", "showPageInfo")
          ),
          radioButtons(
            "paginationType",
            "Pagination type",
            choices = c("page numbers" = "numbers", "page jump" = "jump", "simple")
          ),

          checkboxGroupInput("groupBy", "Group By", choices = c("Species", "Petal.Width")),

          checkboxGroupInput(
            "rowDetails",
            "Row Details",
            choices = c(
              "Show row details" = "showRowDetails",
              "Multiple row details" = "multiRowDetails",
              "Column header" = "columnHeader"
            ),
            selected = "showRowDetails"
          ),

          radioButtons(
            "rowSelection",
            "Row Selection",
            choices = c(
              "None" = "none",
              "Single" = "single",
              "Multiple" = "multiple"
            )
          ),

          checkboxGroupInput(
            "colOptions",
            "Column Options",
            choices = c(
              Formatting = "format",
              "Custom cell renderer" = "cell"
            ),
            selected = c("format", "cell")
          )
      ),
      div(class = "content",
          tabsetPanel(
            tabPanel(
              "Table",
              reactableOutput("table"),
              conditionalPanel(
                "input.rowSelection !== 'none'",
                hr(),
                "Selected rows:",
                verbatimTextOutput("selected")
              )
            ),
            tabPanel(
              "Code",
              htmlOutput("code")
            )
          )
      )
  )
)

server <- function(input, output, session) {
  tblOptions <- reactive({
    list(
      groupBy = input$groupBy,
      columnGroups = if ("columnGroups" %in% input$options) {
        quote(list(
          colGroup(name = "Sepal", columns = c("Sepal.Length", "Sepal.Width")),
          colGroup(name = "Petal", columns = c("Petal.Length", "Petal.Width"))
        ))
      },
      filterable = "filterable" %in% input$options,
      searchable = "searchable" %in% input$options,
      resizable = "resizable" %in% input$options,
      sortable = "sortable" %in% input$options,
      defaultSorted = if ("defaultSorted" %in% input$options) c("Sepal.Length", "Sepal.Width"),
      pagination = "pagination" %in% input$pagination,
      paginationType = input$paginationType,
      showPageSizeOptions = all(c("showPageSizeOptions", "pagination") %in% input$pagination),
      showPageInfo = "showPageInfo" %in% input$pagination,
      selection = if (input$rowSelection != "none") input$rowSelection,
      onClick = if (input$rowSelection != "none") {
        "select"
      } else if (any(c("showRowDetails", "multiRowDetails") %in% input$rowDetails)) {
        "expand"
      },
      outlined = "outlined" %in% input$options,
      bordered = "bordered" %in% input$options,
      borderless = "borderless" %in% input$options,
      striped = "striped" %in% input$options,
      highlight = "highlight" %in% input$options,
      compact = "compact" %in% input$options,
      showSortable = "showSortable" %in% input$options,
      fullWidth = "fullWidth" %in% input$options,
      height = if ("pagination" %in% input$pagination) "auto" else 500,
      theme = if (input$theme != "Default") bquote(.(themes[[input$theme]]))
    )
  })

  colOptions <- reactive({
    list(
      Sepal.Length = list(
        format = if ("format" %in% input$colOptions) {
          quote(colFormat(suffix = " cm", digits = 1))
        }
      ),
      Sepal.Width = list(
        format = if ("format" %in% input$colOptions) {
          quote(list(aggregated = colFormat(suffix = " (avg)", digits = 2)))
        },
        cell = if ("cell" %in% input$colOptions) {
          function(value) {
            if (value >= 3.3) {
              classes <- "tag num-high"
            } else if (value >= 3) {
              classes <- "tag num-med"
            } else {
              classes <- "tag num-low"
            }
            value <- format(value, nsmall = 1)
            span(class = classes, value)
          }
        },
        footer = if ("footer" %in% input$options) {
          function(values) {
            div(tags$b("Average: "), round(mean(values), 1))
          }
        },
        details = if ("multiRowDetails" %in% input$rowDetails) {
          function(index) {
            tagList(
              paste("Details for row:", index),
              pre(paste(capture.output(iris[index, ]), collapse = "\n"))
            )
          }
        }
      )
    )
  })

  rowDetailsOptions <- reactive({
    if (!any(c("showRowDetails", "multiRowDetails") %in% input$rowDetails)) {
      return(NULL)
    }

    details <- function(index) {
      if (index == 3) {
        tabsetPanel(
          tabPanel("plot", plotOutput("plot")),
          tabPanel("subtable", reactable(iris[1:3, 1:2], fullWidth = FALSE))
        )
      } else if (index == 5) {
        paste("Details for row:", index)
      }
    }

    if (!"columnHeader" %in% input$rowDetails) {
      details
    } else {
      bquote(
        colDef(
          details = .(details),
          name = "More",
          width = 70
        )
      )
    }
  })

  code <- reactive({
    opts <- tblOptions()
    colOpts <- colOptions()

    call <- bquote(reactable(
      iris,
      columnGroups = .(opts$columnGroups),
      filterable = .(opts$filterable),
      searchable = .(opts$searchable),
      sortable = .(opts$sortable),
      resizable = .(opts$resizable),
      pagination = .(opts$pagination),
      paginationType = .(opts$paginationType),
      showPageSizeOptions = .(opts$showPageSizeOptions),
      showPageInfo = .(opts$showPageInfo),
      selection = .(opts$selection),
      onClick = .(opts$onClick),
      outlined = .(opts$outlined),
      bordered = .(opts$bordered),
      borderless = .(opts$borderless),
      striped = .(opts$striped),
      highlight = .(opts$highlight),
      compact = .(opts$compact),
      showSortable = .(opts$showSortable),
      groupBy = .(opts$groupBy),
      defaultSorted = .(opts$defaultSorted),
      fullWidth = .(opts$fullWidth),
      height = .(opts$height),
      columns = list(
        Sepal.Length = colDef(
          name = .(if (is.null(opts$columnGroups)) "Sepal Length" else "Length"),
          aggregate = "max",
          format = .(colOpts$Sepal.Length$format)
        ),
        Sepal.Width = colDef(
          name = .(if (is.null(opts$columnGroups)) "Sepal Width" else "Width"),
          defaultSortOrder = "desc",
          aggregate = "mean",
          format = .(colOpts$Sepal.Width$format),
          cell = .(colOpts$Sepal.Width$cell),
          footer = .(colOpts$Sepal.Width$footer),
          details = .(colOpts$Sepal.Width$details)
        ),
        Petal.Length = colDef(
          name = .(if (is.null(opts$columnGroups)) "Petal Length" else "Length"),
          aggregate = "sum"
        ),
        Petal.Width = colDef(
          name = .(if (is.null(opts$columnGroups)) "Petal Width" else "Width"),
          aggregate = "count"
        ),
        Species = colDef(
          aggregate = "frequency"
        )
      ),
      details = .(rowDetailsOptions()),
      theme = .(if (!is.null(opts$theme)) quote(theme))
    ))

    # Omit default reactable args
    defaultArgs <- formals(reactable)
    for (argname in names(call)) {
      if (argname != "" &&
          (is.null(call[[argname]]) || identical(call[[argname]], defaultArgs[[argname]]))) {
        call[[argname]] <- NULL
      }
    }

    # Omit default colDef args
    defaultArgs <- formals(colDef)
    for (colname in names(call$columns)) {
      col <- call$columns[[colname]]
      for (argname in names(col)) {
        if (argname != "" && identical(col[[argname]], defaultArgs[[argname]])) {
          call$columns[[colname]][[argname]] <- NULL
        }
      }
    }

    code <- gsub("\\n", "\n", deparse(call, control = "useSource"), fixed = TRUE)

    if (!is.null(opts$theme)) {
      themeCall <- bquote({ theme <- .(opts$theme) })
      themeCode <- gsub("\\n", "\n", deparse(themeCall[[2]], control = "useSource"), fixed = TRUE)
      code <- c(themeCode, "", code)
    }

    paste(code, collapse = "\n")
  })

  output$code <- renderUI({
    prismCodeBlock(code())
  })

  output$table <- renderReactable({
    eval(parse(text = code()))
  })

  output$selected <- renderPrint({
    print(getReactableState("table", "selected"))
  })

  output$plot <- renderPlot({
    hist(iris$Sepal.Length)
  })

  outputOptions(output, "table", suspendWhenHidden = FALSE)
}

shinyApp(ui, server)

.onLoad <- function(...) {
  registerMethods(list(
    c("htmltools", "as.tags", "reactable"),
    c("knitr", "knit_print", "reactable")
  ))
}


# Reusable function for registering a set of methods with S3 manually. The
# methods argument is a list of character vectors, each of which has the form
# c(package, genname, class).
registerMethods <- function(methods) {
  lapply(methods, function(method) {
    pkg <- method[[1]]
    generic <- method[[2]]
    class <- method[[3]]
    func <- get(paste(generic, class, sep = "."))
    if (pkg %in% loadedNamespaces()) {
      registerS3method(generic, class, func, envir = asNamespace(pkg))
    }
    setHook(
      packageEvent(pkg, "onLoad"),
      function(...) {
        registerS3method(generic, class, func, envir = asNamespace(pkg))
      }
    )
  })
}

# Create client-rendered versions of the Examples and Cookbook docs for comparison
examples_src <- readLines("vignettes/examples.Rmd")
examples_src <- sub("options(reactable.static = TRUE)", "options(reactable.static = FALSE)", examples_src, fixed = TRUE)
writeLines(examples_src, "vignettes/examples-no-static.Rmd")

cookbook_src <- readLines("vignettes/cookbook/cookbook.Rmd")
cookbook_src <- sub("options(reactable.static = TRUE)", "options(reactable.static = FALSE)", cookbook_src, fixed = TRUE)
writeLines(cookbook_src, "vignettes/cookbook/cookbook-no-static.Rmd")

pkgdown::build_site()

unlink("vignettes/examples-no-static.Rmd")
unlink("vignettes/cookbook/cookbook-no-static.Rmd")

# Replace example table image in README with an actual table
index_html <- xml2::read_html("docs/index.html")

example <- xml2::xml_find_first(index_html, ".//*/comment()[contains(., 'pkgdown:example')]/following-sibling::*")

rmarkdown::render("pkgdown/example.Rmd", output_dir = "docs")
example_html <- xml2::read_html("docs/example.html")
widget_html <- xml2::xml_find_first(example_html, ".//*[contains(@class, 'reactable')]")
widget_deps <- xml2::xml_find_all(example_html, ".//script|.//link")

for (dep in rev(widget_deps)) {
  xml2::xml_add_sibling(example, dep)
}
xml2::xml_replace(example, widget_html)

xml2::write_html(index_html, "docs/index.html")

#' Language options
#'
#' @description
#' Use `reactableLang()` to customize the language strings in a table.
#' Language strings include both visible text and accessible labels that can
#' be read by assistive technology, such as screen readers.
#'
#' To set the default language strings for all tables, use the global
#' `reactable.language` option.
#'
#' @param sortLabel Accessible label for column sort buttons.
#'   Takes a `{name}` parameter for the column name.
#'
#' @param filterPlaceholder Placeholder for column filter inputs.
#' @param filterLabel Accessible label for column filter inputs.
#'   Takes a `{name}` parameter for the column name.
#'
#' @param searchPlaceholder Placeholder for the table search input.
#' @param searchLabel Accessible label for the table search input.
#'
#' @param noData Placeholder text when the table has no data.
#'
#' @param pageNext Text for the next page button.
#' @param pagePrevious Text for the previous page button.
#' @param pageNumbers Text for the page numbers info. Only used with the `"jump"` and
#'   `"simple"` pagination types.
#'   Takes the following parameters:
#'   - `{page}` for the current page
#'   - `{pages}` for the total number of pages
#' @param pageInfo Text for the page info.
#'   Takes the following parameters:
#'   - `{rowStart}` for the starting row of the page
#'   - `{rowEnd}` for the ending row of the page
#'   - `{rows}` for the total number of rows
#' @param pageSizeOptions Text for the page size options input.
#'   Takes a `{rows}` parameter for the page size options input.
#' @param pageNextLabel Accessible label for the next page button.
#' @param pagePreviousLabel Accessible label for the previous page button.
#' @param pageNumberLabel Accessible label for the page number buttons.
#'   Only used with the the `"numbers"` pagination type.
#'   Takes a `{page}` parameter for the page number.
#' @param pageJumpLabel Accessible label for the page jump input. Only used with
#'   the `"jump"` pagination type.
#' @param pageSizeOptionsLabel Accessible label for the page size options input.
#'
#' @param groupExpandLabel Accessible label for the row group expand button.
#'
#' @param detailsExpandLabel Accessible label for the row details expand button.
#'
#' @param selectAllRowsLabel Accessible label for the select all rows checkbox.
#' @param selectAllSubRowsLabel Accessible label for the select all sub rows checkbox.
#' @param selectRowLabel Accessible label for the select row checkbox.
#'
#' @param defaultGroupHeader Deprecated and no longer used.
#' @param detailsCollapseLabel Deprecated and no longer used.
#' @param deselectAllRowsLabel Deprecated and no longer used.
#' @param deselectAllSubRowsLabel Deprecated and no longer used.
#' @param deselectRowLabel Deprecated and no longer used.
#'
#' @return A language options object that can be used to customize the language
#'   strings in `reactable()`.
#'
#' @usage reactableLang(
#'   sortLabel = "Sort {name}",
#'   filterPlaceholder = "",
#'   filterLabel = "Filter {name}",
#'   searchPlaceholder = "Search",
#'   searchLabel = "Search",
#'   noData = "No rows found",
#'   pageNext = "Next",
#'   pagePrevious = "Previous",
#'   pageNumbers = "{page} of {pages}",
#'   pageInfo = "{rowStart}\u2013{rowEnd} of {rows} rows",
#'   pageSizeOptions = "Show {rows}",
#'   pageNextLabel = "Next page",
#'   pagePreviousLabel = "Previous page",
#'   pageNumberLabel = "Page {page}",
#'   pageJumpLabel = "Go to page",
#'   pageSizeOptionsLabel = "Rows per page",
#'   groupExpandLabel = "Toggle group",
#'   detailsExpandLabel = "Toggle details",
#'   selectAllRowsLabel = "Select all rows",
#'   selectAllSubRowsLabel = "Select all rows in group",
#'   selectRowLabel = "Select row",
#'   defaultGroupHeader = NULL,
#'   detailsCollapseLabel = NULL,
#'   deselectAllRowsLabel = NULL,
#'   deselectAllSubRowsLabel = NULL,
#'   deselectRowLabel = NULL
#' )
#'
#' @examples
#' reactable(
#'   iris[1:30, ],
#'   searchable = TRUE,
#'   paginationType = "simple",
#'   language = reactableLang(
#'     searchPlaceholder = "Search...",
#'     noData = "No entries found",
#'     pageInfo = "{rowStart}\u2013{rowEnd} of {rows} entries",
#'     pagePrevious = "\u276e",
#'     pageNext = "\u276f",
#'
#'     # Accessible labels for assistive technology, such as screen readers
#'     pagePreviousLabel = "Previous page",
#'     pageNextLabel = "Next page"
#'   )
#' )
#'
#' # Set the default language for all tables
#' options(reactable.language = reactableLang(
#'   searchPlaceholder = "Search...",
#'   noData = "No entries found",
#'   pageInfo = "{rowStart} to {rowEnd} of {rows} entries"
#' ))
#'
#' reactable(iris[1:30, ], searchable = TRUE)
#'
#' @export
reactableLang <- function(
  # Sorting
  sortLabel = "Sort {name}",

  # Filters
  filterPlaceholder = "",
  filterLabel = "Filter {name}",

  # Search
  searchPlaceholder = "Search",
  searchLabel = "Search",

  # Tables
  noData = "No rows found",

  # Pagination
  pageNext = "Next",
  pagePrevious = "Previous",
  pageNumbers = "{page} of {pages}",
  pageInfo = "{rowStart}\u2013{rowEnd} of {rows} rows",
  pageSizeOptions = "Show {rows}",
  pageNextLabel = "Next page",
  pagePreviousLabel = "Previous page",
  pageNumberLabel = "Page {page}",
  pageJumpLabel = "Go to page",
  pageSizeOptionsLabel = "Rows per page",

  # Row grouping
  groupExpandLabel = "Toggle group",

  # Row details
  detailsExpandLabel = "Toggle details",

  # Selection
  selectAllRowsLabel = "Select all rows",
  selectAllSubRowsLabel = "Select all rows in group",
  selectRowLabel = "Select row",

  # Deprecated and no longer used (in v0.3.0)
  defaultGroupHeader = NULL,
  detailsCollapseLabel = NULL,
  deselectAllRowsLabel = NULL,
  deselectAllSubRowsLabel = NULL,
  deselectRowLabel = NULL
) {

  if (!is.null(defaultGroupHeader)) {
    warning(
      "`defaultGroupHeader` is deprecated and no longer used. ",
      "Use the `columnGroups` argument in `reactable()` to customize the column ",
      "group header for `groupBy` columns."
    )
  }
  if (!is.null(detailsCollapseLabel)) {
    warning("`detailsCollapseLabel` is deprecated and no longer used. ",
            "Use the `detailsExpandLabel` argument to customize the accessible label ",
            "for the row details expand button.")
  }
  if (!is.null(deselectAllRowsLabel)) {
    warning("`deselectAllRowsLabel` is deprecated and no longer used. ",
            "Use the `selectAllRowsLabel` argument to customize the accessible label ",
            "for the select all rows checkbox.")
  }
  if (!is.null(deselectAllSubRowsLabel)) {
    warning("`deselectAllSubRowsLabel` is deprecated and no longer used. ",
            "Use the `selectAllSubRowsLabel` argument to customize the accessible label ",
            "for the select all sub rows checkbox.")
  }
  if (!is.null(deselectRowLabel)) {
    warning("`deselectRowLabel` is deprecated and no longer used. ",
            "Use the `selectRowLabel` argument to customize the accessible label ",
            "for the select row checkbox.")
  }

  defaultArgs <- formals()
  args <- as.list(match.call())
  args <- args[names(args) %in% names(defaultArgs)]

  for (arg in names(args)) {
    value <- get(arg)
    if (!is.null(value) && !is.character(value)) {
      stop(sprintf("`%s` must be a character string", arg))
    }
    args[[arg]] <- value
  }
  lang <- filterNulls(args)
  structure(lang, class = "reactableLang")
}

is.reactableLang <- function(x) {
  inherits(x, "reactableLang")
}

test_that("reactableLang", {
  expect_equal(reactableLang(), structure(list(), class = "reactableLang"))
  expect_equal(reactableLang(sortLabel = "_Sort {name}", searchPlaceholder = "", noData = NULL),
               structure(list(sortLabel = "_Sort {name}", searchPlaceholder = ""), class = "reactableLang"))

  lang <- list(
    # Sorting
    sortLabel = "_Sort {name}",

    # Filters
    filterPlaceholder = "_Filter",
    filterLabel = "_Filter {name}",

    # Search
    searchPlaceholder = "_Search",
    searchLabel = "_Search",

    # Tables
    noData = "_No rows found",

    # Pagination
    pageNext = "_Next",
    pagePrevious = "_Previous",
    pageNumbers = "_{page} of {pages}",
    pageInfo = "_{rowStart}–{rowEnd} of {rows} rows",
    pageSizeOptions = "_Show {rows}",
    pageNextLabel = "_Next page",
    pagePreviousLabel = "_Previous page",
    pageNumberLabel = "_Page {page}",
    pageJumpLabel = "_Go to page",
    pageSizeOptionsLabel = "_Rows per page",

    # Row grouping
    groupExpandLabel = "_Toggle group",

    # Row details
    detailsExpandLabel = "_Toggle details",

    # Selection
    selectAllRowsLabel = "_Select all rows",
    selectAllSubRowsLabel = "_Select all rows in group",
    selectRowLabel = "_Select row {row}"
  )
  expect_equal(do.call(reactableLang, lang), structure(lang, class = "reactableLang"))

  # Expressions should work
  expect_equal(reactableLang(noData = paste("no", "data"))$noData, "no data")
  expect_equal(reactableLang(pageInfo = (function() "pageinfo")())$pageInfo, "pageinfo")
  expect_equal(do.call(reactableLang, list(sortLabel = "sort"))$sortLabel, "sort")

  # Errors
  expect_error(reactableLang(noData = 123, detailsExpandLabel = TRUE), "`noData` must be a character string")
  expect_error(reactableLang(selectRowLabel = list()), "`selectRowLabel` must be a character string")

  # Deprecations
  deprecatedArgs <- c("defaultGroupHeader", "detailsCollapseLabel", "deselectAllRowsLabel",
                      "deselectAllSubRowsLabel", "deselectRowLabel")
  for (arg in deprecatedArgs) {
    expect_warning(
      do.call(reactableLang, setNames(list("label"), arg)),
      sprintf("`%s` is deprecated and no longer used", arg)
    )
  }
})

test_that("is.reactableLang", {
  expect_true(is.reactableLang(reactableLang()))
  expect_false(is.reactableLang(list()))
})

#' Update a reactable instance
#'
#' `updateReactable()` updates a reactable instance within a Shiny application.
#'
#' @param outputId The Shiny output ID of the `reactable` instance.
#' @param data Table data. A data frame or matrix.
#'
#'   `data` should have the same columns as the original table data.
#'   When updating `data`, the selected rows, expanded rows, and current page
#'   will reset unless explicitly specified. All other state will persist,
#'   including sorting, filtering, and grouping state.
#' @param selected Selected rows. Either a numeric vector of row indices,
#'   or `NA` to deselect all rows.
#' @param expanded Expanded rows. Either `TRUE` to expand all rows, or `FALSE`
#'   to collapse all rows.
#' @param page The current page. A single, positive integer.
#' @param meta Custom table metadata. Either a named list with new values, or `NA`
#'   to clear all metadata. New values are merged into the current metadata, so only
#'   the values specified in `meta` will be updated.
#' @param session The Shiny session object. Defaults to the current Shiny session.
#' @return None
#'
#' @examples
#' # Run in an interactive R session
#' if (interactive()) {
#'
#' library(shiny)
#' library(reactable)
#'
#' data <- MASS::Cars93[, 1:7]
#'
#' ui <- fluidPage(
#'   actionButton("select_btn", "Select rows"),
#'   actionButton("clear_btn", "Clear selection"),
#'   actionButton("expand_btn", "Expand rows"),
#'   actionButton("collapse_btn", "Collapse rows"),
#'   actionButton("page_btn", "Change page"),
#'   selectInput("filter_type", "Filter type", unique(data$Type), multiple = TRUE),
#'   reactableOutput("table")
#' )
#'
#' server <- function(input, output) {
#'   output$table <- renderReactable({
#'     reactable(
#'       data,
#'       filterable = TRUE,
#'       searchable = TRUE,
#'       selection = "multiple",
#'       details = function(index) paste("Details for row:", index)
#'     )
#'   })
#'
#'   observeEvent(input$select_btn, {
#'     # Select rows
#'     updateReactable("table", selected = c(1, 3, 5))
#'   })
#'
#'   observeEvent(input$clear_btn, {
#'     # Clear row selection
#'     updateReactable("table", selected = NA)
#'   })
#'
#'   observeEvent(input$expand_btn, {
#'     # Expand all rows
#'     updateReactable("table", expanded = TRUE)
#'   })
#'
#'   observeEvent(input$collapse_btn, {
#'     # Collapse all rows
#'     updateReactable("table", expanded = FALSE)
#'   })
#'
#'   observeEvent(input$page_btn, {
#'     # Change current page
#'     updateReactable("table", page = 3)
#'   })
#'
#'   observe({
#'     # Filter data
#'     filtered <- if (length(input$filter_type) > 0) {
#'       data[data$Type %in% input$filter_type, ]
#'     } else {
#'       data
#'     }
#'     updateReactable("table", data = filtered)
#'   })
#' }
#'
#' shinyApp(ui, server)
#' }
#'
#' @export
updateReactable <- function(outputId, data = NULL, selected = NULL, expanded = NULL,
                            page = NULL, meta = NULL, session = NULL) {
  if (is.null(session)) {
    if (requireNamespace("shiny", quietly = TRUE)) {
      session <- shiny::getDefaultReactiveDomain()
    }
    if (is.null(session)) {
      # Not in an active Shiny session
      return(invisible(NULL))
    }
  }

  if (!is.character(outputId)) {
    stop("`outputId` must be a character string")
  }
  outputId <- session$ns(outputId)

  dataKey <- NULL
  if (!is.null(data)) {
    if (!is.data.frame(data) && !is.matrix(data)) {
      stop("`data` must be a data frame or matrix")
    }
    dataKey <- digest::digest(data)
    # Reset selected, expanded, and page state by default
    selected <- if (is.null(selected)) NA else selected
    expanded <- if (is.null(expanded)) FALSE else expanded
    page <- if (is.null(page)) 1 else page
  }

  if (!is.null(selected)) {
    if (!is.numeric(selected) && !is.na(selected)) {
      stop("`selected` must be numeric or NA")
    }
    selected <- stats::na.omit(selected)
    # Convert to 0-based indexing
    selected <- as.list(as.integer(selected) - 1)
  }

  if (!is.null(expanded) && !is.logical(expanded)) {
    stop("`expanded` must be TRUE or FALSE")
  }

  if (!is.null(page)) {
    if (!is.numeric(page) || length(page) != 1 || page <= 0) {
      stop("`page` must be a single, positive integer")
    }
    # Convert to 0-based indexing
    page <- as.integer(page - 1)
  }

  if (!is.null(meta)) {
    if (!isNamedList(meta) && !is.na(meta)) {
      stop("`meta` must be a named list or NA")
    }
    # Allow empty lists, but don't serialize them as an empty array, []
    if (identical(meta, list())) {
      meta <- NULL
    }
  }

  # Get JS evals for meta. Exclude other props like data - although data could
  # potentially have JS() code within list-columns, it's not supported by reactable(), and
  # JS() code just ends up as a string.
  jsEvals <- htmlwidgets::JSEvals(list(meta = meta))
  if (length(jsEvals) == 0) {
    jsEvals <- NULL
  }

  newState <- filterNulls(list(
    data = data,
    dataKey = dataKey,
    selected = selected,
    expanded = expanded,
    page = page,
    meta = meta,
    jsEvals = jsEvals
  ))

  if (length(newState) > 0) {
    session$sendCustomMessage(sprintf("__reactable__%s", outputId), newState)
  }
}

#' Get the state of a reactable instance
#'
#' `getReactableState()` gets the state of a reactable instance within a Shiny application.
#'
#' @param outputId The Shiny output ID of the `reactable` instance.
#' @param name Character vector of state value(s) to get. Values must be one of `"page"`,
#'   `"pageSize"`, `"pages"`, `sorted`, or `"selected"`. If unspecified, all values will
#'   be returned.
#' @param session The Shiny session object. Defaults to the current Shiny session.
#' @return If `name` is specified, one of the following values:
#'
#'   - `page`: the current page
#'   - `pageSize`: the page size
#'   - `pages`: the number of pages
#'   - `sorted`: the sorted columns - a named list of columns with values of `"asc"` for
#'      ascending order or `"desc"` for descending order, or `NULL` if no columns are sorted
#'   - `selected`: the selected rows - a numeric vector of row indices, or `NULL` if no rows are selected
#'
#'  If `name` contains more than one value, `getReactableState()` returns a named list of
#'  the specified values.
#'
#'  If `name` is unspecified, `getReactableState()` returns a named list containing all values.
#'
#'  If the table has not been rendered yet, `getReactableState()` returns `NULL`.
#'
#' @examples
#' # Run in an interactive R session
#' if (interactive()) {
#'
#' library(shiny)
#' library(reactable)
#' library(htmltools)
#'
#' ui <- fluidPage(
#'   actionButton("prev_page_btn", "Previous page"),
#'   actionButton("next_page_btn", "Next page"),
#'   reactableOutput("table"),
#'   verbatimTextOutput("table_state"),
#'   uiOutput("selected_row_details")
#' )
#'
#' server <- function(input, output) {
#'   output$table <- renderReactable({
#'     reactable(
#'       MASS::Cars93[, 1:5],
#'       showPageSizeOptions = TRUE,
#'       selection = "multiple",
#'       onClick = "select"
#'     )
#'   })
#'
#'   output$table_state <- renderPrint({
#'     state <- req(getReactableState("table"))
#'     print(state)
#'   })
#'
#'   observeEvent(input$prev_page_btn, {
#'     # Change to the previous page
#'     page <- getReactableState("table", "page")
#'     if (page > 1) {
#'       updateReactable("table", page = page - 1)
#'     }
#'   })
#'
#'   observeEvent(input$next_page_btn, {
#'     # Change to the next page
#'     state <- getReactableState("table")
#'     if (state$page < state$pages) {
#'       updateReactable("table", page = state$page + 1)
#'     }
#'   })
#'
#'   output$selected_row_details <- renderUI({
#'     selected <- getReactableState("table", "selected")
#'     req(selected)
#'     details <- MASS::Cars93[selected, -c(1:5)]
#'     tagList(
#'       h2("Selected row details"),
#'       tags$pre(
#'         paste(capture.output(print(details, width = 1200)), collapse = "\n")
#'       )
#'     )
#'   })
#' }
#'
#' shinyApp(ui, server)
#' }
#'
#' @export
getReactableState <- function(outputId, name = NULL, session = NULL) {
  if (is.null(session)) {
    if (requireNamespace("shiny", quietly = TRUE)) {
      session <- shiny::getDefaultReactiveDomain()
    }
    if (is.null(session)) {
      # Not in an active Shiny session
      return(NULL)
    }
  }
  if (!is.character(outputId)) {
    stop("`outputId` must be a character string")
  }

  getState <- function(outputId, name) {
    # NOTE: input IDs must always come first to work with Shiny modules
    session$input[[sprintf("%s__reactable__%s", outputId, name)]]
  }

  props <- c("page", "pageSize", "pages", "sorted", "selected")
  if (!is.null(name)) {
    if (!is.character(name) || any(!name %in% props)) {
      stop(paste("`name` values must be one of", paste(sprintf('"%s"', props), collapse = ", ")))
    }
    if (length(name) == 1) {
      return(getState(outputId, name))
    } else {
      props <- name
    }
  }

  state <- stats::setNames(
    lapply(props, function(prop) {
      getState(outputId, prop)
    }),
    props
  )

  if (length(filterNulls(state)) == 0) {
    return(NULL)
  }

  state
}

mockSession <- function(namespace = NULL) {
  e <- new.env()
  e$input <- list()
  e$msgs <- list()
  e$sendCustomMessage <- function(type, message) {
    msg <- list(type = type, message = message)
    e$msgs[[length(e$msgs) + 1]] <- msg
    e$lastMsg <- msg
    msg
  }
  e$ns <- function(id) {
    shiny::NS(namespace, id)
  }
  e
}

test_that("updateReactable", {
  session <- mockSession()
  expect_error(updateReactable(123, session = session),
               "`outputId` must be a character string")
  expect_error(updateReactable("id", data = list(), session = session),
               "`data` must be a data frame or matrix")
  expect_error(updateReactable("id", selected = TRUE, session = session),
               "`selected` must be numeric")
  expect_error(updateReactable("id", expanded = 123, session = session),
               "`expanded` must be TRUE or FALSE")
  expect_error(updateReactable("id", page = TRUE, session = session),
               "`page` must be a single, positive integer")
  expect_error(updateReactable("id", page = c(1, 3), session = session),
               "`page` must be a single, positive integer")
  expect_error(updateReactable("id", page = 0, session = session),
               "`page` must be a single, positive integer")

  expect_null(updateReactable("id"))
  updateReactable("id", session = session)
  expect_null(session$lastMsg)

  # Update data
  updateReactable("mytbl", data = data.frame(x = 1), session = session)
  expected <- list(
    data = data.frame(x = 1),
    dataKey = digest::digest(data.frame(x = 1)),
    selected = list(),
    expanded = FALSE,
    page = 0
  )
  expect_equal(session$lastMsg, list(type = "__reactable__mytbl", message = expected))
  # Override state resets
  updateReactable("mytbl", data = matrix(4), selected = 3, expanded = TRUE,
                  page = 2, session = session)
  expected <- list(
    data = matrix(4),
    dataKey = digest::digest(matrix(4)),
    selected = list(2),
    expanded = TRUE,
    page = 1
  )
  expect_equal(session$lastMsg, list(type = "__reactable__mytbl", message = expected))

  # Update selected rows
  updateReactable("mytbl", selected = 1, session = session)
  expect_equal(session$lastMsg, list(type = "__reactable__mytbl", message = list(selected = list(0))))
  updateReactable("mytbl", selected = c(1, 3, 5), session = session)
  expect_equal(session$lastMsg, list(type = "__reactable__mytbl", message = list(selected = list(0, 2, 4))))
  updateReactable("mytbl", selected = integer(0), session = session)
  expect_equal(session$lastMsg, list(type = "__reactable__mytbl", message = list(selected = list())))
  updateReactable("mytbl", selected = NA, session = session)
  expect_equal(session$lastMsg, list(type = "__reactable__mytbl", message = list(selected = list())))
  updateReactable("mytbl", selected = NA_real_, session = session)
  expect_equal(session$lastMsg, list(type = "__reactable__mytbl", message = list(selected = list())))
  updateReactable("mytbl", selected = c(3, 5, NA), session = session)
  expect_equal(session$lastMsg, list(type = "__reactable__mytbl", message = list(selected = list(2, 4))))

  # Update expanded rows
  updateReactable("mytbl", expanded = TRUE, session = session)
  expect_equal(session$lastMsg, list(type = "__reactable__mytbl", message = list(expanded = TRUE)))
  updateReactable("mytbl", expanded = FALSE, session = session)
  expect_equal(session$lastMsg, list(type = "__reactable__mytbl", message = list(expanded = FALSE)))

  updateReactable("mytbl", selected = c(1, 3), expanded = FALSE, session = session)
  expect_equal(session$lastMsg, list(
    type = "__reactable__mytbl",
    message = list(selected = list(0, 2), expanded = FALSE)
  ))

  # Update current page
  updateReactable("mytbl", page = 2, session = session)
  expect_equal(session$lastMsg, list(type = "__reactable__mytbl", message = list(page = 1)))

  # Update meta
  expect_error(updateReactable("mytbl", meta = TRUE, session = session),
               "`meta` must be a named list or NA")

  updateReactable("mytbl", meta = list(custom = 123, fn = JS("n => n > 30")), session = session)
  expect_equal(session$lastMsg, list(
    type = "__reactable__mytbl",
    message = list(
      meta = list(custom = 123, fn = JS("n => n > 30")),
      jsEvals = I("meta.fn") # Should be wrapped in I() so length-1 arrays serialize as arrays
    )
  ))
  updateReactable("mytbl", meta = NA, session = session)
  expect_equal(session$lastMsg, list(
    type = "__reactable__mytbl",
    message = list(meta = NA)
  ))
  updateReactable("mytbl", meta = list(), page = 1, session = session)
  expect_equal(session$lastMsg, list(
    type = "__reactable__mytbl",
    message = list(page = 0)
  ))

  # JS evals should not include data
  updateReactable("mytbl", data = data.frame(x = I(list(fn = JS("() => {}")))), session = session)
  expect_equal(session$lastMsg$message$jsEvals, NULL)

  # Should work with Shiny modules
  session <- mockSession(namespace = "mod")
  updateReactable("mytbl", selected = 2, session = session)
  expect_equal(session$lastMsg, list(type = "__reactable__mod-mytbl", message = list(selected = list(1))))
})

test_that("getReactableState", {
  session <- mockSession()
  expect_error(getReactableState(123, session = session), "`outputId` must be a character string")
  expect_error(getReactableState("id", "x", session = session), '`name` values must be one of "page", "pageSize", "pages", "sorted", "selected"')

  expect_null(getReactableState("id"))
  updateReactable("id", session = session)

  expect_equal(getReactableState("mytbl", session = session), NULL)

  session$input[["mytbl__reactable__page"]] <- 3
  expect_equal(getReactableState("mytbl", "page", session = session), 3)

  session$input[["mytbl__reactable__pageSize"]] <- 2
  expect_equal(getReactableState("mytbl", "pageSize", session = session), 2)

  session$input[["mytbl__reactable__pages"]] <- 10
  expect_equal(getReactableState("mytbl", "pages", session = session), 10)

  session$input[["mytbl__reactable__sorted"]] <- list(a = "asc", b = "desc")
  expect_equal(getReactableState("mytbl", "sorted", session = session), list(a = "asc", b = "desc"))

  session$input[["mytbl__reactable__selected"]] <- c(1, 5, 7)
  expect_equal(getReactableState("mytbl", "selected", session = session), c(1, 5, 7))

  # Multiple values
  expect_equal(getReactableState("mytbl", c("page", "pageSize"), session = session), list(page = 3, pageSize = 2))

  # All values
  expect_equal(getReactableState("mytbl", session = session), list(
    page = 3,
    pageSize = 2,
    pages = 10,
    sorted = list(a = "asc", b = "desc"),
    selected = c(1, 5, 7)
  ))
})

#' @importFrom htmlwidgets JS
#' @export
htmlwidgets::JS

toJSON <- function(x) {
  jsonlite::toJSON(x, dataframe = "columns", rownames = FALSE, digits = NA,
                   POSIXt = "ISO8601", Date = "ISO8601", UTC = TRUE, force = TRUE,
                   auto_unbox = TRUE, null = "null")
}

mergeLists <- function(a, b) {
  if (is.null(a)) {
    a <- list()
  }
  if (is.null(b)) {
    b <- list()
  }
  for (i in seq_along(b)) {
    name <- names(b[i])
    if (is.null(name) || name == "") {
      a <- c(a, b[i])
    } else if (!is.null(b[[i]])) {
      a[[name]] <- b[[i]]
    }
  }
  a
}

filterNulls <- function(x) {
  Filter(Negate(is.null), x)
}

is.JS <- function(x) {
  inherits(x, class(JS("")))
}

isNamedList <- function(x) {
  if (!is.list(x)) {
    return(FALSE)
  }
  if (length(x) >= 1 && is.null(names(x))) {
    return(FALSE)
  }
  if (any(names(x) == "")) {
    return(FALSE)
  }
  TRUE
}

is.Date <- function(x) {
  inherits(x, "Date")
}

is.POSIXct <- function(x) {
  inherits(x, "POSIXct")
}

is.POSIXlt <- function(x) {
  inherits(x, "POSIXlt")
}

is.tag <- function(x) {
  inherits(x, "shiny.tag")
}

is.htmlwidget <- function(x) {
  inherits(x, "htmlwidget")
}

is.htmlDependency <- function(x) {
  inherits(x, "html_dependency")
}

# Test both shiny.tag.list and regular lists of tags
isTagList <- function(x) {
  inherits(x, "shiny.tag.list") || (is.list(x) && all(sapply(x, is.tag)))
}

asReactTag <- function(x) {
  if (is.htmlwidget(x)) {
    if (inherits(x, "reactable")) {
      # Extract tag for nested tables
      x$x$tag <- htmltools::tagAppendAttributes(x$x$tag, nested = TRUE)
      tag <- asReactTag(x$x$tag)
      tag <- htmltools::attachDependencies(tag, x$dependencies)
      return(tag)
    } else {
      tags <- htmltools::as.tags(x)
      tags <- asReactTag(tags)
      # Add a key for proper widget rerenders
      tags <- reactR::React$WidgetContainer(tags, key = digest::digest(x))
      return(tags)
    }
  }

  # Unnest and wrap tag lists in fragments for proper hydration
  if (isTagList(x)) {
    x <- unnestTagList(x)
    # Preserve HTML dependencies
    deps <- htmltools::htmlDependencies(x)
    # Collect inline HTML dependencies
    inlineDeps <- sapply(x, is.htmlDependency)
    if (any(inlineDeps)) {
      deps <- c(deps, x[inlineDeps])
      x[inlineDeps] <- NULL
    }
    x <- lapply(x, asReactTag)
    x <- do.call(reactR::React$Fragment, x)
    return(htmltools::attachDependencies(x, deps))
  }

  if (!is.tag(x)) {
    # Nodes should be strings for proper hydration
    if (!is.null(x) && (!is.character(x) || is.na(x))) {
      x <- format(x)
    }
    return(x)
  }

  # Unnest tag lists for proper hydration
  x$children <- unnestTagList(x$children)
  # Preserve HTML dependencies
  deps <- htmltools::htmlDependencies(x$children)
  # Collect inline HTML dependencies
  inlineDeps <- sapply(x$children, is.htmlDependency)
  if (any(inlineDeps)) {
    deps <- c(deps, x$children[inlineDeps])
    x$children[inlineDeps] <- NULL
  }
  x <- htmltools::attachDependencies(x, deps, append = TRUE)
  # Filter null elements for proper hydration
  x$children <- filterNulls(x$children)
  x$children <- lapply(x$children, asReactTag)
  x <- normalizeTagAttributes(x)
  x$attribs <- asReactAttributes(x$attribs, x$name)
  x
}

# Recursively unnest tag lists
unnestTagList <- function(x) {
  isList <- function(x) {
    # Ignore HTML tags/widgets/dependencies, which are technically lists
    is.list(x) && !is.tag(x) && !is.htmlwidget(x) && !is.htmlDependency(x)
  }

  if (!isList(x)) {
    return(x)
  }

  # Preserve HTML dependencies of tag lists
  htmlDeps <- htmltools::htmlDependencies(x)

  tags <- Reduce(function(a, b) {
    if (is.null(b)) {
      return(a)
    }
    if (isTagList(b)) {
      # Preserve HTML dependencies of nested tag lists
      deps <- htmltools::htmlDependencies(b)
      if (!is.null(deps)) {
        htmlDeps <<- c(htmlDeps, deps)
      }
      b <- unnestTagList(b)
    }
    # Merge without losing attributes
    if (isList(b)) {
      c(a, b)
    } else {
      c(a, list(b))
    }
  }, x, list())

  htmltools::attachDependencies(tags, htmlDeps)
}

# Normalize tags that may have duplicate attributes, such as duplicate class attributes
# in HTML widgets with htmlwidgets >= 1.6.0.
normalizeTagAttributes <- function(tag) {
  attribNames <- names(tag$attribs)
  for (dupAttrib in attribNames[duplicated(attribNames)]) {
    combinedValue <- htmltools::tagGetAttribute(tag, dupAttrib)
    tag$attribs[attribNames == dupAttrib] <- NULL
    tag$attribs[[dupAttrib]] <- combinedValue
  }
  tag
}

# Transform HTML attributes to React DOM attributes.
# Not all attributes are supported at the moment - notable exceptions are
# some event handler attributes and `selected` attributes for <option> elements.
asReactAttributes <- function(attribs, tagName) {
  htmlAttribs <- list(
    autofocus = "autoFocus",
    autocomplete = "autoComplete",
    autoplay = "autoPlay",
    charset = "charSet",
    class = "className",
    colspan = "colSpan",
    crossorigin = "crossOrigin",
    enctype = "encType",
    "for" = "htmlFor",
    formaction = "formAction",
    formenctype = "formEncType",
    formmethod = "formMethod",
    formnovalidate = "formNoValidate",
    formtarget = "formTarget",
    frameborder = "frameBorder",
    hreflang = "hrefLang",
    "http-equiv" = "httpEquiv",
    inputmode = "inputMode",
    maxlength = "maxLength",
    mediagroup = "mediaGroup",
    minlength = "minLength",
    novalidate = "noValidate",
    radiogroup = "radioGroup",
    readonly = "readOnly",
    rowspan = "rowSpan",
    spellcheck = "spellCheck",
    srcdoc = "srcDoc",
    srclang = "srcLang",
    tabindex = "tabIndex",
    usemap = "useMap"
  )

  eventAttribs <- list(
    onblur = "onBlur",
    onchange = "onChange",
    onclick = "onClick",
    ondblclick = "onDoubleClick",
    onfocus = "onFocus",
    ongotpointercapture = "onGotPointerCapture",
    onlostpointercapture = "onLostPointerCapture",
    oninput = "onInput",
    onkeydown = "onKeyDown",
    onkeypress = "onKeyPress",
    onkeyup = "onKeyUp",
    onload = "onLoad",
    onmousedown = "onMouseDown",
    onmouseenter = "onMouseEnter",
    onmouseleave = "onMouseLeave",
    onmousemove = "onMouseMove",
    onmouseout = "onMouseOut",
    onmouseover = "onMouseOver",
    onmouseup = "onMouseUp",
    onpointercancel = "onPointerCancel",
    onpointerdown = "onPointerDown",
    onpointerenter = "onPointerEnter",
    onpointerleave = "onPointerLeave",
    onpointermove = "onPointerMove",
    onpointerout = "onPointerOut",
    onpointerover = "onPointerOver",
    onpointerup = "onPointerUp",
    onresize = "onResize",
    onselect = "onSelect",
    ontouchcancel = "onTouchCancel",
    ontouchend = "onTouchEnd",
    ontouchmove = "onTouchMove",
    ontouchstart = "onTouchStart"
  )

  for (name in names(attribs)) {
    # Map HTML attributes to React attributes. Not required as React still accepts
    # the standard attribute names.
    if (!is.null(htmlAttribs[[name]])) {
      attribs[[htmlAttribs[[name]]]] <- attribs[[name]]
      attribs[[name]] <- NULL
    }

    # Transform inline event attributes, ensuring `this` and `event` are in scope.
    if (!is.null(eventAttribs[[name]])) {
      attribs[[eventAttribs[[name]]]] <- JS(sprintf(
        "function(_e){(function(event){%s}).apply(event.target,[_e])}",
        attribs[[name]]
      ))
      attribs[[name]] <- NULL
    }
  }

  # Transform form element attributes to their uncontrolled equivalents, since
  # controlled attributes don't make sense outside of React.
  if (tagName %in% c("input", "select", "textarea")) {
    value <- attribs[["value"]]
    if (!is.null(value)) {
      attribs[["defaultValue"]] <- value
      attribs[["value"]] <- NULL
    }
  }

  if (tagName == "input") {
    checked <- attribs[["checked"]]
    if (!is.null(checked)) {
      checked <- if (is.na(checked)) TRUE else checked
      attribs[["defaultChecked"]] <- checked
      attribs[["checked"]] <- NULL
    }
  }

  style <- attribs[["style"]]
  if (!is.null(style) && is.character(style)) {
    attribs[["style"]] <- asReactStyle(style)
  }

  attribs
}

asReactStyle <- function(style) {
  if (!is.character(style)) {
    return(style)
  }
  pairs <- strsplit(unlist(strsplit(style, ";")), ":")
  if (length(pairs) > 0) {
    pairs <- Reduce(function(props, pair) {
      if (length(pair) == 2) {
        name <- trimws(pair[[1]])
        value <- trimws(pair[[2]])
        props[[name]] <- value
      }
      props
    }, pairs, list())
  }
  pairs
}

# Backport for R 3.1
trimws <- function(x) {
  x <- sub("[ \t\r\n]+$", "", x, perl = TRUE)
  sub("^[ \t\r\n]+", "", x, perl = TRUE)
}

# Call function with a variable number of args.
# Args default to NULL if not supplied.
callFunc <- function(func, ...) {
  args <- list(...)
  numArgs <- length(formals(func))
  do.call(func, args[seq_len(numArgs)])
}

library(htmltools)

test_that("mergeLists", {
  a <- list(a = 1, b = "b", c = 3)
  b <- list(a = 2, c = 4, d = "d")
  expect_equal(mergeLists(a, b), list(a = 2, b = "b", c = 4, d = "d"))

  a <- list(a = 1, b = 2)
  b <- list()
  expect_equal(mergeLists(a, b), list(a = 1, b = 2))

  a <- list()
  b <- list(a = 1, b = 2)
  expect_equal(mergeLists(a, b), list(a = 1, b = 2))

  a <- list(a = NULL, b = 2)
  b <- list(a = 1, b = NULL)
  expect_equal(mergeLists(a, b), list(a = 1, b = 2))

  expect_equal(mergeLists(NULL, list(a = 1, b = 2)), list(a = 1, b = 2))
  expect_equal(mergeLists(list(a = 1, b = 2), NULL), list(a = 1, b = 2))

  a <- list(a = NULL, b = 2, 3)
  b <- list(a = 1, b = NULL, 4)
  expect_equal(mergeLists(a, b), list(a = 1, b = 2, 3, 4))

  a <- list(a = NULL, b = 2)
  b <- list(1, 2, 3)
  expect_equal(mergeLists(a, b), list(a = NULL, b = 2, 1, 2, 3))
})

test_that("filterNulls", {
  expect_equal(filterNulls(list(a = 1, b = NULL, c = NULL, d = 2)), list(a = 1, d = 2))
  expect_equal(filterNulls(list(a = 1, b = "b")), list(a = 1, b = "b"))
  expect_equal(filterNulls(list(a = 1, 2, b = NULL)), list(a = 1, 2))
  expect_equal(filterNulls(list(1, NULL, 2)), list(1, 2))
})

test_that("isNamedList", {
  expect_true(isNamedList(list()))
  expect_true(isNamedList(list(a = 1, b = 2)))
  expect_false(isNamedList(list(1)))
  expect_false(isNamedList(list(1, a = 2)))
  expect_false(isNamedList(NULL))
  expect_false(isNamedList("a"))
})

test_that("is.tag", {
  expect_true(is.tag(tags$div()))
  expect_false(is.tag(list()))
})

test_that("is.htmlwidget", {
  expect_true(is.htmlwidget(reactable(data.frame(x = 1))))
  expect_false(is.htmlwidget(div()))
})

test_that("is.htmlDependency", {
  dep <- htmlDependency("dep", "0.1.0", "/path/to/dep")
  expect_true(is.htmlDependency(dep))
  expect_false(is.htmlDependency(div()))
})

test_that("isTagList", {
  expect_true(isTagList(tagList()))
  expect_true(isTagList(tagList("a")))
  expect_true(isTagList(tagList(1, div())))
  expect_true(isTagList(list(div(), span())))
  expect_false(isTagList(div()))
  expect_false(isTagList(list(div(), list())))
})

test_that("asReactTag", {
  # Nodes should be strings
  expect_equal(asReactTag("text"), "text")
  expect_equal(asReactTag("\u2718"), "\u2718")
  expect_equal(asReactTag(123), "123")
  expect_equal(asReactTag(TRUE), "TRUE")
  expect_equal(asReactTag(NA), "NA")  # should be "NA" rather than NA_character_
  expect_equal(asReactTag(NA_character_), "NA")  # should be "NA" rather than NA_character_
  expect_equal(asReactTag(factor("xy")), "xy")
  expect_equal(asReactTag(as.Date("2019-01-03")), "2019-01-03")
  expect_equal(asReactTag(list("text")), "text")
  # NULLs should be left as-is
  expect_equal(asReactTag(NULL), NULL)

  # Tags should be extracted from nested tables
  tag <- asReactTag(reactable(data.frame(x = 1)))
  expect_true(is.tag(tag))
  # Nested tables should be marked
  expect_true(tag$attribs$nested)

  # All other htmlwidgets should be converted to tags
  tbl <- reactable(data.frame(x = 1))
  class(tbl) <- c("my-widget", "htmlwidget")
  tag <- asReactTag(tbl)
  expect_equal(tag$name, "WidgetContainer")
  expect_equal(tag$attribs, list(key = digest::digest(tbl)))
  expect_equal(findDependencies(tag), findDependencies(tbl))
  expect_equal(length(tag$children), 1)
  expect_equal(tag$children[[1]]$name, "Fragment")

  # Tag lists should be unnested and wrapped in fragments
  expect_equal(asReactTag(tagList()), reactR::React$Fragment())
  expect_equal(asReactTag(tagList(div("x"))), reactR::React$Fragment(div("x")))
  expect_equal(asReactTag(tagList(div(), "x")), reactR::React$Fragment(div(), "x"))
  # htmlwidgets in tag lists
  tag <- asReactTag(tagList(reactable(data.frame(x = 1)), "y"))
  expect_equal(length(tag$children), 2)
  expect_true(is.tag(tag$children[[1]]))
  expect_equal(tag$children[[2]], "y")

  # Nested tags should be unnested
  nestedTag <- div(
    list(
      div(),
      div(list(div()))
    )
  )
  expected <- div(
    div(),
    div(div())
  )
  expect_equal(asReactTag(nestedTag), expected)

  nestedTag <- div(
    tagList("a", div(
      tagList("b", span("c", class = "c"))
    ))
  )
  expected <- div("a", div("b", span("c", className = "c")))
  expect_equal(asReactTag(nestedTag), expected)

  nestedTagList <- tagList(
    div(class = "a"),
    tagList(
      div(),
      tagList("x", span("y", class = "y"))
    )
  )
  expected <- reactR::React$Fragment(
    div(className = "a"),
    div(),
    "x",
    span("y", className = "y")
  )
  expect_equal(asReactTag(nestedTagList), expected)

  # Null elements should be pruned
  expect_equal(asReactTag(div(1, NULL, 3)), div("1", "3"))
  expect_equal(asReactTag(tagList(NULL, "a", tagList(NULL, "b", NULL), div(NULL, "c"))),
               reactR::React$Fragment("a", "b", div("c")))

  # Attributes should be converted
  expect_equal(asReactTag(div(style = "color: red", class = "cls")),
               div(style = list(color = "red"), className = "cls"))

  # Attributes should be preserved
  expect_equal(asReactTag(div(factor("xy"))), div("xy"))
  expect_equal(asReactTag(div(div(as.Date("2019-01-03")))), div(div("2019-01-03")))

  # Duplicate attributes should be included and collapsed (e.g., for likelihood of
  # duplicate class attributes in HTML widgets with htmlwidgets >= 1.6.0)
  expect_equal(asReactTag(div(class = "a", class = "b")), div(className = "a b"))
  expect_equal(
    asReactTag(span(class = "a", test = "t", style = list(color = "red"), class = "bb", style = list(color = "blue"))),
    span(test = "t", style = list(list(color = "red"), list(color = "blue")), className = "a bb")
  )
})

test_that("asReactTag preserves HTML dependencies", {
  dep <- htmlDependency("dep", "0.1.0", "/path/to/dep")
  dep2 <- htmlDependency("dep2", "0.5.0", "/path/to/dep2")

  # Single tag
  tag <- attachDependencies(div(div("x")), dep)
  expect_equal(htmlDependencies(asReactTag(tag)), list(dep))

  # Tag w/ nested deps
  tag <- div(attachDependencies(div("x"), dep))
  expect_equal(htmlDependencies(asReactTag(tag)$children[[1]]), list(dep))

  # Multiple nested deps
  tag <- div(attachDependencies(div("x"), dep2), attachDependencies(div("x"), dep))
  expect_equal(findDependencies(asReactTag(tag)), list(dep2, dep))

  # Tag list
  tag <- attachDependencies(tagList(div("x")), dep)
  expect_equal(htmlDependencies(asReactTag(tag)), list(dep))

  # Tag list w/ nested tag deps
  tag <- attachDependencies(tagList(div("x"), attachDependencies(div("y"), dep)), dep2)
  expect_equal(findDependencies(asReactTag(tag)), list(dep, dep2))

  # Tag list w/ nested tag list deps
  tag <- attachDependencies(tagList(div("x"), attachDependencies(tagList("y"), dep)), dep2)
  expect_equal(findDependencies(asReactTag(tag)), list(dep2, dep))

  # Tag w/ nested tag list deps
  tag <- div(attachDependencies(tagList(div("x")), dep), div("y"))
  expect_equal(findDependencies(asReactTag(tag)), list(dep))

  # HTML dependency objects
  tag <- tagList("x", "y", dep)
  expect_equal(asReactTag(tag), attachDependencies(reactR::React$Fragment("x", "y"), dep))
  tag <- div("x", div(), dep, dep2, "z")
  expect_equal(asReactTag(tag), attachDependencies(div("x", div(), "z"), list(dep, dep2)))

  # Nested HTML dependency objects
  tag <- tagList("x", div(dep), span("y"))
  expect_equal(asReactTag(tag), reactR::React$Fragment("x", attachDependencies(div(), dep), span("y")))
  tag <- div("x", tagList(dep), span("y"))
  expect_equal(asReactTag(tag), attachDependencies(div("x", span("y")), dep))

  # HTML dependencies in nested tables
  tbl <- reactable(
    data.frame(x = 1),
    columns = list(x = colDef(cell = function() tagList(dep, dep2)))
  )
  tag <- asReactTag(tbl)
  expect_equal(htmlDependencies(tag), list(dep, dep2))
})

test_that("asReactAttributes", {
  attribs <- list(class = "cls", "for" = "id", tabindex = 1)
  expected <- list(className = "cls", htmlFor = "id", tabIndex = 1)
  expect_equal(asReactAttributes(attribs, "th"), expected)

  attribs <- list(value = "x")
  expect_equal(asReactAttributes(attribs, "input"), list(defaultValue = "x"))
  expect_equal(asReactAttributes(attribs, "select"), list(defaultValue = "x"))
  expect_equal(asReactAttributes(attribs, "textarea"), list(defaultValue = "x"))
  expect_equal(asReactAttributes(attribs, "option"), list(value = "x"))
  expect_equal(asReactAttributes(attribs, "button"), list(value = "x"))

  attribs <- list(checked = NA)
  expect_equal(asReactAttributes(attribs, "input"), list(defaultChecked = TRUE))
  expect_equal(asReactAttributes(attribs, "div"), list(checked = NA))

  attribs <- list(onchange = "onChange(this, event)", onclick = "console.log(this, event);")
  expect_equal(
    asReactAttributes(attribs, "select"),
    list(
      onChange = JS("function(_e){(function(event){onChange(this, event)}).apply(event.target,[_e])}"),
      onClick = JS("function(_e){(function(event){console.log(this, event);}).apply(event.target,[_e])}")
    )
  )

  attribs <- list(style = "border: none; color: red; text-align: left")
  expected <- list(style = list(border = "none", color = "red", "text-align" = "left"))
  expect_equal(asReactAttributes(attribs, "div"), expected)

  attribs <- list(style = list(border = "none"))
  expected <- list(style = list(border = "none"))
  expect_equal(asReactAttributes(attribs, "div"), expected)

  # Non-converted attributes
  expect_equal(asReactAttributes(list("data-attr" = "t"), "div"), list("data-attr" = "t"))
  expect_equal(asReactAttributes(list("aria-label" = "lab"), "div"), list("aria-label" = "lab"))
})

test_that("asReactStyle", {
  expect_equal(asReactStyle("color: red"), list(color = "red"))
  expect_equal(asReactStyle("color: red;"), list(color = "red"))
  expect_equal(asReactStyle("  color: red; margin-bottom:55px ;"),
               list(color = "red", "margin-bottom" = "55px"))
  expect_equal(asReactStyle("  color: red ;; margin-bott"),
               list(color = "red"))
  expect_equal(asReactStyle("color"), list())
  expect_equal(asReactStyle(list(height = 0)), list(height = 0))
})

test_that("trimws", {
  expect_equal(trimws(" "), "")
  expect_equal(trimws("xvz "), "xvz")
  expect_equal(trimws("abd "), "abd")
  expect_equal(trimws("   xvz "), "xvz")
})

test_that("callFunc", {
  expect_equal(callFunc(function(x) x, 5), 5)
  expect_equal(callFunc(function(x) x, 5, "a", "b"), 5)
  expect_equal(callFunc(function(x, y) x + y, 5, 1), 6)
  expect_equal(callFunc(function(x, y) x + y, 5, 1), 6)
  expect_equal(callFunc(function(x) x), NULL)
  expect_equal(callFunc(function(x, y) y, "x"), NULL)
})

library(dplyr)
library(jsonlite)
library(spotifyr)

# Get global top 50 playlist tracks
global_top50 <- get_playlist_tracks("37i9dQZEVXbMDoHDwVN2tF")

# Get daily plays for the global daily top tracks
daily_plays <- read.csv("https://spotifycharts.com/regional/global/daily/latest/download",
                        skip = 1, encoding = "UTF-8", stringsAsFactors = FALSE)

# Make up some trends, since they're not provided by either data source
set.seed(20)
trend <- sample(c("unchanged", "up", "down", "new"), nrow(global_top50), replace = TRUE,
                prob = c(0.2, 0.35, 0.35, 0.1))

top_tracks <- global_top50 %>%
  left_join(daily_plays, by = c("track.external_urls.spotify" = "URL")) %>%
  mutate(trend = trend) %>%
  select(position = Position, trend = trend, name = track.name, artists = track.artists,
         album = track.album.name, explicit = track.explicit, daily_plays = Streams,
         duration_ms = track.duration_ms, url = track.external_urls.spotify,
         album_images = track.album.images)

# This is a nested data structure, so save as JSON
write_json(top_tracks, "top_tracks.json")

#' Theme options
#'
#' @description
#' Use `reactableTheme()` to customize the default styling of a table.
#' You can set theme variables to change the default styles, or add custom CSS
#' to specific elements of the table.
#'
#' The `color` variables are specified as character strings of CSS color values.
#' The `width` and `padding` variables are specified as either character strings
#' of CSS width and padding values, or numeric pixel values. The `style` arguments
#' take custom CSS as named lists of camelCased properties.
#'
#' To set the default theme for all tables, use the global `reactable.theme` option.
#'
#' @details
#' You can use nested CSS selectors in `style` arguments to target
#' the current element, using `&` as the selector, or other child elements
#' (just like in Sass). This is useful for adding pseudo-classes like `&:hover`,
#' or adding styles in a certain context like `.outer-container &`.
#'
#' @param color Default text color.
#' @param backgroundColor Default background color.
#' @param borderColor Default border color.
#' @param borderWidth Default border width.
#' @param stripedColor Default row stripe color.
#' @param highlightColor Default row highlight color.
#' @param cellPadding Default cell padding.
#'
#' @param style Additional CSS for the table.
#'
#' @param tableStyle Additional CSS for the table element (excludes
#'   the pagination bar and search input).
#'
#' @param headerStyle Additional CSS for header cells.
#'
#' @param groupHeaderStyle Additional CSS for group header cells.
#'
#' @param tableBodyStyle Additional CSS for the table body element.
#'
#' @param rowGroupStyle Additional CSS for row groups.
#'
#' @param rowStyle Additional CSS for rows.
#' @param rowStripedStyle Additional CSS for striped rows.
#' @param rowHighlightStyle Additional CSS for highlighted rows.
#' @param rowSelectedStyle Additional CSS for selected rows.
#'
#' @param cellStyle Additional CSS for cells.
#'
#' @param footerStyle Additional CSS for footer cells.
#'
#' @param inputStyle Additional CSS for inputs.
#' @param filterInputStyle Additional CSS for filter inputs.
#' @param searchInputStyle Additional CSS for the search input.
#'
#' @param selectStyle Additional CSS for table select controls.
#'
#' @param paginationStyle Additional CSS for the pagination bar.
#' @param pageButtonStyle,pageButtonHoverStyle,pageButtonActiveStyle,pageButtonCurrentStyle
#'   Additional CSS for page buttons, page buttons with hover or active
#'   states, and the current page button.
#'
#' @return A theme options object that can be used to customize the default
#'   styling in `reactable()`.
#'
#' @examples
#' reactable(
#'   iris[1:30, ],
#'   searchable = TRUE,
#'   striped = TRUE,
#'   highlight = TRUE,
#'   bordered = TRUE,
#'   theme = reactableTheme(
#'     borderColor = "#dfe2e5",
#'     stripedColor = "#f6f8fa",
#'     highlightColor = "#f0f5f9",
#'     cellPadding = "8px 12px",
#'     style = list(
#'       fontFamily = "-apple-system, BlinkMacSystemFont, Segoe UI, Helvetica, Arial, sans-serif"
#'     ),
#'     searchInputStyle = list(width = "100%")
#'   )
#' )
#'
#' # Set the default theme for all tables
#' options(reactable.theme = reactableTheme(
#'   color = "hsl(233, 9%, 87%)",
#'   backgroundColor = "hsl(233, 9%, 19%)",
#'   borderColor = "hsl(233, 9%, 22%)",
#'   stripedColor = "hsl(233, 12%, 22%)",
#'   highlightColor = "hsl(233, 12%, 24%)",
#'   inputStyle = list(backgroundColor = "hsl(233, 9%, 25%)"),
#'   selectStyle = list(backgroundColor = "hsl(233, 9%, 25%)"),
#'   pageButtonHoverStyle = list(backgroundColor = "hsl(233, 9%, 25%)"),
#'   pageButtonActiveStyle = list(backgroundColor = "hsl(233, 9%, 28%)")
#' ))
#'
#' reactable(
#'   iris[1:30, ],
#'   filterable = TRUE,
#'   showPageSizeOptions = TRUE,
#'   striped = TRUE,
#'   highlight = TRUE,
#'   details = function(index) paste("Details for row", index)
#' )
#'
#' # Use nested selectors to highlight headers when sorting
#' reactable(
#'   iris[1:30, ],
#'   columns = list(Sepal.Length = colDef(sortable = FALSE)),
#'   showSortable = TRUE,
#'   theme = reactableTheme(
#'     headerStyle = list(
#'       "&:hover[aria-sort]" = list(background = "hsl(0, 0%, 96%)"),
#'       "&[aria-sort='ascending'], &[aria-sort='descending']" = list(background = "hsl(0, 0%, 96%)"),
#'       borderColor = "#555"
#'     )
#'   )
#' )
#'
#' @export
reactableTheme <- function(
  color = NULL,
  backgroundColor = NULL,
  borderColor = NULL,
  borderWidth = NULL,
  stripedColor = NULL,
  highlightColor = NULL,
  cellPadding = NULL,

  style = NULL,

  tableStyle = NULL,

  headerStyle = NULL,

  groupHeaderStyle = NULL,

  tableBodyStyle = NULL,

  rowGroupStyle = NULL,

  rowStyle = NULL,
  rowStripedStyle = NULL,
  rowHighlightStyle = NULL,
  rowSelectedStyle = NULL,

  cellStyle = NULL,

  footerStyle = NULL,

  inputStyle = NULL,
  filterInputStyle = NULL,
  searchInputStyle = NULL,

  selectStyle = NULL,

  paginationStyle = NULL,
  pageButtonStyle = NULL,
  pageButtonHoverStyle = NULL,
  pageButtonActiveStyle = NULL,
  pageButtonCurrentStyle = NULL
) {

  args <- formals()

  for (arg in names(args)) {
    value <- get(arg)
    if (is.null(value)) {
      next
    }

    if (grepl("style$", arg, ignore.case = TRUE)) {
      if (!isNamedList(value) && !is.character(value)) {
        stop(sprintf("`%s` must be a named list", arg))
      }
    } else if (grepl("width$|padding$", arg, ignore.case = TRUE)) {
      if (!is.character(value) && !is.numeric(value)) {
        stop(sprintf("`%s` must be a character string or number", arg))
      }
    } else if (!is.character(value)) {
      stop(sprintf("`%s` must be a character string", arg))
    }

    args[[arg]] <- value
  }

  theme <- filterNulls(args)
  structure(theme, class = "reactableTheme")
}

is.reactableTheme <- function(x) {
  inherits(x, "reactableTheme")
}

test_that("reactableTheme", {
  expect_equal(reactableTheme(), structure(list(), .Names = character(0), class = "reactableTheme"))
  expect_equal(
    reactableTheme(cellPadding = paste0("3", "px"), style = list(color = "#fff")),
    structure(list(cellPadding = "3px", style = list(color = "#fff")), class = "reactableTheme")
  )

  argTypes <- list(
    color = "string",
    backgroundColor = "string",
    borderColor = "string",
    borderWidth = "number",
    stripedColor = "string",
    highlightColor = "string",
    cellPadding = "number",

    style = "style",

    tableStyle = "style",

    headerStyle = "style",

    groupHeaderStyle = "style",

    tableBodyStyle = "style",

    rowGroupStyle = "style",

    rowStyle = "style",
    rowStripedStyle = "style",
    rowHighlightStyle = "style",
    rowSelectedStyle = "style",

    cellStyle = "style",

    footerStyle = "style",

    inputStyle = "style",
    filterInputStyle = "style",
    searchInputStyle = "style",

    selectStyle = "style",

    paginationStyle = "style",
    pageButtonStyle = "style",
    pageButtonHoverStyle = "style",
    pageButtonActiveStyle = "style",
    pageButtonCurrentStyle = "style"
  )

  makeTheme <- function(arg, value) {
    do.call(reactableTheme, stats::setNames(list(value), arg))
  }

  for (arg in names(argTypes)) {
    if (argTypes[[arg]] == "style") {
      expect_equal(makeTheme(arg, list(color = "red"))[[arg]], list(color = "red"))
      expect_equal(makeTheme(arg, list(color = "red", "&:hover" = list(padding = 1)))[[arg]],
                   list(color = "red", "&:hover" = list(padding = 1)))
      expect_error(makeTheme(arg, 123), sprintf("`%s` must be a named list", arg))
    } else if (argTypes[[arg]] == "number") {
      expect_equal(makeTheme(arg, 34)[[arg]], 34)
      expect_equal(makeTheme(arg, "22px")[[arg]], "22px")
      expect_error(makeTheme(arg, NA), sprintf("`%s` must be a character string or number", arg))
    } else {
      expect_equal(makeTheme(arg, "#fff")[[arg]], "#fff")
      expect_error(makeTheme(arg, 234), sprintf("`%s` must be a character string", arg))
    }
  }
})

test_that("is.reactableTheme", {
  expect_true(is.reactableTheme(reactableTheme()))
  expect_false(is.reactableTheme(list()))
})

#' Column definitions
#'
#' Use `colDef()` to customize the columns in a table.
#'
#' @param name Column header name.
#' @param aggregate Aggregate function to use when rows are grouped. The name
#'   of a built-in aggregate function or a custom [JS()] aggregate function.
#'   Built-in aggregate functions are: `"mean"`, `"sum"`, `"max"`, `"min"`,
#'   `"median"`, `"count"`, `"unique"`, and `"frequency"`.
#'
#'   To enable row grouping, use the `groupBy` argument in [reactable()].
#' @param sortable Enable sorting? Overrides the table option.
#' @param resizable Enable column resizing? Overrides the table option.
#' @param filterable Enable column filtering? Overrides the table option.
#' @param searchable Enable or disable global table searching for this column.
#'   By default, global searching applies to all visible columns. Set this to
#'   `FALSE` to exclude a visible column from searching, or `TRUE` to include a
#'   hidden column in searching.
#' @param filterMethod Custom filter method to use for column filtering.
#'   A [JS()] function that takes an array of row objects, the column ID,
#'   and the filter value as arguments, and returns the filtered array of
#'   row objects.
#' @param show Show the column?
#'
#'   If `FALSE`, this column will be excluded from global table searching by
#'   default. To include this hidden column in searching, set `searchable`
#'   to `TRUE` in [colDef()].
#' @param defaultSortOrder Default sort order. Either `"asc"` for ascending
#'   order or `"desc"` for descending order. Overrides the table option.
#' @param sortNALast Always sort missing values ([NA] or [NaN]) last?
#' @param format Column formatting options. A [colFormat()] object to
#'   format all cells, or a named list of [colFormat()] objects to format standard
#'   cells (`"cell"`) and aggregated cells (`"aggregated"`) separately.
#' @param cell Custom cell renderer. An R function that takes the cell value,
#'   row index, and column name as arguments, or a [JS()] function that takes a
#'   cell info object and table state object as arguments.
#' @param grouped Custom grouped cell renderer. A [JS()] function that takes a
#'   cell info object and table state object as arguments.
#' @param aggregated Custom aggregated cell renderer. A [JS()] function that takes
#'   a cell info object and table state object as arguments.
#' @param header Custom header renderer. An R function that takes the header value
#'   and column name as arguments, or a [JS()] function that takes a column
#'   object and table state object as arguments.
#' @param footer Footer content or render function. Render functions can be an
#'   R function that takes the column values and column name as arguments, or a
#'   [JS()] function that takes a column object and table state object as
#'   arguments.
#' @param details Additional content to display when expanding a row. An R function
#'   that takes the row index and column name as arguments, or a [JS()] function
#'   that takes a row info object and table state object as arguments.
#'   Cannot be used on a `groupBy` column.
#' @param filterInput Custom filter input or render function. Render functions can
#'   be an R function that takes the column values and column name as arguments,
#'   or a [JS()] function that takes a column object and table state object as
#'   arguments.
#' @param html Render content as HTML? Raw HTML strings are escaped by default.
#' @param na String to display for missing values (i.e. [NA] or [NaN]).
#'   By default, missing values are displayed as blank cells.
#' @param rowHeader Mark up cells in this column as row headers?
#'
#'  Set this to `TRUE` to help users navigate the table using assistive technologies.
#'  When cells are marked up as row headers, assistive technologies will read them
#'  aloud while navigating through cells in the table.
#'
#'  Cells in the row names column are automatically marked up as row headers.
#' @param minWidth Minimum width of the column in pixels. Defaults to 100.
#' @param maxWidth Maximum width of the column in pixels.
#' @param width Fixed width of the column in pixels. Overrides `minWidth` and `maxWidth`.
#' @param align Horizontal alignment of content in the column. One of
#'   `"left"`, `"right"`, `"center"`. By default, all numbers are right-aligned,
#'   while all other content is left-aligned.
#' @param vAlign Vertical alignment of content in data cells. One of `"top"`
#'   (the default), `"center"`, `"bottom"`.
#' @param headerVAlign Vertical alignment of content in header cells. One of
#'   `"top"` (the default), `"center"`, `"bottom"`.
#' @param sticky Make the column sticky when scrolling horizontally? Either
#'   `"left"` or `"right"` to make the column stick to the left or right side.
#'
#'   If a sticky column is in a column group, all columns in the group will
#'   automatically be made sticky, including the column group header.
#' @param class Additional CSS classes to apply to cells. Can also be an R function
#'   that takes the cell value, row index, and column name as arguments, or a [JS()]
#'   function that takes a row info object, column object, and table state object
#'   as arguments.
#'
#'   Note that R functions cannot apply classes to aggregated cells.
#' @param style Inline styles to apply to cells. A named list or character string.
#'   Can also be an R function that takes the cell value and row index as arguments,
#'   or a [JS()] function that takes a row info object, column object, and
#'   table state object as arguments.
#'
#'   Note that R functions cannot apply styles to aggregated cells.
#'   If `style` is a named list, property names should be camelCased.
#' @param headerClass Additional CSS classes to apply to the header.
#' @param headerStyle Inline styles to apply to the header. A named list or
#'   character string.
#'
#'   Note that if `headerStyle` is a named list, property names should be camelCased.
#' @param footerClass Additional CSS classes to apply to the footer.
#' @param footerStyle Inline styles to apply to the footer. A named list or
#'   character string.
#'
#'   Note that if `footerStyle` is a named list, property names should be camelCased.
#' @return A column definition object that can be used to customize columns
#'   in `reactable()`.
#'
#' @examples
#' reactable(
#'   iris,
#'   columns = list(
#'     Sepal.Length = colDef(name = "Sepal Length"),
#'     Sepal.Width = colDef(filterable = TRUE),
#'     Petal.Length = colDef(show = FALSE),
#'     Petal.Width = colDef(defaultSortOrder = "desc")
#'   )
#' )
#'
#' @export
colDef <- function(
  name = NULL,
  aggregate = NULL,
  sortable = NULL,
  resizable = NULL,
  filterable = NULL,
  searchable = NULL,
  filterMethod = NULL,
  show = TRUE,
  defaultSortOrder = NULL,
  sortNALast = FALSE,
  format = NULL,
  cell = NULL,
  grouped = NULL,
  aggregated = NULL,
  header = NULL,
  footer = NULL,
  details = NULL,
  filterInput = NULL,
  html = FALSE,
  na = "",
  rowHeader = FALSE,
  minWidth = 100,
  maxWidth = NULL,
  width = NULL,
  align = NULL,
  vAlign = NULL,
  headerVAlign = NULL,
  sticky = NULL,
  class = NULL,
  style = NULL,
  headerClass = NULL,
  headerStyle = NULL,
  footerClass = NULL,
  footerStyle = NULL
) {

  if (!is.null(name) && !is.character(name)) {
    stop("`name` must be a character string")
  }

  if (!is.null(aggregate)) {
    if (is.character(aggregate) && !is.JS(aggregate)) {
      aggregators <- c("mean", "sum", "max", "min", "median", "count",
                       "unique", "frequency")
      if (!aggregate %in% aggregators) {
        stop("`aggregate` must be a valid aggregate function")
      }
    } else if (!is.JS(aggregate)) {
      stop("`aggregate` must be a character string or JS function")
    }
  }

  if (!is.null(sortable) && !is.logical(sortable)) {
    stop("`sortable` must be TRUE or FALSE")
  }

  if (!is.null(resizable) && !is.logical(resizable)) {
    stop("`resizable` must be TRUE or FALSE")
  }

  if (!is.null(filterable) && !is.logical(filterable)) {
    stop("`filterable` must be TRUE or FALSE")
  }

  if (!is.null(searchable) && !is.logical(searchable)) {
    stop("`searchable` must be TRUE or FALSE")
  }

  if (!is.null(filterMethod) && !is.JS(filterMethod)) {
    stop('`filterMethod` must be a JS function')
  }

  if (!is.null(show) && !is.logical(show)) {
    stop("`show` must be TRUE or FALSE")
  }

  if (!is.null(defaultSortOrder) && !isSortOrder(defaultSortOrder)) {
    stop('`defaultSortOrder` must be "asc" or "desc"')
  }

  if (!is.null(sortNALast) && !is.logical(sortNALast)) {
    stop("`sortNALast` must be TRUE or FALSE")
  }

  if (!is.null(format)) {
    if (!is.colFormat(format) && !isNamedList(format)) {
      stop('`format` must be a column formatting option set or named list')
    }
    if (is.colFormat(format)) {
      format <- list(cell = format, aggregated = format)
    }
    if (any(!names(format) %in% c("cell", "aggregated"))) {
      stop('`format` must have names "cell" or "aggregated"')
    }
    for (opts in format) {
      if (!is.colFormat(opts)) {
        stop("`format` must be a list of column formatting options")
      }
    }
  }

  if (!is.null(cell) && !is.JS(cell) && !is.function(cell)) {
    stop("`cell` renderer must be an R function or JS function")
  }

  if (!is.null(grouped) && !is.JS(grouped)) {
    stop("`grouped` renderer must be a JS function")
  }

  if (!is.null(aggregated) && !is.JS(aggregated)) {
    stop("`aggregated` renderer must be a JS function")
  }

  if (!is.null(details) && !is.function(details) && !is.JS(details) && !is.list(details)) {
    stop("`details` renderer must be an R function or JS function")
  }

  if (!is.null(html) && !is.logical(html)) {
    stop("`html` must be TRUE or FALSE")
  }

  if (!is.null(na) && !is.character(na)) {
    stop("`na` must be a character string")
  }

  if (!is.null(rowHeader) && !is.logical(rowHeader)) {
    stop("`rowHeader` must be TRUE or FALSE")
  }

  if (!is.null(minWidth) && !is.numeric(minWidth)) {
    stop("`minWidth` must be numeric")
  }

  if (!is.null(maxWidth) && !is.numeric(maxWidth)) {
    stop("`maxWidth` must be numeric")
  }

  if (!is.null(width) && !is.numeric(width)) {
    stop("`width` must be numeric")
  }

  if (!is.null(align)) {
    if (!isTRUE(align %in% c("left", "right", "center"))) {
      stop('`align` must be one of "left", "right", "center"')
    }
  }

  if (!is.null(vAlign)) {
    if (!isTRUE(vAlign %in% c("top", "center", "bottom"))) {
      stop('`vAlign` must be one of "top", "center", "bottom"')
    }
  }

  if (!is.null(headerVAlign)) {
    if (!isTRUE(headerVAlign %in% c("top", "center", "bottom"))) {
      stop('`headerVAlign` must be one of "top", "center", "bottom"')
    }
  }

  if (!is.null(sticky)) {
    if (!isTRUE(sticky %in% c("left", "right"))) {
      stop('`sticky` must be "left" or "right"')
    }
  }

  if (!is.null(class) && !is.character(class) && !is.JS(class) && !is.function(class)) {
    stop("`class` must be a character string, JS function, or R function")
  }

  if (!is.null(style) && !isNamedList(style) && !is.character(style) &&
      !is.JS(style) && !is.function(style)) {
    stop("`style` must be a named list, character string, JS function, or R function")
  }

  if (!is.null(headerClass) && !is.character(headerClass)) {
    stop("`headerClass` must be a character string")
  }

  if (!is.null(headerStyle) && !isNamedList(headerStyle) && !is.character(headerStyle)) {
    stop("`headerStyle` must be a named list or character string")
  }

  if (!is.null(footerClass) && !is.character(footerClass)) {
    stop("`footerClass` must be a character string")
  }

  if (!is.null(footerStyle) && !isNamedList(footerStyle) && !is.character(footerStyle)) {
    stop("`footerStyle` must be a named list or character string")
  }

  # If an arg with a non-NULL default value wasn't specified by the user, filter
  # them out so they can take on the default from a default column definition.
  userArgs <- names(match.call())[-1]

  structure(
    filterNulls(list(
      name = name,
      aggregate = aggregate,
      sortable = sortable,
      resizable = resizable,
      filterable = filterable,
      searchable = searchable,
      filterMethod = filterMethod,
      show = if ("show" %in% userArgs) show,
      defaultSortDesc = if (!is.null(defaultSortOrder)) isDescOrder(defaultSortOrder),
      sortNALast = if ("sortNALast" %in% userArgs) sortNALast,
      format = format,
      cell = cell,
      grouped = grouped,
      aggregated = aggregated,
      header = header,
      footer = footer,
      details = details,
      filterInput = filterInput,
      html = if ("html" %in% userArgs) html,
      na = if ("na" %in% userArgs) na,
      rowHeader = if ("rowHeader" %in% userArgs) rowHeader,
      minWidth = if ("minWidth" %in% userArgs) minWidth,
      maxWidth = maxWidth,
      width = width,
      align = align,
      vAlign = vAlign,
      headerVAlign = headerVAlign,
      sticky = sticky,
      className = class,
      style = if (is.function(style) || is.JS(style)) style else asReactStyle(style),
      headerClassName = headerClass,
      headerStyle = asReactStyle(headerStyle),
      footerClassName = footerClass,
      footerStyle = asReactStyle(footerStyle)
    )),
    class = "colDef"
  )
}

is.colDef <- function(x) {
  inherits(x, "colDef")
}

isSortOrder <- function(x) {
  is.character(x) && x %in% c("asc", "desc")
}

isDescOrder <- function(x) {
  is.character(x) && x == "desc"
}

#' Column group definitions
#'
#' Use `colGroup()` to create column groups in a table.
#'
#' @param name Column group header name.
#' @param columns Character vector of column names in the group.
#' @param header Custom header renderer. An R function that takes the header value
#'   as an argument, or a [JS()] function that takes a column object and
#'   table state object as arguments.
#' @param html Render header content as HTML? Raw HTML strings are escaped by default.
#' @param align Horizontal alignment of content in the column group header. One of
#'   `"left"`, `"right"`, `"center"` (the default).
#' @param headerVAlign Vertical alignment of content in the column group header. One of
#'   `"top"` (the default), `"center"`, `"bottom"`.
#' @param sticky Make the column group sticky when scrolling horizontally? Either
#'   `"left"` or `"right"` to make the column group stick to the left or right side.
#'
#'   If a column group is sticky, all columns in the group will automatically
#'   be made sticky.
#' @param headerClass Additional CSS classes to apply to the header.
#' @param headerStyle Inline styles to apply to the header. A named list or
#'   character string.
#'
#'   Note that if `headerStyle` is a named list, property names should be camelCased.
#' @return A column group definition object that can be used to create column
#'   groups in `reactable()`.
#'
#' @examples
#' reactable(
#'   iris,
#'   columns = list(
#'     Sepal.Length = colDef(name = "Length"),
#'     Sepal.Width = colDef(name = "Width"),
#'     Petal.Length = colDef(name = "Length"),
#'     Petal.Width = colDef(name = "Width")
#'   ),
#'   columnGroups = list(
#'     colGroup(name = "Sepal", columns = c("Sepal.Length", "Sepal.Width")),
#'     colGroup(name = "Petal", columns = c("Petal.Length", "Petal.Width"))
#'   )
#' )
#'
#' @export
colGroup <- function(
  name = NULL,
  columns = NULL,
  header = NULL,
  html = FALSE,
  align = NULL,
  headerVAlign = NULL,
  sticky = NULL,
  headerClass = NULL,
  headerStyle = NULL
) {
  if (!is.null(name) && !is.character(name)) {
    stop("`name` must be a character string")
  }
  if (!is.null(columns)) {
    if (!is.character(columns)) {
      stop("`columns` must be a character vector")
    } else {
      # Ensure column IDs are serialized as an array
      columns <- as.list(columns)
    }
  }

  # If an arg with a non-NULL default value wasn't specified by the user, filter
  # them out so they can take on the default from a default column definition.
  userArgs <- names(match.call())[-1]

  args <- filterNulls(list(
    name = name,
    header = header,
    html = if ("html" %in% userArgs) html,
    align = align,
    headerVAlign = headerVAlign,
    sticky = sticky,
    headerClass = headerClass,
    headerStyle = headerStyle
  ))

  group <- tryCatch({
    do.call(colDef, args)
  }, error = function(e) e)

  if (inherits(group, "error")) {
    stop(group$message)
  }

  group$columns <- columns
  structure(group, class = "colGroup")
}

is.colGroup <- function(x) {
  inherits(x, "colGroup")
}

#' Column formatting options
#'
#' Use `colFormat()` to add data formatting to a column.
#'
#' @param prefix Prefix string.
#' @param suffix Suffix string.
#' @param digits Number of decimal digits to use for numbers.
#' @param separators Whether to use grouping separators for numbers, such as
#'   thousands separators or thousand/lakh/crore separators. The format is
#'   locale-dependent.
#' @param percent Format number as a percentage? The format is locale-dependent.
#' @param currency Currency format. An ISO 4217 currency code such as `"USD"`
#'   for the US dollar, `"EUR"` for the euro, or `"CNY"` for the Chinese RMB.
#'   The format is locale-dependent.
#' @param datetime Format as a locale-dependent date-time?
#' @param date Format as a locale-dependent date?
#' @param time Format as a locale-dependent time?
#' @param hour12 Whether to use 12-hour time (`TRUE`) or 24-hour time (`FALSE`).
#'   The default time convention is locale-dependent.
#' @param locales Locales to use for number, date, time, and currency formatting.
#'   A character vector of BCP 47 language tags, such as `"en-US"` for English
#'   (United States), `"hi"` for Hindi, or `"sv-SE"` for Swedish (Sweden).
#'   Defaults to the locale of the user's browser.
#'
#'   Multiple locales may be specified to provide a fallback language in case
#'   a locale is unsupported. When multiple locales are specified, the first
#'   supported locale will be used.
#'
#'   See a list of [common BCP 47 language tags](https://learn.microsoft.com/en-us/openspecs/office_standards/ms-oe376/6c085406-a698-4e12-9d4d-c3b0ee3dbc4a)
#'   for reference.
#' @return A column format object that can be used to customize data formatting
#'   in `colDef()`.
#'
#' @seealso Custom cell rendering in [colDef()] to customize data formatting
#'   beyond what the built-in formatters provide.
#'
#' @examples
#' data <- data.frame(
#'   price_USD = c(123456.56, 132, 5650.12),
#'   price_INR = c(350, 23208.552, 1773156.4),
#'   number_FR = c(123456.56, 132, 5650.12),
#'   temp = c(22, NA, 31),
#'   percent = c(0.9525556, 0.5, 0.112),
#'   date = as.Date(c("2019-01-02", "2019-03-15", "2019-09-22"))
#' )
#'
#' reactable(data, columns = list(
#'   price_USD = colDef(format = colFormat(prefix = "$", separators = TRUE, digits = 2)),
#'   price_INR = colDef(format = colFormat(currency = "INR", separators = TRUE, locales = "hi-IN")),
#'   number_FR = colDef(format = colFormat(locales = "fr-FR")),
#'   temp = colDef(format = colFormat(suffix = " \u00b0C")),
#'   percent = colDef(format = colFormat(percent = TRUE, digits = 1)),
#'   date = colDef(format = colFormat(date = TRUE, locales = "en-GB"))
#' ))
#'
#' # Date formatting
#' datetimes <- as.POSIXct(c("2019-01-02 3:22:15", "2019-03-15 09:15:55", "2019-09-22 14:20:00"))
#' data <- data.frame(
#'   datetime = datetimes,
#'   date = datetimes,
#'   time = datetimes,
#'   time_24h = datetimes,
#'   datetime_pt_BR = datetimes
#' )
#'
#' reactable(data, columns = list(
#'   datetime = colDef(format = colFormat(datetime = TRUE)),
#'   date = colDef(format = colFormat(date = TRUE)),
#'   time = colDef(format = colFormat(time = TRUE)),
#'   time_24h = colDef(format = colFormat(time = TRUE, hour12 = FALSE)),
#'   datetime_pt_BR = colDef(format = colFormat(datetime = TRUE, locales = "pt-BR"))
#' ))
#'
#' # Currency formatting
#' data <- data.frame(
#'   USD = c(12.12, 2141.213, 0.42, 1.55, 34414),
#'   EUR = c(10.68, 1884.27, 0.37, 1.36, 30284.32),
#'   INR = c(861.07, 152122.48, 29.84, 110, 2444942.63),
#'   JPY = c(1280, 226144, 44.36, 164, 3634634.61),
#'   MAD = c(115.78, 20453.94, 4.01, 15, 328739.73)
#' )
#'
#' reactable(data, columns = list(
#'   USD = colDef(
#'     format = colFormat(currency = "USD", separators = TRUE, locales = "en-US")
#'   ),
#'   EUR = colDef(
#'     format = colFormat(currency = "EUR", separators = TRUE, locales = "de-DE")
#'   ),
#'   INR = colDef(
#'     format = colFormat(currency = "INR", separators = TRUE, locales = "hi-IN")
#'   ),
#'   JPY = colDef(
#'     format = colFormat(currency = "JPY", separators = TRUE, locales = "ja-JP")
#'   ),
#'   MAD = colDef(
#'     format = colFormat(currency = "MAD", separators = TRUE, locales = "ar-MA")
#'   )
#' ))
#'
#' # Formatting aggregated cells
#' data <- data.frame(
#'   States = state.name,
#'   Region = state.region,
#'   Area = state.area
#' )
#'
#' reactable(
#'   data,
#'   groupBy = "Region",
#'   columns = list(
#'     States = colDef(
#'       aggregate = "count",
#'       format = list(
#'         aggregated = colFormat(suffix = " states")
#'       )
#'     ),
#'     Area = colDef(
#'       aggregate = "sum",
#'       format = colFormat(suffix = " mi\u00b2", separators = TRUE)
#'     )
#'   )
#' )
#'
#' @export
colFormat <- function(prefix = NULL, suffix = NULL, digits = NULL,
                      separators = FALSE, percent = FALSE, currency = NULL,
                      datetime = FALSE, date = FALSE, time = FALSE, hour12 = NULL,
                      locales = NULL) {

  if (!is.null(prefix) && !is.character(prefix)) {
    stop("`prefix` must be a character string")
  }
  if (!is.null(suffix) && !is.character(suffix)) {
    stop("`suffix` must be a character string")
  }
  if (!is.null(digits)) {
    if (!is.numeric(digits) || digits < 0 || digits > 18) {
      stop("`digits` must be a number between 0 and 18")
    }
    digits <- as.integer(digits)
  }
  if (!is.logical(separators)) {
    stop("`separators` must be TRUE or FALSE")
  }
  if (!is.logical(percent)) {
    stop("`percent` must be TRUE or FALSE")
  }
  if (!is.null(currency) && !is.character(currency)) {
    stop("`currency` must be a character string")
  }
  if (!is.logical(datetime)) {
    stop("`datetime` must be TRUE or FALSE")
  }
  if (!is.logical(date)) {
    stop("`date` must be TRUE or FALSE")
  }
  if (!is.logical(time)) {
    stop("`time` must be TRUE or FALSE")
  }
  if (!is.null(hour12) && !is.logical(hour12)) {
    stop("`hour12` must be TRUE or FALSE")
  }
  if (!is.null(locales) && !is.character(locales)) {
    stop("`locales` must be a character string")
  }

  options <- list(
    prefix = prefix,
    suffix = suffix,
    digits = digits,
    separators = if (separators) TRUE,
    percent = if (percent) TRUE,
    currency = currency,
    datetime = if (datetime) TRUE,
    date = if (date) TRUE,
    time = if (time) TRUE,
    hour12 = hour12,
    locales = locales
  )
  options <- filterNulls(options)
  structure(options, class = "colFormat")
}

is.colFormat <- function(x) {
  inherits(x, "colFormat")
}

colType <- function(x) {
  if (is.null(x)) {
    return(NULL)
  }
  if (is.numeric(x)) {
    return("numeric")
  }
  if (is.Date(x) || is.POSIXct(x) || is.POSIXlt(x)) {
    return("Date")
  }
  class(x)
}

test_that("colDef", {
  # Default args
  expect_equal(colDef(), structure(list(), .Names = character(0), class = "colDef"))

  # Valid args
  col <- colDef(name = "col", aggregate = "sum",
                sortable = TRUE, resizable = TRUE,
                defaultSortOrder = "desc",
                format = list(cell = colFormat(), aggregated = colFormat()),
                cell = JS("cellInfo => cellInfo.value"), aggregated = JS("cellInfo => cellInfo.value"),
                footer = "footer", details = function(i) i,
                maxWidth = 250, width = 125,
                align = "right", class = "cell", style = list(color = "a"),
                headerClass = "hdr", headerStyle = list(height = 10),
                footerClass = "ftr", footerStyle = "color:blue")

  expected <- structure(list(
    name = "col", aggregate = "sum",
    sortable = TRUE, resizable = TRUE,
    defaultSortDesc = TRUE,
    format = list(cell = colFormat(), aggregated = colFormat()),
    cell = JS("cellInfo => cellInfo.value"), aggregated = JS("cellInfo => cellInfo.value"),
    footer = "footer", details = function(i) i,
    maxWidth = 250, width = 125,
    align = "right", className = "cell", style = list(color = "a"),
    headerClassName = "hdr", headerStyle = list(height = 10),
    footerClassName = "ftr", footerStyle = list(color = "blue")
  ), class = "colDef")
  expect_equal(col, expected)

  # Invalid args
  invalidArgs <- list(
    name = list(1, FALSE),
    aggregate = list(2, TRUE, function() {}, "fn", "SUM"),
    sortable = list(1, "TRUE"),
    resizable = list(1, "TRUE"),
    defaultSortOrder = list(1, TRUE, "ascending"),
    format = list(23, list(CELL = colFormat()), list(aggregated = list())),
    cell = list("function() {}"),
    aggregated = list(function() {}),
    maxWidth = list("1", FALSE),
    align = list("a", "RIGHT", 1),
    width = list("1", FALSE),
    class = list(1, list()),
    style = list(list("a"), 2),
    headerClass = list(1, list()),
    headerStyle = list(list("a"), 2),
    footerClass = list(1, list()),
    footerStyle = list(list("a"), 2)
  )
  for (arg in names(invalidArgs)) {
    for (val in invalidArgs[[arg]]) {
      expect_error(do.call(colDef, stats::setNames(list(val), arg)))
    }
  }
})

test_that("colDef aggregate function", {
  aggregators <- list(
    JS("function(values, rows) { return 1 }"),
    "mean", "sum", "max", "min", "median", "count", "unique", "frequency"
  )
  for (func in aggregators) {
    col <- colDef(aggregate = func)
    expect_equal(col$aggregate, func)
  }
})

test_that("colDef filterable", {
  expect_error(colDef(filterable = ""), "`filterable` must be TRUE or FALSE")
  expect_equal(colDef()$filterable, NULL)
  expect_equal(colDef(filterable = FALSE)$filterable, FALSE)
  expect_equal(colDef(filterable = TRUE)$filterable, TRUE)
})

test_that("colDef searchable", {
  expect_error(colDef(searchable = ""), "`searchable` must be TRUE or FALSE")
  expect_equal(colDef()$searchable, NULL)
  expect_equal(colDef(searchable = FALSE)$searchable, FALSE)
  expect_equal(colDef(searchable = TRUE)$searchable, TRUE)
})

test_that("colDef filterMethod", {
  expect_error(colDef(filterMethod = "rows => rows"), "`filterMethod` must be a JS function")
  expect_equal(colDef()$filterMethod, NULL)
  expect_equal(
    colDef(filterMethod = JS("(rows, columnId, filterValue) => rows"))$filterMethod,
    JS("(rows, columnId, filterValue) => rows")
  )
})

test_that("colDef show", {
  expect_error(colDef(show = ""), "`show` must be TRUE or FALSE")
  expect_equal(colDef()$show, NULL)
  expect_equal(colDef(show = FALSE)$show, FALSE)
  expect_equal(colDef(show = TRUE)$show, TRUE)
})

test_that("colDef sortNALast", {
  expect_error(colDef(sortNALast = 5), "`sortNALast` must be TRUE or FALSE")
  expect_equal(colDef()$sortNALast, NULL)
  expect_equal(colDef(sortNALast = FALSE)$sortNALast, FALSE)
  expect_equal(colDef(sortNALast = TRUE)$sortNALast, TRUE)
})

test_that("colDef format", {
  format <- colFormat()

  # Default cell format
  col <- colDef(format = format)
  expect_equal(col$format, list(cell = format, aggregated = format))

  # Separate cell and aggregated format/renderer
  col <- colDef(format = list(aggregated = format))
  expect_equal(col$format, list(aggregated = format))
  col <- colDef(format = list(cell = format))
  expect_equal(col$format, list(cell = format))
})

test_that("colDef renderers", {
  # Cell renderer
  col <- colDef(cell = JS("cellInfo => cellInfo.value"))
  expect_equal(col$cell, JS("cellInfo => cellInfo.value"))
  col <- colDef(cell = function(value, index) value)
  expect_equal(col$cell, function(value, index) value)

  # Grouped cell renderer
  col <- colDef(grouped = JS("cellInfo => cellInfo.value"))
  expect_equal(col$grouped, JS("cellInfo => cellInfo.value"))
  expect_error(colDef(grouped = function() {}), "`grouped` renderer must be a JS function")

  # Aggregated cell renderer
  col <- colDef(aggregated = JS("cellInfo => cellInfo.value"))
  expect_equal(col$aggregated, JS("cellInfo => cellInfo.value"))

  # Header renderer
  col <- colDef(header = "header")
  expect_equal(col$header, "header")
  col <- colDef(header = JS("colInfo => colInfo.column.name"))
  expect_equal(col$header, JS("colInfo => colInfo.column.name"))
  col <- colDef(header = function(value) value)
  expect_equal(col$header, function(value) value)

  # Footer renderer
  col <- colDef(footer = "footer")
  expect_equal(col$footer, "footer")
  col <- colDef(footer = JS("colInfo => colInfo.column.name"))
  expect_equal(col$footer, JS("colInfo => colInfo.column.name"))
  col <- colDef(footer = function(values, name) name)
  expect_equal(col$footer, function(values, name) name)

  # Details renderer
  col <- colDef(details = JS("rowInfo => rowInfo.values['value']"))
  expect_equal(col$details, JS("rowInfo => rowInfo.values['value']"))
  col <- colDef(details = function(i, name) i)
  expect_equal(col$details, function(i, name) i)
  col <- colDef(details = list(1, 2, 3))
  expect_equal(col$details, list(1, 2, 3))
  expect_error(colDef(details = "function() {}"), "`details` renderer must be an R function or JS function")
})

test_that("colDef html", {
  expect_error(colDef(html = 5), "`html` must be TRUE or FALSE")
  expect_equal(colDef()$html, NULL)
  expect_equal(colDef(html = FALSE)$html, FALSE)
  expect_equal(colDef(html = TRUE)$html, TRUE)
})

test_that("colDef na", {
  expect_error(colDef(na = TRUE), "`na` must be a character string")
  expect_equal(colDef()$na, NULL)
  expect_equal(colDef(na = "NA")$na, "NA")
})

test_that("colDef rowHeader", {
  expect_error(colDef(rowHeader = 5), "`rowHeader` must be TRUE or FALSE")
  expect_equal(colDef()$rowHeader, NULL)
  expect_equal(colDef(rowHeader = FALSE)$rowHeader, FALSE)
  expect_equal(colDef(rowHeader = TRUE)$rowHeader, TRUE)
})

test_that("colDef minWidth", {
  expect_error(colDef(minWidth = FALSE), "`minWidth` must be numeric")
  expect_equal(colDef()$minWidth, NULL)
  expect_equal(colDef(minWidth = 100)$minWidth, 100)
})

test_that("colDef vAlign", {
  expect_error(colDef(vAlign = TRUE), '`vAlign` must be one of "top", "center", "bottom"')
  expect_null(colDef()$vAlign)
  expect_equal(colDef(vAlign = "top")$vAlign, "top")
  expect_equal(colDef(vAlign = "center")$vAlign, "center")
  expect_equal(colDef(vAlign = "bottom")$vAlign, "bottom")
})

test_that("colDef headerVAlign", {
  expect_error(colDef(headerVAlign = TRUE), '`headerVAlign` must be one of "top", "center", "bottom"')
  expect_null(colDef()$headerVAlign)
  expect_equal(colDef(headerVAlign = "top")$headerVAlign, "top")
  expect_equal(colDef(headerVAlign = "center")$headerVAlign, "center")
  expect_equal(colDef(headerVAlign = "bottom")$headerVAlign, "bottom")
})

test_that("colDef sticky", {
  expect_error(colDef(sticky = TRUE), '`sticky` must be "left" or "right"')
  expect_equal(colDef()$sticky, NULL)
  expect_equal(colDef(sticky = "left")$sticky, "left")
  expect_equal(colDef(sticky = "right")$sticky, "right")
})

test_that("colDef class", {
  col <- colDef(class = "cell")
  expect_equal(col$className, "cell")

  col <- colDef(class = JS("rowInfo => 'cell'"))
  expect_equal(col$className, JS("rowInfo => 'cell'"))

  col <- colDef(class = function(value, index) "cell")
  expect_equal(col$className, function(value, index) "cell")
})

test_that("colDef style", {
  col <- colDef(style = " border-bottom: 1px solid; top: 50px",
                headerStyle = " border: 1px solid; top: 25px;;df",
                footerStyle = " border: 1px solid; top: 25px;;df")
  expect_equal(col$style, list("border-bottom" = "1px solid", top = "50px"))
  expect_equal(col$headerStyle, list("border" = "1px solid", top = "25px"))
  expect_equal(col$footerStyle, list("border" = "1px solid", top = "25px"))

  col <- colDef(style = JS("rowInfo => ({ backgroundColor: 'red' })"))
  expect_equal(col$style, JS("rowInfo => ({ backgroundColor: 'red' })"))

  col <- colDef(style = function(value, index) "background-color: red")
  expect_equal(col$style, function(value, index) "background-color: red")
})

test_that("is.colDef", {
  expect_true(is.colDef(colDef()))
  expect_false(is.colDef(list()))
})

test_that("sort order", {
  expect_true(isSortOrder("asc"))
  expect_true(isSortOrder("desc"))
  expect_false(isSortOrder(FALSE))
  expect_false(isSortOrder(list("asc")))
  expect_true(isDescOrder("desc"))
  expect_false(isDescOrder("asc"))
})

test_that("colGroup", {
  # Default args
  expect_equal(colGroup(), structure(list(), .Names = character(0), class = "colGroup"))

  # Valid args
  group <- colGroup(name = "name", columns = c("colA", "colB"), html = TRUE, align = "right",
                    headerStyle = list(color = "red"), headerClass = "cls")
  expect_equal(group, structure(list(
    name = "name",
    html = TRUE,
    align = "right",
    headerClassName = "cls",
    headerStyle = list(color = "red"),
    columns = list("colA", "colB")
  ), class = "colGroup"))

  # Header renderer
  group <- colGroup(header = "header")
  expect_equal(group$header, "header")
  group <- colGroup(header = JS("colInfo => colInfo.column.name"))
  expect_equal(group$header, JS("colInfo => colInfo.column.name"))
  group <- colGroup(header = function(value) value)
  expect_equal(group$header, function(value) value)

  # html
  expect_error(colGroup(html = 5), "`html` must be TRUE or FALSE")
  expect_equal(colGroup()$html, NULL)
  expect_equal(colGroup(html = FALSE)$html, FALSE)
  expect_equal(colGroup(html = TRUE)$html, TRUE)

  # headerVAlign
  expect_error(colGroup(headerVAlign = TRUE), '`headerVAlign` must be one of "top", "center", "bottom"')
  expect_null(colGroup()$headerVAlign)
  expect_equal(colGroup(headerVAlign = "top")$headerVAlign, "top")
  expect_equal(colGroup(headerVAlign = "center")$headerVAlign, "center")
  expect_equal(colGroup(headerVAlign = "bottom")$headerVAlign, "bottom")

  # sticky
  expect_error(colGroup(sticky = TRUE), '`sticky` must be "left" or "right"')
  group <- colGroup(sticky = "left")
  expect_equal(group$sticky, "left")
  group <- colDef(sticky = "right")
  expect_equal(group$sticky, "right")

  # headerStyle
  group <- colGroup("grp", "col", headerStyle = " border: 1px solid; top: 25px;;df")
  expect_equal(group$headerStyle, list("border" = "1px solid", top = "25px"))

  # Invalid args
  expect_error(colGroup(name = 1, columns = "col"))
  expect_error(colGroup(name = "name", columns = list("col")))
  expect_error(colGroup(name = "name", columns = 123))
  expect_error(colGroup(name = "name", columns = "col", align = "CENTER"))
})

test_that("is.colGroup", {
  expect_true(is.colGroup(colGroup("name", "col")))
  expect_false(is.colGroup(list()))
})

test_that("colFormat", {
  # Default args
  expect_equal(colFormat(), structure(list(), .Names = character(0), class = "colFormat"))

  # Valid args
  options <- colFormat(prefix = "a", suffix = "b", digits = 5, separators = TRUE,
                       percent = TRUE, currency = "INR", datetime = TRUE,
                       date = TRUE, time = TRUE, hour12 = FALSE,
                       locales = c("en-US", "zh-Hans-CN"))
  expect_equal(options, structure(list(
    prefix = "a", suffix = "b", digits = 5, separators = TRUE, percent = TRUE,
    currency = "INR", datetime = TRUE, date = TRUE, time = TRUE, hour12 = FALSE,
    locales = c("en-US", "zh-Hans-CN")
  ), class = "colFormat"))

  # Invalid args
  expect_error(colFormat(prefix = 1))
  expect_error(colFormat(suffix = 1))
  expect_error(colFormat(digits = -1))
  expect_error(colFormat(digits = 19))
  expect_error(colFormat(separators = "sad"))
  expect_error(colFormat(percent = "true"))
  expect_error(colFormat(currency = FALSE))
  expect_error(colFormat(datetime = "false"))
  expect_error(colFormat(date = "false"))
  expect_error(colFormat(time = "false"))
  expect_error(colFormat(hour12 = 24))
  expect_error(colFormat(locales = 123))
})

test_that("is.colFormat", {
  expect_true(is.colFormat(colFormat()))
  expect_false(is.colFormat(list()))
})

test_that("colType", {
  expect_equal(colType(c(1, 2)), "numeric")
  expect_equal(colType(1L), "numeric")
  expect_equal(colType(c("a", "b")), "character")
  expect_equal(colType(as.factor(c("a", "b"))), "factor")
  expect_equal(colType(c(TRUE, FALSE)), "logical")
  expect_equal(colType(as.Date("2018-05-06")), "Date")
  expect_equal(colType(as.POSIXct("2018-05-06")), "Date")
  expect_equal(colType(as.POSIXlt("2018-05-06")), "Date")
})

#' @details
#' See the [online documentation](https://glin.github.io/reactable/) for
#' examples and an extensive usage guide.
#' @keywords internal
#' @import htmlwidgets
#' @import htmltools
#' @name reactable-package
#' @aliases reactable-package
"_PACKAGE"

#' Create an interactive data table
#'
#' `reactable()` creates a data table from tabular data with sorting
#' and pagination by default. The data table is an HTML widget that can be
#' used in R Markdown documents and Shiny applications, or viewed from an
#' R console.
#'
#' @param data A data frame or matrix.
#'
#'   Can also be a [`crosstalk::SharedData`] object that wraps a data frame.
#' @param columns Named list of column definitions. See [colDef()].
#' @param columnGroups List of column group definitions. See [colGroup()].
#' @param rownames Show row names? Defaults to `TRUE` if the data has row names.
#'
#'   To customize the row names column, use `".rownames"` as the column name.
#'
#'   Cells in the row names column are automatically marked up as row headers
#'   for assistive technologies.
#' @param groupBy Character vector of column names to group by.
#'
#'   To aggregate data when rows are grouped, use the `aggregate` argument in [colDef()].
#' @param sortable Enable sorting? Defaults to `TRUE`.
#' @param resizable Enable column resizing?
#' @param filterable Enable column filtering?
#' @param searchable Enable global table searching?
#' @param searchMethod Custom search method to use for global table searching.
#'   A [JS()] function that takes an array of row objects, an array of
#'   column IDs, and the search value as arguments, and returns the filtered
#'   array of row objects.
#' @param defaultColDef Default column definition used by every column. See [colDef()].
#' @param defaultColGroup Default column group definition used by every column group.
#'   See [colGroup()].
#' @param defaultSortOrder Default sort order. Either `"asc"` for ascending
#'   order or `"desc"` for descending order. Defaults to `"asc"`.
#' @param defaultSorted Character vector of column names to sort by default.
#'   Or to customize sort order, a named list with values of `"asc"` or `"desc"`.
#' @param pagination Enable pagination? Defaults to `TRUE`.
#' @param defaultPageSize Default page size for the table. Defaults to 10.
#' @param showPageSizeOptions Show page size options?
#' @param pageSizeOptions Page size options for the table. Defaults to 10, 25, 50, 100.
#' @param paginationType Pagination control to use. Either `"numbers"` for page
#'   number buttons (the default), `"jump"` for a page jump, or `"simple"` to show
#'   'Previous' and 'Next' buttons only.
#' @param showPagination Show pagination? Defaults to `TRUE` if the table has more
#'   than one page.
#' @param showPageInfo Show page info? Defaults to `TRUE`.
#' @param minRows Minimum number of rows to show per page. Defaults to 1.
#' @param paginateSubRows When rows are grouped, paginate sub rows? Defaults to `FALSE`.
#' @param details Additional content to display when expanding a row. An R function
#'   that takes the row index and column name as arguments, or a [JS()] function
#'   that takes a row info object as an argument. Can also be a [colDef()] to
#'   customize the details expander column.
#' @param defaultExpanded Expand all rows by default?
#' @param selection Enable row selection? Either `"multiple"` or `"single"` for
#'   multiple or single row selection.
#'
#'   To get the selected rows in Shiny, use [getReactableState()].
#'
#'   To customize the selection column, use `".selection"` as the column name.
#' @param defaultSelected A numeric vector of default selected row indices.
#' @param onClick Action to take when clicking a cell. Either `"expand"` to expand
#'   the row, `"select"` to select the row, or a [JS()] function that takes a
#'   row info object, column object, and table state object as arguments.
#' @param highlight Highlight table rows on hover?
#' @param outlined Add borders around the table?
#' @param bordered Add borders around the table and every cell?
#' @param borderless Remove inner borders from table?
#' @param striped Add zebra-striping to table rows?
#' @param compact Make tables more compact?
#' @param wrap Enable text wrapping? If `TRUE` (the default), long text will be
#'   wrapped to multiple lines. If `FALSE`, text will be truncated to fit on one line.
#' @param showSortIcon Show a sort icon when sorting columns?
#' @param showSortable Show an indicator on sortable columns?
#' @param class Additional CSS classes to apply to the table.
#' @param style Inline styles to apply to the table. A named list or character string.
#'
#'   Note that if `style` is a named list, property names should be camelCased.
#' @param rowClass Additional CSS classes to apply to table rows. A character
#'   string, a [JS()] function that takes a row info object and table state object
#'   as arguments, or an R function that takes a row index argument.
#' @param rowStyle Inline styles to apply to table rows. A named list, character
#'   string, [JS()] function that takes a row info object and table state object
#'   as arguments, or an R function that takes a row index argument.
#'
#'   Note that if `rowStyle` is a named list, property names should be camelCased.
#'   If `rowStyle` is a [JS()] function, it should return a JavaScript object with
#'   camelCased property names.
#' @param fullWidth Stretch the table to fill the full width of its container?
#'   Defaults to `TRUE`.
#' @param width Width of the table in pixels. Defaults to `"auto"` for automatic sizing.
#'
#'   To set the width of a column, see [colDef()].
#' @param height Height of the table in pixels. Defaults to `"auto"` for automatic sizing.
#' @param theme Theme options for the table, specified by
#'   [reactableTheme()]. Defaults to the global `reactable.theme` option.
#'   Can also be a function that returns a [reactableTheme()] or `NULL`.
#' @param language Language options for the table, specified by
#'   [reactableLang()]. Defaults to the global `reactable.language` option.
#' @param meta Custom metadata to pass to JavaScript render functions or style functions.
#'   A named list of values that can also be [JS()] expressions or functions.
#'   Custom metadata can be accessed using the `state.meta` property, and updated
#'   using `updateReactable()` in Shiny or `Reactable.setMeta()` in the JavaScript API.
#' @param elementId Element ID for the widget.
#' @param static Render the table to static HTML? Defaults to the global
#'  `reactable.static` option. Requires the V8 package, which is not installed
#'   with reactable by default.
#'
#'   With static rendering, tables are pre-rendered to their initial HTML so they appear
#'   immediately without any flash of content. Tables are then made interactive and
#'   subsequently rendered by JavaScript as needed.
#'
#'   Static rendering is **experimental**, and is not supported for tables
#'   rendered via [reactableOutput()] in Shiny.
#' @param selectionId **Deprecated**. Use [getReactableState()] to get the selected rows
#'   in Shiny.
#' @return A `reactable` HTML widget that can be used in R Markdown documents
#'   and Shiny applications, or viewed from an R console.
#'
#' @note
#' See the [online documentation](https://glin.github.io/reactable/) for
#' additional details and examples.
#'
#' @seealso
#' * [renderReactable()] and [reactableOutput()] for using reactable
#'   in Shiny applications or interactive R Markdown documents.
#' * [colDef()], [colFormat()], and [colGroup()] to customize columns.
#' * [reactableTheme()] and [reactableLang()] to customize the table.
#'
#' @examples
#' # Basic usage
#' reactable(iris)
#'
#' # Grouping and aggregation
#' reactable(
#'   iris,
#'   groupBy = "Species",
#'   columns = list(
#'     Sepal.Length = colDef(aggregate = "count"),
#'     Sepal.Width = colDef(aggregate = "mean"),
#'     Petal.Length = colDef(aggregate = "sum"),
#'     Petal.Width = colDef(aggregate = "max")
#'   )
#' )
#'
#' # Row details
#' reactable(iris, details = function(index) {
#'   htmltools::div(
#'     "Details for row: ", index,
#'     htmltools::tags$pre(paste(capture.output(iris[index, ]), collapse = "\n"))
#'   )
#' })
#'
#' # Conditional styling
#' reactable(sleep, columns = list(
#'   extra = colDef(style = function(value) {
#'     if (value > 0) {
#'       color <- "green"
#'     } else if (value < 0) {
#'       color <- "red"
#'     } else {
#'       color <- "#777"
#'     }
#'     list(color = color, fontWeight = "bold")
#'   })
#' ))
#'
#' @export
reactable <- function(
  data,
  columns = NULL,
  columnGroups = NULL,
  rownames = NULL,
  groupBy = NULL,
  sortable = TRUE,
  resizable = FALSE,
  filterable = FALSE,
  searchable = FALSE,
  searchMethod = NULL,
  defaultColDef = NULL,
  defaultColGroup = NULL,
  defaultSortOrder = "asc",
  defaultSorted = NULL,
  pagination = TRUE,
  defaultPageSize = 10,
  showPageSizeOptions = FALSE,
  pageSizeOptions = c(10, 25, 50, 100),
  paginationType = "numbers",
  showPagination = NULL,
  showPageInfo = TRUE,
  minRows = 1,
  paginateSubRows = FALSE,
  details = NULL,
  defaultExpanded = FALSE,
  selection = NULL,
  defaultSelected = NULL,
  onClick = NULL,
  highlight = FALSE,
  outlined = FALSE,
  bordered = FALSE,
  borderless = FALSE,
  striped = FALSE,
  compact = FALSE,
  wrap = TRUE,
  showSortIcon = TRUE,
  showSortable = FALSE,
  class = NULL,
  style = NULL,
  rowClass = NULL,
  rowStyle = NULL,
  fullWidth = TRUE,
  width = NULL,
  height = NULL,
  theme = getOption("reactable.theme"),
  language = getOption("reactable.language"),
  meta = NULL,
  elementId = NULL,
  static = getOption("reactable.static", FALSE),
  selectionId = NULL
) {
  crosstalkKey <- NULL
  crosstalkGroup <- NULL
  dependencies <- list()
  if (requireNamespace("crosstalk", quietly = TRUE)) {
    if (crosstalk::is.SharedData(data)) {
      crosstalkKey <- as.list(data$key())
      crosstalkGroup <- data$groupName()
      data <- data$origData()
      dependencies <- crosstalk::crosstalkLibs()
    }
  }

  if (!(is.data.frame(data) || is.matrix(data))) {
    stop("`data` must be a data frame or matrix")
  } else if (is.matrix(data)) {
    data <- as.data.frame(data, stringsAsFactors = FALSE)
  }
  if (ncol(data) == 0) {
    stop("`data` must have at least one column")
  }

  if (is.null(rownames)) {
    # Check if row names were set. This may not work if row names were set to
    # integers, but it's more reliable than using .row_names_info() on
    # data frames that have been subsetted.
    rownames <- is.character(attr(data, "row.names"))
  }
  if (!is.logical(rownames)) {
    stop("`rownames` must be TRUE or FALSE")
  } else if (rownames) {
    rownamesKey <- ".rownames"
    # Get row names from attribute to preserve type (in case of integer row names).
    # cbind.data.frame() is for consistent behavior with dplyr's grouped_df in R <= 3.6
    data <- cbind.data.frame(
      stats::setNames(data.frame(attr(data, "row.names"), stringsAsFactors = FALSE), rownamesKey),
      data
    )
    rownamesColumn <- colDef(name = "", sortable = FALSE, filterable = FALSE)
    if (rownamesKey %in% names(columns)) {
      rownamesColumn <- mergeLists(rownamesColumn, columns[[rownamesKey]])
    }
    rownamesColumn$rowHeader <- TRUE
    columns[[rownamesKey]] <- rownamesColumn
  }

  if (!is.null(groupBy)) {
    if (!all(groupBy %in% colnames(data))) {
      stop("`groupBy` columns must exist in `data`")
    }
    if (any(sapply(columns[groupBy], function(col) !is.null(col[["details"]])))) {
      stop("`details` cannot be used on a grouping column")
    }
  }

  if (!is.logical(sortable)) {
    stop("`sortable` must be TRUE or FALSE")
  }

  if (!is.logical(resizable)) {
    stop("`resizable` must be TRUE or FALSE")
  }

  if (!is.logical(filterable)) {
    stop("`filterable` must be TRUE or FALSE")
  }

  if (!is.logical(searchable)) {
    stop("`searchable` must be TRUE or FALSE")
  }

  if (!is.null(searchMethod) && !is.JS(searchMethod)) {
    stop('`searchMethod` must be a JS function')
  }

  columnKeys <- colnames(data)

  # Exclude special column for sub rows
  subRowsKey <- ".subRows"
  columnKeys <- columnKeys[columnKeys != subRowsKey]

  if (!is.null(details)) {
    detailsKey <- ".details"
    columnKeys <- c(detailsKey, columnKeys)
    detailsColumn <- colDef(name = "", sortable = FALSE, filterable = FALSE,
                            searchable = FALSE, resizable = FALSE, width = 45,
                            align = "center")
    if (is.colDef(details)) {
      detailsColumn <- mergeLists(detailsColumn, details)
    } else {
      detailsColumn <- mergeLists(detailsColumn, colDef(details = details))
    }
    # Prepend column
    columns <- c(stats::setNames(list(detailsColumn), detailsKey), columns)
  }

  if (!is.null(selection)) {
    selectionKey <- ".selection"
    columnKeys <- c(selectionKey, columnKeys)
    selectionColumn <- colDef(name = "", resizable = FALSE, width = 45)
    selectionColumn$selectable <- TRUE
    if (selectionKey %in% names(columns)) {
      selectionColumn <- mergeLists(selectionColumn, columns[[selectionKey]])
    }
    columns[[selectionKey]] <- selectionColumn
  }

  if (!is.null(defaultColDef)) {
    if (!is.colDef(defaultColDef)) {
      stop("`defaultColDef` must be a column definition")
    }
    columns <- lapply(columnKeys, function(name) {
      mergeLists(defaultColDef, columns[[name]])
    })
    columns <- stats::setNames(columns, columnKeys)
  }

  if (!is.null(defaultColGroup)) {
    if (!is.colGroup(defaultColGroup)) {
      stop("`defaultColGroup` must be a column group definition")
    }
    columnGroups <- lapply(columnGroups, function(group) {
      mergeLists(defaultColGroup, group)
    })
  }

  if (!is.null(columns)) {
    if (!isNamedList(columns) || !all(sapply(columns, is.colDef))) {
      stop("`columns` must be a named list of column definitions")
    }
    if (!all(names(columns) %in% columnKeys)) {
      stop("`columns` names must exist in `data`")
    }
  }

  if (!is.null(columnGroups)) {
    if (!all(sapply(columnGroups, is.colGroup))) {
      stop("`columnGroups` must be a list of column group definitions")
    }
    for (group in columnGroups) {
      if (length(group$columns) == 0) {
        stop("`columnGroups` groups must contain at least one column")
      }
      if (!all(group$columns %in% columnKeys)) {
        stop("`columnGroups` columns must exist in `data`")
      }
    }
  }

  if (!isSortOrder(defaultSortOrder)) {
    stop('`defaultSortOrder` must be "asc" or "desc"')
  }
  if (!is.null(defaultSorted)) {
    if (!is.character(defaultSorted) && !isNamedList(defaultSorted)) {
      stop("`defaultSorted` must be a named list or character vector of column names")
    }
    if (is.character(defaultSorted)) {
      orders <- lapply(defaultSorted, function(name) {
        if (!is.null(columns[[name]]$defaultSortDesc)) {
          if (columns[[name]]$defaultSortDesc) "desc" else "asc"
        } else {
          defaultSortOrder
        }
      })
      defaultSorted <- stats::setNames(orders, defaultSorted)
    }
    if (!all(sapply(defaultSorted, isSortOrder))) {
      stop('`defaultSorted` values must be "asc" or "desc"')
    }
    if (!all(names(defaultSorted) %in% colnames(data))) {
      stop("`defaultSorted` columns must exist in `data`")
    }
  }
  if (!is.logical(pagination)) {
    stop("`pagination` must be TRUE or FALSE")
  }
  if (!is.numeric(defaultPageSize)) {
    stop("`defaultPageSize` must be numeric")
  }
  if (!is.logical(showPageSizeOptions)) {
    stop("`showPageSizeOptions` must be TRUE or FALSE")
  }
  if (!is.null(pageSizeOptions)) {
    if (!is.numeric(pageSizeOptions)) {
      stop("`pageSizeOptions` must be numeric")
    }
    pageSizeOptions <- as.list(pageSizeOptions)
  }
  if (!paginationType %in% c("numbers", "jump", "simple")) {
    stop('`paginationType` must be one of "numbers", "jump", "simple"')
  }
  if (!is.null(showPagination) && !is.logical(showPagination)) {
    stop("`showPagination` must be TRUE or FALSE")
  }
  if (!is.logical(showPageInfo)) {
    stop("`showPageInfo` must be TRUE or FALSE")
  }
  if (!is.numeric(minRows)) {
    stop("`minRows` must be numeric")
  }
  if (!is.logical(paginateSubRows)) {
    stop("`paginateSubRows` must be TRUE or FALSE")
  }
  if (!is.logical(defaultExpanded)) {
    stop("`defaultExpanded` must be TRUE or FALSE")
  }
  if (!is.null(selection) && !selection %in% c("multiple", "single")) {
    stop('`selection` must be "multiple" or "single"')
  }
  if (!is.null(selectionId)) {
    warning("`selectionId` is deprecated. Use `getReactableState()` to get the selected rows in Shiny.")
    if (!is.character(selectionId)) {
      stop("`selectionId` must be a character")
    }
  }
  if (!is.null(defaultSelected)) {
    if (!is.numeric(defaultSelected)) {
      stop("`defaultSelected` must be numeric")
    }
    if (any(defaultSelected < 1 | defaultSelected > nrow(data))) {
      stop("`defaultSelected` row indices must be within range")
    }
    # Convert to 0-based indexing
    defaultSelected <- as.list(defaultSelected - 1)
  }
  if (!is.null(onClick) && !onClick %in% c("expand", "select") && !is.JS(onClick)) {
    stop('`onClick` must be "expand", "select", or a JS function')
  }
  if (!is.logical(highlight)) {
    stop("`highlight` must be TRUE or FALSE")
  }
  if (!is.logical(outlined)) {
    stop("`outlined` must be TRUE or FALSE")
  }
  if (!is.logical(bordered)) {
    stop("`bordered` must be TRUE or FALSE")
  }
  if (!is.logical(borderless)) {
    stop("`borderless` must be TRUE or FALSE")
  }
  if (!is.logical(striped)) {
    stop("`striped` must be TRUE or FALSE")
  }
  if (!is.logical(compact)) {
    stop("`compact` must be TRUE or FALSE")
  }
  if (!is.logical(wrap)) {
    stop("`wrap` must be `TRUE` or `FALSE`")
  }
  if (!is.logical(showSortIcon)) {
    stop("`showSortIcon` must be TRUE or FALSE")
  }
  if (!is.logical(showSortable)) {
    stop("`showSortable` must be TRUE or FALSE")
  }
  if (!is.null(class) && !is.character(class)) {
    stop("`class` must be a character")
  }
  if (!is.null(style) && !isNamedList(style) && !is.character(style)) {
    stop("`style` must be a named list or character string")
  }
  if (!is.null(rowClass)) {
    if (!is.character(rowClass) && !is.JS(rowClass) && !is.function(rowClass)) {
      stop("`rowClass` must be a character, JS function, or R function")
    }
    if (is.function(rowClass)) {
      rowClass <- lapply(seq_len(nrow(data)), function(index) {
        callFunc(rowClass, index)
      })
    }
  }
  if (!is.null(rowStyle)) {
    if (!isNamedList(rowStyle) && !is.character(rowStyle) && !is.JS(rowStyle) && !is.function(rowStyle)) {
      stop("`rowStyle` must be a named list, character string, JS function, or R function")
    }
    if (is.function(rowStyle)) {
      rowStyle <- lapply(seq_len(nrow(data)), function(index) {
        asReactStyle(callFunc(rowStyle, index))
      })
    } else if (is.character(rowStyle) && !is.JS(rowStyle)) {
      rowStyle <- asReactStyle(rowStyle)
    }
  }
  if (!is.logical(fullWidth)) {
    stop("`fullWidth` must be TRUE or FALSE")
  }
  width <- htmltools::validateCssUnit(width)
  height <- htmltools::validateCssUnit(height)

  if (!is.null(theme)) {
    if (is.function(theme)) {
      theme <- callFunc(theme)
    }
    if (!is.null(theme) && !is.reactableTheme(theme)) {
      stop("`theme` must be a reactable theme object")
    }
  }

  if (!is.null(language) && !is.reactableLang(language)) {
    stop("`language` must be a reactable language options object")
  }

  if (!is.null(meta) && !isNamedList(meta)) {
    stop("`meta` must be a named list")
  }

  if (!is.logical(static)) {
    stop("`static` must be TRUE or FALSE")
  }

  addDependencies <- function(x) {
    # Dedupe dependencies
    for (dep in htmltools::findDependencies(x)) {
      dependencies[[sprintf("%s-%s", dep$name, dep$version)]] <<- dep
    }
    dependencies <<- htmltools::resolveDependencies(dependencies)
  }

  cols <- lapply(columnKeys, function(key) {
    column <- list(
      id = key,
      name = key,
      type = colType(data[[key]])
    )

    if (!is.null(columns[[key]])) {
      column <- mergeLists(column, columns[[key]])
    }

    cell <- column[["cell"]]
    if (is.function(cell)) {
      content <- lapply(seq_len(nrow(data)), function(index) {
        value <- data[[key]][[index]]
        callFunc(cell, value, index, key)
      })
      column$cell <- lapply(content, asReactTag)
      addDependencies(column$cell)
    }

    header <- column[["header"]]
    if (!is.null(header)) {
      if (is.function(header)) {
        header <- callFunc(header, column$name, key)
      }
      if (!is.JS(header)) {
        column$header <- asReactTag(header)
        addDependencies(column$header)
      }
    }

    footer <- column[["footer"]]
    if (!is.null(footer)) {
      if (is.function(footer)) {
        values <- data[[key]]
        footer <- callFunc(footer, values, key)
      }
      if (!is.JS(footer)) {
        column$footer <- asReactTag(footer)
        addDependencies(column$footer)
      }
    }

    details <- column[["details"]]
    if (is.function(details)) {
      details <- lapply(seq_len(nrow(data)), function(index) {
        callFunc(details, index, key)
      })
      column$details <- lapply(details, asReactTag)
      addDependencies(column$details)
    }

    filterInput <- column[["filterInput"]]
    if (!is.null(filterInput)) {
      if (is.function(filterInput)) {
        values <- data[[key]]
        filterInput <- callFunc(filterInput, values, key)
      }
      if (!is.JS(filterInput)) {
        column$filterInput <- asReactTag(filterInput)
        addDependencies(column$filterInput)
      }
    }

    className <- column[["className"]]
    if (is.function(className)) {
      classes <- lapply(seq_len(nrow(data)), function(index) {
        value <- data[[key]][[index]]
        callFunc(className, value, index, key)
      })
      column$className <- classes
    }

    style <- column[["style"]]
    if (is.function(style)) {
      style <- lapply(seq_len(nrow(data)), function(index) {
        value <- data[[key]][[index]]
        callFunc(style, value, index, key)
      })
      column$style <- lapply(style, asReactStyle)
    }

    column
  })

  if (!is.null(columnGroups)) {
    columnGroups <- lapply(columnGroups, function(group) {
      header <- group[["header"]]
      if (!is.null(header)) {
        if (is.function(header)) {
          header <- callFunc(header, group$name)
        }
        if (!is.JS(header)) {
          group$header <- asReactTag(header)
          addDependencies(group$header)
        }
      }
      group
    })
  }

  data <- toJSON(data)

  # Create a unique key for the data. The key is used to fully reset state when
  # the data changes (for tables in Shiny).
  dataKey <- digest::digest(list(data, cols))

  # Serialize user-set args only to keep the widget HTML slim
  defaultArgs <- formals()
  args <- as.list(match.call())
  setArgs <- stats::setNames(names(defaultArgs) %in% names(args), names(defaultArgs))

  component <- reactR::component("Reactable", list(
    data = data,
    columns = cols,
    columnGroups = columnGroups,
    groupBy = as.list(groupBy),
    sortable = if (!sortable) FALSE,
    resizable = if (resizable) TRUE,
    filterable = if (filterable) TRUE,
    searchable = if (searchable) TRUE,
    searchMethod = searchMethod,
    defaultSortDesc = if (isDescOrder(defaultSortOrder)) TRUE,
    defaultSorted = columnSortDefs(defaultSorted),
    pagination = if (!pagination) FALSE,
    defaultPageSize = if (setArgs["defaultPageSize"]) defaultPageSize,
    showPageSizeOptions = if (setArgs["showPageSizeOptions"]) showPageSizeOptions,
    pageSizeOptions = if (setArgs["pageSizeOptions"]) pageSizeOptions,
    paginationType = if (setArgs["paginationType"]) paginationType,
    showPagination = if (!is.null(showPagination)) showPagination,
    showPageInfo = if (setArgs["showPageInfo"]) showPageInfo,
    minRows = if (setArgs["minRows"]) minRows,
    paginateSubRows = if (paginateSubRows) TRUE,
    defaultExpanded = if (defaultExpanded) defaultExpanded,
    selection = selection,
    selectionId = selectionId,
    defaultSelected = defaultSelected,
    onClick = onClick,
    highlight = if (highlight) TRUE,
    outlined = if (outlined) TRUE,
    bordered = if (bordered) TRUE,
    borderless = if (borderless) TRUE,
    striped = if (striped) TRUE,
    compact = if (compact) TRUE,
    nowrap = if (!wrap) TRUE,
    showSortIcon = if (!showSortIcon) FALSE,
    showSortable = if (showSortable) TRUE,
    className = class,
    style = asReactStyle(style),
    rowClassName = rowClass,
    rowStyle = rowStyle,
    inline = if (!fullWidth) TRUE,
    width = width,
    height = height,
    theme = theme,
    language = language,
    meta = meta,
    crosstalkKey = crosstalkKey,
    crosstalkGroup = crosstalkGroup,
    elementId = elementId,
    dataKey = dataKey,
    static = static
  ))

  htmlwidgets::createWidget(
    name = "reactable",
    reactR::reactMarkup(component),
    width = width,
    height = height,
    sizingPolicy = htmlwidgets::sizingPolicy(
      defaultWidth = "auto",
      defaultHeight = "auto",
      # Don't limit width when rendered inside an R Notebook
      knitr.figure = FALSE
    ),
    package = "reactable",
    dependencies = dependencies,
    elementId = elementId
  )
}

# Convert named list of column orders to { id, desc } definitions
columnSortDefs <- function(defaultSorted) {
  lapply(names(defaultSorted), function(id) {
    list(id = id, desc = isDescOrder(defaultSorted[[id]]))
  })
}

#' Shiny bindings for reactable
#'
#' Output and render functions for using reactable within Shiny
#' applications and interactive R Markdown documents.
#'
#' @param outputId Output variable to read from.
#' @param width,height A valid CSS unit (like `"100%"`, `"400px"`, `"auto"`)
#'   or a number, which will be coerced to a string and have `"px"` appended.
#' @param inline Use an inline element for the table's container?
#' @param expr An expression that generates a [reactable] widget.
#' @param env The environment in which to evaluate `expr`.
#' @param quoted Is `expr` a quoted expression (with [quote()])? This is useful
#'   if you want to save an expression in a variable.
#' @return `reactableOutput()` returns a `reactable` output element that can be
#'   included in a Shiny UI.
#'
#'   `renderReactable()` returns a `reactable` render function that can be
#'   assigned to a Shiny output slot.
#'
#' @name reactable-shiny
#'
#' @note
#' See the [online demo](https://glin.github.io/reactable/articles/shiny-demo.html)
#' for additional examples of using reactable in Shiny.
#'
#' @seealso [updateReactable()] for updating a reactable instance in Shiny.
#'
#'  [getReactableState()] for getting the state of a reactable instance in Shiny.
#'
#' @examples
#' # Run in an interactive R session
#' if (interactive()) {
#'
#' library(shiny)
#' library(reactable)
#'
#' ui <- fluidPage(
#'  titlePanel("reactable example"),
#'  reactableOutput("table")
#' )
#'
#' server <- function(input, output, session) {
#'   output$table <- renderReactable({
#'    reactable(iris)
#'  })
#' }
#'
#' shinyApp(ui, server)
#' }
#'
#' @export
reactableOutput <- function(outputId, width = "auto", height = "auto", inline = FALSE) {
  output <- htmlwidgets::shinyWidgetOutput(outputId, "reactable", width, height,
                                           inline = inline, package = "reactable")
  # Add attribute to Shiny output containers to differentiate them from static widgets
  addOutputId <- function(x) {
    if (isTagList(x)) {
      x[] <- lapply(x, addOutputId)
    } else if (is.tag(x)) {
      x <- htmltools::tagAppendAttributes(x, "data-reactable-output" = outputId)
    }
    x
  }
  output <- addOutputId(output)
  output
}

#' @rdname reactable-shiny
#' @export
renderReactable <- function(expr, env = parent.frame(), quoted = FALSE) {
  if (!quoted) { expr <- substitute(expr) }
  htmlwidgets::shinyRenderWidget(expr, reactableOutput, env, quoted = TRUE)
}

#' Convert a reactable widget to HTML tags
#'
#' This S3 method exists to enable [reactable()]'s `static` rendering option.
#'
#' @param x a [reactable()] instance.
#' @param standalone Logical value indicating whether the widget is being
#'   rendered in a standalone context.
#' @param ... Additional arguments passed to the S3 method.
#'
#' @keywords internal
#' @export
as.tags.reactable <- function(x, standalone = FALSE, ...) {
  # Only need the static attribute for decided whether to SSR
  attribs <- x$x$tag$attribs
  static <- attribs$static
  x$x$tag$attribs$static <- NULL

  # Should call htmlwidgets:::as.tags.htmlwidget(), which calls htmlwidgets:::toHTML()
  result <- NextMethod("as.tags", x, standalone = standalone, ...)

  if (!isTRUE(static)) {
    return(result)
  }

  if (!requireNamespace("V8", quietly = TRUE)) {
    # Fall back to client-side rendering if V8 isn't installed
    warning('The V8 package must be installed to use `reactable(static = TRUE)`.
Do you need to run `install.packages("V8")`?', call. = FALSE)
    return(result)
  }

  input_json <- toJSON(list(
    props = attribs,
    evals = htmlwidgets::JSEvals(attribs)
  ))

  output <- list()

  tryCatch({
    ctx <- V8::v8()
    ctx$source(system.file("htmlwidgets/reactable.server.js", package = "reactable", mustWork = TRUE))
    output <- ctx$call("Reactable.renderToHTML", input_json)
  }, error = function(e) {
    warning("Failed to render table to static HTML:\n", conditionMessage(e), call. = FALSE)
  })

  if (length(output) == 0) {
    return(result)
  }

  ssrHTML <- HTML(output$html)
  ssrCSS <- NULL

  if (any(nzchar(output$css))) {
    ids <- paste(output$ids, collapse = " ")
    # Make sure to keep this in sync with the Emotion cache key.
    # Since this isn't well documented at https://emotion.sh/docs/ssr#when-using-emotioncss,
    # data-emotion="{cache-key} {space-separated-ids}" allows for hydration to
    # occur automatically without having to call emotion.cache.hydrate(ids),
    # and prevents duplicate styles from being inserted into the page.
    emotionAttr <- sprintf("reactable %s", ids)
    styles <- tags$style(`data-emotion` = emotionAttr, htmltools::HTML(output$css))
    # Use an empty htmlDependency to insert style tags into <head>. This prevents
    # duplicate style tags when rendering duplicate tables and works around a
    # pkgdown issue. pkgdown ignores tags$head() but does use head content from
    # htmlDependencies, although still inserts the head content at the top of
    # <body> instead of <head>.
    ssrCSS <- htmlDependency(
      emotionAttr, "1", "", head = as.character(styles), package = "reactable"
    )
  }

  # Temporarily wrap result in additional tag to avoid this issue
  # https://github.com/rstudio/htmltools/issues/334
  wrapper <- htmltools::tag("WRAPPER", list(result))
  wrapper <- tagAppendChildren(
    wrapper, ssrHTML, ssrCSS, .cssSelector = ".reactable"
  )
  wrapper <- tagAppendAttributes(
    wrapper, `data-react-ssr` = NA, .cssSelector = ".reactable"
  )
  browsable(as.tags(wrapper$children))
}

#' Print a reactable widget for knitr
#'
#' This S3 method exists to enable [reactable()]'s `static` rendering option.
#'
#' @param x A [reactable()] instance.
#' @param ... Additional arguments passed to the S3 method.
#'
#' @keywords internal
#' @export
knit_print.reactable <- function(x, ...) {
  # knitr options (out.width/out.height) are ignored here because as.tags() doesn't
  # pass it to htmlwidgets:::toHTML(), but this is fine because reactable disables
  # knitr.figure in the sizing policy.
  knitr::knit_print(htmltools::as.tags(x, standalone = FALSE), ...)
}

#' Called by HTMLWidgets to produce the widget's root element
#'
#' @param id Element ID.
#' @param style Element style.
#' @param class Element class.
#' @param ... Additional arguments.
#' @keywords internal
widget_html.reactable <- function(id, style, class, ...) {
  # Set text color in R Notebooks to prevent contrast issues when
  # using a dark editor theme and htmltools 0.4.0.
  if (isTRUE(getOption("rstudio.notebook.executing"))) {
    style <- paste0("color: #333;", style)
  }

  htmltools::tags$div(
    id = id, class = class, style = style, reactDependencies()
  )
}

# Make sure react.js come before reactable.js. Note this "workaround"
# wouldn't be needed with ramnathv/htmlwidgets#324
reactDependencies <- function() {
  list(
    reactR::html_dependency_corejs(), # Necessary for RStudio Viewer version < 1.2 and IE11
    reactR::html_dependency_react(),
    reactR::html_dependency_reacttools()
  )
}

# Deprecated convention for htmlwidgets <= 1.5.2 support
reactable_html <- widget_html.reactable

library(htmltools)

test_that("reactable handles invalid args", {
  # NOTE: New arguments should be tested in their own separate test

  df <- data.frame(x = 1)
  expect_error(reactable(df, sortable = "true"))
  expect_error(reactable(df, resizable = "true"))
  expect_error(reactable(df, defaultSortOrder = "ascending"))
  expect_error(reactable(df, defaultSorted = "y"))
  expect_error(reactable(df, defaultSorted = list("x")))
  expect_error(reactable(df, defaultSorted = list(x = "ascending")))
  expect_error(reactable(df, defaultSorted = list(y = "asc")))
  expect_error(reactable(df, defaultExpanded = NULL))
  expect_error(reactable(df, defaultExpanded = 1:3))
  expect_error(reactable(df, defaultSelected = "12"))
  expect_error(reactable(df, onClick = "function() {}"))
  expect_error(reactable(df, highlight = "true"))
  expect_error(reactable(df, outlined = "true"))
  expect_error(reactable(df, bordered = NULL))
  expect_error(reactable(df, borderless = NULL))
  expect_error(reactable(df, striped = "true"))
  expect_error(reactable(df, compact = "true"))
  expect_error(reactable(df, wrap = "true"))
  expect_error(reactable(df, showSortIcon = "true"))
  expect_error(reactable(df, showSortable = "true"))
  expect_error(reactable(df, class = c(1, 5)))
  expect_error(reactable(df, style = 555))
  expect_error(reactable(df, rowClass = 123))
  expect_error(reactable(df, rowStyle = 555))
  expect_error(reactable(df, fullWidth = "yes"))
})

test_that("reactable", {
  # NOTE: New arguments should be tested in their own separate test

  expect_error(reactable(1), "`data` must be a data frame or matrix")

  # Default args
  tbl <- reactable(data.frame(x = 1, y = "b", stringsAsFactors = TRUE))
  attribs <- getAttribs(tbl)
  data <- data.frame(x = 1, y = "b")
  data <- jsonlite::toJSON(data, dataframe = "columns", rownames = FALSE)
  columns <- list(
    list(id = "x", name = "x", type = "numeric"),
    list(id = "y", name = "y", type = "factor")
  )
  expected <- list(
    data = data,
    columns = columns,
    dataKey = digest::digest(list(data, columns)),
    static = FALSE
  )
  expect_equal(attribs, expected)

  # Table options
  tbl <- reactable(data.frame(x = "a", stringsAsFactors = TRUE), rownames = TRUE,
                   sortable = FALSE, resizable = TRUE,
                   defaultSortOrder = "desc", defaultSorted = list(x = "asc"),
                   defaultExpanded = TRUE, highlight = TRUE,
                   outlined = TRUE, bordered = TRUE, borderless = TRUE, striped = TRUE,
                   compact = TRUE, wrap = FALSE, showSortIcon = FALSE,
                   showSortable = TRUE, class = "tbl", style = list(color = "red"),
                   fullWidth = FALSE)
  attribs <- getAttribs(tbl)
  data <- data.frame(.rownames = 1, x = "a")
  data <- jsonlite::toJSON(data, dataframe = "columns", rownames = FALSE)
  columns <- list(
    list(id = ".rownames", name = "", type = "numeric",
         sortable = FALSE, filterable = FALSE, rowHeader = TRUE),
    list(id = "x", name = "x", type = "factor")
  )
  expected <- list(
    data = data,
    columns = columns,
    sortable = FALSE,
    resizable = TRUE,
    defaultSortDesc = TRUE,
    defaultSorted = list(list(id = "x", desc = FALSE)),
    defaultExpanded = TRUE,
    highlight = TRUE,
    outlined = TRUE,
    bordered = TRUE,
    borderless = TRUE,
    striped = TRUE,
    compact = TRUE,
    nowrap = TRUE,
    showSortIcon = FALSE,
    showSortable = TRUE,
    className = "tbl",
    style = list(color = "red"),
    inline = TRUE,
    dataKey = digest::digest(list(data, columns)),
    static = FALSE
  )
  expect_equal(attribs, expected)

  # Style
  tbl <- reactable(data.frame(x = 1), style = " border-bottom: 1px solid; top: 50px")
  attribs <- getAttribs(tbl)
  expect_equal(attribs$style, list("border-bottom" = "1px solid", top = "50px"))
})

test_that("data can be a matrix", {
  data <- matrix(c(1, 2, 3, 4), nrow = 2)
  tbl <- reactable(data)
  attribs <- getAttribs(tbl)
  expect_equal(as.character(attribs$data), '{"V1":[1,2],"V2":[3,4]}')
  expect_length(attribs$columns, 2)

  data <- matrix(c("a", "b", "c", "d"), nrow = 2, dimnames = list(NULL, c("x", "y")))
  tbl <- reactable(data)
  attribs <- getAttribs(tbl)
  expect_equal(as.character(attribs$data), '{"x":["a","b"],"y":["c","d"]}')
  expect_length(attribs$columns, 2)
})

test_that("data should have at least one column", {
  expect_error(reactable(data.frame()), "`data` must have at least one column")

  # Data can have zero rows, as long as it has columns
  tbl <- reactable(data.frame(x = character(0), y = numeric(0)))
  attribs <- getAttribs(tbl)
  expect_equal(as.character(attribs$data), '{"x":[],"y":[]}')
  expect_length(attribs$columns, 2)
})

test_that("data with numbers are serialized with max precision", {
  data <- data.frame(x = 0.123456789012345)  # 16 digits
  tbl <- reactable(data)
  attribs <- getAttribs(tbl)
  expect_equal(as.character(attribs$data), '{"x":[0.123456789012345]}')
})

test_that("data with numeric NA, NaN, Inf, and -Inf are serialized correctly", {
  # jsonlite::toJSON(na = "null") will serialize NA as null, but also discard NaN, Inf, and -Inf
  data <- data.frame(n = c(1, NA, NaN, Inf, -Inf, 0, -1))
  tbl <- reactable(data)
  attribs <- getAttribs(tbl)
  expect_equal(as.character(attribs$data), '{"n":[1,"NA","NaN","Inf","-Inf",0,-1]}')
})

test_that("data with dates/datetimes are serialized in ISO 8601", {
  data <- data.frame(x = as.POSIXct("2019-05-06 3:22:15", tz = "UTC"), y = as.Date("2010-12-30"))
  tbl <- reactable(data)
  attribs <- getAttribs(tbl)
  expect_equal(as.character(attribs$data), '{"x":["2019-05-06T03:22:15Z"],"y":["2010-12-30"]}')
})

test_that("data with custom classes not supported by jsonlite are serialized", {
  data <- data.frame(
    difftime = as.Date("2021-06-11") - as.Date("2021-03-05"),
    custom = I(structure("custom", class = "unsupported_custom_class"))
  )
  tbl <- reactable(data)
  attribs <- getAttribs(tbl)
  serialized <- jsonlite::fromJSON(attribs$data)
  expect_equal(serialized$custom, "custom")
  # jsonlite may support difftime objects at some point, so don't throw an error
  if (serialized$difftime != 98) {
    warning(paste("Unexpected result from serializing difftime:", serialized$difftime))
  }
})

test_that("data with list-columns are serialized correctly", {
  data <- data.frame(
    x = I(list(
      # Length-1 vectors should be unboxed, except when wrapped in I()
      "x",
      I("x"),
      list("x"),
      list(1, 2, 3),
      c("a", "b"),
      list(x = TRUE),
      # Length-1 data frame columns should still be arrays
      data.frame(x = "y"),
      # NULLs should be serialized as null, not jsonlite's default {}
      NULL,
      NA
    ))
  )
  tbl <- reactable(data)
  attribs <- getAttribs(tbl)
  expect_equal(as.character(attribs$data), '{"x":["x",["x"],["x"],[1,2,3],["a","b"],{"x":true},{"x":["y"]},null,null]}')
})

test_that("data supports Crosstalk", {
  data <- crosstalk::SharedData$new(
    data.frame(x = c(1, 2), y = c("a", "b"), stringsAsFactors = FALSE),
    key = ~y,
    group = "group"
  )
  tbl <- reactable(data)
  attribs <- getAttribs(tbl)
  expect_equal(as.character(attribs$data), '{"x":[1,2],"y":["a","b"]}')
  expect_equal(attribs$crosstalkKey, list("a", "b"))
  expect_equal(attribs$crosstalkGroup, "group")

  # Keys with length 1 should be serialized as arrays
  data <- crosstalk::SharedData$new(data.frame(x = 1, y = "2"))
  tbl <- reactable(data)
  attribs <- getAttribs(tbl)
  expect_equal(as.character(attribs$data), '{"x":[1],"y":["2"]}')
  expect_equal(attribs$crosstalkKey, list("1"))
  expect_equal(attribs$crosstalkGroup, data$groupName())

  # Data with list-columns
  data <- crosstalk::SharedData$new(data.frame(x = I(list(list(1,2,3), list(x = 1)))), key = ~x)
  tbl <- reactable(data)
  attribs <- getAttribs(tbl)
  expect_equal(as.character(attribs$data), '{"x":[[1,2,3],{"x":1}]}')
  expect_equal(attribs$crosstalkKey, I(list(list(1,2,3), list(x = 1))))
  expect_equal(attribs$crosstalkGroup, data$groupName())
})

test_that("columns", {
  df <- data.frame(x = 1)
  expect_error(reactable(df, columns = "x"), "`columns` must be a named list of column definitions")
  expect_error(reactable(df, columns = list(list())), "`columns` must be a named list of column definitions")
  expect_error(reactable(df, columns = list(colDef())), "`columns` must be a named list of column definitions")
  expect_error(reactable(df, columns = list(zzzz = colDef())), "`columns` names must exist in `data`")

  df <- data.frame(
    chr = "a",
    num = 1,
    fct = factor("b"),
    lgl = TRUE,
    lst = I(list("a")),
    stringsAsFactors = FALSE
  )
  tbl <- reactable(df)
  columns <- getAttrib(tbl, "columns")
  expect_equal(columns, list(
    list(id = "chr", name = "chr", type = "character"),
    list(id = "num", name = "num", type = "numeric"),
    list(id = "fct", name = "fct", type = "factor"),
    list(id = "lgl", name = "lgl", type = "logical"),
    list(id = "lst", name = "lst", type = "AsIs")
  ))

  # Complex numbers currently fail due to `jsonlite::toJSON(3i, digits = NA)` error
  tryCatch({
    tbl <- reactable(data.frame(cpl = 3i))
    columns <- getAttrib(tbl, "columns")
    expect_equal(columns, list(id = "cpl", name = "cpl", type = "complex"))
    writeLines("complex numbers test passed")
  },
  error = function(e) {
    # Skipped due to error, "invalid value -2147483648 for 'digits' argument"
  })
})

test_that("columnGroups", {
  df <- data.frame(x = 1)
  expect_error(reactable(df, columnGroups = "x"), "`columnGroups` must be a list of column group definitions")
  expect_error(reactable(df, columnGroups = list(colDef())),
               "`columnGroups` must be a list of column group definitions")
  expect_error(reactable(df, columnGroups = list(colGroup(name = "", columns = "y"))),
               "`columnGroups` columns must exist in `data`")
  expect_error(reactable(df, columnGroups = list(colGroup(name = ""))),
               "`columnGroups` groups must contain at least one column")

  df <- data.frame(x = 1, y = 2, z = "z")
  tbl <- reactable(df, columnGroups = list(
    colGroup(columns = c("x", "z")),
    colGroup(name = "y", columns = "y")
  ))
  expect_equal(getAttrib(tbl, "columnGroups"), list(
    colGroup(columns = c("x", "z")),
    colGroup(name = "y", columns = "y")
  ))
})

test_that("rownames", {
  expect_error(reactable(data.frame(x = 1), rownames = "true"),
               "`rownames` must be TRUE or FALSE")

  # Integer row names
  tbl <- reactable(data.frame(x = c(1, 2, 3)), rownames = TRUE)
  attribs <- getAttribs(tbl)
  expect_equal(as.character(attribs$data), '{".rownames":[1,2,3],"x":[1,2,3]}')
  expect_equal(attribs$columns[[1]], list(
    id = ".rownames", name = "",  type = "numeric",
    sortable = FALSE, filterable = FALSE, rowHeader = TRUE
  ))

  # Character row names
  tbl <- reactable(data.frame(x = c(1, 2, 3), row.names = c("a", "b", "c")), rownames = TRUE)
  attribs <- getAttribs(tbl)
  expect_equal(as.character(attribs$data), '{".rownames":["a","b","c"],"x":[1,2,3]}')
  expect_equal(attribs$columns[[1]], list(
    id = ".rownames", name = "",  type = "character",
    sortable = FALSE, filterable = FALSE, rowHeader = TRUE
  ))

  # Custom rownames colDef
  tbl <- reactable(data.frame(x = c(1, 2)), rownames = TRUE, columns = list(
    .rownames = colDef(name = "N", sortable = TRUE, headerClass = "hdr")
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]], list(
    id = ".rownames", name = "N",  type = "numeric",
    sortable = TRUE, filterable = FALSE, headerClassName = "hdr", rowHeader = TRUE
  ))

  # Row names can be part of column groups
  tbl <- reactable(data.frame(x = c(1, 2)), rownames = TRUE, columnGroups = list(
    colGroup("group", c(".rownames", "x"))
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columnGroups[[1]]$columns, list(".rownames", "x"))

  # Can't use rownames column without rownames
  expect_error(reactable(data.frame(1), columns = list(.rownames = colDef())))
  expect_error(reactable(data.frame(1), columnGroups = list(colGroup("", ".rownames"))))

  # If row names present, should be shown by default
  tbl <- reactable(data.frame(x = c(1, 2, 3), row.names = c("a", "b", "c")))
  attribs <- getAttribs(tbl)
  expect_equal(as.character(attribs$data), '{".rownames":["a","b","c"],"x":[1,2,3]}')
  expect_equal(attribs$columns[[1]], list(
    id = ".rownames", name = "",  type = "character",
    sortable = FALSE, filterable = FALSE, rowHeader = TRUE
  ))
  # Handles matrices
  tbl <- reactable(matrix(c(1,2,3), dimnames = list(c(1, 2, 3), "x")))
  attribs <- getAttribs(tbl)
  expect_equal(as.character(attribs$data), '{".rownames":["1","2","3"],"x":[1,2,3]}')
  expect_equal(attribs$columns[[1]], list(
    id = ".rownames", name = "",  type = "character",
    sortable = FALSE, filterable = FALSE, rowHeader = TRUE
  ))

  # If no row names, should not be shown by default
  tbl <- reactable(data.frame(x = c(1, 2, 3)))
  attribs <- getAttribs(tbl)
  expect_equal(as.character(attribs$data), '{"x":[1,2,3]}')
  expect_equal(length(attribs$columns), 1)

  # Should work with grouped_df from dplyr. cbind.data.frame() uses stringsAsFactors, but
  # dplyr:::cbind.grouped_df() ignores it
  grouped_df <- dplyr::grouped_df(data.frame(x = c(1, 2)), "x")
  tbl <- reactable(grouped_df, rownames = TRUE)
  expect_equal(as.character(getAttrib(tbl, "data")), '{".rownames":[1,2],"x":[1,2]}')
})

test_that("groupBy", {
  data <- data.frame(x = 1, y = "a", z = TRUE)

  tbl <- reactable(data)
  expect_equal(getAttrib(tbl, "groupBy"), NULL)

  tbl <- reactable(data, groupBy = "x")
  expect_equal(getAttrib(tbl, "groupBy"), list("x"))

  tbl <- reactable(data, groupBy = c("z", "x"))
  expect_equal(getAttrib(tbl, "groupBy"), list("z", "x"))

  expect_error(reactable(data, groupBy = c("z", "notexists")), "`groupBy` columns must exist in `data`")
  expect_error(reactable(data, groupBy = "x", columns = list(x = colDef(details = function(i) i))),
               "`details` cannot be used on a grouping column")
})

test_that("filterable", {
  data <- data.frame(x = 1)

  tbl <- reactable(data)
  expect_equal(getAttrib(tbl, "filterable"), NULL)

  tbl <- reactable(data, filterable = TRUE)
  expect_equal(getAttrib(tbl, "filterable"), TRUE)

  tbl <- reactable(data, filterable = FALSE)
  expect_equal(getAttrib(tbl, "filterable"), NULL)

  expect_error(reactable(data, filterable = "true"), "`filterable` must be TRUE or FALSE")
})

test_that("searchable", {
  data <- data.frame(x = 1)

  tbl <- reactable(data)
  expect_equal(getAttrib(tbl, "searchable"), NULL)

  tbl <- reactable(data, searchable = TRUE)
  expect_equal(getAttrib(tbl, "searchable"), TRUE)

  tbl <- reactable(data, searchable = FALSE)
  expect_equal(getAttrib(tbl, "searchable"), NULL)

  expect_error(reactable(data, searchable = "true"), "`searchable` must be TRUE or FALSE")
})

test_that("searchMethod", {
  data <- data.frame(x = 1)

  tbl <- reactable(data)
  expect_equal(getAttrib(tbl, "searchMethod"), NULL)

  tbl <- reactable(data, searchMethod = JS("(rows, columnIds, filterValue) => rows"))
  expect_equal(getAttrib(tbl, "searchMethod"), JS("(rows, columnIds, filterValue) => rows"))

  expect_error(reactable(data, searchMethod = "rows => rows"), "`searchMethod` must be a JS function")
})

test_that("defaultColDef", {
  expect_error(reactable(data.frame(x = 1), defaultColDef = list()),
               "`defaultColDef` must be a column definition")

  # Defaults applied
  tbl <- reactable(
    data.frame(x = 1, y = "2"),
    defaultColDef = colDef(
      sortNALast = TRUE,
      html = TRUE,
      width = 22,
      rowHeader = TRUE
    ),
    columns = list(y = colDef(
      class = "cls"
    ))
  )
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]]$sortNALast, TRUE)
  expect_equal(attribs$columns[[2]]$sortNALast, TRUE)
  expect_equal(attribs$columns[[1]]$html, TRUE)
  expect_equal(attribs$columns[[2]]$html, TRUE)
  expect_equal(attribs$columns[[1]]$width, 22)
  expect_equal(attribs$columns[[2]]$width, 22)
  expect_equal(attribs$columns[[1]]$rowHeader, TRUE)
  expect_equal(attribs$columns[[2]]$rowHeader, TRUE)
  expect_equal(attribs$columns[[2]]$class, "cls")

  # Defaults can be overridden
  tbl <- reactable(
    data.frame(x = 1, y = "2"),
    defaultColDef = colDef(
      show = FALSE,
      sortNALast = TRUE,
      html = TRUE,
      na = "na",
      width = 22,
      class = "default-cls",
      rowHeader = TRUE,
      vAlign = "center"
    ),
    columns = list(y = colDef(
      show = TRUE,
      sortNALast = FALSE,
      html = FALSE,
      na = "",
      width = 44,
      rowHeader = FALSE,
      vAlign = "top"
    ))
  )
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]]$show, FALSE)
  expect_equal(attribs$columns[[2]]$show, TRUE)
  expect_equal(attribs$columns[[1]]$sortNALast, TRUE)
  expect_equal(attribs$columns[[2]]$sortNALast, FALSE)
  expect_equal(attribs$columns[[1]]$html, TRUE)
  expect_equal(attribs$columns[[2]]$html, FALSE)
  expect_equal(attribs$columns[[1]]$na, "na")
  expect_equal(attribs$columns[[2]]$na, "")
  expect_equal(attribs$columns[[1]]$width, 22)
  expect_equal(attribs$columns[[2]]$width, 44)
  expect_equal(attribs$columns[[1]]$class, "default-cls")
  expect_equal(attribs$columns[[2]]$class, "default-cls")
  expect_equal(attribs$columns[[1]]$rowHeader, TRUE)
  expect_equal(attribs$columns[[2]]$rowHeader, FALSE)
  expect_equal(attribs$columns[[1]]$vAlign, "center")
  expect_equal(attribs$columns[[2]]$vAlign, "top")

  # Defaults should apply to row name column
  tbl <- reactable(data.frame(x = 1, y = "2"),
                   defaultColDef = colDef(name = "not-row-names", show = FALSE),
                   rownames = TRUE)
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]]$show, FALSE)
  expect_equal(attribs$columns[[2]]$show, FALSE)
  expect_equal(attribs$columns[[3]]$show, FALSE)
  # Defaults don't override row name column defaults
  expect_equal(attribs$columns[[1]]$name, "")
  expect_equal(attribs$columns[[2]]$name, "not-row-names")
  expect_equal(attribs$columns[[3]]$name, "not-row-names")
})

test_that("defaultColGroup", {
  expect_error(reactable(data.frame(x = 1), defaultColGroup = list()),
               "`defaultColGroup` must be a column group definition")

  # Defaults applied
  tbl <- reactable(
    data.frame(x = 1, y = "2"),
    defaultColGroup = colGroup(
      name = "group",
      columns = "x",
      align = "left",
      headerClass = "cls",
      headerStyle = list(color = "blue")
    ),
    columnGroups = list(
      colGroup(columns = "x"),
      colGroup(name = "y", columns = "y")
    )
  )
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columnGroups[[1]]$name, "group")
  expect_equal(attribs$columnGroups[[1]]$columns, list("x"))
  expect_equal(attribs$columnGroups[[1]]$align, "left")
  expect_equal(attribs$columnGroups[[1]]$headerClass, "cls")
  expect_equal(attribs$columnGroups[[1]]$headerStyle, list(color = "blue"))
  expect_equal(attribs$columnGroups[[2]]$name, "y")
  expect_equal(attribs$columnGroups[[2]]$columns, list("y"))
  expect_equal(attribs$columnGroups[[2]]$align, "left")
  expect_equal(attribs$columnGroups[[2]]$headerClass, "cls")
  expect_equal(attribs$columnGroups[[2]]$headerStyle, list(color = "blue"))

  # Defaults can be overrided
  tbl <- reactable(
    data.frame(x = 1, y = "2"),
    defaultColGroup = colGroup(
      align = "left",
      headerClass = "cls",
      headerStyle = list(color = "blue")
    ),
    columnGroups = list(colGroup(
      name = "xy",
      columns = c("x", "y"),
      align = "right",
      headerClass = "xy",
      headerStyle = list(color = "red")
    ))
  )
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columnGroups[[1]]$align, "right")
  expect_equal(attribs$columnGroups[[1]]$headerClass, "xy")
  expect_equal(attribs$columnGroups[[1]]$headerStyle, list(color = "red"))

  # defaultColGroup should still be valid without column groups
  tbl <- reactable(data.frame(x = 1, y = "2"), defaultColGroup = colGroup(headerClass = "cls"))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columnGroups, NULL)
})

test_that("defaultSorted", {
  # Column overrides
  tbl <- reactable(data.frame(x = 1, y = "2"),
                   defaultSorted = c("x", "y"),
                   columns = list(y = colDef(defaultSortOrder = "desc")))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$defaultSorted, list(
    list(id = "x", desc = FALSE),
    list(id = "y", desc = TRUE)
  ))

  # Global defaults w/ column overrides
  tbl <- reactable(data.frame(x = 1, y = "2"),
                   defaultSorted = c("x", "y"),
                   defaultSortOrder = "desc",
                   columns = list(y = colDef(defaultSortOrder = "asc")))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$defaultSorted, list(
    list(id = "x", desc = TRUE),
    list(id = "y", desc = FALSE)
  ))

  # Explicit sort orders aren't overridden
  tbl <- reactable(data.frame(x = 1, y = "2"),
                   defaultSorted = list(x = "asc", y = "desc"),
                   defaultSortOrder = "desc",
                   columns = list(y = colDef(defaultSortOrder = "asc")))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$defaultSorted, list(
    list(id = "x", desc = FALSE),
    list(id = "y", desc = TRUE)
  ))

  # Works with defaultColDef
  tbl <- reactable(data.frame(x = 1, y = "2"),
                   defaultSorted = "x",
                   defaultColDef = colDef(defaultSortOrder = "desc"))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$defaultSorted, list(list(id = "x", desc = TRUE)))
})

test_that("pagination", {
  data <- data.frame(x = 1)

  tbl <- reactable(data)
  expect_equal(getAttrib(tbl, "pagination"), NULL)

  tbl <- reactable(data, pagination = TRUE)
  expect_equal(getAttrib(tbl, "pagination"), NULL)

  tbl <- reactable(data, pagination = FALSE)
  expect_equal(getAttrib(tbl, "pagination"), FALSE)

  expect_error(reactable(data, pagination = "false"), "`pagination` must be TRUE or FALSE")
})

test_that("showPageSizeOptions", {
  data <- data.frame(x = 1)

  tbl <- reactable(data)
  expect_equal(getAttrib(tbl, "showPageSizeOptions"), NULL)

  tbl <- reactable(data, showPageSizeOptions = TRUE)
  expect_equal(getAttrib(tbl, "showPageSizeOptions"), TRUE)

  tbl <- reactable(data, showPageSizeOptions = FALSE)
  expect_equal(getAttrib(tbl, "showPageSizeOptions"), FALSE)

  expect_error(reactable(data, showPageSizeOptions = "true"), "`showPageSizeOptions` must be TRUE or FALSE")
})

test_that("pageSizeOptions", {
  data <- data.frame(x = 1)

  tbl <- reactable(data)
  expect_equal(getAttrib(tbl, "pageSizeOptions"), NULL)

  tbl <- reactable(data, pageSizeOptions = c(10, 25, 50, 100))
  expect_equal(getAttrib(tbl, "pageSizeOptions"), list(10, 25, 50, 100))

  # Length-1 pageSizeOptions should be serialized as a JSON array
  tbl <- reactable(data,  pageSizeOptions = 5)
  expect_equal(getAttrib(tbl, "pageSizeOptions"), list(5))

  expect_error(reactable(data, pageSizeOptions = "1"), "`pageSizeOptions` must be numeric")
})

test_that("paginationType", {
  data <- data.frame(x = 1)

  tbl <- reactable(data)
  expect_equal(getAttrib(tbl, "paginationType"), NULL)

  for (type in c("numbers", "jump", "simple")) {
    tbl <- reactable(data, paginationType = type)
    expect_equal(getAttrib(tbl, "paginationType"), type)
  }

  expect_error(reactable(data, paginationType = "none"),
               '`paginationType` must be one of "numbers", "jump", "simple"')
})

test_that("defaultPageSize", {
  data <- data.frame(x = 1)

  tbl <- reactable(data)
  expect_equal(getAttrib(tbl, "defaultPageSize"), NULL)

  tbl <- reactable(data, defaultPageSize = 10)
  expect_equal(getAttrib(tbl, "defaultPageSize"), 10)

  tbl <- reactable(data, defaultPageSize = 1)
  expect_equal(getAttrib(tbl, "defaultPageSize"), 1)

  expect_error(reactable(data, defaultPageSize = "10"), "`defaultPageSize` must be numeric")
})

test_that("showPagination", {
  data <- data.frame(x = 1)

  tbl <- reactable(data)
  expect_equal(getAttrib(tbl, "showPagination"), NULL)

  tbl <- reactable(data, pagination = FALSE, showPagination = TRUE)
  expect_equal(getAttrib(tbl, "showPagination"), TRUE)

  tbl <- reactable(data, showPagination = FALSE)
  expect_equal(getAttrib(tbl, "showPagination"), FALSE)

  expect_error(reactable(data, showPagination = "false"), "`showPagination` must be TRUE or FALSE")
})

test_that("showPageInfo", {
  data <- data.frame(x = 1)

  tbl <- reactable(data)
  expect_equal(getAttrib(tbl, "showPageInfo"), NULL)

  tbl <- reactable(data, showPageInfo = TRUE)
  expect_equal(getAttrib(tbl, "showPageInfo"), TRUE)

  tbl <- reactable(data, showPageInfo = FALSE)
  expect_equal(getAttrib(tbl, "showPageInfo"), FALSE)

  expect_error(reactable(data, showPageInfo = "true"), "`showPageInfo` must be TRUE or FALSE")
})

test_that("minRows", {
  data <- data.frame(x = 1)

  tbl <- reactable(data)
  expect_equal(getAttrib(tbl, "minRows"), NULL)

  tbl <- reactable(data, minRows = 1)
  expect_equal(getAttrib(tbl, "minRows"), 1)

  tbl <- reactable(data, minRows = 10)
  expect_equal(getAttrib(tbl, "minRows"), 10)

  expect_error(reactable(data, minRows = "2"), "`minRows` must be numeric")
})

test_that("paginateSubRows", {
  data <- data.frame(x = 1)

  tbl <- reactable(data)
  expect_equal(getAttrib(tbl, "paginateSubRows"), NULL)

  tbl <- reactable(data, paginateSubRows = TRUE)
  expect_equal(getAttrib(tbl, "paginateSubRows"), TRUE)

  tbl <- reactable(data, paginateSubRows = FALSE)
  expect_equal(getAttrib(tbl, "paginateSubRows"), NULL)

  expect_error(reactable(data, paginateSubRows = "true"), "`paginateSubRows` must be TRUE or FALSE")
})

test_that("sub rows", {
  data <- data.frame(
    x = c(1, 2),
    .subRows = I(list(
      data.frame(x = c(3, 4)),
      NA
    ))
  )
  tbl <- reactable(data)
  columns <- getAttrib(tbl, "columns")
  expect_equal(length(columns), 1)
  expect_equal(columns[[1]]$id, "x")
})

test_that("column renderers", {
  data <- data.frame(
    x = c(1, 2),
    y = c("a", "b"),
    z = I(list(list(1,2,3), list(4,5,6)))
  )

  # Cell renderers
  tbl <- reactable(data, columns = list(
    x = colDef(cell = function(value) value + 1),
    y = colDef(cell = function(value, index, name) index),
    z = colDef(cell = function(values, index, name) paste(length(values), index, name))
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]]$cell, list("2", "3"))
  expect_equal(attribs$columns[[2]]$cell, list("1", "2"))
  expect_equal(attribs$columns[[3]]$cell, list("3 1 z", "3 2 z"))

  # POSIXlt objects should be handled correctly (mostly just on R <= 3.4)
  data$p <- c(as.POSIXlt("2019-01-01"), as.POSIXlt("2019-05-01"))
  tbl <- reactable(data, columns = list(
    p = colDef(cell = function(value) value)
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[4]]$cell, list("2019-01-01", "2019-05-01"))

  # Header renderers
  tbl <- reactable(data, columns = list(
    x = colDef(header = function(value) paste("header:", value)),
    y = colDef(header = function(value, name) paste("name:", name)),
    z = colDef(name = "ZZ", header = function(value) div(value))
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]][["header"]], "header: x")
  expect_equal(attribs$columns[[2]][["header"]], "name: y")
  expect_equal(attribs$columns[[3]][["header"]], div("ZZ"))

  tbl <- reactable(data, columns = list(
    x = colDef(header = div("header")),
    y = colDef(header = 123)
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]][["header"]], div("header"))
  expect_equal(attribs$columns[[2]][["header"]], "123")

  tbl <- reactable(data, columns = list(
    x = colDef(header = JS("colInfo => 'header'"), headerClass = "header"),
    y = colDef(headerClass = "no-header")
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]][["header"]], JS("colInfo => 'header'"))
  expect_equal(attribs$columns[[2]][["header"]], NULL)

  # Footer renderers
  tbl <- reactable(data, columns = list(
    x = colDef(footer = function(values) paste(values, collapse = " ")),
    y = colDef(footer = function(values, name) name),
    z = colDef(footer = function(values) paste(length(values), sum(unlist(values))))
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]][["footer"]], "1 2")
  expect_equal(attribs$columns[[2]][["footer"]], "y")
  expect_equal(attribs$columns[[3]][["footer"]], "2 21")

  tbl <- reactable(data, columns = list(
    x = colDef(footer = div("footer")),
    y = colDef(footer = 123)
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]][["footer"]], div("footer"))
  expect_equal(attribs$columns[[2]][["footer"]], "123")

  tbl <- reactable(data, columns = list(
    x = colDef(footer = JS("colInfo => 'footer'"), footerClass = "footer"),
    y = colDef(footerClass = "no-footer")
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]][["footer"]], JS("colInfo => 'footer'"))
  expect_equal(attribs$columns[[2]][["footer"]], NULL)

  # Group header renderers
  tbl <- reactable(data, columnGroups = list(
    colGroup(name = "x", header = div("group X"), columns = "x"),
    colGroup(name = "group Y", header = function(value) div("name:", value), columns = "y"),
    colGroup(name = "group Z", header = JS("colInfo => 'header'"), columns = "z")
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columnGroups[[1]][["header"]], div("group X"))
  expect_equal(attribs$columnGroups[[2]][["header"]], div("name:", "group Y"))
  expect_equal(attribs$columnGroups[[3]][["header"]], JS("colInfo => 'header'"))

  tbl <- reactable(data, columns = list(
    x = colDef(header = div("header")),
    y = colDef(header = 123)
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]][["header"]], div("header"))
  expect_equal(attribs$columns[[2]][["header"]], "123")

  tbl <- reactable(data, columns = list(
    x = colDef(header = JS("colInfo => 'header'"), headerClass = "header"),
    y = colDef(headerClass = "no-header")
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]][["header"]], JS("colInfo => 'header'"))
  expect_equal(attribs$columns[[2]][["header"]], NULL)
})

test_that("row details", {
  data <- data.frame(x = c(1, 2), y = c("a", "b"), stringsAsFactors = FALSE)

  # R renderer
  tbl <- reactable(data, details = function(i) if (i == 1) data[i, "y"])
  attribs <- getAttribs(tbl)
  expect_equal(
    attribs$columns[[1]],
    list(
      id = ".details", name = "", type = NULL, sortable = FALSE,
      resizable = FALSE, filterable = FALSE, searchable = FALSE, width = 45,
      align = "center", details = list("a", NULL)
    )
  )

  tbl <- reactable(
    data,
    defaultColDef = colDef(details = function(i, name) paste(i, name))
  )
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]]$details, list("1 x", "2 x"))

  # JS renderer
  tbl <- reactable(data, details = JS("rowInfo => rowInfo.y"))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]]$details, JS("rowInfo => rowInfo.y"))

  # List
  tbl <- reactable(data, details = list("a", "b"))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]]$details, list("a", "b"))

  # Custom details column definition
  tbl <- reactable(data, details = colDef(details = function(i) data[i, "y"],
                                          width = 125, class = "my-details"))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]]$details, list("a", "b"))
  expect_equal(attribs$columns[[1]]$width, 125)
  expect_equal(attribs$columns[[1]]$className, "my-details")

  # Details column definition works with defaultColDef
  tbl <- reactable(
    data,
    defaultColDef = colDef(headerClass = "header", sortable = TRUE, filterable = TRUE),
    details = colDef(details = function(i) data[i, "y"], class = "my-details")
  )
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]]$details, list("a", "b"))
  expect_equal(attribs$columns[[1]]$className, "my-details")
  expect_equal(attribs$columns[[1]]$headerClass, "header")
  expect_equal(attribs$columns[[1]]$sortable, FALSE)
  expect_equal(attribs$columns[[1]]$filterable, FALSE)

  # Column row details
  tbl <- reactable(data, columns = list(y = colDef(details = function(i) data[i, "y"])))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[2]]$details, list("a", "b"))

  # Details column can be part of column groups
  tbl <- reactable(data.frame(x = c(1, 2)), details = function(i) i, columnGroups = list(
    colGroup("group", c(".details", "x"))
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columnGroups[[1]]$columns, list(".details", "x"))

  expect_error(reactable(data, details = "details"), "`details` renderer must be an R function or JS function")
})

test_that("custom filter inputs", {
  data <- data.frame(x = c(1, 2), y = c("a", "b"), z = c("c", "d"), stringsAsFactors = FALSE)

  tbl <- reactable(data, columns = list(
    x = colDef(filterInput = JS("(filterValue, setFilter, column) => filterInput")),
    y = colDef(filterInput = function(values, name) {
      htmltools::tags$input(type = "text", class = paste(c(values, name), collapse = "_"))
    }),
    z = colDef(filterInput = '<input type="text">', html = TRUE)
  ))

  columns <- getAttrib(tbl, "columns")
  expect_equal(columns[[1]]$filterInput, JS("(filterValue, setFilter, column) => filterInput"))
  expect_equal(columns[[2]]$filterInput, htmltools::tags$input(type = "text", className = "a_b_y"))
  expect_equal(columns[[3]]$filterInput, '<input type="text">')
})

test_that("html dependencies from rendered content are passed through", {
  dep <- htmlDependency("dep", "0.1.0", "/path/to/dep")
  dep2 <- htmlDependency("dep2", "0.5.0", "/path/to/dep2")

  # Cell renderers
  data <- data.frame(x = c(1, 2), y = c("a", "b"))
  tbl <- reactable(data, columns = list(
    x = colDef(cell = function(value) {
      attachDependencies(div(value), dep)
    }),
    y = colDef(cell = function(value, index) {
      attachDependencies(tagList(value), dep2)
    })
  ))
  expect_equal(tbl$dependencies, list(dep, dep2))

  # Header renderers
  tbl <- reactable(data, columns = list(
    x = colDef(header = function(value) {
      attachDependencies(div("x"), dep)
    }),
    y = colDef(header = attachDependencies(tagList(), dep2))
  ))
  expect_equal(tbl$dependencies, list(dep, dep2))

  # Footer renderers
  tbl <- reactable(data, columnGroups = list(
    colGroup(columns = "x", header = function(value) attachDependencies(div("x"), dep)),
    colGroup(columns = "y", header = attachDependencies(tagList(), dep2))
  ))
  expect_equal(tbl$dependencies, list(dep, dep2))

  # Group header renderers
  tbl <- reactable(data, columns = list(
    x = colDef(footer = function(values) {
      attachDependencies(div("x"), dep)
    }),
    y = colDef(footer = attachDependencies(tagList(), dep2))
  ))
  expect_equal(tbl$dependencies, list(dep, dep2))

  # Details column definition works with defaultColDef
  tbl <- reactable(
    data,
    details = colDef(details = function(i) {
      attachDependencies(div(i), dep2)
    }),
    columns = list(
      x = colDef(details = function(i) {
        attachDependencies(tagList(i), dep)
      })
    )
  )
  expect_equal(tbl$dependencies, list(dep2, dep))

  # No dependencies
  tbl <- reactable(data)
  expect_equal(tbl$dependencies, list())
})

test_that("row selection", {
  data <- data.frame(x = c(1, 2, 3))
  tbl <- reactable(data, selection = "single")
  attribs <- getAttribs(tbl)
  expect_equal(attribs$selection, "single")
  expect_equal(attribs$columns[[1]], list(
    id = ".selection", name = "", type = NULL, resizable = FALSE,
    width = 45, selectable = TRUE
  ))

  tbl <- reactable(data, selection = "multiple", defaultSelected = c(1, 3, 2))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$selection, "multiple")
  expect_equal(attribs$defaultSelected, list(0, 2, 1))

  # defaultSelected should be serialized as an array
  tbl <- reactable(data, selection = "single", defaultSelected = 3)
  attribs <- getAttribs(tbl)
  expect_equal(attribs$defaultSelected, list(2))

  # Selection column should be customizable
  tbl <- reactable(data, selection = "single", columns = list(
    .selection = colDef(width = 100, class = "my-cls", resizable = TRUE)
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]], list(
    id = ".selection", name = "", type = NULL, resizable = TRUE,
    width = 100, selectable = TRUE, className = "my-cls"
  ))

  # Selection column can be part of column groups
  tbl <- reactable(data.frame(x = c(1, 2)), selection = "multiple", columnGroups = list(
    colGroup("group", c(".selection", "x"))
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columnGroups[[1]]$columns, list(".selection", "x"))

  expect_error(reactable(data, selection = "none"), '`selection` must be "multiple" or "single"')

  # Out of bounds errors
  expect_error(reactable(data, selection = "single", defaultSelected = c(0, 1)))
  expect_error(reactable(data, selection = "multiple", defaultSelected = c(2, 4)))
})

test_that("selectionId", {
  expect_warning(
    reactable(data.frame(x = 1), selectionId = "selected"),
    "`selectionId` is deprecated."
  )
})

test_that("onClick", {
  data <- data.frame(x = c(1, 2), y = c("a", "b"))
  tbl <- reactable(data, onClick = "expand")
  attribs <- getAttribs(tbl)
  expect_equal(attribs$onClick, "expand")

  tbl <- reactable(data, onClick = "select")
  attribs <- getAttribs(tbl)
  expect_equal(attribs$onClick, "select")

  tbl <- reactable(data, onClick = JS("(rowInfo, column, state) => {}"))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$onClick, JS("(rowInfo, column, state) => {}"))
})

test_that("column class functions", {
  data <- data.frame(
    x = c("a", "b", "c"),
    y = c(2, 4, 6),
    z = I(list(list(1,2,3), list(4,5,6), list(0,0,0)))
  )

  # R functions
  tbl <- reactable(data, columns = list(
    x = colDef(class = function(value) if (value != "a") paste0(value, "-cls")),
    y = colDef(class = function(value, index, name) sprintf("%s-%s-%s", value, index, name)),
    z = colDef(class = function(values, index) sprintf("%s-%s", length(values), index))
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]]$className, list(NULL, "b-cls", "c-cls"))
  expect_equal(attribs$columns[[2]]$className, list("2-1-y", "4-2-y", "6-3-y"))
  expect_equal(attribs$columns[[3]]$className, list("3-1", "3-2", "3-3"))

  # POSIXlt objects should be handled correctly (mostly just on R <= 3.4)
  data$p <- c(as.POSIXlt("2019-01-01"), as.POSIXlt("2019-05-01"), as.POSIXlt("2019-07-05"))
  tbl <- reactable(data, columns = list(
    p = colDef(class = function(value) value)
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[4]]$className,
               # Use as.list(c(...)) instead of list() to preserve tzone attributes
               as.list(c(as.POSIXlt("2019-01-01"), as.POSIXlt("2019-05-01"), as.POSIXlt("2019-07-05"))))

  # JS functions
  tbl <- reactable(data, columns = list(
    x = colDef(class = JS("rowInfo => 'cls'"))
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]]$className, JS("rowInfo => 'cls'"))
})

test_that("column style functions", {
  data <- data.frame(
    x = c("a", "b", "c"),
    y = c(2, 4, 6),
    z = I(list(list(1,2,3), list(4,5,6), list(0,0,0)))
  )

  # R functions
  tbl <- reactable(data, columns = list(
    x = colDef(style = function(value) if (value != "a") "background-color: red"),
    y = colDef(style = function(value, index, name) list(color = sprintf("%s-%s-%s", value, index, name))),
    z = colDef(style = function(values, index) list(content = sprintf("%s-%s", length(values), index)))
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]]$style,
               list(NULL, list("background-color" = "red"), list("background-color" = "red")))
  expect_equal(attribs$columns[[2]]$style,
               list(list(color = "2-1-y"), list(color = "4-2-y"), list(color = "6-3-y")))
  expect_equal(attribs$columns[[3]]$style,
               list(list(content = "3-1"), list(content = "3-2"), list(content = "3-3")))

  # POSIXlt objects should be handled correctly (mostly just on R <= 3.4)
  data$p <- c(as.POSIXlt("2019-01-01"), as.POSIXlt("2019-05-01"), as.POSIXlt("2019-07-05"))
  tbl <- reactable(data, columns = list(
    p = colDef(style = function(value) value)
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[4]]$style,
               # Use as.list(c(...)) instead of list() to preserve tzone attributes
               as.list(c(as.POSIXlt("2019-01-01"), as.POSIXlt("2019-05-01"), as.POSIXlt("2019-07-05"))))

  # JS functions
  tbl <- reactable(data, columns = list(
    x = colDef(style = JS("rowInfo => ({ backgroundColor: 'red' })"))
  ))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$columns[[1]]$style, JS("rowInfo => ({ backgroundColor: 'red' })"))
})

test_that("rowClass and rowStyle", {
  # rowClass
  tbl <- reactable(data.frame(x = 1), rowClass = "cls")
  attribs <- getAttribs(tbl)
  expect_equal(attribs$rowClassName, "cls")

  tbl <- reactable(data.frame(x = 1), rowClass = JS("(rowInfo, state) => 'cls'"))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$rowClassName, JS("(rowInfo, state) => 'cls'"))

  tbl <- reactable(data.frame(x = c("a", "b", "c")),
                   rowClass = function(index) paste0("row-", index))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$rowClassName, list("row-1", "row-2", "row-3"))

  # rowStyle
  tbl <- reactable(data.frame(x = 1), rowStyle = " border-bottom: 1px solid; top: 50px")
  attribs <- getAttribs(tbl)
  expect_equal(attribs$rowStyle, list("border-bottom" = "1px solid", top = "50px"))

  tbl <- reactable(data.frame(x = 1), rowStyle = list(color = "red"))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$rowStyle, list(color = "red"))

  tbl <- reactable(data.frame(x = 1), rowStyle = JS("(rowInfo, state) => ({ color: 'red' })"))
  attribs <- getAttribs(tbl)
  expect_equal(attribs$rowStyle, JS("(rowInfo, state) => ({ color: 'red' })"))

  tbl <- reactable(
    data.frame(x = c("a", "b", "c")),
    rowStyle = function(index) {
      if (index == 1) "background-color: red;"
      else if (index == 2) list(color = "red")
    }
  )
  attribs <- getAttribs(tbl)
  expect_equal(attribs$rowStyle,
               list(list("background-color" = "red"), list(color = "red"), NULL))
})

test_that("width, height, and sizingPolicy", {
  data <- data.frame(x = 1)
  expect_error(reactable(data, width = "not a valid CSS unit"))
  expect_error(reactable(data, height = "not a valid CSS unit"))

  tbl <- reactable(data)
  expect_equal(tbl$width, NULL)
  expect_equal(tbl$height, NULL)

  tbl <- reactable(data, width = "400px", height = "100%")
  expect_equal(tbl$width, "400px")
  expect_equal(tbl$height, "100%")

  expect_equal(tbl$sizingPolicy$knitr$figure, FALSE)
  expect_equal(tbl$sizingPolicy$defaultWidth, "auto")
  expect_equal(tbl$sizingPolicy$defaultHeight, "auto")
})

test_that("theme", {
  data <- data.frame(x = 1)

  tbl <- reactable(data, theme = reactableTheme())
  attribs <- getAttribs(tbl)
  expect_equal(attribs$theme, NULL)

  theme <- reactableTheme(style = list(color = "red"), borderColor = "#555")
  tbl <- reactable(data, theme = theme)
  attribs <- getAttribs(tbl)
  expect_equal(attribs$theme, theme)

  # Theme function
  theme <- function() reactableTheme(cellPadding = 13)
  tbl <- reactable(data, theme = theme)
  attribs <- getAttribs(tbl)
  expect_equal(attribs$theme, theme())

  theme <- function() NULL
  tbl <- reactable(data, theme = theme)
  attribs <- getAttribs(tbl)
  expect_equal(attribs$theme, NULL)

  # Global theme option
  theme <- reactableTheme(style = list(color = "red"), borderColor = "#555")
  old <- options(reactable.theme = theme)
  on.exit(options(old))
  tbl <- reactable(data)
  attribs <- getAttribs(tbl)
  expect_equal(attribs$theme, theme)

  # Table theme should override global option
  theme <- reactableTheme(borderWidth = "3px")
  tbl <- reactable(data, theme = theme)
  attribs <- getAttribs(tbl)
  expect_equal(attribs$theme, theme)

  # Errors
  expect_error(reactable(data, theme = list()),
               "`theme` must be a reactable theme object")
})

test_that("language", {
  data <- data.frame(x = 1)

  tbl <- reactable(data, language = reactableLang())
  attribs <- getAttribs(tbl)
  expect_equal(attribs$language, NULL)

  language <- reactableLang(pageNext = "_Next", searchPlaceholder = "_Search", sortLabel = "_Sort {name}")
  tbl <- reactable(data, language = language)
  attribs <- getAttribs(tbl)
  expect_equal(attribs$language, language)

  # Global language option
  old <- options(reactable.language = language)
  on.exit(options(old))
  tbl <- reactable(data)
  attribs <- getAttribs(tbl)
  expect_equal(attribs$language, language)

  # Table language should override global option
  language <- reactableLang(selectAllRowsLabel = "_Select all rows")
  tbl <- reactable(data, language = language)
  attribs <- getAttribs(tbl)
  expect_equal(attribs$language, language)

  # Errors
  expect_error(reactable(data, language = list()),
               "`language` must be a reactable language options object")
})

test_that("meta", {
  data <- data.frame(x = 1)

  expect_error(reactable(data, meta = "meta"), "`meta` must be a named list")

  tbl <- reactable(data)
  expect_equal(getAttrib(tbl, "meta"), NULL)

  # Empty lists should be omitted, not serialized as an empty array through jsonlite
  tbl <- reactable(data, meta = list())
  expect_equal(getAttrib(tbl, "meta"), NULL)

  meta <- list(
    number = 30,
    str = "str",
    df = data.frame(y = 2),
    func = JS("value => value > 30"),
    na = NA,
    naInteger = NA_integer_,
    null = NULL,
    date = as.POSIXct("2019-01-02 3:22:15"),
    array = c(2, 4, 6),
    arrayLength1 = 1,
    list = list(1),
    emptyList = list()  # will be serialized as []
  )
  tbl <- reactable(data, meta = meta)
  expect_equal(getAttrib(tbl, "meta"), meta)
})

test_that("elementId", {
  data <- data.frame(x = 1)

  tbl <- reactable(data)
  expect_equal(tbl$elementId, NULL)
  expect_equal(getAttribs(tbl)$elementId, NULL)

  tbl <- reactable(data, elementId = "my-tbl")
  expect_equal(tbl$elementId, "my-tbl")
  expect_equal(getAttribs(tbl)$elementId, "my-tbl")
})

test_that("columnSortDefs", {
  defaultSorted <- list(x = "asc", y = "desc")
  expected <- list(list(id = "x", desc = FALSE), list(id = "y", desc = TRUE))
  expect_equal(columnSortDefs(defaultSorted), expected)
})

test_that("reactableOutput", {
  output <- reactableOutput("mytbl")

  # HTML dependencies should be intact
  deps <- htmltools::htmlDependencies(output)
  expect_true(length(deps) > 0)

  # Output container should have data-reactable-output ID set
  expect_equal(output[[1]]$attribs[["data-reactable-output"]], "mytbl")
  expect_equal(output[[1]]$name, "div")
  expect_equal(output[[1]]$attribs$id, "mytbl")
})

test_that("reactable_html", {
  html <- reactable_html("id", "color: red", "class")
  expect_equal(html, htmltools::tags$div(id = "id", class = "class", style = "color: red", reactDependencies()))

  # Text color should be set in R Notebooks
  old <- options(rstudio.notebook.executing = TRUE)
  on.exit(options(old))
  html <- reactable_html(NULL, "color: red", NULL)
  expect_equal(html, htmltools::tags$div(style = "color: #333;color: red", reactDependencies()))
})

test_that("reactable.yaml widget dependencies are included with correct version", {
  deps <- htmlwidgets::getDependency("reactable")
  reactableDep <- Find(function(x) x$name == "reactable", deps)
  expect_true(!is.null(reactableDep))
  # Keep widget dependency version in sync with package version
  expect_equal(package_version(reactableDep$version), packageVersion("reactable"))
  # Ensure reactable.css is included and named correctly, as it's also used as
  # the insertion point for theme style injection
  expect_equal(reactableDep$stylesheet, "reactable.css")
})

# Static rendering (experimental)
test_that("static rendering", {
  # Skip static rendering tests if V8 was built against a version of libv8 without
  # ICU / i18n support. If running on Fedora, this is an issue with Fedora's V8 package
  # that was fixed in Fedora 37. The workaround on Fedora <= 36 is to build the V8
  # R package using DOWNLOAD_STATIC_LIBV8=true.
  # https://github.com/jeroen/V8/issues/65
  # https://bugzilla.redhat.com/show_bug.cgi?id=1755114
  #
  # On CRAN's fedora-clang platform, where V8 is installed with g++ and not clang, all
  # uncaught exceptions cause V8 to seg fault, which causes this check and other error handling
  # tests to fail. Skip all tests if running on a platform with this issue.
  # See notes on CRAN's fedora-clang platform: https://www.stats.ox.ac.uk/pub/bdr/Rconfig/r-devel-linux-x86_64-fedora-clang
  #
  # The test for the uncaught exception seg fault must be run in a separate R process.
  # If a seg fault occurs, the output is typically: "free(): double free detected in tcache 2"
  code <- "V8::new_context()$eval('not_defined')"
  output <- suppressWarnings(
    system2(R.home("bin/R"), c("-e", shQuote(code)), stdout = TRUE, stderr = TRUE)
  )
  if (attr(output, "status") > 0 && !grepl("ReferenceError", paste(output, collapse = "\n"))) {
    skip(sprintf("V8 crashed on an uncaught exception, unable to run error handling tests.\nThis is probably V8 installed with g++ on fedora-clang. Test output:\n%s", paste(output, collapse = "\n")))
  }

  tryCatch({
    V8::new_context()$eval("''.localeCompare('')")
  }, error = function(e) {
    # If libv8 lacks i18n support, the error is typically: "RangeError: Internal error: Icu error."
    skip("V8 was built against a version of libv8 without i18n support")
  })

  data <- data.frame(
    x = c(1, 2),
    y = c("a", "column-y-cell"),
    z = I(list(list(1,2,3), list(4,5,6))),
    stringsAsFactors = FALSE
  )

  expect_error(reactable(data, static = "true"), "`static` must be TRUE or FALSE")

  # expect_snapshot() uses capture.output(), which forces UTF-8 characters to latin-1 on
  # Windows R <= 4.1 or other non-UTF-8 platforms. expect_snapshot_value() is harder to read
  # but handles UTF-8 characters.
  # For any errors like `Error: lexical error: invalid char in json text.`, delete the snapshot
  # entirely and regenerate.
  expect_snapshot_html_with_utf8 <- function(x) expect_snapshot_value(as.character(x))

  tbl <- reactable(
    data,
    static = TRUE,
    elementId = "stable-id-static-rendering"
  )
  rendered <- htmltools::renderTags(tbl)
  expect_true(grepl("data-react-ssr", rendered$html))
  expect_true(grepl(">column-y-cell<", rendered$html))
  expect_snapshot(cat(rendered$html))

  # JS evals should always be serialized as an array
  expect_true(grepl('"evals":[]', rendered$html, fixed = TRUE))

  # Themes critical CSS should be included in <head>
  tbl <- reactable(
    data,
    static = TRUE,
    theme = reactableTheme(color = "blue", borderWidth = 3),
    elementId = "stable-id-theme-critical-css"
  )
  rendered <- htmltools::renderTags(tbl)
  styleDep <- Find(
    function(dep) !is.null(dep$head) && grepl('<style data-emotion="reactable .+color:blue;', dep$head),
    rendered$dependencies
  )
  expect_true(!is.null(styleDep))
  expect_snapshot(cat(styleDep$head))
  expect_snapshot(cat(rendered$html))

  # Custom render functions and JS evals should work
  tbl <- reactable(
    data,
    columns = list(
      x = colDef(cell = JS("cellInfo => `js-rendered_${cellInfo.value}_`")),
      y = colDef(cell = function(value) htmltools::tags$b(value))
    ),
    static = TRUE,
    elementId = "stable-id-custom-js-evals"
  )
  rendered <- htmltools::renderTags(tbl)
  expect_true(grepl("js-rendered_2_", rendered$html))
  expect_true(grepl("<b>column-y-cell</b>", rendered$html))
  expect_snapshot(cat(rendered$html))

  # Custom render functions and JS evals that call React externally should work
  tbl <- reactable(
    data,
    columns = list(
      y = colDef(cell = JS("cellInfo => React.createElement('b', null, cellInfo.value)"))
    ),
    static = TRUE,
    elementId = "stable-id-external-React"
  )
  rendered <- htmltools::renderTags(tbl)
  expect_true(grepl("<b>column-y-cell</b>", rendered$html))
  expect_snapshot(cat(rendered$html))

  # Known limitation: default expanded rows with defaultExpanded = TRUE is not currently supported.
  # Note: add a test for nested tables if it ever does become supported.
  tbl <- reactable(
    data.frame(x = 1),
    details = function(i) "row-details",
    static = TRUE,
    elementId = "stable-id-row-details"
  )
  rendered <- htmltools::renderTags(tbl)
  expect_false(grepl(">row-details<", rendered$html))
  expect_snapshot_html_with_utf8(rendered$html)

  # Embedded HTML widgets' root elements and widget scripts should not be statically rendered
  tbl <- reactable(
    data.frame(x = 1),
    columns = list(x = colDef(cell = function() {
      sparkline::sparkline(c(1, 2), elementId = "stable-id-sparkline")
    })),
    static = TRUE,
    elementId = "stable-id-html-widgets"
  )
  rendered <- htmltools::renderTags(tbl)
  expect_false(grepl("sparkline", strsplit(rendered$html, "<script")[[1]][[1]]))
  expect_equal(lengths(gregexpr("<script ", rendered$html)), 1) # Should only be one <script> for reactable
  expect_snapshot(cat(rendered$html))

  # Column formatting features that depend on Intl polyfills should work
  formatData <- data.frame(
    str = "str",
    pct = 0.75,
    # Test for floating-point precision issues: should not be formatted as 52.90000000000001%
    pct_digits = 0.529,
    currency_USD = 10,
    currency_EUR = 11.123,
    date = as.POSIXct("2019-05-06 3:22:15", tz = "UTC"),
    time = as.POSIXct("2019-05-06 3:22:15", tz = "UTC"),
    num = 1234.1234,
    locale_hi_IN = 1234567.4,
    stringsAsFactors = FALSE
  )
  tbl <- reactable(
    formatData,
    columns = list(
      str = colDef(format = colFormat(prefix = "pre_", suffix = "_suffix")),
      pct = colDef(format = colFormat(percent = TRUE)),
      pct_digits = colDef(format = colFormat(percent = TRUE)),
      currency_USD = colDef(format = colFormat(currency = "USD")),
      currency_EUR = colDef(format = colFormat(currency = "EUR")),
      date = colDef(format = colFormat(datetime = TRUE, prefix = "_date_", suffix = "_date_")),
      time = colDef(format = colFormat(time = TRUE, prefix = "_time_", suffix = "_time_")),
      num = colDef(format = colFormat(digits = 1, separators = TRUE)),
      # Current limitation: locales other than "en" aren't supported for now
      locale_hi_IN = colDef(format = colFormat(locales = "hi-IN", currency = "INR", separators = TRUE))
    ),
    static = TRUE,
    elementId = "stable-id-formatting-intl-polyfills"
  )
  rendered <- htmltools::renderTags(tbl)
  html <- rendered$html
  expect_true(grepl("pre_str_suffix", html, fixed = TRUE))
  expect_true(grepl(">75%<", html, fixed = TRUE))
  expect_true(grepl(">52.9%<", html, fixed = TRUE))
  expect_true(grepl(">$10.00<", html, fixed = TRUE))
  expect_true(grepl(">€11.12<", html, fixed = TRUE))
  expect_true(grepl(">1,234.1<", html, fixed = TRUE))
  expect_true(grepl(">₹1,234,567.40<", html, fixed = TRUE))
  # Date/time formatting depends on the local timezone, which can't easily be controlled in tests
  expect_false(grepl("_date_2019-05-06T03:22:15Z_date_", html))
  html <- sub(">_date_.+_date_<", ">_date_replaced_date_<", html)
  expect_false(grepl("_time_2019-05-06T03:22:15Z_time_", html))
  html <- sub(">_time_.+_time_<", ">_time_replaced_time_<", html)
  expect_snapshot_html_with_utf8(html)

  # Should fall back to client-side rendering when there are rendering errors
  tbl <- reactable(
    data,
    columns = list(x = colDef(cell = JS("throw new Error('error rendering JS')"))),
    static = TRUE,
    elementId = "stable-id-CSR-fallback"
  )
  expect_warning({ rendered <- htmltools::renderTags(tbl) }, "Failed to render table to static HTML:\nError: error rendering JS")
  expect_false(grepl("data-react-ssr", rendered$html))
  expect_false(grepl(">column-y-cell<", rendered$html))
  expect_snapshot(cat(rendered$html))

  # Custom knit_print method should work
  tbl <- reactable(data, static = TRUE)
  output <- knitr::knit_print(tbl, options = list(screenshot.force = FALSE))
  expect_true(grepl("data-react-ssr", output))

  tbl <- reactable(data)
  output <- knitr::knit_print(tbl, options = list(screenshot.force = FALSE))
  expect_false(grepl("data-react-ssr", output))
})

---
title: NBA Box Score
output: html_document
resource_files:
  - '.'
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)

options(reactable.static = TRUE)
```

```{css ref.label='styles', echo=FALSE}
```

```{r ref.label='font_styles', echo=FALSE}
```

```{css ref.label='pkgdown_styles', echo=FALSE}
```

```{r box_score, echo=FALSE}
library(reactable)
library(htmltools)

player_stats <- read.csv("player_stats.csv", stringsAsFactors = FALSE)
team_stats <- read.csv("team_stats.csv", stringsAsFactors = FALSE)
line_score <- read.csv("line_score.csv", stringsAsFactors = FALSE)

line_score <- line_score[, c("TEAM_ID", "TEAM_CITY_NAME", "TEAM_NICKNAME", "TEAM_WINS_LOSSES",
                             "PTS_QTR1", "PTS_QTR2", "PTS_QTR3", "PTS_QTR4", "PTS")]

line_score_tbl <- reactable(
  line_score,
  sortable = FALSE,
  defaultColDef = colDef(headerClass = "line-score-header", align = "center", minWidth = 50),
  columns = list(
    TEAM_ID = colDef(show = FALSE),
    TEAM_CITY_NAME = colDef(
      name = "",
      align = "left",
      minWidth = 250,
      cell = function(value, index) {
        team_url <- sprintf("https://stats.nba.com/team/%s/traditional", line_score[index, "TEAM_ID"])
        team_name <- paste(value, line_score[index, "TEAM_NICKNAME"])
        team_record <- line_score[index, "TEAM_WINS_LOSSES"]
        tagList(
          tags$a(class = "team-name", href = team_url, target = "_blank", team_name),
          span(class = "team-record", team_record)
        )
      }
    ),
    TEAM_NICKNAME = colDef(show = FALSE),
    TEAM_WINS_LOSSES = colDef(show = FALSE),
    PTS_QTR1 = colDef(name = "Q1"),
    PTS_QTR2 = colDef(name = "Q2"),
    PTS_QTR3 = colDef(name = "Q3"),
    PTS_QTR4 = colDef(name = "Q4"),
    PTS = colDef(name = "Final", class = "line-score-final")
  ),
  class = "line-score-tbl"
)

box_score_tbl <- function(player_stats, team_stats, team) {
  # Convert M:SS strings to datetimes for proper sorting
  player_stats$MIN_STR <- player_stats$MIN
  player_stats$MIN <- strptime(player_stats$MIN, format = "%M:%S")

  cols <- c("PLAYER_ID", "PLAYER_NAME", "START_POSITION", "MIN", "MIN_STR",
          "FGM", "FGA", "FG_PCT", "FG3M", "FG3A", "FG3_PCT", "FTM", "FTA",
          "FT_PCT", "OREB", "DREB", "REB", "AST", "STL", "BLK", "TO", "PF",
          "PTS", "PLUS_MINUS")
  stats <- player_stats[player_stats$TEAM_ABBREVIATION == team, cols]
  team_stats <- team_stats[team_stats$TEAM_ABBREVIATION == team, ]

  reactable(
    stats,
    pagination = FALSE,
    defaultSortOrder = "desc",
    defaultSorted = "PTS",
    defaultColDef = colDef(
      sortNALast = TRUE,
      minWidth = 45,
      class = JS("function(rowInfo, column, state) {
        // Highlight sorted columns
        for (let i = 0; i < state.sorted.length; i++) {
          if (state.sorted[i].id === column.id) {
            return 'sorted'
          }
        }
      }"),
      headerClass = "box-score-header",
      footer = function(values, name) {
        value <- team_stats[[name]]
        # Format shots made-attempted
        if (name %in% c("FGM", "FG3M", "FTM")) {
          attempted_name <- c(FGM = "FGA", FG3M = "FG3A", FTM = "FTA")[name]
          value <- sprintf("%s-%s", value, team_stats[[attempted_name]])
        }
        # Format percentages
        if (name %in% c("FG_PCT", "FG3_PCT", "FT_PCT")) {
          value <- paste0(value * 100, "%")
        }
        # Format +/-
        if (name == "PLUS_MINUS") {
          value <- sprintf("%+d", value)
        }
        value
      }
    ),
    columns = list(
      PLAYER_ID = colDef(show = FALSE),
      PLAYER_NAME = colDef(
        name = "Player",
        defaultSortOrder = "asc",
        width = 130,
        cell = function(value, index) {
          player_id <- stats[index, "PLAYER_ID"]
          player_url <- sprintf("https://stats.nba.com/player/%s", player_id)
          start_position <- stats[index, "START_POSITION"]
          if (start_position != "") {
            value <- tagList(value, " ", tags$sup(start_position))
          }
          tags$a(href = player_url, target = "_blank", value)
        },
        footer = span(class = "box-score-total", "Totals")
      ),
      START_POSITION = colDef(show = FALSE),
      MIN = colDef(name = "Min", minWidth = 60, align = "right", cell = function(value, index) {
        if (!is.na(value)) stats[index, "MIN_STR"] else "DNP"
      }),
      MIN_STR = colDef(show = FALSE),
      FGM = colDef(name = "FG", minWidth = 55, cell = function(value, index) {
        if (!is.na(value)) sprintf("%s-%s", value, stats[index, "FGA"])
      }),
      FGA = colDef(show = FALSE),
      FG_PCT = colDef(name = "FG%", minWidth = 55, format = colFormat(percent = TRUE)),
      FG3M = colDef(name = "3P", minWidth = 55, cell = function(value, index) {
        if (!is.na(value)) sprintf("%s-%s", value, stats[index, "FG3A"])
      }),
      FG3A = colDef(name = "3PA", show = FALSE),
      FG3_PCT = colDef(name = "3P%", minWidth = 55, format = colFormat(percent = TRUE)),
      FTM = colDef(name = "FT", minWidth = 55, cell = function(value, index) {
        if (!is.na(value)) sprintf("%s-%s", value, stats[index, "FTA"])
      }),
      FTA = colDef(show = FALSE),
      FT_PCT = colDef(name = "FT%", minWidth = 55, format = colFormat(percent = TRUE)),
      OREB = colDef(name = "ORB"),
      DREB = colDef(name = "DRB"),
      PLUS_MINUS = colDef(name = "+/-", cell = function(value) {
        if (is.na(value)) "" else sprintf("%+d", value)
      })
    ),
    showSortIcon = FALSE,
    highlight = TRUE,
    striped = TRUE,
    class = "box-score-tbl",
    theme = reactableTheme(cellPadding = "8px")
  )
}

div(class = "box-score",
  h2(class = "header", "Raptors vs. Warriors:",
     tags$a(class = "game-date", href="https://stats.nba.com/game/0041800403", target = "_blank", "Jun 5, 2019")),

  div(class = "line-score", line_score_tbl),

  div(class = "box-score-title", "Toronto Raptors"),
  box_score_tbl(player_stats, team_stats, "TOR"),

  div(class = "box-score-title", "Golden State Warriors"),
  box_score_tbl(player_stats, team_stats, "GSW")
)
```

---

Source: [NBA.com](https://stats.nba.com/game/0041800403)

Raw data: [`line_score.csv`](line_score.csv), [`player_stats.csv`](player_stats.csv), [`team_stats.csv`](team_stats.csv)

<h2>Source Code</h2>

```{r ref.label="box_score", eval=FALSE}
```

```{r font_styles, eval=FALSE}
htmltools::tags$link(href = "https://fonts.googleapis.com/css?family=Roboto:400,500&display=fallback", rel = "stylesheet")
```

```{css styles, eval=FALSE}
.box-score {
  font-family: 'Roboto', Helvetica, Arial, sans-serif;
}

.box-score a {
  color: #337ab7;
  text-decoration: none;
}

.box-score a:hover,
.box-score a:focus {
  text-decoration: underline;
  text-decoration-thickness: max(1px, 0.0625rem);
}

.header {
  text-align: center;
  font-size: 1.25rem;
}

.game-date {
  font-size: 1rem;
}

.line-score {
  margin-top: 1.5rem;
  text-align: center;
}

.line-score-tbl {
  margin: 0 auto;
  max-width: 32rem;
  font-size: 0.9375rem;
}

.line-score-header {
  font-size: 0.8125rem;
  font-weight: 400;
}

.line-score-final {
  font-weight: 500;
}

.team-name {
  font-weight: 500;
}

.team-record {
  margin-left: 0.375rem;
  color: hsl(0, 0%, 45%);
  font-size: 0.75rem;
}

.box-score-title {
  margin-top: 1.5rem;
  padding: 0.5rem;
  background-color: hsl(205, 100%, 36%);
  color: hsl(0, 0%, 98%);
  font-size: 0.9375rem;
  font-weight: 400;
}

.box-score-tbl {
  font-size: 0.75rem;
  letter-spacing: 0.2px;
}

.box-score-header {
  border-bottom-width: 1px;
  background-color: hsl(205, 93%, 16%);
  color: hsl(0, 0%, 98%);
  font-weight: 400;
  font-size: 0.7rem;
  text-transform: uppercase;
  transition: box-shadow 0.3s cubic-bezier(0.175, 0.885, 0.32, 1.275);
}

.box-score-header:hover,
.box-score-header:focus,
.box-score-header[aria-sort="ascending"],
.box-score-header[aria-sort="descending"] {
  background-color: hsl(205, 100%, 36%);
}

.box-score-header[aria-sort="ascending"] {
  box-shadow: inset 0 10px 0 -6px #efaa10;
}

.box-score-header[aria-sort="descending"] {
  box-shadow: inset 0 -10px 0 -6px #efaa10;
}

.sorted {
  background-color: hsla(0, 0%, 60%, 0.1);
}

.box-score-total {
  font-size: 0.8125rem;
  font-weight: 500;
}
```

```{css pkgdown_styles, include=FALSE}
/* rmarkdown html documents */
.main-container {
  max-width: 1024px !important;
}

/* pkgdown articles */
.row > main {
  max-width: 1024px;
}
```

---
title: "Extra Custom Filtering Examples"
output:
  html_document:
    toc: true
    toc_float:
      smooth_scroll: false
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)

library(reactable)
```

More examples of [custom filtering](./custom-filtering.html). Note that the
examples on this page do not support IE 11.

## Material UI Range Filter

Range filter using [Material UI's Slider component](https://mui.com/material-ui/react-slider/),
adapted from an [example by @timelyportfolio](https://github.com/glin/reactable/issues/9#issuecomment-1008424287).

```{js}
// JavaScript code for the Material UI range filter. This is a js language
// chunk, but you could also inline this in R using `tags$script(HTML("..."))`

const muiRangeFilter = (column, state) => {
  // Get min and max values from raw table data and memoize so it doesn't
  // have to be recalculated each time
  const range = React.useMemo(() => {
    let min = Infinity
    let max = -Infinity
    state.data.forEach(row => {
      const value = row[column.id]
      if (value < min) {
        min = Math.floor(value)
      } else if (value > max) {
        max = Math.ceil(value)
      }
    })
    return [min, max]
  }, [state.data])

  const value = column.filterValue ? column.filterValue : range
  const valueLabel = `${value[0]}...${value[1]}`

  return React.createElement(
    'div',
    // Add some margin so the tooltips don't get cut off
    { style: { margin: '0 8px' } },
    [
      valueLabel,
      React.createElement(
        MaterialUI.Slider,
        {
          value: value,
          min: range[0],
          max: range[1],
          valueLabelDisplay: 'auto',
          onChange: (e, val) => column.setFilter(val),
          'aria-label': `Filter ${column.name}`
        }
      )
    ]
  )
}

const filterRange = (rows, columnId, filterValue) => {
  const [min, max] = filterValue
  return rows.filter(row => {
    const value = row.values[columnId]
    return value >= min && value <= max
  })
}
```

```{r}
library(htmltools)

data <- MASS::Cars93[, c("Manufacturer", "Model", "Type", "Price")]

muiDependency <- function() {
  list(
    # Material UI requires React
    reactR::html_dependency_react(),
    htmlDependency(
      name = "mui",
      version = "5.6.3",
      src = c(href = "https://unpkg.com/@mui/material@5.6.3/umd/"),
      script = "material-ui.production.min.js"
    )
  )
}

browsable(tagList(
  muiDependency(),

  reactable(
    data,
    columns = list(
      Price = colDef(
        filterable = TRUE,
        filterMethod = JS("filterRange"),
        filterInput = JS("muiRangeFilter")
      )
    ),
    defaultPageSize = 5,
    minRows = 5
  )
))
```


```{css echo=FALSE}
/* rmarkdown html documents */
.main-container {
  font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Helvetica, Arial, sans-serif;
}

.main-container blockquote {
  font-size: inherit;
}
```

---
output:
  html_document:
    theme: null
    highlight: null
    mathjax: null
    self_contained: false
---

```{r, include=FALSE}
options(reactable.static = TRUE)
```

```{r, echo=FALSE}
library(reactable)

reactable(iris)
```

---
title: "Popular Movies"
output: html_document
resource_files:
  - '.'
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

```{r table, echo=FALSE, message=FALSE}
library(reactable)
library(jsonlite)
library(dplyr)
library(crosstalk)
library(htmltools)

# tmdb_movies.json is a JSON dataset with 2000 rows and 22 columns.
# Some of these columns contain nested lists (e.g., multiple genres per movie).
movies <- fromJSON("tmdb_movies.json") %>%
  mutate(
    genres = lapply(genres, as.list),
    popularity_rank = rank(-popularity, ties.method = "first"),
    year = substr(release_date, 1, 4),
    score = vote_average * 10
  ) %>%
  arrange(popularity_rank)

tbl_data <- movies %>%
  select(popularity_rank, title, release_date, score, id, poster_path, certification, genres, runtime)

# Precalculate user score colors to be used for custom cell rendering
get_score_color <- function(score) {
  blue_pal <- function(x) rgb(colorRamp(c("#9fc7df", "#416ea4"))(x), maxColorValue = 255)
  normalized <- (score - min(score)) / (max(score) - min(score))
  blue_pal(normalized)
}
tbl_data <- mutate(tbl_data, score_color = get_score_color(score))

# Shared data for the table
shared_data <- SharedData$new(tbl_data)

# Shared data for the Crosstalk filters. This is a separate shared data object
# so we can filter on columns that aren't in the table. The same group is used
# to link the two datasets together.
shared_movies <- SharedData$new(movies, group = shared_data$groupName())

tbl <- reactable(
  shared_data,
  defaultColDef = colDef(vAlign = "center", headerClass = "header"),
  columns = list(
    popularity_rank = colDef(
      name = "Rank",
      align = "center",
      minWidth = 70,
      maxWidth = 90
    ),
    title = colDef(
      name = "Movie",
      # Since there are so many rows, we use a JS render function to keep the
      # page size down. This would be much easier to code in R, but that would
      # significantly increase the file size.
      cell = JS("renderMovie"),
      html = TRUE,
      minWidth = 250
    ),
    release_date = colDef(
      name = "Year",
      defaultSortOrder = "desc",
      cell = JS("cellInfo => cellInfo.value.slice(0, 4)"),
      minWidth = 70,
      maxWidth = 90
    ),
    score = colDef(
      name = "User Score",
      defaultSortOrder = "desc",
      # Show the user score in a donut chart like TMDb does. Since donut charts
      # are hard to compare, apply a color scale as well.
      cell = JS("renderUserScore"),
      html = TRUE,
      align = "center",
      width = 140,
      class = "user-score"
    ),
    id = colDef(show = FALSE),
    poster_path = colDef(show = FALSE),
    certification = colDef(show = FALSE),
    genres = colDef(show = FALSE),
    runtime = colDef(show = FALSE),
    score_color = colDef(show = FALSE)
  ),
  highlight = TRUE,
  language = reactableLang(
    noData = "No movies found",
    pageInfo = "{rowStart}\u2013{rowEnd} of {rows} movies"
  ),
  theme = reactableTheme(
    highlightColor = "#f3fafb",
    borderColor = "hsl(0, 0%, 93%)",
    headerStyle = list(borderColor = "hsl(0, 0%, 90%)")
  ),
  class = "movies-tbl"
)
```

```{js table_js, echo=FALSE}
// Custom JavaScript cell renderer for the Movie column
function renderMovie(cellInfo) {
  const url = 'https://www.themoviedb.org/movie/' + cellInfo.row['id']
  const imageSrc = 'https://www.themoviedb.org/t/p/w45' + cellInfo.row['poster_path']
  const altText = cellInfo.value + ' movie poster'
  const poster = `<a href="${url}"><img src="${imageSrc}" class="movie-poster" alt="${altText}"></a>`

  const title = `<a href="${url}">${cellInfo.value}</a>`

  let rating = cellInfo.row['certification']
  if (rating) {
    rating = `<span class="movie-rating">${rating}</span>`
  } else {
    rating = ''
  }

  const genres = ' ' + cellInfo.row['genres'].join(', ')

  let runtime = cellInfo.row['runtime']
  if (runtime != null) {
    const hours = Math.floor(runtime / 60)
    const mins = runtime % 60
    runtime = [hours > 0 ? hours + 'h' : '', mins > 0 ? mins + 'm' : ''].join(' ')
    runtime = `<span aria-hidden="true"> • </span>` +
      `<div class="movie-runtime">${runtime}</div>`
  } else {
    runtime = ''
  }

  const details = `<div class="movie-info-details">${rating}${genres}${runtime}</div>`
  const text = `<div class="movie-info-text">${title}${details}</div>`
  return `<div class="movie-info">${poster}${text}</div>`
}

// Custom JavaScript cell renderer for the User Score column
function renderUserScore(cellInfo) {
  return donutChart(cellInfo.value, cellInfo.row['score_color'])
}

// Generates HTML for a donut chart given a percentage value (out of 100) and color
function donutChart(value, color) {
  // All units are in rem for relative scaling
  const radius = 1.5
  const diameter = 3.75
  const center = diameter / 2
  const width = 0.25
  const sliceLength = 2 * Math.PI * radius
  const sliceOffset = sliceLength * (1 - value / 100)
  const donutChart = `
    <svg width="${diameter}rem" height="${diameter}rem" style="transform: rotate(-90deg)" focusable="false">
      <circle cx="${center}rem" cy="${center}rem" r="${radius}rem" fill="none" stroke-width="${width}rem" stroke="rgba(0,0,0,0.1)"></circle>
      <circle cx="${center}rem" cy="${center}rem" r="${radius}rem" fill="none" stroke-width="${width}rem" stroke="${color}"
       stroke-dasharray="${sliceLength}rem" stroke-dashoffset="${sliceOffset}rem"></circle>
    </svg>
  `
  const label = `
    <div style="position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%)">
      ${value}%
    </div>
  `
  return `
    <div style="display: inline-flex; position: relative">
      ${donutChart}
      ${label}
    </div>
  `
}
```

```{r filters, echo=FALSE}
# Custom Crosstalk select filter. This is a single-select input that works
# on columns containing multiple values per row (list columns).
select_filter <- function(id, label, shared_data, group, choices = NULL,
                          width = "100%", class = "filter-input") {
  values <- shared_data$data()[[group]]
  keys <- shared_data$key()
  if (is.list(values)) {
    # Multiple values per row
    flat_keys <- unlist(mapply(rep, keys, sapply(values, length)))
    keys_by_value <- split(flat_keys, unlist(values), drop = TRUE)
    choices <- if (is.null(choices)) sort(unique(unlist(values))) else choices
  } else {
    # Single value per row
    keys_by_value <- split(seq_along(keys), values, drop = TRUE)
    choices <- if (is.null(choices)) sort(unique(values)) else choices
  }

  script <- sprintf("
    window['__ct__%s'] = (function() {
      const handle = new window.crosstalk.FilterHandle('%s')
      const keys = %s
      return {
        filter: function(value) {
          if (!value) {
            handle.clear()
          } else {
            handle.set(keys[value])
          }
        }
      }
    })()
  ", id, shared_data$groupName(), toJSON(keys_by_value))

  div(
    class = class,
    tags$label(`for` = id, label),
    tags$select(
      id = id,
      onchange = sprintf("window['__ct__%s'].filter(this.value)", id),
      style = sprintf("width: %s", validateCssUnit(width)),
      tags$option(value = "", "All"),
      lapply(choices, function(value) tags$option(value = value, value))
    ),
    tags$script(HTML(script))
  )
}

# Custom Crosstalk search filter. This is a free-form text field that does
# case-insensitive text searching on a single column.
search_filter <- function(id, label, shared_data, group, width = "100%", class = "filter-input") {
  values <- as.list(shared_data$data()[[group]])
  values_by_key <- setNames(values, shared_data$key())
  
  script <- sprintf("
    window['__ct__%s'] = (function() {
      const handle = new window.crosstalk.FilterHandle('%s')
      const valuesByKey = %s
      return {
        filter: function(value) {
          if (!value) {
            handle.clear()
          } else {
            // Escape special characters in the search value for regex matching
            value = value.replace(/[.*+?^${}()|[\\]\\\\]/g, '\\\\$&')
            const regex = new RegExp(value, 'i')
            const filtered = Object.keys(valuesByKey).filter(function(key) {
              const value = valuesByKey[key]
              if (Array.isArray(value)) {
                for (let i = 0; i < value.length; i++) {
                  if (regex.test(value[i])) {
                    return true
                  }
                }
              } else {
                return regex.test(value)
              }
            })
            handle.set(filtered)
          }
        }
      }
    })()
  ", id, shared_data$groupName(), toJSON(values_by_key))
  
  div(
    class = class,
    tags$label(`for` = id, label),
    tags$input(
      id = id,
      type = "search",
      oninput = sprintf("window['__ct__%s'].filter(this.value)", id),
      style = sprintf("width: %s", validateCssUnit(width))
    ),
    tags$script(HTML(script))
  )
}

# Custom Crosstalk range filter. This is a simple range input that only filters
# minimum values of a column.
range_filter <- function(id, label, shared_data, group, min = NULL, max = NULL,
                         step = NULL, suffix = "", width = "100%", class = "filter-input") {
  values <- shared_data$data()[[group]]
  values_by_key <- setNames(as.list(values), shared_data$key())

  script <- sprintf("
    window['__ct__%s'] = (function() {
      const handle = new window.crosstalk.FilterHandle('%s')
      const valuesByKey = %s
      return {
        filter: function(value) {
          const filtered = Object.keys(valuesByKey).filter(function(key) {
            return valuesByKey[key] >= value
          })
          handle.set(filtered)
        }
      }
    })()
  ", id, shared_data$groupName(), toJSON(values_by_key))

  min <- if (!is.null(min)) min else min(values)
  max <- if (!is.null(max)) max else max(values)
  value <- min

  oninput <- paste(
    sprintf("document.getElementById('%s__value').textContent = this.value + '%s';", id, suffix),
    sprintf("window['__ct__%s'].filter(this.value)", id)
  )

  div(
    class = class,
    tags$label(`for` = id, label),
    div(
      tags$input(
        id = id,
        type = "range",
        min = min,
        max = max,
        step = step,
        value = value,
        oninput = oninput,
        onchange = oninput, # For IE11 support
        style = sprintf("width: %s", validateCssUnit(width))
      )
    ),
    span(id = paste0(id, "__value"), paste0(value, suffix)),
    tags$script(HTML(script))
  )
}
```

```{r output, echo=FALSE}
div(
  class = "movies",
  div(
    class = "filters",
    search_filter("filter_title", "Search titles", shared_movies, "title"),
    select_filter("filter_genres", "Genre", shared_movies, "genres"),
    select_filter("filter_year", "Year", shared_movies, "year",
                  choices = sort(unique(movies$year), decreasing = TRUE)),
    select_filter("filter_language", "Language", shared_movies, "original_language"),
    select_filter("filter_rating", "Rating", shared_movies, "certification",
                  choices = c("G", "PG", "PG-13", "R", "NC-17", "NR")),
    range_filter("filter_score", "Min Score", shared_movies, "score", suffix = "%")
  ),
  tags$hr(),
  tbl
)
```

---

Source: [The Movie Database (TMDb)](https://www.themoviedb.org/) and the
[TMDb API](https://developers.themoviedb.org/3/)

Raw data: [`tmdb_movies.json`](tmdb_movies.json)

This product uses the TMDb API but is not endorsed or certified by TMDb.

<h2>Source Code</h2>

```{r ref.label="table", eval=FALSE}
```

```{js ref.label="table_js", eval=FALSE}
```

```{r ref.label="filters", eval=FALSE}
```

```{r ref.label="output", eval=FALSE}
```

```{css}
/* Font from https://fontsarena.com/hanken-grotesk-by-hanken-design-co/ */
@font-face {
  font-family: 'Hanken Grotesk';
  font-style: normal;
  font-weight: 400;
  src: url("fonts/HKGrotesk-Regular.woff2") format("woff2"),
       url("fonts/HKGrotesk-Regular.woff") format("woff");
}

@font-face {
  font-family: 'Hanken Grotesk';
  font-style: normal;
  font-weight: 600;
  src: url("fonts/HKGrotesk-SemiBold.woff2") format("woff2"),
       url("fonts/HKGrotesk-SemiBold.woff") format("woff");
}

@font-face {
  font-family: 'Hanken Grotesk';
  font-style: normal;
  font-weight: 700;
  src: url("fonts/HKGrotesk-Bold.woff2") format("woff2"),
       url("fonts/HKGrotesk-Bold.woff") format("woff");
}

.movies {
  font-family: 'Hanken Grotesk', Helvetica, Arial, sans-serif;
}

.movies h2 {
  font-weight: 600;
}

.movies a {
  color: #007899;
  text-decoration: none;
}

.movies a:hover,
.movies a:focus {
  text-decoration: underline;
  text-decoration-thickness: max(1px, 0.0625rem);
}

.movies-tbl {
  margin-top: 1rem;
  font-size: 1rem;
}

.header {
  color: hsl(0, 0%, 45%);
  font-weight: 700;
  font-size: 0.8125rem;
  letter-spacing: 0.4px;
  text-transform: uppercase;
}

.header:hover[aria-sort],
.header[aria-sort='ascending'],
.header[aria-sort='descending'] {
  color: hsl(0, 0%, 5%);
}

.movie-info {
  display: flex;
  align-items: center;
}

.movie-info-text {
  margin-left: 0.75rem;
  font-weight: 600;
  overflow: hidden;
  text-overflow: ellipsis;
}

.movie-info-details {
  margin-top: 0.125rem;
  font-size: 0.875rem;
  font-weight: 400;
  color: hsl(0, 0%, 40%);
  overflow: hidden;
  text-overflow: ellipsis;
}

.movie-poster {
  width: 45px;
  height: 68px;
  box-shadow: 0 0 0 1px hsl(0, 0%, 95%);
}

.movie-runtime {
  display: inline-block;
}

.movie-rating {
  margin-right: 0.25rem;
  padding: 0 0.25rem;
  border: 1px solid hsl(0, 0%, 75%);
  border-radius: 2px;
}

.user-score {
  font-weight: 600;
}

.filters {
  display: flex;
  flex-wrap: wrap;
  margin-top: 1rem;
  margin-left: -32px;
}

.filter-input {
  margin-top: 0.5rem;
  margin-left: 32px;
  flex: 1;
  min-width: 250px;
}

.filter-input label {
  color: hsl(0, 0%, 45%);
  font-weight: 700;
  font-size: 0.8125rem;
  letter-spacing: 0.4px;
  text-transform: uppercase;
}

.filter-input select,
.filter-input input[type="search"] {
  padding: 0 0.375rem;
  height: 2rem;
}

.filter-input input[type="search"] {
  /* Revert Bootstrap 5's Reboot styles, which change native search input styling */
  -webkit-appearance: searchfield;
  outline-offset: revert;
  border-color: revert;
}
```

```{css echo=FALSE}
/* pkgdown articles */
.row > main {
  max-width: 940px;
}
```

---
title: "CRAN Packages"
output: html_document
resource_files:
  - '.'
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

```{r table, echo=FALSE}
library(reactable)
library(htmltools)

pkgs <- read.csv("cran_pkgs.csv", stringsAsFactors = FALSE)
versions <- jsonlite::read_json("versions.json", simplifyDataFrame = TRUE)

row_details <- function(index) {
  pkg <- pkgs[index, ]
  urls <- unlist(strsplit(gsub(",", " , ", pkg$URL, perl = TRUE), "[ \n]"))

  pkg_field <- function(name, ...) {
    if (any(is.na(...))) NULL
    else tagList(div(class = "detail-label", name), ...)
  }

  detail <- div(
    class = "package-detail",
    div(class = "detail-header", pkg$Package, span(class = "detail-title", pkg$Title)),
    div(class = "detail-description", pkg$Description),
    pkg_field("Version", pkg$Version),
    pkg_field("Depends", pkg$Depends),
    pkg_field("Imports", pkg$Imports),
    pkg_field("Suggests", pkg$Suggests),
    pkg_field("Author", pkg$Author),
    pkg_field("License", pkg$License),
    pkg_field("URL", lapply(urls, function(url) {
      if (grepl("https?://", url)) tags$a(href = url, url)
      else if (identical(url, ",")) ", "
      else url
    })),
    pkg_field("System Requirements", pkg$SystemRequirements)
  )

  if (length(versions[[pkg$Package]]) > 0) {
    archived <- pkg_field(
      "Archived Versions",
      reactable(
        versions[[pkg$Package]],
        pagination = FALSE,
        defaultColDef = colDef(headerClass = "header"),
        columns = list(
          Date = colDef(name = "Published", align = "right", cell = function(value) {
            strftime(value, format = "%b %d, %Y")
          })
        ),
        class = "archived-table",
        theme = reactableTheme(cellPadding = "8px 12px")
      )
    )
    detail <- tagAppendChild(detail, archived)
  }

  detail
}

tbl <- reactable(
  pkgs[, c("Package", "Version", "Title", "Date", "Downloads")],
  defaultSorted = "Downloads",
  defaultPageSize = 20,
  showPageSizeOptions = TRUE,
  pageSizeOptions = c(10, 20, 50, 100),
  onClick = "expand",
  resizable = TRUE,
  defaultColDef = colDef(headerClass = "header"),
  columns = list(
    Title = colDef(minWidth = 250, class = "package-title", cell = function(value) {
      span(title = value, value)
    }),
    Date = colDef(name = "Published", align = "right", cell = function(value) {
      strftime(value, format = "%b %d, %Y")
    }),
    Downloads = colDef(defaultSortOrder = "desc", cell = function(value) {
      value <- format(round(value / 1e6, 1), nsmall = 1)
      tagList(value, span(class = "units", "M"))
    })
  ),
  details = row_details,
  wrap = FALSE,
  class = "packages-table",
  rowStyle = list(cursor = "pointer"),
  theme = reactableTheme(cellPadding = "8px 12px")
)

div(class = "cran-packages",
  h2(class = "title", "Top CRAN Packages of 2019"),
  tbl
)
```

---

Source: [crandb](https://github.com/r-hub/crandb), [cranlogs](https://github.com/r-hub/cranlogs)

<h2>Source Code</h2>

```{r ref.label="table", eval=FALSE}
```

```{r}
tags$link(href = "https://fonts.googleapis.com/css?family=Nunito:400,600,700&display=fallback", rel = "stylesheet")
```

```{css}
.cran-packages {
  font-family: Nunito, "Helvetica Neue", Helvetica, Arial, sans-serif;
}

.title {
  font-size: 1.5rem;
}

.packages-table {
  margin-top: 1rem;
  font-size: 0.9375rem;
  border: 1px solid hsl(213, 33%, 93%);
  border-radius: 4px;
  box-shadow: 0 4px 8px 0 rgba(0, 0, 0, 0.1);
}

.header {
  background-color: hsl(213, 45%, 97%);
  border-bottom-color: hsl(213, 33%, 93%);
  border-bottom-width: 1px;
  color: hsl(213, 13%, 33%);
}

.header[aria-sort]:hover,
.header[aria-sort]:focus {
  color: hsl(213, 55%, 50%);
}

.units {
  margin-left: 0.15em;
  color: rgba(0, 0, 0, 0.6);
}

.package-title {
  font-size: 0.875rem;
}

.package-detail {
  padding: 24px;
  box-shadow: inset 0 1px 3px #dbdbdb;
  background: hsl(213, 20%, 99%);
}

.detail-label {
  margin: 1.25rem 0 0.25rem;
  font-size: 0.875rem;
  color: rgba(0, 0, 0, 0.6);
}

.detail-header {
  margin-bottom: 1rem;
  font-size: 1.25rem;
  font-weight: 600;
}

.detail-title {
  margin-left: 1rem;
  font-size: 0.875rem;
  font-weight: 400;
  color: rgba(0, 0, 0, 0.6);
}

.detail-description {
  font-size: 0.875rem;
}

.archived-table {
  max-width: 15rem;
  border: 1px solid hsl(213, 33%, 93%);
  border-radius: 4px;
  box-shadow: 0 2px 7px 0 rgba(0, 0, 0, 0.05);
  font-size: 0.875rem;
}
```


```{css echo=FALSE}
/* rmarkdown html documents */
.main-container {
  max-width: 1024px !important;
}

h1.title {
  display: none;
}

/* pkgdown articles */
.row > main {
  max-width: 1024px;
}
```

---
title: "Women's World Cup Predictions"
output: html_document
resource_files:
  - '.'
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)

options(reactable.static = TRUE)
```

```{css ref.label="styles", echo=FALSE}
```

```{r ref.label="font_styles", echo=FALSE}
```

```{css ref.label="pkgdown_styles", echo=FALSE}
```

```{r table, echo=FALSE}
library(reactable)
library(htmltools)

forecasts <- read.csv("wwc_forecasts.csv", stringsAsFactors = FALSE)

rating_cols <- c("spi", "global_o", "global_d")
group_cols <- c("group_1", "group_2", "group_3")
knockout_cols <- c("make_round_of_16", "make_quarters", "make_semis", "make_final", "win_league")
forecasts <- forecasts[, c("team", "points", "group", rating_cols, group_cols, knockout_cols)]

rating_column <- function(maxWidth = 55, ...) {
  colDef(maxWidth = maxWidth, align = "center", class = "cell number", ...)
}

group_column <- function(class = NULL, ...) {
  colDef(cell = format_pct, maxWidth = 70, align = "center", class = paste("cell number", class), ...)
}

knockout_column <- function(maxWidth = 70, class = NULL, ...) {
  colDef(
    cell = format_pct,
    maxWidth = maxWidth,
    class = paste("cell number", class),
    style = function(value) {
      # Lighter color for <1%
      if (value < 0.01) {
        list(color = "#aaa")
      } else {
        list(color = "#111", background = knockout_pct_color(value))
      }
    },
    ...
  )
}

format_pct <- function(value) {
  if (value == 0) "  \u2013 "    # en dash for 0%
  else if (value == 1) "\u2713"  # checkmark for 100%
  else if (value < 0.01) " <1%"
  else if (value > 0.99) ">99%"
  else formatC(paste0(round(value * 100), "%"), width = 4)
}

make_color_pal <- function(colors, bias = 1) {
  get_color <- colorRamp(colors, bias = bias)
  function(x) rgb(get_color(x), maxColorValue = 255)
}

off_rating_color <- make_color_pal(c("#ff2700", "#f8fcf8", "#44ab43"), bias = 1.3)
def_rating_color <- make_color_pal(c("#ff2700", "#f8fcf8", "#44ab43"), bias = 0.6)
knockout_pct_color <- make_color_pal(c("#ffffff", "#f2fbd2", "#c9ecb4", "#93d3ab", "#35b0ab"), bias = 2)

tbl <- reactable(
  forecasts,
  pagination = FALSE,
  defaultSorted = "win_league",
  defaultSortOrder = "desc",
  defaultColGroup = colGroup(headerVAlign = "bottom"),
  columnGroups = list(
    colGroup(name = "Team Rating", columns = rating_cols),
    colGroup(name = "Chance of Finishing Group Stage In ...", columns = group_cols),
    colGroup(name = "Knockout Stage Chances", columns = knockout_cols)
  ),
  defaultColDef = colDef(
    vAlign = "center",
    headerVAlign = "bottom",
    class = "cell",
    headerClass = "header"
  ),
  columns = list(
    team = colDef(
      defaultSortOrder = "asc",
      minWidth = 200,
      headerStyle = list(fontWeight = 700),
      cell = function(value, index) {
        div(
          class = "team",
          img(class = "team-flag", alt = paste(value, "flag"), src = sprintf("images/%s.png", value)),
          div(
            span(class = "team-name", value),
            span(class = "team-record", sprintf("%s pts.", forecasts[index, "points"]))
          )
        )
      }
    ),
    points = colDef(show = FALSE),
    group = colDef(defaultSortOrder = "asc", align = "center", maxWidth = 75,
                   class = "cell group", headerStyle = list(fontWeight = 700)),
    spi = rating_column(format = colFormat(digits = 1)),
    global_o = rating_column(
      name = "Off.",
      cell = function(value) {
        scaled <- (value - min(forecasts$global_o)) / (max(forecasts$global_o) - min(forecasts$global_o))
        color <- off_rating_color(scaled)
        value <- format(round(value, 1), nsmall = 1)
        div(class = "spi-rating", style = list(background = color), value)
      }
    ),
    global_d = rating_column(
      name = "Def.", 
      defaultSortOrder = "asc",
      cell = function(value) {
        scaled <- 1 - (value - min(forecasts$global_d)) / (max(forecasts$global_d) - min(forecasts$global_d))
        color <- def_rating_color(scaled)
        value <- format(round(value, 1), nsmall = 1)
        div(class = "spi-rating", style = list(background = color), value)
      }
    ),
    group_1 = group_column(name = "1st Place", class = "border-left"),
    group_2 = group_column(name = "2nd Place"),
    group_3 = group_column(name = "3rd Place"),
    make_round_of_16 = knockout_column(name = "Make Round of 16", class = "border-left"),
    make_quarters = knockout_column(name = "Make Qtr-Finals"),
    make_semis = knockout_column(name = "Make Semifinals", maxWidth = 90),
    make_final = knockout_column(name = "Make Final"),
    win_league = knockout_column(name = "Win World Cup")
  ),
  # Emphasize borders between groups when sorting by group
  rowClass = JS("
    function(rowInfo, state) {
      const firstSorted = state.sorted[0]
      if (firstSorted && firstSorted.id === 'group') {
        const nextRow = state.pageRows[rowInfo.viewIndex + 1]
        if (nextRow && rowInfo.values['group'] !== nextRow.group) {
          return 'group-last'
        }
      }
    }"
  ),
  showSortIcon = FALSE,
  borderless = TRUE,
  class = "standings-table"
)

div(class = "standings",
  div(class = "title",
    h2("2019 Women's World Cup Predictions"),
    "Soccer Power Index (SPI) ratings and chances of advancing for every team"
  ),
  tbl,
  "Forecast from before 3rd group matches"
)
```

---

Source: [FiveThirtyEight](https://projects.fivethirtyeight.com/2019-womens-world-cup-predictions/)

Raw data: [`wwc_forecasts.csv`](wwc_forecasts.csv)

<h2>Source Code</h2>

Full source: [`vignettes/womens-world-cup/womens-world-cup.Rmd`](https://github.com/glin/reactable/blob/main/vignettes/womens-world-cup/womens-world-cup.Rmd)

```{r ref.label="table", eval=FALSE}
```

```{r font_styles, eval=FALSE}
htmltools::tags$link(
  href = "https://fonts.googleapis.com/css?family=Karla:400,700|Fira+Mono&display=fallback",
  rel = "stylesheet"
)
```

```{css styles, eval=FALSE}
.standings {
  font-family: Karla, "Helvetica Neue", Helvetica, Arial, sans-serif;
  font-size: 0.875rem;
}

.title {
  margin-top: 2rem;
  margin-bottom: 1.125rem;
  font-size: 1rem;
}

.title h2 {
  font-size: 1.25rem;
  font-weight: 600;
}

.standings-table {
  margin-bottom: 1.25rem;
}

.header {
  border-bottom-color: #555;
  font-size: 0.8125rem;
  font-weight: 400;
  text-transform: uppercase;
}

/* Highlight headers when sorting */
.header:hover,
.header:focus,
.header[aria-sort="ascending"],
.header[aria-sort="descending"] {
  background-color: #eee;
}

.border-left {
  border-left: 2px solid #555;
}

/* Use box-shadow to create row borders that appear behind vertical borders */
.cell {
  box-shadow: inset 0 -1px 0 rgba(0, 0, 0, 0.15);
}

.group-last .cell {
  box-shadow: inset 0 -2px 0 #555;
}

.team {
  display: flex;
  align-items: center;
}

.team-flag {
  height: 1.3rem;
  border: 1px solid #f0f0f0;
}

.team-name {
  margin-left: 0.5rem;
  font-size: 1.125rem;
  font-weight: 700;
}

.team-record {
  margin-left: 0.35rem;
  color: hsl(0, 0%, 45%);
  font-size: 0.8125rem;
}

.group {
  font-size: 1.1875rem;
}

.number {
  font-family: "Fira Mono", Consolas, Monaco, monospace;
  font-size: 1rem;
  white-space: pre;
}

.spi-rating {
  display: flex;
  align-items: center;
  justify-content: center;
  margin: auto;
  width: 1.875rem;
  height: 1.875rem;
  border: 1px solid rgba(0, 0, 0, 0.1);
  border-radius: 50%;
  color: #000;
  font-size: 0.8125rem;
  letter-spacing: -1px;
}
```


```{css pkgdown_styles, include=FALSE}
/* rmarkdown html documents */
.main-container {
  max-width: 1054px !important;
}

/* pkgdown articles */
.row > main {
  max-width: 1054px;
}

.page-header {
  display: none;
}
```

---
title: "Static Rendering"
output:
  html_document:
    toc: true
    toc_float:
      smooth_scroll: false
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)

library(reactable)
```

::: {.callout}
New in v0.4.0
:::

## Introduction

Static rendering is an optional feature that allows you to render tables to static HTML in R.

By default, tables are rendered completely in the user's browser by JavaScript, and
there's no actual HTML being generated when you run `reactable()` in R.

This means when you load a page, tables will be blank at first, and then appear a short time
later as they get rendered. In extreme cases, it can take a noticeably long time for tables to appear,
like with the heavy [Examples without static rendering](./examples-no-static.html) page.

With static rendering, tables are pre-rendered to their initial HTML so they appear
immediately without any flash of content. Tables are then made interactive and subsequently
rendered by JavaScript as needed. You can try refreshing on the
[Examples with static rendering](./examples.html) page for a comparison.

**Advantages**:

- Better user experience in many cases, as tables appear immediately and don't cause
  any layout shift on page load.
- Better support for indexing of table content by search engines. Even crawlers that
  support JavaScript like Googlebot may have a crawl budget that limits crawling of
  JavaScript-heavy pages.
- Non-interactive tables (without pagination, sorting, filtering) no longer require
  JavaScript to view, which may be useful when JavaScript is unavailable, like in
  emails.

Static rendering uses the V8 JavaScript engine provided by the [V8 package](https://github.com/jeroen/V8).
V8 is not installed with reactable by default, so it must be installed before
using static rendering.

::: {.callout-note}
**Note:** Static rendering is currently experimental, and is not supported for tables
rendered via `reactableOutput()` in Shiny. There are also other [limitations and tradeoffs](#limitations-and-tradeoffs)
to consider, and static rendering may not make sense for every table.
:::

## Usage

To use static rendering, ensure that the [V8 package](https://github.com/jeroen/V8)
is installed first. V8 is not installed with reactable by default.

```{r, eval=FALSE}
install.packages("V8")
```

Then, use `reactable(static = TRUE)` to render a single table to static HTML:

```{r}
data <- MASS::Cars93[, c("Manufacturer", "Model", "Type", "Price")]

reactable(data, defaultPageSize = 5, static = TRUE)
```

Or use `options(reactable.static = TRUE)` to enable static rendering for all tables:

```{r}
options(reactable.static = TRUE)

reactable(data, defaultPageSize = 5)
```

```{r, echo=FALSE}
options(reactable.static = FALSE)
```

## Examples

The reactable package documentation uses static rendering in several articles, and
provides alternate versions of these pages without static rendering for comparison:

- [Examples with static rendering](./examples.html) vs. [Examples without static rendering](./examples-no-static.html)
- [Demo Cookbook with static rendering](./cookbook/cookbook.html) vs. [Demo Cookbook without static rendering](./cookbook/cookbook-no-static.html)

## Limitations and Tradeoffs

Although static rendering might seem useful to enable for all tables, there are some
limitations and tradeoffs to consider, and it may not make sense to use for every table.

- Static rendering isn't supported for tables rendered via `reactableOutput()` in Shiny.
  It is still supported for `reactable()` tables specified in a Shiny app's `ui` definition, however.
- Some features in reactable do not fully support static rendering yet:
  - `colFormat()` only supports `en` locales. Formatting with other locales will still
    work in the browser, but tables may initially show data formatted in an `en` locale.
  - Default expanded rows and row details with `reactable(defaultExpanded = TRUE)`
    are not statically rendered.
- There may be a noticeable period where tables are visible but not yet interactable,
  as the client-side JavaScript that makes tables interactive is still running. In these
  cases, the user experience may be better with fully client-side rendered tables.
- HTML documents will be larger, which could affect users with slower download speeds.
  Static rendering is especially not recommended for unpaginated tables with a large number of rows.
- Custom JavaScript render functions and style functions won't have access to the DOM,
  as they won't be run in a browser environment. If your custom JavaScript functions
  require the DOM (e.g., `window` or `document`), you can defer rendering until the DOM
  is available to prevent rendering failures:
  ```{r, eval=FALSE}
  colDef(
    cell = JS("
      function(cellInfo) {
        if (typeof document !== 'undefined') {
          // Not in a browser environment, defer rendering until running in a browser
          return null
        }
      }
    ")
  )
  ```
- Custom JavaScript render functions and style functions won't have access to external
  JavaScript code. Any external JavaScript has to be included inline for now, but this
  may change in the future.
- Tables using `reactableTheme()` may initially appear unstyled when rendered in pkgdown
  websites. Statically rendered theme styles are usually placed in the document `<head>`
  to prevent problem, but pkgdown renders all user-specified `<head>` content into the
  document body, which means tables will appear before browsers process the theme styles.


```{css echo=FALSE}
/* rmarkdown html documents */
.main-container {
  font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Helvetica, Arial, sans-serif;
}

.main-container blockquote {
  font-size: inherit;
}
```

---
title: "Internationalization"
output: html_document
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)

library(reactable)
```

```{r}
reactable(
  iris[1:30, ],
  searchable = TRUE,
  showPageSizeOptions = TRUE,
  language = reactableLang(
    searchPlaceholder = "搜索",
    searchLabel = "搜索",
    noData = "没有匹配结果",
    pageSizeOptions = "显示 {rows}",
    pageInfo = "{rowStart} 至 {rowEnd} 项结果，共 {rows} 项",
    pagePrevious = "上页",
    pageNext = "下页",
    pagePreviousLabel = "",
    pageNextLabel = ""
  )
)
```

```{css echo=FALSE}
/* rmarkdown html documents */
.main-container {
  font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Helvetica, Arial, sans-serif;
}
```

---
title: "Examples"
output:
  html_document:
    toc: true
    toc_float:
      smooth_scroll: false
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)

library(reactable)

options(reactable.static = TRUE)
```

## Basic Usage
To create a data table, use `reactable()` on a data frame or matrix.
The table will be sortable and paginated by default:
```{r}
library(reactable)

reactable(iris)
```

### Column definitions
Columns can be customized by providing a named list of column definitions
created by `colDef()` to `columns`:
```{r}
reactable(
  iris[1:5, ],
  columns = list(
    Sepal.Length = colDef(name = "Sepal Length"),
    Sepal.Width = colDef(name = "Sepal Width"),
    Species = colDef(align = "center")
  )
)
```

For convenience, you can also specify a default `colDef()` to use for all
columns in `defaultColDef`:
```{r}
reactable(
  iris[1:5, ],
  defaultColDef = colDef(
    header = function(value) gsub(".", " ", value, fixed = TRUE),
    cell = function(value) format(value, nsmall = 1),
    align = "center",
    minWidth = 70,
    headerStyle = list(background = "#f7f7f8")
  ),
  columns = list(
    Species = colDef(minWidth = 140)  # overrides the default
  ),
  bordered = TRUE,
  highlight = TRUE
)
```

## Sorting

Tables are sortable by default. You can sort a column by clicking on its
header, or sort multiple columns by holding the shift key while sorting.

Sorting toggles between ascending and descending order by default. To clear
the sort, hold the shift key while sorting, and the sorting will additionally
toggle between ascending, descending, and unsorted order.

::: {.callout-note}
**Note:** Ascending order means the lowest, first, or earliest values will appear first.
Descending order means the largest, last, or latest values will appear first.
:::

### Default sorted columns

You can set the default sorted columns by providing a vector of column names
to `defaultSorted`:

```{r}
reactable(iris[48:52, ], defaultSorted = c("Species", "Petal.Length"))
```

You can also provide a named list to customize the default sort orders.
Use `"asc"` for ascending order, or `"desc"` for descending order:

```{r}
reactable(iris[48:52, ], defaultSorted = list(Species = "asc", Petal.Length = "desc"))
```

### Default sort order

Columns are sorted in ascending order first by default. To change the default sort
order for all columns in the table, set `defaultSortOrder` in `reactable()` to
`"asc"` for ascending order, or `"desc"` for descending order.

To change the sort order of an individual column, set `defaultSortOrder` in
its `colDef()` to `"asc"` or `"desc"`. The default sort order of the column takes
precedence over the table.

```{r}
reactable(
  iris[48:52, ],
  defaultSortOrder = "desc",
  columns = list(
    Species = colDef(defaultSortOrder = "asc")
  ),
  defaultSorted = c("Species", "Petal.Length")
)
```

### Sort missing values last

You can ignore missing values when sorting by setting `sortNALast` on a column:

```{r}
reactable(
  data.frame(
    n = c(1, 2, 3, -Inf, Inf),
    x = c(2, 3, 1, NA, NaN),
    y = c("aa", "cc", "bb", NA, NA)
  ),
  defaultColDef = colDef(sortNALast = TRUE),
  defaultSorted = "x"
)
```

### No sorting

You can disable sorting by setting `sortable` to `FALSE` on the table or column.
When only some columns are sortable, it can help to indicate sortable
columns using `showSortable`:

```{r}
reactable(
  iris[1:5, ],
  sortable = FALSE,
  showSortable = TRUE,
  columns = list(
    Petal.Width = colDef(sortable = TRUE),
    Petal.Length = colDef(sortable = TRUE)
  )
)
```

### Hide sort icons

You can hide sort icons by setting `showSortIcon` to `FALSE`. This is only
recommended when you want to use a [custom sort indicator](cookbook/cookbook.html#custom-sort-indicators).
```{r}
reactable(iris[1:5, ], showSortIcon = FALSE)
```


## Filtering

You can make columns filterable by setting `filterable = TRUE` in `reactable()`:

```{r}
data <- MASS::Cars93[1:20, c("Manufacturer", "Model", "Type", "AirBags", "Price")]

reactable(data, filterable = TRUE, minRows = 10)
```

To make specific columns filterable (or not), set `filterable` to `TRUE` or `FALSE` in `colDef()`:

```{r}
reactable(
  data,
  filterable = TRUE,
  columns = list(
    Price = colDef(filterable = FALSE)
  ),
  defaultPageSize = 5
)
```

### Custom filtering

Column filtering can be customized using the `filterMethod` and `filterInput` arguments
in `colDef()`. See the [Custom Filtering](custom-filtering.html) guide for more details
and examples.

This example shows basic usage of a custom filter method, changing filtering
on the `Manufacturer` column to be case-sensitive rather than case-insensitive.
(Try filtering for "bmw" and then "BMW").

```{r}
data <- MASS::Cars93[, c("Manufacturer", "Model", "Type", "Price")]

reactable(
  data,
  columns = list(
    Manufacturer = colDef(
      filterable = TRUE,
      # Filter by case-sensitive text match
      filterMethod = JS("function(rows, columnId, filterValue) {
        return rows.filter(function(row) {
          return row.values[columnId].indexOf(filterValue) !== -1
        })
      }")
    )
  ),
  defaultPageSize = 5
)
```

## Searching

You can make the entire table searchable by setting `searchable = TRUE` in `reactable()`:

```{r}
data <- MASS::Cars93[1:20, c("Manufacturer", "Model", "Type", "AirBags", "Price")]

reactable(data, searchable = TRUE, minRows = 10)
```

### Custom searching

The table search method can be customized using the `searchMethod` argument in `reactable()`.
See the [Custom Filtering](custom-filtering.html) guide for details and examples.

## Pagination

You can change the default page size by configuring `defaultPageSize`:

```{r}
reactable(iris[1:6, ], defaultPageSize = 4)
```

You can also set the minimum rows per page using `minRows`. This may be
useful when rows don't completely fill the page, or if the table has filtering:

```{r}
reactable(iris[1:6, ], defaultPageSize = 4, minRows = 4, searchable = TRUE)
```

### Page size options

You can show a dropdown of page sizes for users to choose from using `showPageSizeOptions`.
The page size options can be customized through `pageSizeOptions`:

```{r}
reactable(
  iris[1:12, ],
  showPageSizeOptions = TRUE,
  pageSizeOptions = c(4, 8, 12),
  defaultPageSize = 4
)
```

### Alternative pagination types

You can use an alternative pagination type by setting `paginationType` to:

- `"jump"` to show a page jump
- `"simple"` to show previous/next buttons only

#### Page jump

```{r}
reactable(iris[1:50, ], paginationType = "jump", defaultPageSize = 4)
```

#### Simple

```{r}
reactable(iris[1:50, ], paginationType = "simple", defaultPageSize = 4)
```

### Hide page info

You can hide page info by setting `showPageInfo` to `FALSE`:

```{r}
reactable(iris[1:12, ], showPageInfo = FALSE, defaultPageSize = 4)

reactable(iris[1:12, ], showPageInfo = FALSE, showPageSizeOptions = TRUE, defaultPageSize = 4)
```

### Always show pagination

By default, pagination is hidden if the table only has one page. To keep the
pagination shown, set `showPagination` to `TRUE`. This is especially useful if
you want to keep the page info showing the number of rows in the table.

```{r}
reactable(iris[1:5, ], showPagination = TRUE)
```

### No pagination

Tables are paginated by default, but you can disable pagination by setting `pagination` to `FALSE`:

```{r}
reactable(iris[1:20, ], pagination = FALSE, highlight = TRUE, height = 250)
```

::: {.callout-tip}
**Tip:** Disabling pagination is not recommended for large tables with many
interactive elements (such as links, expand buttons, or selection checkboxes),
as that can make it difficult for keyboard users to navigate the page.
:::

## Grouping and Aggregation

You can group rows in a table by specifying one or more columns in `groupBy`:

```{r}
data <- MASS::Cars93[10:22, c("Manufacturer", "Model", "Type", "Price", "MPG.city")]

reactable(data, groupBy = "Manufacturer")
```

When rows are grouped, you can aggregate data in a column using an `aggregate` function:

```{r}
data <- MASS::Cars93[14:38, c("Type", "Price", "MPG.city", "DriveTrain", "Man.trans.avail")]

reactable(
  data,
  groupBy = "Type",
  columns = list(
    Price = colDef(aggregate = "max"),
    MPG.city = colDef(aggregate = "mean", format = colFormat(digits = 1)),
    DriveTrain = colDef(aggregate = "unique"),
    Man.trans.avail = colDef(aggregate = "frequency")
  )
)
```

You can use one of the built-in aggregate functions:
```r
colDef(aggregate = "sum")        # Sum of numbers
colDef(aggregate = "mean")       # Mean of numbers
colDef(aggregate = "max")        # Maximum of numbers
colDef(aggregate = "min")        # Minimum of numbers
colDef(aggregate = "median")     # Median of numbers
colDef(aggregate = "count")      # Count of values
colDef(aggregate = "unique")     # Comma-separated list of unique values
colDef(aggregate = "frequency")  # Comma-separated counts of unique values
```

Or a custom aggregate function in JavaScript:
```r
colDef(
  aggregate = JS("
    function(values, rows) {
      // input:
      //  - values: an array of all values in the group
      //  - rows: an array of row data values for all rows in the group (optional)
      //
      // output:
      //  - an aggregated value, e.g. a comma-separated list
      return values.join(', ')
    }
  ")
)
```

### Multiple groups
```{r}
data <- data.frame(
  State = state.name,
  Region = state.region,
  Division = state.division,
  Area = state.area
)

reactable(
  data,
  groupBy = c("Region", "Division"),
  columns = list(
    Division = colDef(aggregate = "unique"),
    Area = colDef(aggregate = "sum", format = colFormat(separators = TRUE))
  ),
  bordered = TRUE
)
```

### Custom aggregate function
Custom aggregate functions are useful when none of the built-in aggregate functions
apply, or when you want to aggregate values from multiple columns. For example,
when calculating aggregate averages or percentages.

Within a custom aggregate function, you can access the values in the column using
the `values` argument, and the values in other columns using the `rows` argument:
```js
columns = list(
  Price = colDef(
    aggregate = JS("function(values, rows) {
      values
      // [46.8, 27.6, 57]

      rows
      // [
      //   { "Model": "Dynasty", "Manufacturer": "Dodge", "Price": 46.8, "Units": 2 },
      //   { "Model": "Colt", "Manufacturer": "Dodge", "Price": 27.6, "Units": 5 },
      //   { "Model": "Caravan", "Manufacturer": "Dodge", "Price": 57, "Units": 5 }
      // ]
    }")
  )
)
```

Here's an example that calculates an aggregate average price by dividing the
the sum of two columns, `Price` and `Units`:

```{r, message=FALSE}
library(dplyr)

set.seed(10)

data <- sample_n(MASS::Cars93[23:40, ], 30, replace = TRUE) %>%
  mutate(Price = Price * 3, Units = sample(1:5, 30, replace = TRUE)) %>%
  mutate(Avg.Price = Price / Units) %>%
  select(Model, Manufacturer, Price, Units, Avg.Price)

reactable(
  data,
  groupBy = "Manufacturer",
  columns = list(
    Price = colDef(aggregate = "sum", format = colFormat(currency = "USD")),
    Units = colDef(aggregate = "sum"),
    Avg.Price = colDef(
      # Calculate the aggregate Avg.Price as `sum(Price) / sum(Units)`
      aggregate = JS("function(values, rows) {
        let totalPrice = 0
        let totalUnits = 0
        rows.forEach(function(row) {
          totalPrice += row['Price']
          totalUnits += row['Units']
        })
        return totalPrice / totalUnits
      }"),
      format = colFormat(currency = "USD")
    )
  )
)
```

### Include sub rows in pagination

By default, sub rows are excluded from pagination and always shown on the same
page when expanded. To include sub rows in pagination, you can set `paginateSubRows`
to `TRUE`. This is recommended for grouped tables with a large number of rows
where expanded rows may not all fit on one page.

```{r}
data <- MASS::Cars93[, c("Manufacturer", "Model", "Type", "Price", "MPG.city")]

reactable(data, groupBy = "Type", paginateSubRows = TRUE)
```


## Column Formatting

You can format data in a column by providing `colFormat()` options to the `format`
argument in `colDef()`.

The formatters for numbers, dates, times, and currencies are locale-sensitive and
automatically adapt to language preferences of the user's browser. This means, for
example, that users will see dates formatted in their own timezone or numbers formatted
in their own locale.

To use a specific locale for data formatting, provide a vector of BCP 47 language
tags in the `locales` argument. See a list of
[common BCP 47 language tags](https://learn.microsoft.com/en-us/openspecs/office_standards/ms-oe376/6c085406-a698-4e12-9d4d-c3b0ee3dbc4a)
for reference.

::: {.callout-note}
**Note:** Column formatters change how data is displayed without affecting the underlying
data. Sorting, filtering, and grouping will still work on the original data.
:::

```{r}
data <- data.frame(
  price_USD = c(123456.56, 132, 5650.12),
  price_INR = c(350, 23208.552, 1773156.4),
  number_FR = c(123456.56, 132, 5650.12),
  temp = c(22, NA, 31),
  percent = c(0.9525556, 0.5, 0.112),
  date = as.Date(c("2019-01-02", "2019-03-15", "2019-09-22"))
)

reactable(data, columns = list(
  price_USD = colDef(format = colFormat(prefix = "$", separators = TRUE, digits = 2)),
  price_INR = colDef(format = colFormat(currency = "INR", separators = TRUE, locales = "hi-IN")),
  number_FR = colDef(format = colFormat(locales = "fr-FR")),
  temp = colDef(format = colFormat(suffix = " °C")),
  percent = colDef(format = colFormat(percent = TRUE, digits = 1)),
  date = colDef(format = colFormat(date = TRUE, locales = "en-GB"))
))
```

### Date formatting

```{r}
datetimes <- as.POSIXct(c("2019-01-02 3:22:15", "2019-03-15 09:15:55", "2019-09-22 14:20:00"),
                        tz = "America/New_York")
data <- data.frame(
  datetime = datetimes,
  date = datetimes,
  time = datetimes,
  time_24h = datetimes,
  datetime_pt_BR = datetimes
)

reactable(data, columns = list(
  datetime = colDef(format = colFormat(datetime = TRUE)),
  date = colDef(format = colFormat(date = TRUE)),
  time = colDef(format = colFormat(time = TRUE)),
  time_24h = colDef(format = colFormat(time = TRUE, hour12 = FALSE)),
  datetime_pt_BR = colDef(format = colFormat(datetime = TRUE, locales = "pt-BR"))
))
```

### Currency formatting

```{r}
data <- data.frame(
  USD = c(12.12, 2141.213, 0.42, 1.55, 34414),
  EUR = c(10.68, 1884.27, 0.37, 1.36, 30284.32),
  INR = c(861.07, 152122.48, 29.84, 110, 2444942.63),
  JPY = c(1280, 226144, 44.36, 164, 3634634.61),
  MAD = c(115.78, 20453.94, 4.01, 15, 328739.73)
)

reactable(data, columns = list(
  USD = colDef(
    format = colFormat(currency = "USD", separators = TRUE, locales = "en-US")
  ),
  EUR = colDef(
    format = colFormat(currency = "EUR", separators = TRUE, locales = "de-DE")
  ),
  INR = colDef(
    format = colFormat(currency = "INR", separators = TRUE, locales = "hi-IN")
  ),
  JPY = colDef(
    format = colFormat(currency = "JPY", separators = TRUE, locales = "ja-JP")
  ),
  MAD = colDef(
    format = colFormat(currency = "MAD", separators = TRUE, locales = "ar-MA")
  )
))
```

### Formatting aggregated cells

Column formatters apply to both standard and aggregated cells by default. If you
want to format aggregated cells separately, provide a named list of `cell` and
`aggregated` options:

```{r, eval=FALSE}
colDef(
  format = list(
    cell = colFormat(...),       # Standard cells
    aggregated = colFormat(...)  # Aggregated cells
  )
)
```

For example, only the aggregated `States` are formatted here:

```{r}
data <- data.frame(
  States = state.name,
  Region = state.region,
  Area = state.area
)

reactable(
  data,
  groupBy = "Region",
  columns = list(
    States = colDef(
      aggregate = "count",
      format = list(
        aggregated = colFormat(suffix = " states")
      )
    ),
    Area = colDef(
      aggregate = "sum",
      format = colFormat(suffix = " mi²", separators = TRUE)
    )
  )
)
```

### Displaying missing values

Missing values are ignored by formatters and shown as empty cells by default.
You can customize their display text by setting `na` on a column:

```{r}
reactable(
  data.frame(
    n = c(1, 2, NA, 4, 5),
    x = c(55, 27, NA, NaN, 19),
    y = c(1, NA, 0.25, 0.55, NA)
  ),
  columns = list(
    x = colDef(na = "–", format = colFormat(prefix = "$")),
    y = colDef(na = "NA", format = colFormat(percent = TRUE))
  )
)
```

### Custom data formatting

If none of the built-in formatters apply to your data, you can use a
[custom cell renderer](#cell-rendering) instead.


## Custom Rendering

You can customize how data is displayed using an R or JavaScript function
that returns custom content. R render functions support
[Shiny HTML tags](https://shiny.rstudio.com/articles/tag-glossary.html) (or
[`htmltools`](https://unleash-shiny.rinterface.com/htmltools-overview.html)) and
[HTML widgets](https://www.htmlwidgets.org/),
while JavaScript render functions allow for more dynamic behavior.

You can also render content as HTML using `colDef(html = TRUE)`. Note that all
raw HTML is escaped by default.

See [Custom Rendering](custom-rendering.html) for details on how to
use render functions, and the [Demo Cookbook](cookbook/cookbook.html)
for even more examples of custom rendering.

::: {.callout-note}
**Note:** Custom rendering changes how data is displayed without affecting the underlying
data. Sorting, filtering, and grouping will still work on the original data.
:::

### Cell rendering

#### R render function

```{r}
data <- MASS::Cars93[1:5, c("Manufacturer", "Model", "Type", "AirBags", "Price")]

reactable(data, columns = list(
  Model = colDef(cell = function(value, index) {
    # Render as a link
    url <- sprintf("https://wikipedia.org/wiki/%s_%s", data[index, "Manufacturer"], value)
    htmltools::tags$a(href = url, target = "_blank", as.character(value))
  }),
  AirBags = colDef(cell = function(value) {
    # Render as an X mark or check mark
    if (value == "None") "\u274c No" else "\u2714\ufe0f Yes"
  }),
  Price = colDef(cell = function(value) {
    # Render as currency
    paste0("$", format(value * 1000, big.mark = ","))
  })
))
```

#### JavaScript render function

```{r}
data <- MASS::Cars93[1:5, c("Manufacturer", "Model", "Type", "AirBags", "Price")]

reactable(data, columns = list(
  Model = colDef(html = TRUE, cell = JS('
    function(cellInfo) {
      // Render as a link
      const url = `https://wikipedia.org/wiki/${cellInfo.row["Manufacturer"]}_${cellInfo.value}`
      return `<a href="${url}" target="_blank">${cellInfo.value}</a>`
    }
  ')),
  AirBags = colDef(cell = JS("
    function(cellInfo) {
      // Render as an X mark or check mark
      return cellInfo.value === 'None' ? '\u274c No' : '\u2714\ufe0f Yes'
    }
  ")),
  Price = colDef(cell = JS("
    function(cellInfo) {
      // Render as currency
      return '$' + (cellInfo.value * 1000).toLocaleString()
    }
  "))
))
```

#### Embedding HTML widgets

```{r message=FALSE}
library(dplyr)
library(sparkline)

data <- chickwts %>%
  group_by(feed) %>%
  summarise(weight = list(weight)) %>%
  mutate(boxplot = NA, sparkline = NA)

reactable(data, columns = list(
  weight = colDef(cell = function(values) {
    sparkline(values, type = "bar", chartRangeMin = 0, chartRangeMax = max(chickwts$weight))
  }),
  boxplot = colDef(cell = function(value, index) {
    sparkline(data$weight[[index]], type = "box")
  }),
  sparkline = colDef(cell = function(value, index) {
    sparkline(data$weight[[index]])
  })
))
```

### Grouped cell rendering

```{r}
data <- MASS::Cars93[10:22, c("Manufacturer", "Model", "Type", "Price", "MPG.city")]

reactable(
  data,
  groupBy = c("Manufacturer", "Type"),
  columns = list(
    Manufacturer = colDef(
      # Render grouped cells without the row count
      grouped = JS("function(cellInfo) {
        return cellInfo.value
      }")
    ),
    Type = colDef(
      # Render grouped cells with the row count, only if there are multiple sub rows
      grouped = JS("function(cellInfo) {
        if (cellInfo.subRows.length > 1) {
          return cellInfo.value + ' (' + cellInfo.subRows.length + ')'
        }
        return cellInfo.value
      }")
    )
  )
)
```

### Aggregated cell rendering

```{r, message=FALSE}
library(dplyr)

set.seed(10)

data <- sample_n(tail(MASS::Cars93, 9), 30, replace = TRUE) %>%
  select(Manufacturer, Model, Type, Sales = Price)

reactable(
  data,
  groupBy = "Manufacturer",
  searchable = TRUE,
  columns = list(
    Model = colDef(aggregate = "unique"),
    Type = colDef(
      # Render aggregated value as a comma-separated list of unique values
      aggregated = JS("function(cellInfo) {
        const values = cellInfo.subRows.map(function(row) { return row['Type'] })
        const unique = values.reduce(function(obj, v) { obj[v] = true; return obj }, {})
        return Object.keys(unique).join(', ')
      }")
    ),
    Sales = colDef(
      aggregate = "sum",
      # Render aggregated cell as currency
      aggregated = JS("function(cellInfo) {
        return '$' + cellInfo.value.toFixed(2)
      }")
    )
  )
)
```

### Header rendering

::: {.callout}
This example requires reactable v0.3.0 or above.
:::

```{r}
library(htmltools)

reactable(
  iris[1:5, ],
  defaultColDef = colDef(header = function(value) {
    units <- div(style = "color: #737373", "cm")
    div(title = value, value, units)
  }),
  columns = list(
    Petal.Width = colDef(
      name = "Petal Width",
      html = TRUE,
      align = "left",
      header = JS('function(column) {
        return column.name + `<div style="color: #737373">cm</div>`
      }')
    ),
    Species = colDef(header = function(value) {
      tags$a(href = "https://wikipedia.org/wiki/List_of_Iris_species", value)
    })
  )
)
```

### Custom metadata {#custom-meta-rendering}

::: {.callout}
New in v0.4.0
:::

You can pass arbitrary data from R to JavaScript render functions using the `meta`
argument in `reactable()`.

`meta` should be a named list of values that can also be `JS()` expressions or functions.
Custom metadata can be accessed from JavaScript using the `state.meta` property, and updated
using `updateReactable()` in Shiny or [`Reactable.setMeta()`](./javascript-api.html#reactable-setmeta)
in the JavaScript API.

Use custom metadata to:

- Simplify JavaScript render functions that need access to data outside of the table
- Dynamically change how data is formatted without rerendering the table
- Share JavaScript code or data between different render functions

```{r}
library(htmltools)

data <- MASS::Cars93[1:6, c("Manufacturer", "Model", "Type", "Price", "MPG.city")]

exchange_rates <- list(
  USD = 1,
  CAD = 1.30,
  JPY = 137.56
)

tbl <- reactable(
  data,
  columns = list(
    Price = colDef(
      cell = JS("function(cellInfo, state) {
        const { currency, exchangeRates } = state.meta
        const converted = cellInfo.value * exchangeRates[currency]
        return converted.toLocaleString(undefined, { style: 'currency', currency: currency })
      }")
    )
  ),
  meta = list(
    currency = "USD",
    exchangeRates = exchange_rates
  ),
  elementId = "cars-currency-table"
)

browsable(
  tagList(
    tags$label(
      "Currency",
      tags$select(
        onchange = "Reactable.setMeta('cars-currency-table', { currency: event.target.value })",
        lapply(names(exchange_rates), tags$option)
      )
    ),

    tags$hr("aria-hidden" = "true"),

    tbl
  )
)
```


## Footers

You can add column footers using the `footer` argument in `colDef()`.

`footer` can either be custom content to render (e.g., a character string or HTML tag),
or a custom render function. See [Custom Rendering](custom-rendering.html) to learn
more about using custom render functions.

### R render function {#footers-r-render-function}

```{r, message=FALSE}
library(dplyr)
library(htmltools)

data <- MASS::Cars93[18:47, ] %>%
  select(Manufacturer, Model, Type, Sales = Price)

reactable(
  data,
  defaultPageSize = 5,
  columns = list(
    Manufacturer = colDef(footer = "Total"),
    Sales = colDef(footer = function(values) sprintf("$%.2f", sum(values)))
  ),
  defaultColDef = colDef(footerStyle = list(fontWeight = "bold"))
)
```

### JavaScript render function {#footers-js-render-function}

::: {.callout}
This example requires reactable v0.3.0 or above.
:::

```{r}
reactable(
  data,
  searchable = TRUE,
  defaultPageSize = 5,
  minRows = 5,
  columns = list(
    Manufacturer = colDef(footer = "Total"),
    Sales = colDef(
      footer = JS("function(column, state) {
        let total = 0
        state.sortedData.forEach(function(row) {
          total += row[column.id]
        })
        return '$' + total.toFixed(2)
      }")
    )
  ),
  defaultColDef = colDef(footerStyle = list(fontWeight = "bold"))
)
```

### Embedding HTML widgets {#embedding-html-widgets-footers}

```{r}
library(sparkline)

reactable(
  iris[1:20, ],
  defaultPageSize = 5,
  bordered = TRUE,
  defaultColDef = colDef(footer = function(values) {
    if (!is.numeric(values)) return()
    sparkline(values, type = "box", width = 100, height = 30)
  })
)
```


## Expandable Row Details

::: {.callout}
This example requires reactable v0.3.0 or above.
:::

You can make rows expandable with additional content through `details`, which
takes an R or JavaScript render function. See [Custom Rendering](custom-rendering.html)
for details on how to use render functions.

```{r}
reactable(iris[1:5, ], details = function(index) {
  htmltools::div(
    "Details for row: ", index,
    htmltools::tags$pre(paste(capture.output(iris[index, ]), collapse = "\n"))
  )
})
```

The details column can be customized by providing a `colDef()` instead. This can
be used to add a column name, render HTML content, or change the column width:

```{r}
reactable(iris[1:5, ], details = colDef(
  name = "More",
  details = JS("function(rowInfo) {
    return `Details for row: ${rowInfo.index}` +
      `<pre>${JSON.stringify(rowInfo.values, null, 2)}</pre>`
  }"),
  html = TRUE,
  width = 60
))
```

### Nested tables
With R render functions, you can render HTML tags, HTML widgets, and even nested tables:
```{r}
data <- unique(CO2[, c("Plant", "Type")])

reactable(data, details = function(index) {
  plant_data <- CO2[CO2$Plant == data$Plant[index], ]
  htmltools::div(style = "padding: 1rem",
    reactable(plant_data, outlined = TRUE)
  )
})
```

### Conditional row details
R render functions support conditional rendering. If a render function
returns `NULL`, the row won't be expandable:
```{r}
reactable(iris[1:5, ], details = function(index) {
  if (index %in% c(3, 5)) {
    reactable(data.frame(x = c(1, 2, 3), y = c("a", "b", "c")), fullWidth = FALSE)
  }
})
```

### Multiple row details

::: {.callout}
This example requires reactable v0.3.0 or above.
:::

You can add `details` to individual columns, and even show multiple details for a row:

```{r}
reactable(iris[1:5, ],
  details = function(index) {
    if (index %in% c(3, 5)) {
      reactable(data.frame(x = c(1, 2, 3), y = c("a", "b", "c")), fullWidth = FALSE)
    }
  },
  columns = list(
    Petal.Length = colDef(details = function(index) {
      paste("Petal.Length: ", iris[index, "Petal.Length"])
    }),
    Sepal.Length = colDef(format = colFormat(digits = 1), details = JS("
      function(rowInfo) {
        return 'Sepal.Length: ' + rowInfo.values['Sepal.Length']
      }
    "))
  )
)
```

### Default expanded rows

You can expand all rows by default by setting `defaultExpanded` to `TRUE`:

```{r}
reactable(
  iris[1:12, ],
  defaultPageSize = 4,
  details = function(index) paste("Details for row:", index),
  defaultExpanded = TRUE
)
```


## Conditional Styling

You can conditionally style a table using functions that return inline styles
or CSS classes. Just like with custom rendering, style functions can either be
in R or JavaScript.

See [Conditional Styling](conditional-styling.html) for details on how to
use style functions, and the [Demo Cookbook](cookbook/cookbook.html)
for even more examples of conditional styling.

### Cell styling

#### R style function

```{r}
reactable(sleep[1:6, ], columns = list(
  extra = colDef(
    style = function(value) {
      if (value > 0) {
        color <- "#008000"
      } else if (value < 0) {
        color <- "#e00000"
      } else {
        color <- "#777"
      }
      list(color = color, fontWeight = "bold")
    }
  )
))
```

#### JavaScript style function

::: {.callout}
This example requires reactable v0.3.0 or above.
:::

```{r}
reactable(sleep[1:6, ], columns = list(
  extra = colDef(
    style = JS("function(rowInfo) {
      const value = rowInfo.values['extra']
      let color
      if (value > 0) {
        color = '#008000'
      } else if (value < 0) {
        color = '#e00000'
      } else {
        color = '#777'
      }
      return { color: color, fontWeight: 'bold' }
    }")
  )
))
```

### Row styling

#### R style function

```{r row_styling, eval=FALSE}
reactable(sleep[1:6, ], 
  rowStyle = function(index) {
    if (sleep[index, "extra"] < -1) {
      list(background = "rgba(0, 0, 0, 0.05)")
    }
  },
  rowClass = function(index) {
    if (sleep[index, "extra"] < -1) {
      "bold"
    }
  }
)
```

```{css}
.bold {
  font-weight: bold;
}
```

```{r ref.label="row_styling", echo=FALSE}
```

#### JavaScript style function

::: {.callout}
This example requires reactable v0.3.0 or above.
:::

```{r}
reactable(sleep[1:6, ],
  rowStyle = JS("function(rowInfo) {
    if (rowInfo.values['extra'] < -1) {
      return { background: 'rgba(0, 0, 0, 0.05)' }
    }
  }"),
  rowClass = JS("function(rowInfo) {
    if (rowInfo.values['extra'] < -1) {
      return 'bold'
    }
  }")
)
```

### Custom metadata {#custom-meta-styling}

::: {.callout}
New in v0.4.0
:::

You can pass arbitrary data from R to JavaScript style functions using the `meta`
argument in `reactable()`.

`meta` should be a named list of values that can also be `JS()` expressions or functions.
Custom metadata can be accessed from JavaScript using the `state.meta` property, and updated
using `updateReactable()` in Shiny or [`Reactable.setMeta()`](./javascript-api.html#reactable-setmeta)
in the JavaScript API.

Use custom metadata to:

- Simplify JavaScript style functions that need access to data outside of the table
- Dynamically change how data is styled without rerendering the table
- Share JavaScript code or data between different style functions

```{r}
library(htmltools)

data <- MASS::Cars93[1:6, c("Manufacturer", "Model", "Type", "Price", "MPG.city")]

mpg_normalized <- (data$MPG.city - min(data$MPG.city)) / (max(data$MPG.city) - min(data$MPG.city))
mpg_colors <-  rgb(colorRamp(c("#ffe4cc", "#ff9f1a"))(mpg_normalized), maxColorValue = 255)

tbl <- reactable(
  data,
  columns = list(
    MPG.city = colDef(
      style = JS("function(rowInfo, column, state) {
        const { showColors, mpgColors } = state.meta
        if (showColors) {
          return { backgroundColor: mpgColors[rowInfo.index] }
        }
      }")
    )
  ),
  meta = list(
    mpgColors = mpg_colors,
    showColors = TRUE
  ),
  elementId = "cars-colors-table"
)

browsable(
  tagList(
    tags$label(
      tags$input(
        type = "checkbox",
        checked = NA,
        onclick = "Reactable.setMeta('cars-colors-table', function(prevMeta) {
          return { showColors: !prevMeta.showColors }
        })"
      ),
      "Show color scale"
    ),

    tags$hr("aria-hidden" = "true"),

    tbl
  )
)
```


## Table Styling

You can customize table styling using several options, which can all be combined:

### Highlight rows on hover

```{r}
reactable(iris[1:5, ], highlight = TRUE)
```

### Bordered

```{r}
reactable(iris[1:5, ], bordered = TRUE)
```

### Borderless

```{r}
reactable(iris[1:5, ], borderless = TRUE)
```

### Outlined

```{r}
reactable(iris[1:5, ], outlined = TRUE)
```

### Striped

```{r}
reactable(iris[1:5, ], striped = TRUE)
```

### Bordered + striped + highlighting

```{r}
reactable(iris[1:5, ], bordered = TRUE, striped = TRUE, highlight = TRUE)
```

### Outlined + borderless

```{r}
reactable(iris[1:5, ], outlined = TRUE, borderless = TRUE)
```

### Compact

```{r}
reactable(iris[1:5, ], compact = TRUE)
```

### No text wrapping

Long text is wrapped by default, but you can force text to fit on a single line
by setting `wrap` to `FALSE`:

```{r}
data <- aggregate(. ~ Species, iris, toString)

reactable(
  data,
  wrap = FALSE,
  resizable = TRUE,
  bordered = TRUE,
  columns = list(Petal.Length = colDef(name = "Petal Length (cm)", minWidth = 50))
)
```

### Fixed height + sticky header/footer

You can make tables scrollable by setting a fixed height or width.
Headers and footers are sticky by default, so they stay in place when scrolling.

Scrollable tables are automatically made focusable when navigating using a keyboard to
ensure that they're always accessible for keyboard users.

```{r}
reactable(
  iris[1:20, ],
  height = 270,
  striped = TRUE,
  defaultColDef = colDef(
    footer = function(values, name) {
      htmltools::div(name, style = list(fontWeight = 600))
    }
  )
)
```

### Column widths

By default, columns have a minimum width of 100px and stretch to fill the table.
You can control the width of a column using the following arguments in `colDef()`:

- `minWidth` - minimum width of the column in pixels (defaults to `100`)
- `maxWidth` - maximum width of the column in pixels
- `width` - fixed width of the column in pixels (overrides `minWidth` and `maxWidth`)

When columns stretch, `minWidth` also controls the ratio at which columns grow.
For example, if a table consists of 3 columns having `minWidth = 100` each,
the columns will stretch at a ratio of `100:100:100`. Each column will take up 1/3
of the table's width and not shrink below 100px.

Another example: if a table consists of three columns having minimum widths of
200px, 100px, and 100px, the columns will take up 50%, 25%, and 25% of the table's
width respectively:

```{r}
reactable(
  MASS::Cars93[1:6, c("Make", "Type", "Weight")],
  columns = list(
    Make = colDef(minWidth = 200),   # 50% width, 200px minimum
    Type = colDef(minWidth = 100),   # 25% width, 100px minimum
    Weight = colDef(minWidth = 100)  # 25% width, 100px minimum
  ),
  bordered = TRUE
)
```

### No full width

Tables are full width by default, but you can shrink the table to fit its
contents by setting `fullWidth` to `FALSE`:

```{r no_full_width_1, eval=FALSE}
reactable(
  MASS::Cars93[1:5, 1:5],
  fullWidth = FALSE,
  bordered = TRUE,
  defaultColDef = colDef(minWidth = 120)
)
```

<div style="overflow: auto;">
```{r ref.label="no_full_width_1", echo=FALSE}
```
</div>

You can also set a maximum or fixed width on the table:

```{r no_full_width_2, eval=FALSE}
reactable(
  MASS::Cars93[1:5, 1:5],
  bordered = TRUE,
  defaultColDef = colDef(minWidth = 120),
  # Set a maximum width on the table:
  style = list(maxWidth = 650),
  # Or a fixed width:
  width = 650
)
```

<div style="overflow: auto;">
```{r ref.label="no_full_width_2", echo=FALSE}
```
</div>

### Vertical alignment

You can change the vertical alignment of cell content using the `vAlign` or
`headerVAlign` arguments in `colDef()` and `colGroup()`. `vAlign` controls the
alignment of data cells, while `headerVAlign` controls the alignment of header cells.
Possible options are `"top"` (the default), `"center"`, and `"bottom"`.

```{r}
library(dplyr)
library(htmltools)

data <- starwars[1:6, ] %>%
  select(character = name, height, mass, gender, homeworld, species)

reactable(
  data,
  columns = list(
    character = colDef(
      name = "Character / Species",
      # Show species under character names
      cell = function(value, index) {
        species <- data$species[index]
        species <- if (!is.na(species)) species else "Unknown"
        div(
          div(style = list(fontWeight = 600), value),
          div(style = list(fontSize = "0.75rem"), species)
        )
      }
    ),
    species = colDef(show = FALSE)
  ),
  # Vertically center cells and bottom-align headers
  defaultColDef = colDef(vAlign = "center", headerVAlign = "bottom"),
  bordered = TRUE
)
```

### Custom CSS

For more control over styling, you can add custom class names to the table and
apply your own CSS:

```{r}
reactable(
  iris[1:18, ],
  defaultPageSize = 6,
  borderless = TRUE,
  class = "my-tbl",
  defaultColDef = colDef(headerClass = "my-header"),
  columns = list(
    Sepal.Width = colDef(class = "my-col"),
    Petal.Width = colDef(class = "my-col")
  ),
  rowClass = "my-row"
)
```

```{css, echo=FALSE}
.my-tbl {
  border: 1px solid rgba(0, 0, 0, 0.1);
}

.my-header {
  border-width: 1px;
}

.my-col {
  border-right: 1px solid rgba(0, 0, 0, 0.05);
}

.my-row:hover {
  background-color: #f5f8ff;
}
```

In R Markdown documents, you can embed CSS using a `css` language chunk:

```{r, echo=FALSE, comment=NA}
cat('```{css, echo=FALSE}
.my-tbl {
  border: 1px solid rgba(0, 0, 0, 0.1);
}

.my-header {
  border-width: 1px;
}

.my-col {
  border-right: 1px solid rgba(0, 0, 0, 0.05);
}

.my-row:hover {
  background-color: #f5f8ff;
}
```')
```

The examples here embed CSS for demonstration, but it's sometimes better to add CSS
through an external style sheet. To learn more about adding custom CSS through an
external style sheet:

- [Using custom CSS in R Markdown documents](https://bookdown.org/yihui/rmarkdown/html-document.html#custom-css)
- [Using custom CSS in Shiny apps](https://shiny.rstudio.com/articles/css.html)

::: {.callout-note}
**Note:** If you inspect a table's HTML, you might find CSS classes like `.rt-table` on
different elements of the table. These CSS classes are undocumented and subject to change,
so we recommend adding your own custom class names, or using [themes](#theming) to customize
parts of the table that aren't covered by the custom class names.
:::

## Theming

Themes provide a powerful way to customize table styling that can be reused
across tables. You can either set theme variables to change the default styles
(e.g., row stripe color), or add your own custom CSS to specific elements of the table.

To apply a theme, provide a `reactableTheme()` to `theme`:
```{r}
reactable(
  iris[1:30, ],
  searchable = TRUE,
  striped = TRUE,
  highlight = TRUE,
  bordered = TRUE,
  theme = reactableTheme(
    borderColor = "#dfe2e5",
    stripedColor = "#f6f8fa",
    highlightColor = "#f0f5f9",
    cellPadding = "8px 12px",
    style = list(fontFamily = "-apple-system, BlinkMacSystemFont, Segoe UI, Helvetica, Arial, sans-serif"),
    searchInputStyle = list(width = "100%")
  )
)
```

### Global theme

To set the default theme for all tables, use the global `reactable.theme` option:

```{r}
options(reactable.theme = reactableTheme(
  color = "hsl(233, 9%, 87%)",
  backgroundColor = "hsl(233, 9%, 19%)",
  borderColor = "hsl(233, 9%, 22%)",
  stripedColor = "hsl(233, 12%, 22%)",
  highlightColor = "hsl(233, 12%, 24%)",
  inputStyle = list(backgroundColor = "hsl(233, 9%, 25%)"),
  selectStyle = list(backgroundColor = "hsl(233, 9%, 25%)"),
  pageButtonHoverStyle = list(backgroundColor = "hsl(233, 9%, 25%)"),
  pageButtonActiveStyle = list(backgroundColor = "hsl(233, 9%, 28%)")
))

reactable(
  iris[1:30, ],
  filterable = TRUE,
  showPageSizeOptions = TRUE,
  striped = TRUE,
  highlight = TRUE,
  details = function(index) paste("Details for row", index)
)
```

```{r, echo=FALSE}
options(reactable.theme = NULL)
```

### Nested selectors

You can use nested CSS selectors in theme styles to target the current element,
using `&` as the selector, or other child elements (just like in Sass).
This is useful for adding pseudo-classes like `&:hover`, or adding styles in a
certain context like `.outer-container &`.

For example, to highlight headers when sorting:

```{r}
reactable(
  iris[1:5, ],
  columns = list(Sepal.Length = colDef(sortable = FALSE)),
  showSortable = TRUE,
  theme = reactableTheme(
    headerStyle = list(
      "&:hover[aria-sort]" = list(background = "hsl(0, 0%, 96%)"),
      "&[aria-sort='ascending'], &[aria-sort='descending']" = list(background = "hsl(0, 0%, 96%)"),
      borderColor = "#555"
    )
  )
)
```

Or to apply a dark theme when a parent element has a certain class, like `.dark`:

```{r theme_toggle, eval=FALSE}
theme <- reactableTheme(
  style = list(".dark &" = list(color = "#fff", background = "#282a36")),
  cellStyle = list(".dark &" = list(borderColor = "rgba(255, 255, 255, 0.15)")),
  headerStyle = list(".dark &" = list(borderColor = "rgba(255, 255, 255, 0.15)")),
  paginationStyle = list(".dark &" = list(borderColor = "rgba(255, 255, 255, 0.15)")),
  rowHighlightStyle = list(".dark &" = list(background = "rgba(255, 255, 255, 0.04)")),
  pageButtonHoverStyle = list(".dark &" = list(background = "rgba(255, 255, 255, 0.08)")),
  pageButtonActiveStyle = list(".dark &" = list(background = "rgba(255, 255, 255, 0.1)"))
)

tbl <- reactable(iris[1:12, ], highlight = TRUE, defaultPageSize = 6, theme = theme)

# Simple theme toggle button
tags$button(onclick = "document.querySelector('.themeable-tbl').classList.toggle('dark')",
            "Toggle light/dark")

# Start with the dark theme enabled
div(class = "themeable-tbl dark", tbl)
```

```{r ref.label="theme_toggle", echo=FALSE}
```

### Dynamic theming

Themes can also be functions that return a `reactableTheme()` for
context-specific styling.

For example, to style tables in RStudio R Notebooks only when a dark
editor theme is active:

```{r, eval=FALSE}
options(reactable.theme = function() {
  theme <- reactableTheme(
    color = "hsl(233, 9%, 85%)",
    backgroundColor = "hsl(233, 9%, 19%)",
    borderColor = "hsl(233, 9%, 22%)",
    stripedColor = "hsl(233, 12%, 22%)",
    highlightColor = "hsl(233, 12%, 24%)",
    inputStyle = list(backgroundColor = "hsl(233, 9%, 25%)"),
    selectStyle = list(backgroundColor = "hsl(233, 9%, 25%)"),
    pageButtonHoverStyle = list(backgroundColor = "hsl(233, 9%, 25%)"),
    pageButtonActiveStyle = list(backgroundColor = "hsl(233, 9%, 28%)")
  )

  if (isTRUE(getOption("rstudio.notebook.executing"))) {
    if (requireNamespace("rstudioapi", quietly = TRUE) && rstudioapi::getThemeInfo()$dark) {
      return(theme)
    }
  }
})
```


## Column Groups
You can create column groups by passing a list of `colGroup()` definitions to `columnGroups`:
```{r}
reactable(
  iris[1:5, ],
  columns = list(
    Sepal.Length = colDef(name = "Length"),
    Sepal.Width = colDef(name = "Width"),
    Petal.Length = colDef(name = "Length"),
    Petal.Width = colDef(name = "Width")
  ),
  columnGroups = list(
    colGroup(name = "Sepal", columns = c("Sepal.Length", "Sepal.Width")),
    colGroup(name = "Petal", columns = c("Petal.Length", "Petal.Width"))
  )
)
```


## Column Resizing
You can make columns resizable by setting `resizable` to `TRUE`:
```{r}
reactable(MASS::Cars93[1:5, ], resizable = TRUE, wrap = FALSE, bordered = TRUE)
```


## Sticky Columns

You can make columns sticky when scrolling horizontally using the `sticky`
argument in `colDef()` or `colGroup()`. Set `sticky` to either `"left"`
or `"right"` to make the column stick to the left or right side.

```{r}
reactable(
  MASS::Cars93[1:5, ],
  columns = list(
    Manufacturer = colDef(
      sticky = "left",
      # Add a right border style to visually distinguish the sticky column
      style = list(borderRight = "1px solid #eee"),
      headerStyle = list(borderRight = "1px solid #eee")
    ),
    Make = colDef(
      sticky = "right",
      # Add a left border style to visually distinguish the sticky column
      style = list(borderLeft = "1px solid #eee"),
      headerStyle = list(borderLeft = "1px solid #eee")
    )
  ),
  defaultColDef = colDef(minWidth = 150)
)
```

### Multiple sticky columns

```{r}
# Background style to visually distinguish sticky columns
sticky_style <- list(backgroundColor = "#f7f7f7")

reactable(
  MASS::Cars93[1:5, ],
  columns = list(
    Manufacturer = colDef(
      sticky = "left",
      style = sticky_style,
      headerStyle = sticky_style
    ),
    Model = colDef(
      sticky = "left",
      style = sticky_style,
      headerStyle = sticky_style
    ),
    Type = colDef(
      sticky = "left",
      style = sticky_style,
      headerStyle = sticky_style
    )
  ),
  resizable = TRUE,
  wrap = FALSE,
  bordered = TRUE
)
```

### Sticky column groups

If a column group is sticky, all columns in the group will automatically be
made sticky.

```{r}
reactable(
  MASS::Cars93[1:5, ],
  columnGroups = list(
    colGroup("Make", columns = c("Manufacturer", "Model"), sticky = "left"),
    colGroup("Price", columns = c("Min.Price", "Price", "Max.Price"), sticky = "left")
  ),
  defaultColDef = colDef(footer = "Footer"),
  resizable = TRUE,
  wrap = FALSE,
  bordered = TRUE
)
```


## Row Names and Row Headers

### Row names

Row names are shown by default if present. You can customize the row names column
using `".rownames"` as the column name:

```{r}
reactable(
  USPersonalExpenditure,
  columns = list(
    .rownames = colDef(name = "Category", sortable = TRUE)
  )
)
```

If row names haven't been set explicitly, you can force them to show by setting
`rownames` to `TRUE`:

```{r}
reactable(iris[1:5, ], rownames = TRUE)
```

### Row headers

You can mark up cells in a column as row headers by setting `rowHeader` to `TRUE`
in `colDef()`.

Use this to help users navigate the table using assistive technologies. When cells
are marked up as row headers, assistive technologies will read them aloud while
navigating through cells in the table.

Cells in the row names column are automatically marked up as row headers.

```{r}
data <- MASS::Cars93[1:5, c("Make", "Type", "Price", "MPG.city", "AirBags")]

reactable(
  data,
  columns = list(
    Make = colDef(rowHeader = TRUE, style = list(fontWeight = 600))
  ),
  bordered = TRUE
)
```


## Cell Click Actions
You can add cell click actions using the `onClick` argument, which accepts the
following values:

- `"expand"` to expand the row
- `"select"` to select the row
- A JavaScript function for a custom action, e.g., sending the click event to Shiny

### Expand on click
```{r}
reactable(
  iris[48:52, ],
  groupBy = "Species",
  details = function(index) paste("Details for row:", index),
  onClick = "expand",
  # Give rows a pointer cursor to indicate that they're clickable
  rowStyle = list(cursor = "pointer")
)
```

### Select on click
```{r}
reactable(iris[1:5, ], selection = "multiple", onClick = "select")
```

### Custom action

::: {.callout}
This example requires reactable v0.3.0 or above.
:::

This example uses a custom click action to create custom "show details" action
buttons in each row of the table:

```{r}
data <- cbind(
  MASS::Cars93[1:5, c("Manufacturer", "Model", "Type", "Price")],
  details = NA
)

reactable(
  data,
  columns = list(
    # Render a "show details" button in the last column of the table.
    # This button won't do anything by itself, but will trigger the custom
    # click action on the column.
    details = colDef(
      name = "",
      sortable = FALSE,
      cell = function() htmltools::tags$button("Show details")
    )
  ),
  onClick = JS("function(rowInfo, column) {
    // Only handle click events on the 'details' column
    if (column.id !== 'details') {
      return
    }

    // Display an alert dialog with details for the row
    window.alert('Details for row ' + rowInfo.index + ':\\n' + JSON.stringify(rowInfo.values, null, 2))

    // Send the click event to Shiny, which will be available in input$show_details
    // Note that the row index starts at 0 in JavaScript, so we add 1
    if (window.Shiny) {
      Shiny.setInputValue('show_details', { index: rowInfo.index + 1 }, { priority: 'event' })
    }
  }")
)
```

::: {.callout-warning}
**Warning:** Custom click actions are currently not accessible to keyboard
users, and are generally not recommended. If they must be used, ensure that
they can be triggered by a keyboard through other means, such as a button
in the example above.
:::

## Language Options

You can customize the language in the table by providing a set of `reactableLang()`
options to `language`:
```{r}
reactable(
  iris[1:30, ],
  searchable = TRUE,
  paginationType = "simple",
  language = reactableLang(
    searchPlaceholder = "Search...",
    noData = "No entries found",
    pageInfo = "{rowStart} to {rowEnd} of {rows} entries",
    pagePrevious = "\u276e",
    pageNext = "\u276f",

    # Accessible labels for assistive technologies such as screen readers.
    # These are already set by default, but don't forget to update them when
    # changing visible text.
    pagePreviousLabel = "Previous page",
    pageNextLabel = "Next page"
  )
)
```

### Global language options
To set the default language strings for all tables, use the global
`reactable.language` option:
```{r}
options(reactable.language = reactableLang(
  pageSizeOptions = "\u663e\u793a {rows}",
  pageInfo = "{rowStart} \u81f3 {rowEnd} \u9879\u7ed3\u679c,\u5171 {rows} \u9879",
  pagePrevious = "\u4e0a\u9875",
  pageNext = "\u4e0b\u9875"
))

reactable(iris[1:12, ], defaultPageSize = 4, showPageSizeOptions = TRUE)
```

```{r, echo=FALSE}
options(reactable.language = NULL)
```

## Shiny

To use reactable in Shiny apps, use `renderReactable()` and `reactableOutput()`:

```{r, eval=FALSE}
library(shiny)
library(reactable)

ui <- fluidPage(
  titlePanel("reactable example"),
  reactableOutput("table")
)

server <- function(input, output, session) {
  output$table <- renderReactable({
    reactable(iris)
  })
}

shinyApp(ui, server)
```

### Row selection

You can enable row selection by setting `selection` to `"single"`
for single selection, or `"multiple"` for multiple selection.

To get the selected rows in Shiny, use `getReactableState()`.
The selected rows are given as a vector of row indices
(e.g. `c(1, 6, 4)`) or `NULL` if no rows are selected.

```{r, eval=FALSE}
library(shiny)
library(reactable)

ui <- fluidPage(
  titlePanel("row selection example"),
  reactableOutput("table"),
  verbatimTextOutput("selected")
)

server <- function(input, output, session) {
  selected <- reactive(getReactableState("table", "selected"))

  output$table <- renderReactable({
    reactable(iris, selection = "multiple", onClick = "select")
  })

  output$selected <- renderPrint({
    print(selected())
  })

  observe({
    print(iris[selected(), ])
  })
}

shinyApp(ui, server)
```

#### Default selected rows

You can preselect rows by specifying a vector of row indices in `defaultSelected`:
```{r}
reactable(iris[1:4, ], selection = "multiple", defaultSelected = c(1, 3))
```

#### Style selected rows

You can style selected rows using `rowSelectedStyle` in `reactableTheme()`:
```{r}
reactable(
  iris[1:4, ],
  selection = "multiple",
  defaultSelected = c(1, 3),
  borderless = TRUE,
  onClick = "select",
  theme = reactableTheme(
    rowSelectedStyle = list(backgroundColor = "#eee", boxShadow = "inset 2px 0 0 0 #ffa62d")
  )
)
```

Or using a `rowStyle` or `rowClass` JavaScript function:
```{r}
reactable(
  MASS::Cars93[10:22, c("Manufacturer", "Model", "Type", "Price", "MPG.city")],
  groupBy = "Manufacturer",
  selection = "multiple",
  defaultSelected = c(1, 2),
  borderless = TRUE,
  onClick = "select",
  rowStyle = JS("function(rowInfo) {
    if (rowInfo && rowInfo.selected) {
      return { backgroundColor: '#eee', boxShadow: 'inset 2px 0 0 0 #ffa62d' }
    }
  }")
)
```

#### Customize the selection column

You can customize the selection column using `".selection"` as the column name:
```{r}
reactable(
  MASS::Cars93[1:4, ],
  columns = list(
    .selection = colDef(
      width = 80,
      sticky = "left",
      style = list(cursor = "pointer"),
      headerStyle = list(cursor = "pointer")
    )
  ),
  selection = "multiple",
  onClick = "select",
  resizable = TRUE,
  wrap = FALSE,
  bordered = TRUE
)
```

### Update a reactable instance

You can update the selected rows, expanded rows, current page, or data using
`updateReactable()`:

```{r, eval=FALSE}
library(shiny)
library(reactable)

data <- MASS::Cars93[, 1:7]

ui <- fluidPage(
  actionButton("select_btn", "Select rows"),
  actionButton("clear_btn", "Clear selection"),
  actionButton("expand_btn", "Expand rows"),
  actionButton("collapse_btn", "Collapse rows"),
  actionButton("page_btn", "Change page"),
  selectInput("filter_type", "Filter type", unique(data$Type), multiple = TRUE),
  reactableOutput("table")
)

server <- function(input, output) {
  output$table <- renderReactable({
    reactable(
      data,
      filterable = TRUE,
      searchable = TRUE,
      selection = "multiple",
      details = function(index) paste("Details for row:", index)
    )
  })

  observeEvent(input$select_btn, {
    # Select rows
    updateReactable("table", selected = c(1, 3, 5))
  })

  observeEvent(input$clear_btn, {
    # Clear row selection
    updateReactable("table", selected = NA)
  })

  observeEvent(input$expand_btn, {
    # Expand all rows
    updateReactable("table", expanded = TRUE)
  })

  observeEvent(input$collapse_btn, {
    # Collapse all rows
    updateReactable("table", expanded = FALSE)
  })

  observeEvent(input$page_btn, {
    # Change current page
    updateReactable("table", page = 3)
  })

  observe({
    # Filter data
    filtered <- if (length(input$filter_type) > 0) {
      data[data$Type %in% input$filter_type, ]
    } else {
      data
    }
    updateReactable("table", data = filtered)
  })
}

shinyApp(ui, server)
```

### Get the state of a reactable instance

You can get the current state of a table using `getReactableState()`.

By default, `getReactableState()` returns a named list with the following values:

- `page`: the current page
- `pageSize`: the page size
- `pages`: the number of pages
- `sorted`: the sorted columns - a named list of columns with values of `"asc"` for
  ascending order or `"desc"` for descending order, or `NULL` if no columns are sorted
- `selected`: the selected rows - a numeric vector of row indices, or `NULL` if no rows are selected

To only watch for changes on a specific value, you can use the optional `name` argument, like
`getReactableState(outputId, "selected")`.

```{r, eval=FALSE}
library(shiny)
library(reactable)
library(htmltools)

ui <- fluidPage(
  actionButton("prev_page_btn", "Previous page"),
  actionButton("next_page_btn", "Next page"),
  reactableOutput("table"),
  verbatimTextOutput("table_state"),
  uiOutput("selected_row_details")
)

server <- function(input, output) {
  output$table <- renderReactable({
    reactable(
      MASS::Cars93[, 1:5],
      showPageSizeOptions = TRUE,
      selection = "multiple",
      onClick = "select"
    )
  })

  output$table_state <- renderPrint({
    state <- req(getReactableState("table"))
    print(state)
  })

  observeEvent(input$prev_page_btn, {
    # Change to the previous page
    page <- getReactableState("table", "page")
    if (page > 1) {
      updateReactable("table", page = page - 1)
    }
  })

  observeEvent(input$next_page_btn, {
    # Change to the next page
    state <- getReactableState("table")
    if (state$page < state$pages) {
      updateReactable("table", page = state$page + 1)
    }
  })
  
  output$selected_row_details <- renderUI({
    selected <- getReactableState("table", "selected")
    req(selected)
    details <- MASS::Cars93[selected, -c(1:5)]
    tagList(
      h2("Selected row details"),
      tags$pre(
        paste(capture.output(print(details, width = 1200)), collapse = "\n")
      )
    )
  })
}

shinyApp(ui, server)
```


## Cross-Widget Interactions

You can link selection and filtering with other HTML widgets in an R Markdown
document or Shiny app using [Crosstalk](https://rstudio.github.io/crosstalk).
To get started, install the `crosstalk` package and wrap your data frame in a
`crosstalk::SharedData` object:

```{r, eval=FALSE}
install.packages("crosstalk")

library(crosstalk)

data <- SharedData$new(iris)
```

Then, pass the shared data to `reactable()` and any other Crosstalk-compatible
HTML widget or filter input:

```{r, eval=FALSE}
reactable(data)

filter_slider("sepal_length", "Sepal Length", data, ~Sepal.Length)
```

For more examples and a list of Crosstalk-compatible widgets, check out
[Using Crosstalk](https://rstudio.github.io/crosstalk/using.html)
in the Crosstalk package documentation.

### Filtering

Tables can be filtered by widgets that support Crosstalk's filtering API,
such as Crosstalk's `filter_checkbox()`, `filter_slider()`, and `filter_select()` inputs:

```{r}
library(crosstalk)

cars <- MASS::Cars93[1:20, c("Manufacturer", "Model", "Type", "Price")]
data <- SharedData$new(cars)

shiny::fluidRow(
  shiny::column(
    4,
    filter_checkbox("type", "Type", data, ~Type),
    filter_slider("price", "Price", data, ~Price, width = "100%"),
    filter_select("mfr", "Manufacturer", data, ~Manufacturer)
  ),
  shiny::column(
    8,
    reactable(data, minRows = 10)
  )
)
```

::: {.callout-note}
**Note:** This example uses `shiny::fluidRow()` and `shiny::column()` to create
a Bootstrap grid layout, which works with all Bootstrap versions. `crosstalk::bscols()`
can also create a grid, but is only compatible with Bootstrap 3. If you're not using
Bootstrap, here's an alternative way to create a responsive grid using CSS grid.

<details>
<summary>Example: grid layout using CSS grid</summary>
```{r, eval=FALSE}
library(crosstalk)

cars <- MASS::Cars93[1:20, c("Manufacturer", "Model", "Type", "Price")]
data <- SharedData$new(cars)

div(
  style = "display: grid; grid-template-columns: repeat(auto-fit, minmax(200px, 1fr)); gap: 0.75rem;",
  div(
    filter_checkbox("type", "Type", data, ~Type),
    filter_slider("price", "Price", data, ~Price, width = "100%"),
    filter_select("mfr", "Manufacturer", data, ~Manufacturer)
  ),
  div(
    style = "grid-column: span 2;",
    reactable(data, minRows = 10)
  )
)
```
</details>
:::

### Linked selection

Table selection state can be linked with other widgets that support Crosstalk's
linked selection (or linked brushing) API.

In this example, you can select rows to highlight points on the map, or select areas
on the map to highlight rows in the table.

```{r, message=FALSE}
library(crosstalk)
library(leaflet)
library(dplyr)

# A SpatialPointsDataFrame for the map.
# Set a group name to share data points with the table.
brew_sp <- SharedData$new(breweries91, group = "breweries")

# A regular data frame (without coordinates) for the table.
# Use the same group name as the map data.
brew_data <- as_tibble(breweries91) %>%
  select(brewery, address, village, founded) %>%
  SharedData$new(group = "breweries")

map <- leaflet(brew_sp) %>%
  addTiles() %>%
  addMarkers()

tbl <- reactable(
  brew_data,
  selection = "multiple",
  onClick = "select",
  rowStyle = list(cursor = "pointer"),
  minRows = 10
)

htmltools::browsable(
  htmltools::tagList(map, tbl)
)
```

## JavaScript API

You can use the JavaScript API to create custom interactive controls for your table
without the use of Shiny, or add cross-widget interactions beyond what Crosstalk provides.

See the [JavaScript API guide](javascript-api.html) for details on how to use the
JavaScript API in R Markdown documents or Shiny apps, and the full API reference.

### CSV download button

[`Reactable.downloadDataCSV()`](./javascript-api.html#reactable-downloaddatacsv) downloads table data to a CSV file,
including any filters that have been applied. See the JavaScript API guide for more details on usage, including
how to customize the field separator, decimal separator, included columns, and more.

```{r}
library(htmltools)
library(fontawesome)

data <- MASS::Cars93[1:15, c("Manufacturer", "Model", "Type", "Price")]

htmltools::browsable(
  tagList(
    tags$button(
      tagList(fontawesome::fa("download"), "Download as CSV"),
      onclick = "Reactable.downloadDataCSV('cars-download-table', 'cars.csv')"
    ),

    reactable(
      data,
      searchable = TRUE,
      defaultPageSize = 5,
      elementId = "cars-download-table"
    )
  )
)
```

### CSV download button in Shiny

While you can create download buttons in Shiny using `shiny::downloadButton()`, you may
still prefer to use the JavaScript API, as [`Reactable.downloadDataCSV()`](./javascript-api.html#reactable-downloaddatacsv)
automatically applies any client-side filtering that has been done to the table.

```{r, eval=FALSE}
library(shiny)
library(reactable)
library(htmltools)

csvDownloadButton <- function(id, filename = "data.csv", label = "Download as CSV") {
  tags$button(
    tagList(icon("download"), label),
    onclick = sprintf("Reactable.downloadDataCSV('%s', '%s')", id, filename)
  )
}

ui <- fluidPage(
  csvDownloadButton("cars_table", filename = "cars.csv"),
  reactableOutput("cars_table")
)

server <- function(input, output) {
  output$cars_table <- renderReactable({
    reactable(
      MASS::Cars93[, c("Manufacturer", "Model", "Type", "Price")],
      searchable = TRUE
    )
  })
}

shinyApp(ui, server)
```

### Custom column filter

```{r}
library(htmltools)

data <- MASS::Cars93[1:15, c("Manufacturer", "Model", "Type", "Price")]

htmltools::browsable(
  tagList(
    div(
      div(tags$label("Filter Type", `for` = "cars-type-filter")),
      tags$select(
        id = "cars-type-filter",
        onchange = "Reactable.setFilter('cars-filter-table', 'Type', this.value)",
        tags$option("All", value = ""),
        lapply(unique(data$Type), tags$option)
      )
    ),
    
    tags$hr("aria-hidden" = "true"),

    reactable(data, defaultPageSize = 5, elementId = "cars-filter-table")
  )
)
```

### Custom search input

```{r}
library(htmltools)

data <- MASS::Cars93[1:15, c("Manufacturer", "Model", "Type", "Price")]

htmltools::browsable(
  tagList(
    div(
      style = "margin-bottom: 0.75rem",
      tags$input(
        type = "text",
        placeholder = "Search for cars...",
        style = "padding: 0.25rem 0.5rem; width: 100%",
        oninput = "Reactable.setSearch('cars-search-table', this.value)"
      )
    ),

    reactable(data, defaultPageSize = 5, elementId = "cars-search-table")
  )
)
```

### Column grouping select

```{r}
library(dplyr)
library(htmltools)

set.seed(10)

data <- sample_n(tail(MASS::Cars93, 9), 30, replace = TRUE) %>%
  select(Manufacturer, Model, Type, Sales = Price)

htmltools::browsable(
  tagList(
    div(tags$label("Group by", `for` = "cars-grouping-select")),
    tags$select(
      id = "cars-grouping-select",
      onchange = "Reactable.setGroupBy('cars-grouping-table', this.value ? [this.value] : [])",
      tags$option("None", value = ""),
      lapply(c("Manufacturer", "Model", "Type"), tags$option)
    ),

    tags$hr("aria-hidden" = "true"),

    reactable(
      data,
      columns = list(
        Manufacturer = colDef(aggregate = "unique"),
        Model = colDef(aggregate = "unique"),
        Type = colDef(aggregate = "unique"),
        Sales = colDef(aggregate = "sum", format = colFormat(currency = "USD"))
      ),
      defaultPageSize = 5,
      minRows = 5,
      elementId = "cars-grouping-table"
    )
  )
)
```

### Row expansion toggle button

```{r}
library(htmltools)

data <- MASS::Cars93[1:5, c("Manufacturer", "Model", "Type", "Price")]

htmltools::browsable(
  tagList(
    tags$button(
      "Expand/collapse all",
      onclick = "Reactable.toggleAllRowsExpanded('cars-expansion-table')"
    ),

    reactable(
      data,
      groupBy = "Manufacturer",
      defaultPageSize = 5,
      elementId = "cars-expansion-table"
    )
  )
)
```

### Column visibility toggle button

::: {.callout}
New in v0.4.0
:::

```{r}
library(htmltools)

data <- MASS::Cars93[1:5, c("Manufacturer", "Model", "Type", "Price",
                            "Passengers", "DriveTrain", "Cylinders", "EngineSize")]

htmltools::browsable(
  tagList(
    tags$button(
      "Show/hide more columns",
      onclick = "Reactable.setHiddenColumns('cars-vis-table', prevColumns => {
        return prevColumns.length === 0 ? ['Passengers', 'DriveTrain', 'Cylinders', 'EngineSize'] : []
      })"
    ),
    reactable(
      data,
      columns = list(
        Passengers = colDef(show = FALSE),
        DriveTrain = colDef(show = FALSE),
        Cylinders = colDef(show = FALSE),
        EngineSize = colDef(show = FALSE)
      ),
      elementId = "cars-vis-table"
    )
  )
)
```


```{css echo=FALSE}
/* rmarkdown html documents */
.main-container {
  font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Helvetica, Arial, sans-serif;
}

.main-container blockquote {
  font-size: inherit;
}
```

---
title: "100k Rows"
output: html_document
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)

set.seed(5)

options(reactable.static = TRUE)
```

```{r}
library(reactable)

rows <- 100000
dates <- seq.Date(as.Date("2018-01-01"), as.Date("2018-12-01"), "day")
data <- data.frame(
  index = seq_len(rows),
  date = sample(dates, rows, replace = TRUE),
  city = sample(names(precip), rows, replace = TRUE),
  state = sample(rownames(USArrests), rows, replace = TRUE),
  temp = round(runif(rows, 0, 100), 1),
  stringsAsFactors = FALSE
)

reactable(
  data,
  filterable = TRUE,
  searchable = TRUE,
  minRows = 10,
  highlight = TRUE,
  columns = list(
    state = colDef(
      html = TRUE,
      cell = JS('function(cell) {
        return `<a href="https://wikipedia.org/wiki/${cell.value}">${cell.value}</a>`
      }')
    )
  ),
  details = colDef(
    html = TRUE,
    details = JS("function(rowInfo) {
      return `Details for row: ${rowInfo.index}` +
        `<pre>${JSON.stringify(rowInfo.values, null, 2)}</pre>`
    }")
  )
)
```

```{css echo=FALSE}
/* rmarkdown html documents */
.main-container {
  font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Helvetica, Arial, sans-serif;
}
```

---
title: "Sticky Selection Column"
output: html_document
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)

library(reactable)
```

### Using inline styles

```{r}
reactable(
  MASS::Cars93,
  selection = "multiple",
  columns = list(
    .selection = colDef(
      style = list(position = "sticky", left = 0, background = "#fff", zIndex = 1),
      headerStyle = list(position = "sticky", left = 0, background = "#fff", zIndex = 1),
      width = 45
    ),
    Manufacturer = colDef(
      style = list(position = "sticky", left = 45, background = "#fff", zIndex = 1),
      headerStyle = list(position = "sticky", left = 45, background = "#fff", zIndex = 1)
    )
  ),
  onClick = "select",
  wrap = FALSE
)
```

### Using CSS

```{r sticky, eval=FALSE}
reactable(
  MASS::Cars93,
  selection = "multiple",
  columns = list(
    .selection = colDef(
      class = "sticky left-col-1",
      headerClass = "sticky left-col-1",
      width = 45
    ),
    Manufacturer = colDef(
      class = "sticky left-col-2",
      headerClass = "sticky left-col-2"
    ),
    Model = colDef(
      class = "sticky left-col-3",
      headerClass = "sticky left-col-3"
    )
  ),
  onClick = "select",
  wrap = FALSE
)
```

```{css}
.sticky {
  position: sticky !important;
  background: #fff;
  z-index: 1;
}

.left-col-1 {
  left: 0;
}

.left-col-2 {
  left: 45px;
}

.left-col-3 {
  left: 145px;
  border-right: 1px solid #eee !important;
}
```

```{r ref.label="sticky", echo=FALSE}
```


```{css echo=FALSE}
/* rmarkdown html documents */
.main-container {
  font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Helvetica, Arial, sans-serif;
}

.main-container blockquote {
  font-size: inherit;
}
```

---
title: "Building the Twitter Followers Demo"
output: html_document
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

```{r, echo=FALSE}
library(reactable)
library(htmltools)

data <- read.csv("twitter-followers/twitter_followers.csv", stringsAsFactors = FALSE)
```

```{r final_table, echo=FALSE}
tbl <- reactable(
  data,
  pagination = FALSE,
  defaultSorted = "exclusive_followers_pct",
  defaultColDef = colDef(headerClass = "header", align = "left"),
  columns = list(
    account = colDef(
      cell = function(value) {
        url <- paste0("https://twitter.com/", value)
        tags$a(href = url, target = "_blank", paste0("@", value))
      },
      width = 150
    ),
    followers = colDef(
      defaultSortOrder = "desc",
      cell = function(value) {
        width <- paste0(value * 100 / max(data$followers), "%")
        value <- format(value, big.mark = ",")
        value <- format(value, width = 9, justify = "right")
        bar <- div(
          class = "bar-chart",
          style = list(marginRight = "0.375rem"),
          div(class = "bar", style = list(width = width, backgroundColor = "#3fc1c9"))
        )
        div(class = "bar-cell", span(class = "number", value), bar)
      }
    ),
    exclusive_followers_pct = colDef(
      name = "Exclusive Followers",
      defaultSortOrder = "desc",
      cell = JS('function(cellInfo) {
        // Format as percentage
        const pct = (cellInfo.value * 100).toFixed(1) + "%"
        // Pad single-digit numbers
        let value = pct.padStart(5)
        // Show % on first row only
        if (cellInfo.viewIndex > 0) {
          value = value.replace("%", " ")
        }
        // Render bar chart
        return `
          <div class="bar-cell">
            <span class="number">${value}</span>
            <div class="bar-chart" style="background-color: #e1e1e1">
              <div class="bar" style="width: ${pct}; background-color: #fc5185"></div>
            </div>
          </div>
        `
      }'),
      html = TRUE
    )
  ),
  compact = TRUE,
  class = "followers-tbl"
)
```

```{r, echo=FALSE}
show_more <- function(id, ..., height = "20rem") {
  tagList(
    div(id = id, class = "show-hidden", ..., div(class = "show-overlay")),
    div(
      style = "text-align: center",
      tags$button(class = "show-btn", onclick = sprintf("toggleShow('%s', this)", id), "Show more")
    ),
    tags$script("function toggleShow(id, btn) {
      let container = document.getElementById(id)
      if (container.classList.contains('show-hidden')) {
        container.classList.remove('show-hidden')
        btn.textContent = 'Show fewer'
      } else {
        container.classList.add('show-hidden')
        btn.textContent = 'Show more'
      }
    }"),
    tags$style(sprintf("
      #%s {
        overflow: hidden;
      }
      
      #%1$s.show-hidden {
        max-height: %2$s;
      }

      #%1$s.show-hidden .show-overlay {
        position: sticky;
        bottom: 0;
        height: 5rem;
        width: 100%%;
        background-image: linear-gradient(to bottom, rgba(255, 255, 255, 0), #fff);
        pointer-events: none;
      }

      .show-btn {
        background: transparent;
        border: 1px solid hsl(0, 0%%, 80%%);
      }

      .show-btn:hover {
        background: hsl(0, 0%%, 93%%);
      }

      .show-btn:active {
        background: hsl(0, 0%%, 90%%);
      }
    ", id, height))
  )
}

show_more(
  id = "followers-show-more",
  div(class = "twitter-followers",
      div(class = "followers-header",
          h2(class = "followers-title", "Candidates whose followers are loyal only to them"),
          "Share of each 2020 candidate's followers who don't follow any other candidates"
      ),
      tbl
  )
)
```

---

The [Twitter Followers](twitter-followers/twitter-followers.html) demo is a
re-creation of the table from the FiveThirtyEight article,
[Which 2020 Candidates Have The Most In Common ... On Twitter?](https://fivethirtyeight.com/features/which-2020-candidates-have-the-most-in-common-on-twitter/)

It's an interactive HTML table with sorting, data formatting, embedded bar charts,
and custom styling. In this article, we'll walk through how we made this table using
reactable, and show a typical workflow for building tables.

## Get the data

FiveThirtyEight shares the data for many of their articles online at
https://data.fivethirtyeight.com, licensed under [CC by 4.0](https://github.com/fivethirtyeight/data/blob/master/LICENSE).

You can download the [raw data](https://github.com/fivethirtyeight/twitter-overlap) for this article,
but we'll conveniently begin working with a cleaned CSV file:
[`twitter_followers.csv`](https://glin.github.io/reactable/articles/twitter-followers/twitter_followers.csv).

```{r}
data <- read.csv("https://glin.github.io/reactable/articles/twitter-followers/twitter_followers.csv",
                 stringsAsFactors = FALSE)

dplyr::glimpse(data)
```

## Create a basic table

The first thing we'll do is create a basic table using `reactable()`:
```{r}
library(reactable)

reactable(data)
```

You can already sort the table, but there's no default sorting on the
"exclusive followers" column. The numeric columns are still unformatted and
sort in ascending order (smallest to largest) by default.

Let's customize the default sorting, add proper column names, and format the data.

We'll use reactable's [built-in column formatters](examples.html#column-formatting)
to add an @ symbol to the Twitter handles, add thousands separators to the
follower counts, and format the percentages with 1 decimal place.

```{r}
reactable(
  data,
  defaultSorted = "exclusive_followers_pct",
  columns = list(
    account = colDef(
      name = "Account",
      format = colFormat(prefix = "@")
    ),
    followers = colDef(
      name = "Followers",
      defaultSortOrder = "desc",
      format = colFormat(separators = TRUE)
    ),
    exclusive_followers_pct = colDef(
      name = "Exclusive Followers",
      defaultSortOrder = "desc",
      format = colFormat(percent = TRUE, digits = 1)
    )
  )
)
```

## Add bar charts

Next, we'll add bar charts to the numeric and percentage columns. The
FiveThirtyEight table uses pure HTML and CSS to create these bar charts,
so we'll do something similar using a method based on CSS flexbox (and also shown
in the [Demo Cookbook](cookbook/cookbook.html#bar-charts)).

We'll generate the bar chart HTML with help from the `htmltools` package,
and render them in the cells via [custom render functions](examples.html#custom-rendering).

Since we're taking over cell rendering with custom render functions, we'll also
have to manually format the numbers and percentages now. Custom cell renderers
override column formatters, but this may change in the future.

::: {.callout-tip}
**Tip:** If you're ever curious to see how an HTML table was made, you can open
your [browser's developer tools](https://developer.mozilla.org/en-US/docs/Learn/Common_questions/What_are_browser_developer_tools)
and inspect the HTML and CSS behind the table. This is how we figured out
how the bar charts were made, which colors and fonts were used, and so on.
:::

```{r}
library(htmltools)

# Render a bar chart with a label on the left
bar_chart <- function(label, width = "100%", height = "0.875rem", fill = "#00bfc4", background = NULL) {
  bar <- div(style = list(background = fill, width = width, height = height))
  chart <- div(style = list(flexGrow = 1, marginLeft = "0.375rem", background = background), bar)
  div(style = list(display = "flex", alignItems = "center"), label, chart)
}

reactable(
  data,
  defaultSorted = "exclusive_followers_pct",
  columns = list(
    account = colDef(
      name = "Account",
      format = colFormat(prefix = "@")
    ),
    followers = colDef(
      name = "Followers",
      defaultSortOrder = "desc",
      # Render the bar charts using a custom cell render function
      cell = function(value) {
        width <- paste0(value * 100 / max(data$followers), "%")
        # Add thousands separators
        value <- format(value, big.mark = ",")
        bar_chart(value, width = width, fill = "#3fc1c9")
      },
      # And left-align the columns
      align = "left"
    ),
    exclusive_followers_pct = colDef(
      name = "Exclusive Followers",
      defaultSortOrder = "desc",
      # Render the bar charts using a custom cell render function
      cell = function(value) {
        # Format as percentages with 1 decimal place
        value <- paste0(format(value * 100, nsmall = 1), "%")
        bar_chart(value, width = value, fill = "#fc5185", background = "#e1e1e1")
      },
      # And left-align the columns
      align = "left"
    )
  )
)
```

The bar charts look good, but they aren't aligned because the numbers
have different widths. Let's fix this by giving each numeric label the same width.
One way to do this would be to format the labels as fixed-width strings, and use
a monospaced font so that each character takes up the same width. (An alternate
method is shown in the [Demo Cookbook](cookbook/cookbook.html#units-on-first-row-only).)

::: {.callout-note}
**Note:** Some fonts have numerals that are all equal in width, but others do not.
In tables with numeric columns, using a font with tabular or monospaced
figures can make the numbers easier to align and read. You can learn more
about the different types of fonts in this article,
[Proportional vs. Tabular Figures](https://www.fonts.com/content/learning/fontology/level-3/numbers/proportional-vs-tabular-figures).
:::

```{r}
reactable(
  data,
  defaultSorted = "exclusive_followers_pct",
  columns = list(
    account = colDef(
      name = "Account",
      format = colFormat(prefix = "@")
    ),
    followers = colDef(
      name = "Followers",
      defaultSortOrder = "desc",
      cell = function(value) {
        width <- paste0(value * 100 / max(data$followers), "%")
        value <- format(value, big.mark = ",")
        # Fix each label using the width of the widest number (incl. thousands separators)
        value <- format(value, width = 9, justify = "right")
        bar_chart(value, width = width, fill = "#3fc1c9")
      },
      align = "left",
      # Use the operating system's default monospace font, and
      # preserve white space to prevent it from being collapsed by default
      style = list(fontFamily = "monospace", whiteSpace = "pre")
    ),
    exclusive_followers_pct = colDef(
      name = "Exclusive Followers",
      defaultSortOrder = "desc",
      cell = function(value) {
        value <- paste0(format(value * 100, nsmall = 1), "%")
        # Fix width here to align single and double-digit percentages
        value <- format(value, width = 5, justify = "right")
        bar_chart(value, width = value, fill = "#fc5185", background = "#e1e1e1")
      },
      align = "left",
      style = list(fontFamily = "monospace", whiteSpace = "pre")
    )
  )
)
```

## Dynamic formatting

The FiveThirtyEight table has a nifty little detail of only showing the percent
sign in first row of the "exclusive followers" column to reduce repetition. If
you sort the table, you'll notice that the percent always appears in the first row
regardless of row order.

To achieve dynamic behavior like this, we'll have to write some JavaScript.
We need access to the client-side (browser) state of the table to know which row
is first after sorting. This isn't possible to do with R (at least without Shiny),
so we'll render the cells using a custom JavaScript render function
as shown in the [Demo Cookbook](cookbook/cookbook.html#units-on-first-row-only).

Since we're switching to a JavaScript render function, we'll unfortunately have
to reformat the data and recreate the bar chart in JavaScript. We'll generate
the same bar chart HTML by concatenating strings, and it'll be kind of ugly
written as a character string in R.

```{r}
reactable(
  data,
  defaultSorted = "exclusive_followers_pct",
  columns = list(
    account = colDef(
      name = "Account",
      format = colFormat(prefix = "@")
    ),
    followers = colDef(
      name = "Followers",
      defaultSortOrder = "desc",
      cell = function(value) {
        width <- paste0(value * 100 / max(data$followers), "%")
        value <- format(value, big.mark = ",")
        value <- format(value, width = 9, justify = "right")
        bar_chart(value, width = width, fill = "#3fc1c9")
      },
      align = "left",
      style = list(fontFamily = "monospace", whiteSpace = "pre")
    ),
    exclusive_followers_pct = colDef(
      name = "Exclusive Followers",
      defaultSortOrder = "desc",
      # Format and render the cell with a JavaScript render function
      cell = JS('function(cellInfo) {
        // Format as a percentage with 1 decimal place
        const pct = (cellInfo.value * 100).toFixed(1) + "%"
        // Fix width of numeric labels
        let value = pct.padStart(5)
        // Show percent sign on first row only
        if (cellInfo.viewIndex > 0) {
          value = value.replace("%", " ")
        }
        // Render bar chart
        return `
          <div style="display: flex; align-items: center;">
            <span style="font-family: monospace; white-space: pre;">${value}</span>
            <div style="flex-grow: 1; margin-left: 0.375rem; height: 0.875rem; background-color: #e1e1e1">
              <div style="height: 100%; width: ${pct}; background-color: #fc5185"></div>
            </div>
          </div>
        `
      }'),
      # Render this column as HTML
      html = TRUE,
      align = "left"
    )
  )
)
```

## Finishing touches

Finally, we'll style the table and add some extra niceties.

We'll display everything on one page using `pagination = FALSE` and reduce
the white space in the table with `compact = TRUE`. We'll apply custom styling
by adding our own CSS classes to the table and column headers. This will be used
to fix the table width, highlight column headers on hover, and change the font
of the table.

FiveThirtyEight uses two commercial fonts for their table: Atlas Grotesk for
text, and Decima Mono for numbers. We'll use similar-looking free fonts from
Google Fonts instead: [Karla](https://fonts.google.com/specimen/Karla) for text,
and [Fira Mono](https://fonts.google.com/specimen/Fira+Mono) for numbers.

We'll also insert links to Twitter accounts using custom cell renderers
and move the bar chart styles to CSS for better organization.

While we won't go into detail here, you can check out the Examples and
Demo Cookbook to learn more about:

- [Adding custom CSS to a table](examples.html#custom-css)
- [Adding custom web fonts to an R Markdown document](cookbook/cookbook.html#custom-fonts)
- [Highlighting column headers on hover](cookbook/cookbook.html#highlight-sorted-headers)
- [Inserting links in a table](cookbook/cookbook.html#insert-links)

The final table and code is shown below. You can also find the source code
for the demo at [`vignettes/twitter-followers/twitter-followers.Rmd`](https://github.com/glin/reactable/blob/main/vignettes/twitter-followers/twitter-followers.Rmd).

---

```{r ref.label="final_demo", echo=FALSE}
```

```{r ref.label="final_table"}
```

```{r final_demo, eval=FALSE}
# Add the title and subtitle
div(class = "twitter-followers",
    div(class = "followers-header",
        h3(class = "followers-title", "Candidates whose followers are loyal only to them"),
        "Share of each 2020 candidate's followers who don't follow any other candidates"
    ),
    tbl
)
```

```{r}
# Add Google Fonts to the page
tags$link(href = "https://fonts.googleapis.com/css?family=Karla:400,700|Fira+Mono&display=fallback",
          rel = "stylesheet")
```

```{css}
/* CSS for the R Markdown document, inserted through a ```{css} code chunk */

/* Styles for the table container, title, and subtitle */
.twitter-followers {
  /* Center the table */
  margin: 0 auto;
  /* Reduce the table width */
  width: 575px;
  font-family: Karla, "Helvetica Neue", Helvetica, Arial, sans-serif;
}

.followers-header {
  margin: 1.125rem 0;
  font-size: 1rem;
}

.followers-title {
  font-size: 1.25rem;
  font-weight: 600;
}

/* Styles for the table */
.followers-tbl {
  font-size: 0.875rem;
  line-height: 1.125rem;
}

.followers-tbl a {
  color: inherit;
  text-decoration: none;
}

.followers-tbl a:hover,
.followers-tbl a:focus {
  text-decoration: underline;
  text-decoration-thickness: max(1px, 0.0625rem);
}

/* Styles for the column headers */
.header {
  border-bottom: 2px solid #555;
  font-size: 0.8125rem;
  font-weight: 400;
  text-transform: uppercase;
}

.header:hover {
  background-color: #eee;
}

/* Styles for the bar charts */
.bar-cell {
  display: flex;
  align-items: center;
}

.number {
  font-family: "Fira Mono", Consolas, Monaco, monospace;
  font-size: 0.84375rem;
  white-space: pre;
}

.bar-chart {
  flex-grow: 1;
  margin-left: 0.375rem;
  height: 0.875rem;
}

.bar {
  height: 100%;
}
```

```{css echo=FALSE}
/* rmarkdown html documents */
.main-container {
  font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Helvetica, Arial, sans-serif;
}

.main-container blockquote {
  font-size: inherit;
}
```

---
title: "Custom Rendering"
output:
  html_document:
    toc: true
    toc_float:
      smooth_scroll: false
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)

library(reactable)
library(htmltools)

propsTable <- function(props) {
  tags$div(
    style = "overflow: auto;",
    tabindex = "0",
    tags$table(
      class = "props-tbl",
      tags$thead(
        tags$tr(
          lapply(colnames(props), function(name) tags$th(name))
        )
      ),
      tags$tbody(
        apply(props, 1, function(row) {
          tags$tr(
            tags$th(scope = "row", tags$code(row[["Property"]])),
            tags$td(tags$code(row[["Example"]])),
            tags$td(row[["Description"]])
          )
        })
      )
    )
  )
}
```


You may want to customize how your data is displayed beyond what the built-in
formatters provide. For example, inserting a link, combining data from multiple
columns, or showing a column total that updates on filtering.

In reactable, you can customize data rendering using either an R or JavaScript
function that returns custom content:

<table style="table-layout: fixed;">
<thead><tr><th>R render functions</th><th>JavaScript render functions</th></tr></thead>
<tbody>
<tr><td>

```{r eval=FALSE}
reactable(iris, columns = list(
  Species = colDef(
    cell = function(value) {
      htmltools::tags$b(value)
    }
  )
))
```

</td><td>

```{r eval=FALSE}
reactable(iris, columns = list(
  Species = colDef(
    cell = JS("function(cellInfo) {
      return `<b>${cellInfo.value}</b>'
    }")
  )
))
```

</td></tr>
<tr><td>

- Easier to use but more static
- Render once, when the table is created
- Supports [`htmltools`](https://shiny.rstudio.com/articles/tag-glossary.html) and [`htmlwidgets`](https://www.htmlwidgets.org/)

</td><td>

- Harder to use but more dynamic
- Render on the fly, based on client-side state
- Can be more efficient in large tables

</td></tr>
</tbody></table>

Whichever one to use depends on the situation and personal preference.
You might prefer to use R render functions except in cases where you need more
dynamic behavior (e.g., render based on filtered state) or have a very large table.

#### Example: column total with filtering

::: {.callout}
This example requires reactable v0.3.0 or above.
:::

For example, you can easily add a column total using an R render function:

```{r}
data <- MASS::Cars93[20:24, c("Manufacturer", "Model", "Type", "Price")]

reactable(
  data,
  searchable = TRUE,
  columns = list(
    Price = colDef(footer = function(values) {
      htmltools::tags$b(sprintf("$%.2f", sum(values)))
    }),
    Manufacturer = colDef(footer = htmltools::tags$b("Total"))
  )
)
```

However, the column total doesn't update with filtering. For that, you need a
JavaScript render function with access to the client-side filtered state:

```{r}
data <- MASS::Cars93[20:24, c("Manufacturer", "Model", "Type", "Price")]

reactable(
  data,
  searchable = TRUE,
  columns = list(
    Price = colDef(
      html = TRUE,
      footer = JS("function(column, state) {
        let total = 0
        state.sortedData.forEach(function(row) {
          total += row[column.id]
        })
        return '<b>$' + total.toFixed(2) + '</b>'
      }")
    ),
    Manufacturer = colDef(html = TRUE, footer = "<b>Total</b>")
  )
)
```


## Cells

### R render function {#cells-r-render-function}

To customize cell rendering, provide an R function with up to 3 optional arguments:

```{r, eval=FALSE}
colDef(
  cell = function(value, index, name) {
    # input:
    #   - value, the cell value
    #   - index, the row index (optional)
    #   - name, the column name (optional)
    #
    # output:
    #   - content to render (e.g. an HTML tag or widget)
    htmltools::div(style = "color: red", toupper(value))
  }
)
```

### JavaScript render function {#cells-js-render-function}

Or a JavaScript function, wrapped in `JS()`, with up to 2 optional arguments:

```{r, eval=FALSE}
colDef(
  cell = JS("
    function(cellInfo, state) {
      // input:
      //  - cellInfo, an object containing cell info
      //  - state, an object containing the table state (optional)
      //
      // output:
      //  - content to render (e.g. an HTML string)
      return `<div>${cellInfo.value}</div>`
    }
  "),
  html = TRUE  # to render as HTML
)
```

With JavaScript functions, you can also customize rendering of grouped cells and
aggregated cells:

```{r, eval=FALSE}
colDef(
  grouped = JS("function(cellInfo, state) {
    return cellInfo.value
  }"),
  aggregated = JS("function(cellInfo, state) {
    return cellInfo.value
  }")
)
```

#### `cellInfo` properties

```{r, echo=FALSE, asis=TRUE}
cellInfoProps <- dplyr::tribble(
  ~Property, ~Example, ~Description,
  "value", '"setosa"', "cell value",
  "row", '{ Petal.Length: 1.7, Species: "setosa" }', "row data",
  "column", '{ id: "Petal.Length" }', "column info object",
  "index", "20", "row index (zero-based)",
  "viewIndex", "0", "row index within the page (zero-based)",
  "aggregated", "true", "whether the row is aggregated",
  "expanded", "true", "whether the row is expanded",
  "filterValue", '"petal"', "column filter value",
  "subRows", '[{ Petal.Length: 1.7, Species: "setosa" }, ...]', "sub rows data (aggregated cells only)",
  "level", "0", "row nesting depth (zero-based)",
  "selected", "true", "whether the row is selected"
)

propsTable(cellInfoProps)
```

#### `state` properties

```{r, echo=FALSE, asis=TRUE}
stateProps <- dplyr::tribble(
  ~Property, ~Example, ~Description,
  "sorted", '[{ id: "Petal.Length", desc: true }, ...]', "columns being sorted in the table",
  "page", "2", "page index (zero-based)",
  "pageSize", "10", "page size",
  "pages", "5", "number of pages",
  "filters", '[{ id: "Species", value: "petal" }]', "column filter values",
  "searchValue", '"petal"', "table search value",
  "selected", '[0, 1, 4]', "selected row indices (zero-based)",
  "pageRows", '[{ Petal.Length: 1.7, Species: "setosa" }, ...]', "current row data on the page",
  "sortedData", '[{ Petal.Length: 1.7, Species: "setosa" }, ...]', "current row data in the table (after sorting, filtering, grouping)",
  "data", '[{ Petal.Length: 1.7, Species: "setosa" }, ...]', "original row data in the table",
  "meta", '{ custom: 123 }', tagList("custom table metadata from", tags$code("reactable()"), "(new in v0.4.0)"),
  "hiddenColumns", '["Petal.Length"]', "columns being hidden in the table"
)

propsTable(stateProps)
```

## Headers

### R render function {#headers-r-render-function}

To customize header rendering, provide an R function with up to 2 optional arguments:

```{r, eval=FALSE}
colDef(
  header = function(value, name) {
    # input:
    #   - value, the header value
    #   - name, the column name (optional)
    #
    # output:
    #   - content to render (e.g. an HTML tag or widget)
    htmltools::div(value)
  }
)
```

### JavaScript render function {#headers-js-render-function}

Or a JavaScript function with up to 2 optional arguments:

```{r, eval=FALSE}
colDef(
  header = JS("
    function(column, state) {
      // input:
      //  - column, an object containing column properties
      //  - state, an object containing the table state (optional)
      //
      // output:
      //  - content to render (e.g. an HTML string)
      return `<div>${column.name}</div>`
    }
  "),
  html = TRUE  # to render as HTML
)
```

#### `column` properties

```{r, echo=FALSE, asis=TRUE}
columnProps <- dplyr::tribble(
  ~Property, ~Example, ~Description,
  "id", '"Petal.Length"', "column ID",
  "name", '"Petal Length"', "column display name",
  "filterValue", '"petal"', "column filter value",
  "setFilter", 'function setFilter(value: any)', tagList("function to set the column filter value",  "(set to ", tags$code("undefined"), " to clear the filter)"),
  "column", '{ id: "Petal.Length", name: "Petal Length", filterValue: "petal" }', "column info object (deprecated in v0.3.0)",
  "data", '[{ Petal.Length: 1.7, Petal.Width: 0.2, _subRows: [] }, ...]', "current row data in the table (deprecated in v0.3.0)"
)

propsTable(columnProps)
```

#### `state` properties

```{r, echo=FALSE, asis=TRUE}
propsTable(stateProps)
```

## Footers

### R render function {#footers-r-render-function}

To add footer content, provide an R function with up to 2 optional arguments:

```{r, eval=FALSE}
colDef(
  footer = function(values, name) {
    # input:
    #   - values, the column values
    #   - name, the column name (optional)
    #
    # output:
    #   - content to render (e.g. an HTML tag or widget)
    htmltools::div(paste("Total:", sum(values)))
  }
)
```

### JavaScript render function {#footers-js-render-function}

Or a JavaScript function with up to 2 optional arguments:

```{r, eval=FALSE}
colDef(
  footer = JS("
    function(column, state) {
      // input:
      //  - column, an object containing column properties
      //  - state, an object containing the table state (optional)
      //
      // output:
      //  - content to render (e.g. an HTML string)
      return `<div>Rows: ${state.sortedData.length}</div>`
    }
  "),
  html = TRUE  # to render as HTML
)
```

#### `column` properties

```{r, echo=FALSE, asis=TRUE}
propsTable(columnProps)
```

#### `state` properties

```{r, echo=FALSE, asis=TRUE}
propsTable(stateProps)
```

## Expandable Row Details

### R render function {#details-r-render-function}

To add expandable row details, provide an R function with up to 2 optional arguments:

```{r, eval=FALSE}
reactable(
  details = function(index, name) {
    # input:
    #   - index, the row index
    #   - name, the column name (optional)
    #
    # output:
    #   - content to render (e.g., an HTML tag or nested table), or NULL to hide details
    htmltools::div(
      paste("Details for row:", index),
      reactable(data[index, ])
    )
  }
)
```

### JavaScript render function {#details-js-render-function}

Or a JavaScript function with up to 2 optional arguments:

```{r, eval=FALSE}
reactable(
  details = JS("
    function(rowInfo, state) {
      // input:
      //  - rowInfo, an object containing row info
      //  - state, an object containing the table state (optional)
      //
      // output:
      //  - content to render (e.g. an HTML string)
      return `<div>Details for row: ${rowInfo.index}</div>`
    }
  ")
)
```

#### `rowInfo` properties

```{r, echo=FALSE, asis=TRUE}
rowInfoProps <- dplyr::tribble(
  ~Property, ~Example, ~Description,
  "values", '{ Petal.Length: 1.7, Species: "setosa" }', "row data values",
  "row", '{ Petal.Length: 1.7, Species: "setosa" }', tagList("same as ", tags$code("values"), " (deprecated in v0.3.0)"),
  "index", "20", "row index (zero-based)",
  "viewIndex", "0", "row index within the page (zero-based)",
  "expanded", "true", "whether the row is expanded",
  "level", "0", "row nesting depth (zero-based)",
  "selected", "true", "whether the row is selected"
)

propsTable(rowInfoProps)
```

#### `state` properties

```{r, echo=FALSE, asis=TRUE}
propsTable(stateProps)
```


```{css echo=FALSE}
/* rmarkdown html documents */
.main-container {
  font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Helvetica, Arial, sans-serif;
}

.main-container blockquote {
  font-size: inherit;
}
```

---
title: "Custom Sort Icons"
output:
  html_document:
    toc: true
    toc_float:
      smooth_scroll: false
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)

library(reactable)
```

### Custom sort icons using CSS pseudo-elements

```{r custom_sort, eval=FALSE}
library(reactable)
library(htmltools)

reactable(
  MASS::Cars93[, 1:5],
  defaultSorted = "Min.Price",
  defaultColDef = colDef(
    header = function(value) {
      # Add custom sort icons to the right of each column header.
      # They don't have any content by themselves, but we'll insert
      # text content (arrow symbols) in them using CSS.
      # Also, ensure these are hidden from screen readers using aria-hidden="true".
      sort_icon <- span(class = "my-sort-icon", "aria-hidden" = TRUE)
      tagList(value, sort_icon)
    }
  ),
  # Hide the built-in sort icon
  showSortIcon = FALSE
)
```

Sortable column headers will have an [`aria-sort`](https://www.digitala11y.com/aria-sort-properties/)
attribute set to either `none` (unsorted), `ascending` (ascending sort),
or `descending` (descending sort).

You can use the `aria-sort` state, along with CSS
[pseudo-elements](https://developer.mozilla.org/en-US/docs/Web/CSS/::after),
to insert custom arrow arrow symbols (or any other symbol, emoji, text)
for each sorting state.

```{css}
/* Styling for all sort icons */
[aria-sort] .my-sort-icon {
  margin-left: 5px;
  color: #3f51b5;
}

/* Styling for ascending sort icon */
[aria-sort='ascending'] .my-sort-icon::after {
  /* Up arrow (Unicode) */
  content: '\2bc5';
}

/* Styling for descending sort icon */
[aria-sort='descending'] .my-sort-icon::after {
  /* Down arrow (Unicode) */
  content: '\2bc6';
}

/* Styling for unsorted sort icon */
[aria-sort='none'] .my-sort-icon::after {
  /* Up-and-down arrow (Unicode) */
  content: '\2195';
  opacity: 0.8;
}
```

To find other Unicode arrows and symbols, see https://en.wikipedia.org/wiki/Arrow_(symbol)#Arrows_in_Unicode or
https://www.fileformat.info/info/unicode/block/geometric_shapes/list.htm.

```{r ref.label="custom_sort", echo=FALSE}
```


```{css echo=FALSE}
/* rmarkdown html documents */
.main-container {
  font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Helvetica, Arial, sans-serif;
}

.main-container blockquote {
  font-size: inherit;
}
```

---
title: "Twitter Followers"
output: html_document
resource_files:
  - '.'
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)

options(reactable.static = TRUE)
```

```{css ref.label="styles", echo=FALSE}
```

```{r ref.label="font_styles", echo=FALSE}
```

```{r table, echo=FALSE}
library(reactable)
library(htmltools)

data <- read.csv("twitter_followers.csv", stringsAsFactors = FALSE)

tbl <- reactable(
  data,
  pagination = FALSE,
  defaultSorted = "exclusive_followers_pct",
  defaultColDef = colDef(headerClass = "header", align = "left"),
  columns = list(
    account = colDef(
      cell = function(value) {
        url <- paste0("https://twitter.com/", value)
        tags$a(href = url, target = "_blank", paste0("@", value))
      },
      width = 150
    ),
    followers = colDef(
      defaultSortOrder = "desc",
      cell = function(value) {
        width <- paste0(value * 100 / max(data$followers), "%")
        value <- format(value, big.mark = ",")
        value <- format(value, width = 9, justify = "right")
        bar <- div(
          class = "bar-chart",
          style = list(marginRight = "0.375rem"),
          div(class = "bar", style = list(width = width, backgroundColor = "#3fc1c9"))
        )
        div(class = "bar-cell", span(class = "number", value), bar)
      }
    ),
    exclusive_followers_pct = colDef(
      name = "Exclusive Followers",
      defaultSortOrder = "desc",
      cell = JS('function(cellInfo) {
        // Format as percentage
        const pct = (cellInfo.value * 100).toFixed(1) + "%"
        // Pad single-digit numbers
        let value = pct.padStart(5)
        // Show % on first row only
        if (cellInfo.viewIndex > 0) {
          value = value.replace("%", " ")
        }
        // Render bar chart
        return `
          <div class="bar-cell">
            <span class="number">${value}</span>
            <div class="bar-chart" style="background-color: #e1e1e1">
              <div class="bar" style="width: ${pct}; background-color: #fc5185"></div>
            </div>
          </div>
        `
      }'),
      html = TRUE
    )
  ),
  compact = TRUE,
  class = "followers-tbl"
)

div(class = "twitter-followers",
  div(class = "followers-header",
    h2(class = "followers-title", "Candidates whose followers are loyal only to them"),
    "Share of each 2020 candidate's followers who don't follow any other candidates"
  ),
  tbl
)
```

---

Source: [FiveThirtyEight](https://fivethirtyeight.com/features/which-2020-candidates-have-the-most-in-common-on-twitter/)

Raw data: [`twitter_followers.csv`](twitter_followers.csv)

How it was made: [Building the Twitter Followers Demo](../building-twitter-followers.html)

<h2>Source Code</h2>

```{r ref.label="table", eval=FALSE}
```

```{r font_styles, eval=FALSE}
htmltools::tags$link(href = "https://fonts.googleapis.com/css?family=Karla:400,700|Fira+Mono&display=fallback", rel = "stylesheet")
```

```{css styles, eval=FALSE}
.twitter-followers {
  margin: 0 auto;
  width: 575px;
  font-family: Karla, "Helvetica Neue", Helvetica, Arial, sans-serif;
}

.followers-header {
  margin: 1.125rem 0;
  font-size: 1rem;
}

.followers-title {
  font-size: 1.25rem;
  font-weight: 600;
}

.followers-tbl {
  font-size: 0.875rem;
  line-height: 1.125rem;
}

.followers-tbl a {
  color: inherit;
  text-decoration: none;
}

.followers-tbl a:hover,
.followers-tbl a:focus {
  text-decoration: underline;
  text-decoration-thickness: max(1px, 0.0625rem);
}

.header {
  border-bottom: 2px solid #555;
  font-size: 0.8125rem;
  font-weight: 400;
  text-transform: uppercase;
}

.header:hover {
  background-color: #eee;
}

.bar-cell {
  display: flex;
  align-items: center;
}

.number {
  font-family: "Fira Mono", Consolas, Monaco, monospace;
  font-size: 0.84375rem;
  white-space: pre;
}

.bar-chart {
  flex-grow: 1;
  margin-left: 0.375rem;
  height: 0.875rem;
}

.bar {
  height: 100%;
}
```

# Contributing

## Development

### Prerequisites

- [R](https://www.r-project.org/) >= 3.1
- [Node.js](https://nodejs.org) LTS

### Installation

Install dependencies for R:

```r
# install.packages("devtools")
devtools::install_dev_deps()
```

Install dependencies for JavaScript:

```bash
npm install
```

### Building

Build the JS/CSS bundle (outputs to `inst/htmlwidgets`):

```bash
npm run build
```

Then load (`devtools::load_all()`) or reinstall the package.

These generated files should be added in their own separate commit, preferably
only once per branch.

### Testing

Run R tests:

```r
devtools::test()

# With test coverage (requires DT)
# install.packages("DT")
covr::report()
```

Run JavaScript tests:

```bash
npm test

# With test coverage
npm run test:cover

# Update test snapshots
npm run test:update
```

### Linting and Formatting

Lint and format the JS/CSS:

```bash
npm run lint
npm run format
```

---
title: "Demo Cookbook"
output:
  html_document:
    toc: true
    toc_float:
      smooth_scroll: false
    self_contained: false
resource_files:
  - images
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)

library(reactable)

options(reactable.static = TRUE)
```

A collection of recipes used to create the reactable demos.

## Insert links

```{r}
data <- data.frame(
  Address = c("https://google.com", "https://yahoo.com", "https://duckduckgo.com"),
  Site = c("Google", "Yahoo", "DuckDuckGo")
)

reactable(
  data,
  columns = list(
    # Using htmltools to render a link
    Address = colDef(cell = function(value) {
      htmltools::tags$a(href = value, target = "_blank", value)
    }),
    # Or using raw HTML
    Site = colDef(html = TRUE, cell = function(value, index) {
      sprintf('<a href="%s" target="_blank">%s</a>', data$Address[index], value)
    })
  )
)
```

## Conditional formatting

### Color scales

To add color scales, you can use [R's built-in color utilities](https://bookdown.org/rdpeng/exdata/plotting-and-color-in-r.html#color-utilities-in-r)
(or other color manipulation package):

```{r}
data <- iris[10:29, ]
orange_pal <- function(x) rgb(colorRamp(c("#ffe4cc", "#ff9500"))(x), maxColorValue = 255)

reactable(
  data,
  columns = list(
    Petal.Length = colDef(style = function(value) {
      normalized <- (value - min(data$Petal.Length)) / (max(data$Petal.Length) - min(data$Petal.Length))
      color <- orange_pal(normalized)
      list(background = color)
    })
  )
)
```

```{r}
dimnames <- list(start(nottem)[1]:end(nottem)[1], month.abb)
temps <- matrix(nottem, ncol = 12, byrow = TRUE, dimnames = dimnames)

# ColorBrewer-inspired 3-color scale
BuYlRd <- function(x) rgb(colorRamp(c("#7fb7d7", "#ffffbf", "#fc8d59"))(x), maxColorValue = 255)

reactable(
  temps,
  defaultColDef = colDef(
    style = function(value) {
      if (!is.numeric(value)) return()
      normalized <- (value - min(nottem)) / (max(nottem) - min(nottem))
      color <- BuYlRd(normalized)
      list(background = color)
    },
    format = colFormat(digits = 1),
    minWidth = 50
  ),
  columns = list(
    .rownames = colDef(name = "Year", sortable = TRUE, align = "left")
  ),
  bordered = TRUE
)
```

### Formatting changes

```{r}
stocks <- data.frame(
  Symbol = c("GOOG", "FB", "AMZN", "NFLX", "TSLA"),
  Price = c(1265.13, 187.89, 1761.33, 276.82, 328.13),
  Change = c(4.14, 1.51, -19.45, 5.32, -12.45)
)

reactable(
  stocks,
  columns = list(
    Change = colDef(
      cell = function(value) {
        if (value >= 0) paste0("+", value) else value
      },
      style = function(value) {
        color <- if (value > 0) {
          "#008000"
        } else if (value < 0) {
          "#e00000"
        }
        list(fontWeight = 600, color = color)
      }
    )
  )
)
```

### Tags and badges

```{r, echo=FALSE}
set.seed(20)
```

```{r tags, eval=FALSE}
library(htmltools)

orders <- data.frame(
  Order = 2300:2304,
  Created = seq(as.Date("2019-04-01"), by = "day", length.out = 5),
  Customer = sample(rownames(MASS::painters), 5),
  Status = sample(c("Pending", "Paid", "Canceled"), 5, replace = TRUE),
  stringsAsFactors = FALSE
)

reactable(
  orders,
  columns = list(
    Status = colDef(cell = function(value) {
      class <- paste0("tag status-", tolower(value))
      div(class = class, value)
    })
  )
)
```

```{css}
.tag {
  display: inline-block;
  padding: 0.125rem 0.75rem;
  border-radius: 15px;
  font-weight: 600;
  font-size: 0.75rem;
}

.status-paid {
  background: hsl(116, 60%, 90%);
  color: hsl(116, 30%, 25%);
}

.status-pending {
  background: hsl(230, 70%, 90%);
  color: hsl(230, 45%, 30%);
}

.status-canceled {
  background: hsl(350, 70%, 90%);
  color: hsl(350, 45%, 30%);
}
```

```{r ref.label="tags", echo=FALSE}
```

```{r}
library(htmltools)

status_badge <- function(color = "#aaa", width = "0.55rem", height = width) {
  span(style = list(
    display = "inline-block",
    marginRight = "0.5rem",
    width = width,
    height = height,
    backgroundColor = color,
    borderRadius = "50%"
  ))
}

reactable(
  orders,
  columns = list(
    Status = colDef(cell = function(value) {
      color <- switch(
        value,
        Paid = "hsl(214, 45%, 50%)",
        Pending = "hsl(30, 97%, 70%)",
        Canceled = "hsl(3, 69%, 50%)"
      )
      badge <- status_badge(color = color)
      tagList(badge, value)
    })
  )
)
```

## Bar charts

There are many ways to create bar charts using HTML and CSS, but here's one way
inspired by [Making Charts with CSS](https://css-tricks.com/making-charts-with-css/).

```{r}
library(htmltools)

# Render a bar chart with a label on the left
bar_chart <- function(label, width = "100%", height = "1rem", fill = "#00bfc4", background = NULL) {
  bar <- div(style = list(background = fill, width = width, height = height))
  chart <- div(style = list(flexGrow = 1, marginLeft = "0.5rem", background = background), bar)
  div(style = list(display = "flex", alignItems = "center"), label, chart)
}

data <- MASS::Cars93[20:49, c("Make", "MPG.city", "MPG.highway")]

reactable(
  data,
  columns = list(
    MPG.city = colDef(name = "MPG (city)", align = "left", cell = function(value) {
      width <- paste0(value / max(data$MPG.city) * 100, "%")
      bar_chart(value, width = width)
    }),
    MPG.highway = colDef(name = "MPG (highway)", align = "left", cell = function(value) {
      width <- paste0(value / max(data$MPG.highway) * 100, "%")
      bar_chart(value, width = width, fill = "#fc5185", background = "#e1e1e1")
    })
  )
)
```

### Positive and negative values

```{r}
library(htmltools)

# Render a bar chart with positive and negative values
bar_chart_pos_neg <- function(label, value, max_value = 1, height = "1rem",
                              pos_fill = "#005ab5", neg_fill = "#dc3220") {
  neg_chart <- div(style = list(flex = "1 1 0"))
  pos_chart <- div(style = list(flex = "1 1 0"))
  width <- paste0(abs(value / max_value) * 100, "%")

  if (value < 0) {
    bar <- div(style = list(marginLeft = "0.5rem", background = neg_fill, width = width, height = height))
    chart <- div(
      style = list(display = "flex", alignItems = "center", justifyContent = "flex-end"),
      label,
      bar
    )
    neg_chart <- tagAppendChild(neg_chart, chart)
  } else {
    bar <- div(style = list(marginRight = "0.5rem", background = pos_fill, width = width, height = height))
    chart <- div(style = list(display = "flex", alignItems = "center"), bar, label)
    pos_chart <- tagAppendChild(pos_chart, chart)
  }

  div(style = list(display = "flex"), neg_chart, pos_chart)
}

data <- data.frame(
  company = sprintf("Company%02d", 1:10),
  profit_chg = c(0.2, 0.685, 0.917, 0.284, 0.105, -0.701, -0.528, -0.808, -0.957, -0.11)
)

reactable(
  data,
  bordered = TRUE,
  columns = list(
    company = colDef(name = "Company", minWidth = 100),
    profit_chg = colDef(
      name = "Change in Profit",
      defaultSortOrder = "desc",
      cell = function(value) {
        label <- paste0(round(value * 100), "%")
        bar_chart_pos_neg(label, value)
      },
      align = "center",
      minWidth = 400
    )
  )
)
```

### Background bar charts

Another way to create bar charts is to render them as background images.
This example creates bar images using the
[`linear-gradient()` CSS function](https://developer.mozilla.org/en-US/docs/Web/CSS/linear-gradient()),
inspired by an [example from the DT package](https://rstudio.github.io/DT/010-style.html).

```{r}
# Render a bar chart in the background of the cell
bar_style <- function(width = 1, fill = "#e6e6e6", height = "75%",
                      align = c("left", "right"), color = NULL) {
  align <- match.arg(align)
  if (align == "left") {
    position <- paste0(width * 100, "%")
    image <- sprintf("linear-gradient(90deg, %1$s %2$s, transparent %2$s)", fill, position)
  } else {
    position <- paste0(100 - width * 100, "%")
    image <- sprintf("linear-gradient(90deg, transparent %1$s, %2$s %1$s)", position, fill)
  }
  list(
    backgroundImage = image,
    backgroundSize = paste("100%", height),
    backgroundRepeat = "no-repeat",
    backgroundPosition = "center",
    color = color
  )
}

data <- mtcars[, 1:4]

reactable(
  data,
  columns = list(
    mpg = colDef(
      style = function(value) {
        bar_style(width = value / max(data$mpg), fill = "#2c5e77", color = "#fff")
      },
      align = "left",
      format = colFormat(digits = 1)
    ),
    disp = colDef(
      style = function(value) {
        bar_style(width = value / max(data$disp), fill = "hsl(208, 70%, 90%)")
      }
    ),
    hp = colDef(
      style = function(value) {
        bar_style(width = value / max(data$hp), height = "90%", align = "right")
      }
    )
  ),
  bordered = TRUE
)
```

## Embed images

To embed an image, render an [`<img>` element](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/img)
into the table. Be sure to add [`alt` text](https://webaim.org/techniques/alttext/)
for accessibility, even if the image is purely decorative
(use a null `alt=""` attribute in this case).

### External image files

```{r}
library(htmltools)

data <- data.frame(
  Animal = c("beaver", "cow", "wolf", "goat"),
  Body = c(1.35, 465, 36.33, 27.66),
  Brain = c(8.1, 423, 119.5, 115)
)

reactable(
  data,
  columns = list(
    Animal = colDef(cell = function(value) {
      image <- img(src = sprintf("images/%s.png", value), style = "height: 24px;", alt = value)
      tagList(
        div(style = "display: inline-block; width: 45px;", image),
        value
      )
    }),
    Body = colDef(name = "Body (kg)"),
    Brain = colDef(name = "Brain (g)")
  )
)
```

If the image file is local, ensure the image can be found from the rendered document:

- In pkgdown, you can include local images in your document using the `resource_files`
YAML field. See [External files in pkgdown](https://pkgdown.r-lib.org/reference/build_articles.html#external-files)
for details.
- In Shiny, you can include local images in your app using either the
`www/` directory or `shiny::addResourcePath()` function.
See [Resource Publishing in Shiny](https://shiny.rstudio.com/reference/shiny/latest/resourcePaths.html)
for details.

### Inline embedded images

Images can also be embedded into documents as a
[base64-encoded data URL](https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/Data_URIs)
using `knitr::image_uri()`. This can be more portable, but is usually only
recommended for small image files.

```{r}
library(htmltools)

data <- data.frame(
  Animal = c("beaver", "cow", "wolf", "goat"),
  Body = c(1.35, 465, 36.33, 27.66),
  Brain = c(8.1, 423, 119.5, 115)
)

reactable(
  data,
  columns = list(
    Animal = colDef(cell = function(value) {
      img_src <- knitr::image_uri(sprintf("images/%s.png", value))
      image <- img(src = img_src, style = "height: 24px;", alt = value)
      tagList(
        div(style = "display: inline-block; width: 45px", image),
        value
      )
    })
  )
)
```

## Rating stars

This example uses [Font Awesome icons
(via Shiny)](https://shiny.rstudio.com/reference/shiny/latest/icon.html)
to render rating stars in a table.

To make the rating star icons accessible to users of assistive technology,
the icons are marked up as an image using the
[ARIA `img` role](https://developer.mozilla.org/en-US/docs/Web/Accessibility/ARIA/Roles/Role_Img),
and alternative text is added using an `aria-label` or `title` attribute.

```{r}
library(htmltools)

rating_stars <- function(rating, max_rating = 5) {
  star_icon <- function(empty = FALSE) {
    tagAppendAttributes(shiny::icon("star"),
      style = paste("color:", if (empty) "#edf0f2" else "orange"),
      "aria-hidden" = "true"
    )
  }
  rounded_rating <- floor(rating + 0.5)  # always round up
  stars <- lapply(seq_len(max_rating), function(i) {
    if (i <= rounded_rating) star_icon() else star_icon(empty = TRUE)
  })
  label <- sprintf("%s out of %s stars", rating, max_rating)
  div(title = label, role = "img", stars)
}

ratings <- data.frame(
  Movie = c("Silent Serpent", "Nowhere to Hyde", "The Ape-Man Goes to Mars", "A Menace in Venice"),
  Rating = c(3.65, 2.35, 4.5, 1.4),
  Votes = c(115, 37, 60, 99)
)

reactable(ratings, columns = list(
  Rating = colDef(cell = function(value) rating_stars(value))
))
```

## Show data from other columns

::: {.callout}
This example requires reactable v0.3.0 or above.
:::

To access data from another column, get the current row data using the
row index argument in an R render function, or `cellInfo.row` in a
JavaScript render function. This example shows both ways.

```{r, message=FALSE}
library(dplyr)
library(htmltools)

data <- starwars %>%
  select(character = name, height, mass, gender, homeworld, species)
```

### R render function {#show-data-from-other-columns-r}

```{r}
reactable(
  data,
  columns = list(
    character = colDef(
      # Show species under character names
      cell = function(value, index) {
        species <- data$species[index]
        species <- if (!is.na(species)) species else "Unknown"
        div(
          div(style = "font-weight: 600", value),
          div(style = "font-size: 0.75rem", species)
        )
      }
    ),
    species = colDef(show = FALSE)
  ),
  # Vertically center cells
  defaultColDef = colDef(vAlign = "center"),
  defaultPageSize = 6
)
```

### JavaScript render function {#show-data-from-other-columns-js}

```{r}
reactable(
  data,
  columns = list(
    character = colDef(
      # Show species under character names
      cell = JS('function(cellInfo) {
        const species = cellInfo.row["species"] || "Unknown"
        return `
          <div>
            <div style="font-weight: 600">${cellInfo.value}</div>
            <div style="font-size: 0.75rem">${species}</div>
          </div>
        `
      }'),
      html = TRUE
    ),
    species = colDef(show = FALSE)
  ),
  # Vertically center cells
  defaultColDef = colDef(vAlign = "center"),
  defaultPageSize = 6
)
```

## Total rows

```{r, message=FALSE}
library(dplyr)
library(htmltools)

data <- MASS::Cars93[18:47, ] %>%
  select(Manufacturer, Model, Type, Sales = Price)

reactable(
  data,
  defaultPageSize = 5,
  columns = list(
    Manufacturer = colDef(footer = "Total"),
    Sales = colDef(footer = sprintf("$%.2f", sum(data$Sales)))
  ),
  defaultColDef = colDef(footerStyle = list(fontWeight = "bold"))
)
```

### Dynamic totals

::: {.callout}
This example requires reactable v0.3.0 or above.
:::

To update the total when filtering the table, calculate the total in a
JavaScript render function:

```{r}
reactable(
  data,
  searchable = TRUE,
  defaultPageSize = 5,
  minRows = 5,
  columns = list(
    Manufacturer = colDef(footer = "Total"),
    Sales = colDef(
      footer = JS("function(column, state) {
        let total = 0
        state.sortedData.forEach(function(row) {
          total += row[column.id]
        })
        return '$' + total.toFixed(2)
      }")
    )
  ),
  defaultColDef = colDef(footerStyle = list(fontWeight = "bold"))
)
```

#### Totals with aggregated rows

```{r}
reactable(
  data,
  groupBy = "Manufacturer",
  searchable = TRUE,
  columns = list(
    Manufacturer = colDef(footer = "Total"),
    Sales = colDef(
      aggregate = "sum",
      format = colFormat(currency = "USD"),
      footer = JS("function(column, state) {
        let total = 0
        state.sortedData.forEach(function(row) {
          total += row[column.id]
        })
        return '$' + total.toFixed(2)
      }")
    )
  ),
  defaultColDef = colDef(footerStyle = list(fontWeight = "bold"))
)
```

## Nested tables

To create nested tables, use `reactable()` in a row details renderer:

```{r, message=FALSE}
library(dplyr)

data <- MASS::Cars93[18:47, ] %>%
  mutate(ID = as.character(18:47), Date = seq(as.Date("2019-01-01"), by = "day", length.out = 30)) %>%
  select(ID, Date, Manufacturer, Model, Type, Price)

sales_by_mfr <- group_by(data, Manufacturer) %>%
  summarize(Quantity = n(), Sales = sum(Price))

reactable(
  sales_by_mfr,
  details = function(index) {
    sales <- filter(data, Manufacturer == sales_by_mfr$Manufacturer[index]) %>% select(-Manufacturer)
    tbl <- reactable(sales, outlined = TRUE, highlight = TRUE, fullWidth = FALSE)
    htmltools::div(style = list(margin = "12px 45px"), tbl)
  },
  onClick = "expand",
  rowStyle = list(cursor = "pointer")
)
```

## Units on first row only

To display a label on the first row only (even when sorting), use a JavaScript render function
to add the label when the [cell's `viewIndex` property](../custom-rendering.html#cellinfo-properties)
is `0`.

If the label breaks the alignment of values in the column, realign the values
by adding white space to the cells without units. Two ways to do this are shown below.

```{r units_first_row, eval=FALSE}
data <- MASS::Cars93[40:44, c("Make", "Length", "Luggage.room")]

reactable(
  data,
  class = "car-specs",
  columns = list(
    # Align values using white space (and a monospaced font)
    Length = colDef(
      cell = JS("function(cellInfo) {
        const units = cellInfo.viewIndex === 0 ? '\u2033' : ' '
        return cellInfo.value + units
      }"),
      class = "number"
    ),
    # Align values using a fixed-width container for units
    Luggage.room = colDef(
      name = "Luggage Room",
      cell = JS('function(cellInfo) {
        const units = cellInfo.viewIndex === 0 ? " ft³" : ""
        return cellInfo.value + `<div class="units">${units}</div>`
      }'),
      html = TRUE
    )
  )
)
```

```{css}
.car-specs .number {
  font-family: "Courier New", Courier, monospace;
  white-space: pre;
}

.car-specs .units {
  display: inline-block;
  width: 1.125rem;
}
```

```{r ref.label="units_first_row", echo=FALSE}
```

## Tooltips

To add tooltips to a column header, you can render the header as an
[`<abbr>` element](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/abbr)
with a `title` attribute:

```{r}
library(htmltools)
library(dplyr)

data <- as_tibble(mtcars[1:6, ], rownames = "car") %>%
  select(car:hp)

with_tooltip <- function(value, tooltip) {
  tags$abbr(style = "text-decoration: underline; text-decoration-style: dotted; cursor: help",
            title = tooltip, value)
}

reactable(
  data,
  columns = list(
    mpg = colDef(header = with_tooltip("mpg", "Miles per US gallon")),
    cyl = colDef(header = with_tooltip("cyl", "Number of cylinders")),
    disp = colDef(header = with_tooltip("disp", "Displacement (cubic inches)")),
    hp = colDef(header = with_tooltip("hp", "Gross horsepower"))
  )
)
```

The [`title` attribute is inaccessible](https://developer.paciellogroup.com/blog/2013/01/using-the-html-title-attribute-updated/)
to most keyboard, mobile, and screen reader users, however, so creating tooltips
like this is generally discouraged.

An alternate method would be to use the [tippy package](https://tippy.john-coene.com/),
which provides a JavaScript-based tooltip that supports keyboard, touch, and screen
reader use.

```{r}
library(htmltools)
library(dplyr)
library(tippy)

data <- as_tibble(mtcars[1:6, ], rownames = "car") %>%
  select(car:hp)

# See the ?tippy documentation to learn how to customize tooltips
with_tooltip <- function(value, tooltip, ...) {
  div(style = "text-decoration: underline; text-decoration-style: dotted; cursor: help",
      tippy(value, tooltip, ...))
}

reactable(
  data,
  columns = list(
    mpg = colDef(header = with_tooltip("mpg", "Miles per US gallon")),
    cyl = colDef(header = with_tooltip("cyl", "Number of cylinders"))
  )
)
```

## Highlight cells

```{r}
data <- MASS::road[11:17, ]

reactable(
  data,
  defaultColDef = colDef(
    style = function(value, index, name) {
      if (is.numeric(value) && value == max(data[[name]])) {
        list(fontWeight = "bold")
      }
    }
  )
)
```

## Highlight columns

```{r}
reactable(
  iris[1:5, ],
  columns = list(
    Petal.Length = colDef(style = list(background = "rgba(0, 0, 0, 0.03)"))
  )
)
```

## Highlight rows

```{r}
reactable(
  iris[1:5, ],
  rowStyle = function(index) {
    if (index == 2) list(fontWeight = "bold")
    else if (iris[index, "Petal.Length"] >= 1.5) list(background = "rgba(0, 0, 0, 0.05)")
  }
)
```

## Highlight sorted headers

To style sortable headers on hover, select headers with an `aria-sort` attribute
and `:hover` pseudo-class in CSS:

```{r sort_highlight, eval=FALSE}
reactable(iris[1:5, ], defaultColDef = colDef(headerClass = "sort-header"))
```

```{css}
.sort-header[aria-sort]:hover {
  background: rgba(0, 0, 0, 0.03);
}
```

To style sorted headers, select headers with either an `aria-sort="ascending"`
or `aria-sort="descending"` attribute:

```{css}
.sort-header[aria-sort="ascending"],
.sort-header[aria-sort="descending"] {
  background: rgba(0, 0, 0, 0.03);
}
```

```{r ref.label="sort_highlight", echo=FALSE}
```

## Highlight sorted columns

To style sorted columns, use a JavaScript function to style columns based on the
table's sorted state:

```{r}
reactable(
  iris[1:5, ],
  defaultSorted = "Sepal.Width",
  defaultColDef = colDef(
    style = JS("function(rowInfo, column, state) {
      // Highlight sorted columns
      for (let i = 0; i < state.sorted.length; i++) {
        if (state.sorted[i].id === column.id) {
          return { background: 'rgba(0, 0, 0, 0.03)' }
        }
      }
    }")
  )
)
```

## Borders between groups of data

::: {.callout}
This example requires reactable v0.3.0 or above.
:::

To add borders between groups, use an R or JavaScript function to style rows
based on the previous or next row's data. If the table can be sorted, use a
JavaScript function to style rows only when the groups are sorted.

```{r, message=FALSE}
library(dplyr)

data <- as_tibble(MASS::painters, rownames = "Painter") %>%
  filter(School %in% c("A", "B", "C")) %>%
  mutate(School = recode(School, A = "Renaissance", B = "Mannerist", C = "Seicento")) %>%
  select(Painter, School, everything()) %>%
  group_by(School) %>%
  slice(1:3)

reactable(
  data,
  defaultSorted = list(School = "asc", Drawing = "desc"),
  borderless = TRUE,
  rowStyle = JS("
    function(rowInfo, state) {
      // Ignore padding rows
      if (!rowInfo) return

      // Add horizontal separators between groups when sorting by school
      const firstSorted = state.sorted[0]
      if (firstSorted && firstSorted.id === 'School') {
        const nextRow = state.pageRows[rowInfo.viewIndex + 1]
        if (nextRow && rowInfo.values['School'] !== nextRow['School']) {
          // Use box-shadow to add a 2px border without taking extra space
          return { boxShadow: 'inset 0 -2px 0 rgba(0, 0, 0, 0.1)' }
        }
      }
    }
  ")
)
```

## Merge cells

::: {.callout}
This example requires reactable v0.3.0 or above.
:::

You can give the appearance of merged cells by hiding cells based on the
previous row's data. Just like with the example above, you'll need a JavaScript
style function for grouping to work with sorting, filtering, and pagination.

```{r, message=FALSE}
library(dplyr)

data <- as_tibble(MASS::painters, rownames = "Painter") %>%
  filter(School %in% c("A", "B", "C")) %>%
  mutate(School = recode(School, A = "Renaissance", B = "Mannerist", C = "Seicento")) %>%
  select(School, Painter, everything()) %>%
  group_by(School) %>%
  slice(1:3)

reactable(
  data,
  columns = list(
    School = colDef(
      style = JS("function(rowInfo, column, state) {
        const firstSorted = state.sorted[0]
        // Merge cells if unsorted or sorting by school
        if (!firstSorted || firstSorted.id === 'School') {
          const prevRow = state.pageRows[rowInfo.viewIndex - 1]
          if (prevRow && rowInfo.values['School'] === prevRow['School']) {
            return { visibility: 'hidden' }
          }
        }
      }")
    )
  ),
  outlined = TRUE
)
```

## Borders between columns

```{r}
reactable(
  iris[1:5, ],
  columns = list(
    Sepal.Width = colDef(style = list(borderRight = "1px solid rgba(0, 0, 0, 0.1)")),
    Petal.Width = colDef(style = list(borderRight = "1px solid rgba(0, 0, 0, 0.1)"))
  ),
  borderless = TRUE
)
```

## Style nested rows

To style nested rows, use a JavaScript function to style rows based on their
[nesting `level` property](../conditional-styling.html#rowinfo-properties):

```{r}
data <- MASS::Cars93[4:8, c("Type", "Price", "MPG.city", "DriveTrain", "Man.trans.avail")]

reactable(
  data,
  groupBy = "Type",
  columns = list(
    Price = colDef(aggregate = "max"),
    MPG.city = colDef(aggregate = "mean", format = colFormat(digits = 1)),
    DriveTrain = colDef(aggregate = "unique"),
    Man.trans.avail = colDef(aggregate = "frequency")
  ),
  rowStyle = JS("function(rowInfo) {
    if (rowInfo.level > 0) {
      return { background: '#eee', borderLeft: '2px solid #ffa62d' }
    } else {
      return { borderLeft: '2px solid transparent' }
    }
  }"),
  defaultExpanded = TRUE
)
```

## Custom fonts

Tables don't have a default font, and just inherit the font properties
from their parent elements. (This may explain why tables look different in
R Markdown documents or Shiny apps vs. standalone pages).

To customize the table font, you can set a font on the page, or on the table itself:

```{r}
reactable(
  iris[1:5, ],
  style = list(fontFamily = "Work Sans, sans-serif", fontSize = "0.875rem"),
  defaultSorted = "Species"
)
```

To use a custom font that's not installed on your users' systems by default, use the [`@font-face`](https://developer.mozilla.org/en-US/docs/Web/CSS/@font-face)
CSS rule to add the font and specify where to download it from.

Online font services such as [Google Fonts](https://fonts.google.com/) can make this easier by hosting
custom fonts and providing 1-2 lines of HTML to copy into your document to use those fonts.

For example, to include a font from Google Fonts in an R Markdown document, add a `<link>` tag pointing to the
font stylesheet somewhere:

````{verbatim, lang = "markdown"}
```{r, echo=FALSE}
# Add a custom font from Google Fonts
htmltools::tags$link(href = "https://fonts.googleapis.com/css?family=Work+Sans:400,600,700&display=fallback",
                     rel = "stylesheet")
```
````

```{r, echo=FALSE}
# Add a custom font from Google Fonts
htmltools::tags$link(href = "https://fonts.googleapis.com/css?family=Work+Sans:400,600,700&display=fallback",
                     rel = "stylesheet")
```

Or in Shiny apps, the `<link>` tag can be included in the `<head>` of the page via `ui`:

```{r, eval=FALSE}
library(shiny)
library(reactable)

ui <- fluidPage(
  tags$head(
    tags$link(
      href = "https://fonts.googleapis.com/css?family=Work+Sans:400,600,700&display=fallback",
      rel = "stylesheet"
    ),
    tags$style("
      body {
        font-family: Work Sans, sans-serif;
      }
    ")
  ),
  reactable(
    MASS::Cars93[, 1:5],
    defaultSorted = "Price"
  )
)

server <- function(input, output) {}

shinyApp(ui, server)
```

For an example of using self-hosted custom fonts, see the [Popular Movies](../popular-movies/popular-movies.html) demo.

::: {.callout-tip}
**Tip:** The reactable package documentation uses the default system fonts
installed on your operating system (also known as a
[system font stack](https://css-tricks.com/snippets/css/system-font-stack)), which
load fast and look familiar:

```css
font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Helvetica, Arial, sans-serif;
```

[Bootstrap 5 also uses a system font stack](https://getbootstrap.com/docs/5.0/content/reboot/#native-font-stack)
by default.
:::

## Custom sort indicators

To use a custom sort indicator, you can hide the default sort icon using
`reactable(showSortIcon = FALSE)` and add your own sort indicator.

This also hides the sort icon when a header is focused, so be sure to add a
visual focus indicator to ensure your table is accessible to keyboard users
(to test this, click the first table header then press the Tab key to navigate
to other headers).

Here's an example that changes the sort indicator to a bar on the top or bottom
of the header (indicating an ascending or descending sort), and adds a light background
to headers when hovered or focused.

This example adds sort indicators using only CSS, and takes advantage of the
[`aria-sort`](https://developer.mozilla.org/en-US/docs/Web/Accessibility/ARIA/Attributes/aria-sort)
attribute on table headers to style based on whether the column is sorted in ascending
or descending order.

```{r bar_sort, eval=FALSE}
reactable(
  MASS::Cars93[1:5, c("Manufacturer", "Model", "Type", "Min.Price", "Price")],
  showSortIcon = FALSE,
  bordered = TRUE,
  defaultSorted = "Type",
  defaultColDef = colDef(headerClass = "bar-sort-header")
)
```

```{css}
.bar-sort-header:hover,
.bar-sort-header:focus {
  background: rgba(0, 0, 0, 0.03);
}

/* Add a top bar on ascending sort */
.bar-sort-header[aria-sort="ascending"] {
  box-shadow: inset 0 3px 0 0 rgba(0, 0, 0, 0.6);
}

/* Add a bottom bar on descending sort */
.bar-sort-header[aria-sort="descending"] {
  box-shadow: inset 0 -3px 0 0 rgba(0, 0, 0, 0.6);
}

/* Add an animation when toggling between ascending and descending sort */
.bar-sort-header {
  transition: box-shadow 0.3s cubic-bezier(0.175, 0.885, 0.32, 1.275);
}
```

```{r ref.label="bar_sort", echo=FALSE}
```

```{css echo=FALSE}
/* rmarkdown html documents */
.main-container {
  font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Helvetica, Arial, sans-serif;
}

.main-container blockquote {
  font-size: inherit;
}
```

---
title: "Custom Filtering"
output:
  html_document:
    toc: true
    toc_float:
      smooth_scroll: false
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)

library(reactable)
library(htmltools)

propsTable <- function(props) {
  tags$div(
    style = "overflow: auto;",
    tabindex = "0",
    tags$table(
      class = "props-tbl",
      tags$thead(
        tags$tr(
          lapply(colnames(props), function(name) tags$th(name))
        )
      ),
      tags$tbody(
        apply(props, 1, function(row) {
          tags$tr(
            tags$th(scope = "row", tags$code(row[["Property"]])),
            tags$td(tags$code(row[["Example"]])),
            tags$td(row[["Description"]])
          )
        })
      )
    )
  )
}
```

Custom filter methods and custom filter inputs let you change how filtering is done
in reactable. By default, all filter inputs are text inputs that filter data using
a case-insensitive text match, or for numeric columns, a prefix match.

Column filter methods, table search methods, and column filter inputs can all be
customized separately for flexibility. In some cases, you may want to keep the
default filter input but change the filter method, or in other cases, vice versa.

- [Column Filter Methods](#column-filter-methods)
- [Table Search Methods](#table-search-methods)
- [Column Filter Inputs](#column-filter-inputs)
- [Examples](#examples)

## Column Filter Methods

Column filter methods can be customized using the `filterMethod` argument in `colDef()`:

```{r, eval=FALSE}
colDef(
  filterMethod = JS("
    function(rows, columnId, filterValue) {
      /* ... */
      return filteredRows
    }
  ")
)
```

`filterMethod` should be a JavaScript function, wrapped in `JS()`, that takes
3 arguments — the rows, column ID, and filter value — and returns the filtered
array of rows.

```ts
filterMethod(
  rows: Array<Row>,
  columnId: string,
  filterValue: any
): Array<Row>
```

#### Arguments {#filter-method-js-arguments}

- **`rows`**, an array of row objects. `rows` consists of data rows only, and does
  not include aggregated rows when the table is grouped. Each `Row` has the following
  properties:

    ```{r, echo=FALSE, asis=TRUE}
    rowProps <- dplyr::tribble(
      ~Property, ~Example, ~Description,
      "values", '{ Petal.Length: 1.7, Species: "setosa" }', "row data values",
      "index", "20", "row index (zero-based)"
    )
    
    propsTable(rowProps)
    ```

- **`columnId`**, the column ID. Use this to access the cell value, like
  `row.values[columnId]`.
- **`filterValue`**, the column filter value. This will be a string when using the
  default filter inputs.

#### Details {#filter-method-js-details}

`filterMethod` is not called when the filter value is cleared, so you don't need to
specifically handle the case where `filterValue` is unset (`undefined`).

`filterMethod` only applies to individual column filters, and does not affect global
table searching.


## Table Search Methods

The global table search method can be customized using the `searchMethod` argument
in `reactable()`:

```{r, eval=FALSE}
reactable(
  searchMethod = JS("
    function(rows, columnIds, filterValue) {
      /* ... */
      return filteredRows
    }
  ")
)
```

`searchMethod` should be a JavaScript function, wrapped in `JS()`, that takes
3 arguments — the rows, column IDs, and search value — and returns the filtered
array of rows.

```ts
searchMethod(
  rows: Array<Row>,
  columnIds: Array<string>,
  searchValue: any
): Array<Row>
```

#### Arguments {#search-method-js-arguments}

- **`rows`**, an array of row objects. `rows` consists of data rows only, and does
  not include aggregated rows when the table is grouped. Each `Row` has the following
  properties:

    ```{r, echo=FALSE, asis=TRUE}
    rowProps <- dplyr::tribble(
      ~Property, ~Example, ~Description,
      "values", '{ Petal.Length: 1.7, Species: "setosa" }', "row data values",
      "index", "20", "row index (zero-based)"
    )
    
    propsTable(rowProps)
    ```

- **`columnIds`**, an array of column IDs. Use this to access the cell values
  for the columns being searched, like `row.values[columnId]`.
- **`searchValue`**, the search value. This will be a string when using the
  default search input.

#### Details {#search-method-js-details}

`searchMethod` is not called when the search value is cleared, so you don't need to
specifically handle the case where `searchValue` is unset (`undefined`).


## Column Filter Inputs

Column filter inputs are customized using the `filterInput` argument in `colDef()`.
`filterInput` can either be an element or a function that returns an element to be
rendered in place of the default column filter.

Unlike custom rendering of other table elements, custom filter inputs must communicate
back to the table on filtering changes, which can be tricky.

In many simpler cases, you can write your custom filter input in R and use
[`Reactable.setFilter()`](javascript-api.html#reactable-setfilter) from the
reactable JavaScript API to notify the table of filter changes. Note that the table
must have a unique `elementId` to use `Reactable.setFilter()` — see
[Using the JavaScript API](./javascript-api#using-the-javascript-api) for more details.

For more advanced customization, you can write your filter input entirely in
JavaScript, using the React JavaScript library to create an element that gets and sets
the filter value. In reactable, JavaScript render functions can return React elements,
and reactable comes with React as a dependency via the reactR package.
Use `reactR::html_dependency_react()` to explicitly include this dependency or find the
version of React in use.

For more information on React or using React from R, see the
[React documentation](https://reactjs.org/) and
[reactR package documentation](https://react-r.github.io/reactR/index.html).

::: {.callout-tip}
**Tip:** Many column filter inputs do not have a visible text label, including
the default text inputs. If your custom filter does not have a visible text label,
be sure to give it an accessible name using the
[`aria-label` attribute or similar technique](https://developer.mozilla.org/en-US/docs/Web/Accessibility/ARIA/Attributes/aria-label).
:::

### R render function {#filter-r-render-function}

R render functions take up to 2 optional arguments — the column values and
column name — and return the element to render.

```{r, eval=FALSE}
colDef(
  filterInput = function(values, name) {
    # input:
    #   - values, the column values (optional)
    #   - name, the column name  (optional)
    #
    # output:
    #   - element to render (e.g. an HTML tag or HTML string)
    htmltools::tags$input(
      type = "text",
      onchange = sprintf("Reactable.setFilter('tbl', '%s', this.value)", name),
      "aria-label" = sprintf("Filter %s", name),
      style = "width: 100%;"
    )
  }
)
```

### JavaScript render function {#filter-js-render-function}

JavaScript render functions, wrapped in `JS()`, take up to 2 optional arguments — a column object
and a table state object — and return the element to render.

```{r, eval=FALSE}
colDef(
  filterInput = JS("
    function(column, state) {
      // input:
      //  - column, an object containing column properties
      //  - state, an object containing the table state
      //
      // output:
      //  - element to render (e.g. an HTML string or React element)
      return React.createElement('input', {
        type: 'text',
        value: column.filterValue,
        onChange: function(e) {
          return column.setFilter(e.target.value || undefined)
        },
        'aria-label': 'Filter ' + column.name
      })
    }
  ")
)
```

#### `column` properties

```{r, echo=FALSE, asis=TRUE}
columnProps <- dplyr::tribble(
  ~Property, ~Example, ~Description,
  "id", '"Petal.Length"', "column ID",
  "name", '"Petal Length"', "column display name",
  "filterValue", '"petal"', "column filter value",
  "setFilter", 'function setFilter(value: any)', tagList("function to set the column filter value",  "(set to ", tags$code("undefined"), " to clear the filter)")
)

propsTable(columnProps)
```

#### `state` properties

```{r, echo=FALSE, asis=TRUE}
stateProps <- dplyr::tribble(
  ~Property, ~Example, ~Description,
  "sorted", '[{ id: "Petal.Length", desc: true }, ...]', "columns being sorted in the table",
  "page", "2", "page index (zero-based)",
  "pageSize", "10", "page size",
  "pages", "5", "number of pages",
  "filters", '[{ id: "Species", value: "petal" }]', "column filter values",
  "searchValue", '"petal"', "table search value",
  "selected", '[0, 1, 4]', "selected row indices (zero-based)",
  "pageRows", '[{ Petal.Length: 1.7, Species: "setosa" }, ...]', "current row data on the page",
  "sortedData", '[{ Petal.Length: 1.7, Species: "setosa" }, ...]', "current row data in the table (after sorting, filtering, grouping)",
  "data", '[{ Petal.Length: 1.7, Species: "setosa" }, ...]', "original row data in the table",
  "hiddenColumns", '["Petal.Length"]', "columns being hidden in the table"
)

propsTable(stateProps)
```

## Examples

### Basic custom filter method

This example shows basic usage of a custom filter method, changing filtering
on the `Manufacturer` column to be case-sensitive rather than case-insensitive.
(Try filtering for "bmw" and then "BMW").

```{r}
data <- MASS::Cars93[, c("Manufacturer", "Model", "Type", "Price")]

reactable(
  data,
  columns = list(
    Manufacturer = colDef(
      filterable = TRUE,
      # Filter by case-sensitive text match
      filterMethod = JS("function(rows, columnId, filterValue) {
        return rows.filter(function(row) {
          return row.values[columnId].indexOf(filterValue) !== -1
        })
      }")
    )
  ),
  defaultPageSize = 5
)
```

### Basic custom search method

This example shows basic usage of a custom search method, changing global searching
to be a case-sensitive text match on all columns. (Try searching for "bmw" and then "BMW").

Note that some columns may be numeric or another non-string type, so you can use
[`String()`](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String/String)
to convert values to strings before calling string methods like
[`indexOf()`](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/String/indexOf).

```{r}
data <- MASS::Cars93[, c("Manufacturer", "Model", "Type", "Price")]

reactable(
  data,
  searchable = TRUE,
  # Search by case-sensitive text match
  searchMethod = JS("function(rows, columnIds, searchValue) {
    return rows.filter(function(row) {
      return columnIds.some(function(columnId) {
        return String(row.values[columnId]).indexOf(searchValue) !== -1
      })
    })
  }"),
  defaultPageSize = 5
)
```

### Numeric value filtering

This example shows how you can filter a numeric column based on a minimum value.
(Try filtering the `Price` column for `30`).

```{r}
data <- MASS::Cars93[, c("Manufacturer", "Model", "Type", "Price")]

reactable(
  data,
  columns = list(
    Price = colDef(
      filterable = TRUE,
      # Filter by minimum price
      filterMethod = JS("function(rows, columnId, filterValue) {
        return rows.filter(function(row) {
          return row.values[columnId] >= filterValue
        })
      }")
    )
  ),
  defaultPageSize = 5
)
```

### Regular expression pattern filtering

This example shows how you can filter a column based on a regular expression
pattern. Note that the regular expression is not escaped here — see
[regular expression escaping](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Guide/Regular_Expressions#escaping)
for an example of how to escape special characters in regular expressions.

```{r}
data <- MASS::Cars93[, c("Manufacturer", "Model", "Type", "Price")]

reactable(
  data,
  columns = list(
    Manufacturer = colDef(
      filterable = TRUE,
      # Filter by case-insensitive text match
      filterMethod = JS("function(rows, columnId, filterValue) {
        const pattern = new RegExp(filterValue, 'i')

        return rows.filter(function(row) {
          return pattern.test(row.values[columnId])
        })
      }")
    )
  ),
  defaultPageSize = 5
)
```

### Fuzzy text searching

This example uses the [match-sorter](https://github.com/kentcdodds/match-sorter)
JavaScript library to add fuzzy searching to a table. (Try searching "adi" to match
both "Cadillac" and "Audi").

This also shows how you can create reusable search or filter methods that can be shared
across multiple tables or columns.

```{r}
library(htmltools)

data <- MASS::Cars93[, c("Manufacturer", "Model", "Type", "Price")]

# match-sorter library dependency. Include this anywhere in your document or app.
matchSorterDep <- htmlDependency(
  "match-sorter", 
  "1.8.0",
  c(href = "https://unpkg.com/match-sorter@1.8.0/dist/umd/"),
  script = "match-sorter.min.js"
)

# Fuzzy search method based on match-sorter
# See https://github.com/kentcdodds/match-sorter for advanced customization
matchSorterSearchMethod <-  JS("function(rows, columnIds, searchValue) {
  const keys = columnIds.map(function(id) {
    return function(row) {
      return row.values[id]
    }
  })
  return matchSorter(rows, searchValue, { keys: keys })
}")

browsable(tagList(
  matchSorterDep,

  reactable(
    data,
    searchable = TRUE,
    searchMethod = matchSorterSearchMethod,
    defaultPageSize = 5
  )
))
```

### Select input filter

This example shows how you can render a custom
[`<select>`](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/select)
input filter in R.

The `<select>` input filters the `Manufacturer` column from a set of unique values,
and includes an additional "All" option to clear the filter.

```{r}
library(htmltools)

data <- MASS::Cars93[, c("Manufacturer", "Model", "Type", "Price")]

reactable(
  data,
  filterable = TRUE,
  columns = list(
    Manufacturer = colDef(
      filterInput = function(values, name) {
        tags$select(
          # Set to undefined to clear the filter
          onchange = sprintf("Reactable.setFilter('cars-select', '%s', event.target.value || undefined)", name),
          # "All" has an empty value to clear the filter, and is the default option
          tags$option(value = "", "All"),
          lapply(unique(values), tags$option),
          "aria-label" = sprintf("Filter %s", name),
          style = "width: 100%; height: 28px;"
        )
      }
    )
  ),
  defaultPageSize = 5,
  elementId = "cars-select"
)
```

### Data list filter

The [`<datalist>` element](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/datalist)
is like a native text input, but with an autocomplete feature that lets you choose from a
set of unique options. (Try searching for "ac" in the `Manufacturer` column.)

This example also shows how you can create reusable filter inputs,
or set default custom filters based on column type.

```{r}
library(htmltools)

data <- MASS::Cars93[, c("Manufacturer", "Model", "Type", "Price")]

# Creates a data list column filter for a table with the given ID
dataListFilter <- function(tableId, style = "width: 100%; height: 28px;") {
  function(values, name) {
    dataListId <- sprintf("%s-%s-list", tableId, name)
    tagList(
      tags$input(
        type = "text",
        list = dataListId,
        oninput = sprintf("Reactable.setFilter('%s', '%s', event.target.value || undefined)", tableId, name),
        "aria-label" = sprintf("Filter %s", name),
        style = style
      ),
      tags$datalist(
        id = dataListId,
        lapply(unique(values), function(value) tags$option(value = value))
      )
    )
  }
}

reactable(
  data,
  filterable = TRUE,
  columns = list(
    # Use data list filter for a specific column
    Manufacturer = colDef(
      filterInput = dataListFilter("cars-list")
    )
  ),
  # Or use data list filter as the default for all factor columns
  defaultColDef = colDef(
    filterInput = function(values, name) {
      if (is.factor(values)) {
        dataListFilter("cars-list")(values, name)
      }
    }
  ),
  defaultPageSize = 5,
  elementId = "cars-list"
)
```

### Range filter

This is a basic example of a native
[`<input type="range">` element](https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input/range)
for numeric filtering. The `Price` column is filtered by minimum value.

```{r}
library(htmltools)

data <- MASS::Cars93[, c("Manufacturer", "Model", "Type", "Price")]

reactable(
  data,
  columns = list(
    Price = colDef(
      filterable = TRUE,
      filterMethod = JS("function(rows, columnId, filterValue) {
        return rows.filter(function(row) {
          return row.values[columnId] >= filterValue
        })
      }"),
      filterInput = function(values, name) {
        oninput <- sprintf("Reactable.setFilter('cars-range', '%s', this.value)", name)
        tags$input(
          type = "range",
          min = floor(min(values)),
          max = ceiling(max(values)),
          value = floor(min(values)),
          oninput = oninput,
          onchange = oninput, # For IE11 support
          "aria-label" = sprintf("Filter by minimum %s", name)
        )
      }
    )
  ),
  defaultPageSize = 5,
  elementId = "cars-range"
)
```

### External range filter

This example shows how you can create custom filter inputs outside the table,
using a more complex version of the range filter from above.

```{r}
library(htmltools)

# Custom range input filter with label and value
rangeFilter <- function(tableId, columnId, label, min, max, value = NULL, step = NULL, width = "200px") {
  value <- if (!is.null(value)) value else min
  inputId <- sprintf("filter_%s_%s", tableId, columnId)
  valueId <- sprintf("filter_%s_%s__value", tableId, columnId)
  oninput <- paste(
    sprintf("document.getElementById('%s').textContent = this.value;", valueId),
    sprintf("Reactable.setFilter('%s', '%s', this.value)", tableId, columnId)
  )
  
  div(
    tags$label(`for` = inputId, label),
    div(
      style = sprintf("display: flex; align-items: center; width: %s", validateCssUnit(width)),
      tags$input(
        id = inputId,
        type = "range",
        min = min,
        max = max,
        step = step,
        value = value,
        oninput = oninput,
        onchange = oninput, # For IE11 support
        style = "width: 100%;"
      ),
      span(id = valueId, style = "margin-left: 8px;", value)
    )
  )
}

# Filter method that filters numeric columns by minimum value
filterMinValue <- JS("function(rows, columnId, filterValue) {
  return rows.filter(function(row) {
    return row.values[columnId] >= filterValue
  })
}")
```

```{r}
data <- MASS::Cars93[, c("Manufacturer", "Model", "Type", "Price")]

browsable(tagList(
  rangeFilter(
    "cars-ext-range",
    "Price",
    "Filter by Minimum Price",
    floor(min(data$Price)),
    ceiling(max(data$Price))
  ),
  
  reactable(
    data,
    columns = list(
      Price = colDef(filterMethod = filterMinValue)
    ),
    defaultPageSize = 5,
    elementId = "cars-ext-range"
  )
))
```

### External checkbox filter

This example shows how you can filter a column with logical values using an
external checkbox input.

The data contains a lot of missing values, and we want a checkbox that allows you
to show just the rows with missing values. So we add a hidden column that indicates
with `TRUE` or `FALSE` whether any values in the row are missing, and filter for
`true` values when the checkbox is checked.

```{r}
library(htmltools)

data <- MASS::Cars93[, c("Manufacturer", "Model", "Type", "Price")]

# Add missing values to the data
set.seed(123)
data[] <- lapply(data, function(x) {
  x[sample(1:length(x), length(x) / 3)] <- NA
  x
})

# Indicates with TRUE or FALSE whether any values in the row are missing
data$has_missing <- !complete.cases(data)

browsable(tagList(
  tags$label(
    tags$input(
      type = "checkbox",
      onclick = "Reactable.setFilter('cars-missing', 'has_missing', event.target.checked)"
    ),
    "Show missing values only"
  ),
  
  reactable(
    data,
    columns = list(
      # Hidden column for filtering missing values
      has_missing = colDef(
        show = FALSE,
        filterMethod = JS("function(rows, columnId, filterValue) {
          if (filterValue === true) {
            return rows.filter(function(row) {
              const hasMissing = row.values[columnId]
              return hasMissing
            })
          }
          return rows
        }")
      )
    ),
    defaultColDef = colDef(na = "-"),
    defaultPageSize = 5,
    elementId = "cars-missing"
  )
))
```

### Basic custom filter input (React)

This example shows how you could write your custom filter input entirely in
JavaScript, using React. It renders a basic text input filter for the
`Manufacturer` column.

```{r}
data <- MASS::Cars93[, c("Manufacturer", "Model", "Type", "Price")]

reactable(
  data,
  columns = list(
    Manufacturer = colDef(
      filterable = TRUE,
      filterInput = JS("function(column) {
        return React.createElement('input', {
          type: 'text',
          value: column.filterValue,
          onChange: function(event) {
            // Set to undefined to clear the filter
            return column.setFilter(event.target.value || undefined)
          },
          'aria-label': 'Filter ' + column.name,
          style: { width: '100%' }
        })
      }")
    )
  ),
  defaultPageSize = 5
)
```

### Range filter (React)

Here's a more complex example of a React-based filter input, with filter values
that depend on the column data. The min and max values of the column are found
dynamically from `state.data`.

The JavaScript code is embedded as a separate
[`js` language chunk](https://bookdown.org/yihui/rmarkdown/language-engines.html#javascript-and-css)
to make it easier to work with. but it could also be included through an external
JavaScript file, or inlined in R using `htmltools::tags$script()`.

```{js}
// Custom range filter with value label
function rangeFilter(column, state) {
  // Get min and max values from raw table data
  let min = Infinity
  let max = 0
  state.data.forEach(function(row) {
    const value = row[column.id]
    if (value < min) {
      min = Math.floor(value)
    } else if (value > max) {
      max = Math.ceil(value)
    }
  })

  const filterValue = column.filterValue || min
  const input = React.createElement('input', {
    type: 'range',
    value: filterValue,
    min: min,
    max: max,
    onChange: function(event) {
      // Set to undefined to clear the filter
      column.setFilter(event.target.value || undefined)
    },
    style: { width: '100%', marginRight: '8px' },
    'aria-label': 'Filter ' + column.name
  })

  return React.createElement(
    'div',
    { style: { display: 'flex', alignItems: 'center', height: '100%' } },
    [input, filterValue]
  )
}

// Filter method that filters numeric columns by minimum value
function filterMinValue(rows, columnId, filterValue) {
  return rows.filter(function(row) {
    return row.values[columnId] >= filterValue
  })
}
```

```{r}
data <- MASS::Cars93[, c("Manufacturer", "Model", "Type", "Price")]

reactable(
  data,
  filterable = TRUE,
  columns = list(
    Price = colDef(
      filterMethod = JS("filterMinValue"),
      filterInput = JS("rangeFilter")
    )
  ),
  defaultPageSize = 5,
  minRows = 5
)
```


```{css echo=FALSE}
/* rmarkdown html documents */
.main-container {
  font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Helvetica, Arial, sans-serif;
}

.main-container blockquote {
  font-size: inherit;
}
```

---
title: "Using Observable with reactable in Quarto"
format:
  html:
    echo: false
    code-tools: true
---

Examples of using Observable with reactable in Quarto, based on the [Observable JS Penguins example](https://quarto.org/docs/interactive/ojs/examples/penguins.html)
from the Quarto documentation.

:::{.callout-note}
This is a work in progress, and requires new features in reactable v0.4.0.
:::

Source code: [observable-reactable.qmd](https://github.com/glin/reactable/blob/main/vignettes/quarto/observable-reactable.qmd)

## Using Observable Inputs to filter reactable

```{ojs}
//| panel: input
viewof billLengthMin = Inputs.range(
  [32, 50], 
  { value: 35, step: 1, label: "Bill length (min):" }
)

viewof islands = Inputs.checkbox(
  ["Torgersen", "Biscoe", "Dream"], 
  { value: ["Torgersen", "Biscoe"], label: "Islands:" }
)
```

```{ojs}
//| include: false
data = FileAttachment("palmer-penguins.csv").csv({ typed: true })

filtered = data.filter(penguin => {
  return penguin.bill_length > billLengthMin && islands.includes(penguin.island)
})

// Update table data when filtered data changes
Reactable.setData('tbl', filtered)
```

```{r}
library(reactable)

data <- read.csv("palmer-penguins.csv")

reactable(
  data,
  wrap = FALSE,
  resizable = TRUE,
  minRows = 10,
  elementId = "tbl"
)
```

## Using reactable to filter Observable charts

```{ojs}
// Create an Observable value that automatically tracks the table's filtered data
filteredData = Generators.observe(change => {
  return Reactable.onStateChange('tbl-input', state => {
    change(state.sortedData)
  })
})
```

```{r}
library(reactable)
library(htmltools)

data <- read.csv("palmer-penguins.csv")

# Select input filter with an "All" default option
selectFilter <- function(tableId, style = "width: 100%; height: 100%;") {
  function(values, name) {
    tags$select(
      # Set to undefined to clear the filter
      onchange = sprintf("
        const value = event.target.value
        Reactable.setFilter('%s', '%s', value === '__ALL__' ? undefined : value)
      ", tableId, name),
      # "All" has a special value to clear the filter, and is the default option
      tags$option(value = "__ALL__", "All"),
      lapply(unique(values), tags$option),
      "aria-label" = sprintf("Filter %s", name),
      style = style
    )
  }
}

# Min range filter input that handles NaNs
minRangeFilter <- function(tableId, style = "width: 100%;") {
  function(values, name) {
    values <- na.omit(values)
    oninput <- sprintf("Reactable.setFilter('%s', '%s', this.value)", tableId, name)
    tags$input(
      type = "range",
      min = floor(min(values)),
      max = ceiling(max(values)),
      value = floor(min(values)),
      oninput = oninput,
      style = style,
      "aria-label" = sprintf("Filter by minimum %s", name)
    )
  }
}

# Min value filter method that handles NaNs
filterMinValue <- JS("(rows, columnId, filterValue) => {
  return rows.filter(row => {
    const value = row.values[columnId]
    return !isNaN(value) && value >= filterValue
  })
}")

reactable(
  data,
  columns = list(
    species = colDef(
      filterInput = selectFilter("tbl-input")
    ),
    island = colDef(
      filterInput = selectFilter("tbl-input")
    ),
    bill_length = colDef(
      filterMethod = filterMinValue,
      filterInput = minRangeFilter("tbl-input")
    ),
    sex = colDef(
      filterInput = selectFilter("tbl-input"),
      # Exact match filter method
      filterMethod = JS("(rows, columnId, filterValue) => {
        return rows.filter(row => row.values[columnId] === filterValue)
      }")
    )
  ),
  filterable = TRUE,
  wrap = FALSE,
  resizable = TRUE,
  defaultPageSize = 5,
  minRows = 5,
  elementId = "tbl-input"
)
```

### Penguin body mass by sex and species

```{ojs}
Plot.rectY(filteredData, 
  Plot.binX(
    { y: "count" }, 
    { x: "body_mass", fill: "species", thresholds: 20 }
  ))
  .plot({
    facet: {
      data: filteredData,
      x: "sex",
      y: "species",
      marginRight: 80
    },
    marks: [
      Plot.frame(),
    ]
  }
)
```

---
title: "Spotify Charts"
output: html_document
resource_files:
  - '.'
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```


```{r table, echo=FALSE}
library(reactable)
library(htmltools)

top_tracks <- jsonlite::fromJSON("top_tracks.json")
top_tracks <- top_tracks[, c("position", "trend", "name", "explicit", "artists", "daily_plays", "url")]

# artists is a list-column of data frames containing artist metadata (one artist per row).
# Convert this to a character string of artist names so it can be searched,
# and attach the metadata as an attribute so we can insert links later.
top_tracks$artists <- lapply(top_tracks$artists, function(x) {
  structure(paste(x$name, collapse = ", "), metadata = x)
})

tracks_table <- function(data) {
  reactable(
    data,
    searchable = TRUE,
    highlight = TRUE,
    wrap = FALSE,
    paginationType = "simple",
    minRows = 10,
    columns = list(
      position = colDef(
        header = tagList(
          span("#", "aria-hidden" = "true", title = "Position"),
          # Column label for screen readers (sr-only comes from Bootstrap)
          span("Position", class = "sr-only")
        ),
        width = 40
      ),
      trend = colDef(
        header = span("Trend", class = "sr-only"),
        sortable = FALSE,
        align = "center",
        width = 40,
        cell = function(value) trend_indicator(value)
      ),
      name = colDef(
        name = "Title",
        resizable = TRUE,
        cell = function(value, index) {
          title <- tags$a(href = data[index, "url"], value)
          if (data[index, "explicit"]) {
            explicit_tag <- div(style = list(float = "right"), span(class = "tag", "Explicit"))
            title <- tagList(title, explicit_tag)
          }
          title
        },
        minWidth = 200
      ),
      artists = colDef(
        name = "Artist",
        resizable = TRUE,
        html = TRUE,
        cell = function(value) {
          # Create comma-separated list of artist links
          metadata <- attr(value, "metadata")
          links <- apply(metadata, 1, function(artist) {
            sprintf('<a href="%s">%s</a>', artist["external_urls.spotify"], artist["name"])
          })
          paste(links, collapse = ", ")
        },
        minWidth = 100
      ),
      daily_plays = colDef(
        name = "Daily Plays",
        defaultSortOrder = "desc",
        format = colFormat(separators = TRUE),
        width = 120,
        class = "number"
      ),
      explicit = colDef(show = FALSE),
      url = colDef(show = FALSE)
    ),
    language = reactableLang(
      searchPlaceholder = "Filter tracks",
      noData = "No tracks found",
      pageInfo = "{rowStart}\u2013{rowEnd} of {rows} tracks",
      pagePrevious = "\u276e",
      pageNext = "\u276f",
    ),
    theme = spotify_theme()
  )
}

# Icon to indicate trend: unchanged, up, down, or new
trend_indicator <- function(value = c("unchanged", "up", "down", "new")) {
  value <- match.arg(value)
  label <- switch(value,
                  unchanged = "Unchanged", up = "Trending up",
                  down = "Trending down", new = "New")

  # Add img role and tooltip/label for accessibility
  args <- list(role = "img", title = label)

  if (value == "unchanged") {
    args <- c(args, list("–", style = "color: #666; font-weight: 700"))
  } else if (value == "up") {
    args <- c(args, list(shiny::icon("caret-up"), style = "color: #1ed760"))
  } else if (value == "down") {
    args <- c(args, list(shiny::icon("caret-down"), style = "color: #cd1a2b"))
  } else {
    args <- c(args, list(shiny::icon("circle"), style = "color: #2e77d0; font-size: 0.6rem"))
  }
  do.call(span, args)
}

spotify_theme <- function() {
  search_icon <- function(fill = "none") {
    # Icon from https://boxicons.com
    svg <- sprintf('<svg xmlns="http://www.w3.org/2000/svg" width="24" height="24"><path fill="%s" d="M10 18c1.85 0 3.54-.64 4.9-1.69l4.4 4.4 1.4-1.42-4.39-4.4A8 8 0 102 10a8 8 0 008 8.01zm0-14a6 6 0 11-.01 12.01A6 6 0 0110 4z"/></svg>', fill)
    sprintf("url('data:image/svg+xml;charset=utf-8,%s')", URLencode(svg))
  }

  text_color <- "hsl(0, 0%, 95%)"
  text_color_light <- "hsl(0, 0%, 70%)"
  text_color_lighter <- "hsl(0, 0%, 55%)"
  bg_color <- "hsl(0, 0%, 10%)"

  reactableTheme(
    color = text_color,
    backgroundColor = bg_color,
    borderColor = "hsl(0, 0%, 16%)",
    borderWidth = "1px",
    highlightColor = "rgba(255, 255, 255, 0.1)",
    cellPadding = "10px 8px",
    style = list(
      fontFamily = "Work Sans, Helvetica Neue, Helvetica, Arial, sans-serif",
      fontSize = "0.875rem",
      "a" = list(
        color = text_color,
        textDecoration = "none",
        "&:hover, &:focus" = list(
          textDecoration = "underline",
          textDecorationThickness = "1px"
        )
      ),
      ".number" = list(
        color = text_color_light,
        fontFamily = "Source Code Pro, Consolas, Monaco, monospace"
      ),
      ".tag" = list(
        padding = "0.125rem 0.25rem",
        color = "hsl(0, 0%, 40%)",
        fontSize = "0.75rem",
        border = "1px solid hsl(0, 0%, 24%)",
        borderRadius = "2px",
        textTransform = "uppercase"
      )
    ),
    headerStyle = list(
      color = text_color_light,
      fontWeight = 400,
      fontSize = "0.75rem",
      letterSpacing = "1px",
      textTransform = "uppercase",
      "&:hover, &:focus" = list(color = text_color)
    ),
    rowHighlightStyle = list(
      ".tag" = list(color = text_color, borderColor = text_color_lighter)
    ),
    # Full-width search bar with search icon
    searchInputStyle = list(
      paddingLeft = "1.9rem",
      paddingTop = "0.5rem",
      paddingBottom = "0.5rem",
      width = "100%",
      border = "none",
      backgroundColor = bg_color,
      backgroundImage = search_icon(text_color_light),
      backgroundSize = "1rem",
      backgroundPosition = "left 0.5rem center",
      backgroundRepeat = "no-repeat",
      "&:focus" = list(backgroundColor = "rgba(255, 255, 255, 0.1)", border = "none"),
      "&:hover, &:focus" = list(backgroundImage = search_icon(text_color)),
      "::placeholder" = list(color = text_color_lighter),
      "&:hover::placeholder, &:focus::placeholder" = list(color = text_color)
    ),
    paginationStyle = list(color = text_color_light),
    pageButtonHoverStyle = list(backgroundColor = "hsl(0, 0%, 20%)"),
    pageButtonActiveStyle = list(backgroundColor = "hsl(0, 0%, 24%)")
  )
}


div(class = "spotify-charts",
  h2(class = "title", "Global Top 50"),
  tracks_table(top_tracks)
)
```

---

Source: [Spotify Charts](https://spotifycharts.com/regional/global/weekly/latest),
[Spotify](https://open.spotify.com/playlist/37i9dQZEVXbMDoHDwVN2tF) via [spotifyr](https://github.com/charlie86/spotifyr)

Raw data: [`top_tracks.json`](top_tracks.json), [`get-top-tracks.R`](get-top-tracks.R)

<h2>Source Code</h2>

```{r ref.label="table", eval=FALSE}
```

```{r}
tags$link(href = "https://fonts.googleapis.com/css?family=Work+Sans:400,600,700|Source+Code+Pro:400&display=fallback",
          rel = "stylesheet")
```

```{css}
.spotify-charts {
  padding: 0.75rem 0.75rem 0;
  font-family: "Work Sans", "Helvetica Neue", Helvetica, Arial, sans-serif;
  color: hsl(0, 0%, 95%);
  background-color: hsl(0, 0%, 10%);
}

.title {
  margin: 0.75rem 0.375rem 1.5rem;
  padding: 0;
  font-size: 1.5rem;
  font-weight: 600;
}
```

```{css, echo=FALSE}
/* Bootstrap's sr-only CSS. Add this if Bootstrap styles aren't included. */
.sr-only {
  position: absolute !important;
  width: 1px !important;
  height: 1px !important;
  padding: 0 !important;
  margin: -1px !important;
  overflow: hidden !important;
  clip: rect(0, 0, 0, 0) !important;
  white-space: nowrap !important;
  border: 0 !important;
}
```

```{css echo=FALSE}
/* rmarkdown html documents */
h1.title {
  display: none;
}

/* pkgdown articles */
.row > main {
  max-width: 960px;
}
```

---
title: "Sticky Columns (Old Method)"
output:
  html_document:
    toc: true
    toc_float:
      smooth_scroll: false
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)

library(reactable)
```

::: {.callout-note}
**Note:** Sticky columns are now supported using a `sticky` argument in `colDef()`
and `colGroup()`, so use that if possible.
See [Sticky Columns](./examples.html#sticky-columns) for examples.
:::

### Basic sticky columns

You can make a single left or right column sticky using the [`position: sticky`](https://developer.mozilla.org/en-US/docs/Web/CSS/position) CSS property.

Note that `position: sticky` is supported by Chrome, Firefox, Safari, and Edge, but **not** IE11.
IE11 does fail gracefully though -- the columns just don't stick.

`background: #fff` and `z-index: 1` are added to properly hide scrolling content.

```{r}
reactable(
  MASS::Cars93,
  columns = list(
    Manufacturer = colDef(
      style = list(position = "sticky", left = 0, background = "#fff", zIndex = 1),
      headerStyle = list(position = "sticky", left = 0, background = "#fff", zIndex = 1)
    ),
    Make = colDef(
      style = list(position = "sticky", right = 0, background = "#fff", zIndex = 1),
      headerStyle = list(position = "sticky", right = 0, background = "#fff", zIndex = 1)
    )
  ),
  defaultColDef = colDef(minWidth = 150)
)
```

Adding a border style and reducing some of the repetition:

```{r}
sticky_style <- function(left = TRUE) {
  style <- list(position = "sticky", background = "#fff", zIndex = 1)
  if (left) {
    style <- c(style, list(left = 0, borderRight = "1px solid #eee"))
  } else {
    style <- c(style, list(right = 0, borderLeft = "1px solid #eee"))
  }
  style
}

reactable(
  MASS::Cars93,
  columns = list(
    Manufacturer = colDef(
      style = sticky_style(),
      headerStyle = sticky_style()
    ),
    Make = colDef(
      style = sticky_style(left = FALSE),
      headerStyle = sticky_style(left = FALSE)
    )
  ),
  defaultColDef = colDef(minWidth = 150)
)
```

### Multiple sticky columns

To make multiple columns sticky, you have to set the [`left`](https://developer.mozilla.org/en-US/docs/Web/CSS/left)
or `right` positions of the columns to where they'll stick. In this case, you'll
need to know the column widths -- columns are `100px` wide by default, so
the 2nd column from left needs `left: 100px;`, the 3rd column from left needs
`left: 200px;`, etc.

```{r sticky_multi, eval=FALSE}
reactable(
  MASS::Cars93,
  columns = list(
    Manufacturer = colDef(
      class = "sticky left-col-1",
      headerClass = "sticky left-col-1"
    ),
    Model = colDef(
      class = "sticky left-col-2",
      headerClass = "sticky left-col-2"
    ),
    Type = colDef(
      class = "sticky left-col-3",
      headerClass = "sticky left-col-3"
    ),
    Make = colDef(
      class = "sticky right-col-1",
      headerClass = "sticky right-col-1"
    )
  ),
  wrap = FALSE
)
```

```{css}
.sticky {
  position: sticky !important;
  background: #fff;
  z-index: 1;
}

.left-col-1 {
  left: 0;
}

.left-col-2 {
  left: 100px;
}

.left-col-3 {
  left: 200px;
  border-right: 1px solid #eee !important;
}

.right-col-1 {
  right: 0;
  border-left: 1px solid #eee !important;
}
```

```{r ref.label="sticky_multi", echo=FALSE}
```

### Sticky headers and footers

Sticky columns should work fine with sticky headers and footers:

```{r}
reactable(
  MASS::Cars93[1:20, ],
  pagination = FALSE,
  height = 400,
  wrap = FALSE,
  columns = list(
    Manufacturer = colDef(
      class = "sticky left-col-1",
      headerClass = "sticky left-col-1",
      footerClass = "sticky left-col-1"
    ),
    Model = colDef(
      class = "sticky left-col-2",
      headerClass = "sticky left-col-2",
      footerClass = "sticky left-col-2"
    ),
    Type = colDef(
      class = "sticky left-col-3",
      headerClass = "sticky left-col-3",
      footerClass = "sticky left-col-3"
    ),
    Make = colDef(
      class = "sticky right-col-1",
      headerClass = "sticky right-col-1",
      footerClass = "sticky right-col-1"
    )
  ),
  defaultColDef = colDef(footer = "Footer")
)
```

```{r eval=FALSE, echo=FALSE}
# Proof-of-concept of using row highlighting with sticky columns.
# WARNING: This relies on undocumented CSS classes and is not guaranteed
# to work in a future release. Last tested with reactable 0.2.3.
sticky_style <- function(left = TRUE) {
  style <- list(position = "sticky", background = "#fff", zIndex = 1)
  if (left) {
    style <- c(style, list(left = 0, borderRight = "1px solid #eee"))
  } else {
    style <- c(style, list(right = 0, borderLeft = "1px solid #eee"))
  }
  style
}

sticky_theme <- reactableTheme(
  cellStyle = list(
    ".rt-tr-highlight:hover &" = list(backgroundColor = "hsl(0, 0%, 95%) !important")
  )
)

reactable(
  MASS::Cars93,
  columns = list(
    Manufacturer = colDef(
      style = sticky_style(),
      headerStyle = sticky_style()
    ),
    Make = colDef(
      style = sticky_style(left = FALSE),
      headerStyle = sticky_style(left = FALSE)
    )
  ),
  defaultColDef = colDef(minWidth = 150),
  highlight = TRUE,
  theme = sticky_theme
)
```

```{css echo=FALSE}
/* rmarkdown html documents */
.main-container {
  font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Helvetica, Arial, sans-serif;
}

.main-container blockquote {
  font-size: inherit;
}
```

---
title: "Conditional Styling"
output:
  html_document:
    toc: true
    toc_float:
      smooth_scroll: false
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)

library(reactable)
library(htmltools)

propsTable <- function(props) {
  tags$div(
    style = "overflow: auto;",
    tabindex = "0",
    tags$table(
      class = "props-tbl",
      tags$thead(
        tags$tr(
          lapply(colnames(props), function(name) tags$th(name))
        )
      ),
      tags$tbody(
        apply(props, 1, function(row) {
          tags$tr(
            tags$th(scope = "row", tags$code(row[["Property"]])),
            tags$td(tags$code(row[["Example"]])),
            tags$td(row[["Description"]])
          )
        })
      )
    )
  )
}
```


You can conditionally style a table using functions that return inline styles
or CSS classes. Just like with [custom rendering](custom-rendering.html),
style functions can either be in R or JavaScript:

<table style="table-layout: fixed;">
<thead><tr><th>R functions</th><th>JavaScript functions</th></tr></thead>
<tbody>
<tr><td>

```{r eval=FALSE}
reactable(
  iris,
  rowStyle = function(index) {
    if (iris[index, "Sepal.Width"] > 3.5) {
      list(fontWeight = "bold")
    }
  }
)
```

</td><td>

```{r eval=FALSE}
reactable(
  iris,
  rowStyle = JS("function(rowInfo) {
    if (rowInfo.values['Sepal.Width'] > 3.5) {
      return { fontWeight: 'bold' }
    }
  }")
)
```

</td></tr>
<tr><td>

- Easier to use but more static
- Style once, when the table is created

</td><td>

- Harder to use but more dynamic
- Style on the fly, based on client-side state

</td></tr>
</tbody></table>

Whichever one to use depends on the situation and personal preference.
You might prefer to use R functions except when you need more dynamic behavior
(e.g., style based on sorted state).

#### Example: color scales

We can use R's built-in
[color utilities](https://bookdown.org/rdpeng/exdata/plotting-and-color-in-r.html#color-utilities-in-r)
to apply a color scale to a column:
```{r}
data <- iris[1:5, ]
orange_pal <- function(x) rgb(colorRamp(c("#ffe4cc", "#ffb54d"))(x), maxColorValue = 255)

reactable(
  data,
  columns = list(
    Petal.Length = colDef(
      style = function(value) {
        normalized <- (value - min(data$Petal.Length)) / (max(data$Petal.Length) - min(data$Petal.Length))
        color <- orange_pal(normalized)
        list(background = color)
      }
    )
  )
)
```

#### Example: highlight sorted columns

To style sorted columns, we need to use a JavaScript function to determine
whether a column is currently being sorted:

```{r highlight_sorted, eval=FALSE}
reactable(
  iris[1:5, ],
  defaultSorted = "Petal.Length",
  defaultColDef = colDef(
    class = JS("function(rowInfo, column, state) {
      // Highlight sorted columns
      for (let i = 0; i < state.sorted.length; i++) {
        if (state.sorted[i].id === column.id) {
          return 'sorted'
        }
      }
    }")
  )
)
```

```{css}
.sorted {
  background: rgba(0, 0, 0, 0.03);
}
```

```{r ref.label="highlight_sorted", echo=FALSE}
```


## Cell Styling

### R functions {#cell-r-functions}

Both `style` and `class` take an R function with up to 3 optional arguments:

```{r, eval=FALSE}
colDef(
  style = function(value, index, name) {
    # input:
    #   - value, the cell value
    #   - index, the row index (optional)
    #   - name, the column name (optional)
    #
    # output:
    #   - a named list with camelCased property names
    list(color = "red", marginLeft = "30px")
    #   - or an inline style string
    "color: red; margin-left: 30px;"
  },
  class = function(value, index, name) {
    # input:
    #   - value, the cell value
    #   - index, the row index (optional)
    #   - name, the column name (optional)
    #
    # output:
    #   - CSS class names
    "class1 class2"
  }
)
```

::: {.callout-note}
**Note:** R functions cannot apply styles to aggregated cells.
:::

### JavaScript functions {#cell-js-functions}

Or a JavaScript function, wrapped in `JS()`, with up to 3 optional arguments:

```{r, eval=FALSE}
colDef(
  style = JS("
    function(rowInfo, column, state) {
      // input:
      //  - rowInfo, an object containing row info
      //  - column, an object containing column properties (optional)
      //  - state, an object containing the table state (optional)
      //
      // output:
      //  - a style object with camelCased property names
      return { backgroundColor: 'gray' }
    }
  "),
  class = JS("
    function(rowInfo, column, state) {
      // input:
      //  - rowInfo, an object containing row info
      //  - column, an object containing column properties (optional)
      //  - state, an object containing the table state (optional)
      //
      // output:
      //  - CSS class names
      return 'class1 class2'
    }
  ")
)
```

#### `rowInfo` properties

```{r, echo=FALSE, asis=TRUE}
rowInfoProps <- dplyr::tribble(
  ~Property, ~Example, ~Description,
  "values", '{ Petal.Length: 1.7, Species: "setosa" }', "row data values",
  "row", '{ Petal.Length: 1.7, Species: "setosa" }', tagList("same as ", tags$code("values"), " (deprecated in v0.3.0)"),
  "index", "20", "row index (zero-based)",
  "viewIndex", "0", "row index within the page (zero-based)",
  "aggregated", "true", "whether the row is aggregated",
  "expanded", "true", "whether the row is expanded",
  "subRows", '[{ Petal.Length: 1.7, Species: "setosa" }, ...]', "sub rows data (aggregated rows only)",
  "level", "0", "row nesting depth (zero-based)",
  "selected", "true", "whether the row is selected"
)

propsTable(rowInfoProps)
```

#### `column` properties

```{r, echo=FALSE, asis=TRUE}
columnProps <- dplyr::tribble(
  ~Property, ~Example, ~Description,
  "id", '"Petal.Length"', "column ID",
  "name", '"Petal Length"', "column display name",
  "filterValue", '"petal"', "column filter value ",
  "setFilter", 'function setFilter(value: any)', tagList("function to set the column filter value",  "(set to ", tags$code("undefined"), " to clear the filter)")
)

propsTable(columnProps)
```

#### `state` properties

```{r, echo=FALSE, asis=TRUE}
stateProps <- dplyr::tribble(
  ~Property, ~Example, ~Description,
  "sorted", '[{ id: "Petal.Length", desc: true }, ...]', "columns being sorted in the table",
  "page", "2", "page index (zero-based)",
  "pageSize", "10", "page size",
  "pages", "5", "number of pages",
  "filters", '[{ id: "Species", value: "petal" }]', "column filter values",
  "searchValue", '"petal"', "table search value",
  "selected", '[0, 1, 4]', "selected row indices (zero-based)",
  "pageRows", '[{ Petal.Length: 1.7, Species: "setosa" }, ...]', "current row data on the page",
  "sortedData", '[{ Petal.Length: 1.7, Species: "setosa" }, ...]', "current row data in the table (after sorting, filtering, grouping)",
  "data", '[{ Petal.Length: 1.7, Species: "setosa" }, ...]', "original row data in the table",
  "meta", '{ custom: 123 }', tagList("custom table metadata from", tags$code("reactable()"), "(new in v0.4.0)"),
  "hiddenColumns", '["Petal.Length"]', "columns being hidden in the table"
)

propsTable(stateProps)
```


## Row Styling

### R functions {#row-r-functions}

Both `rowStyle` and `rowClass` take an R function with a single argument:

```{r, eval=FALSE}
reactable(
  rowStyle = function(index) {
    # input:
    #   - index, the row index
    #
    # output:
    #   - a named list with camelCased property names
    list(color = "red", marginLeft = "30px")
    #   - or an inline style string
    "color: red; margin-left: 30px;"
  },
  rowClass = function(index) {
    # input:
    #   - index, the row index
    #
    # output:
    #   - CSS class names
    "class1 class2"
  }
)
```

::: {.callout-note}
**Note:** R functions cannot apply styles to aggregated rows.
:::

### JavaScript functions {#row-js-functions}

Or a JavaScript function with up to 2 optional arguments:

```{r, eval=FALSE}
reactable(
  rowStyle = JS("
    function(rowInfo, state) {
      // input:
      //  - rowInfo, an object containing row info
      //  - state, an object containing the table state (optional)
      //
      // output:
      //  - a style object with camelCased properties
      return { backgroundColor: 'gray' }
    }
  "),
  rowClass = JS("
    function(rowInfo, state) {
      // input:
      //  - rowInfo, an object containing row info
      //  - state, an object containing the table state (optional)
      //
      // output:
      //  - CSS class names
      return 'class1 class2'
    }
  ")
)
```

#### `rowInfo` properties

```{r, echo=FALSE, asis=TRUE}
propsTable(rowInfoProps)
```

#### `state` properties

```{r, echo=FALSE, asis=TRUE}
propsTable(stateProps)
```


```{css echo=FALSE}
/* rmarkdown html documents */
.main-container {
  font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Helvetica, Arial, sans-serif;
}

.main-container blockquote {
  font-size: inherit;
}
```

---
title: 'JavaScript API'
output:
  html_document:
    toc: true
    toc_float:
      smooth_scroll: false
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)

library(reactable)
library(htmltools)

propsTable <- function(props) {
  tags$div(
    style = "overflow: auto;",
    tabindex = "0",
    tags$table(
      class = "props-tbl",
      tags$thead(
        tags$tr(
          lapply(colnames(props), function(name) tags$th(name))
        )
      ),
      tags$tbody(
        apply(props, 1, function(row) {
          tags$tr(
            tags$th(scope = "row", tags$code(row[["Property"]])),
            tags$td(tags$code(row[["Example"]])),
            tags$td(row[["Description"]])
          )
        })
      )
    )
  )
}
```

## Introduction

The JavaScript API lets you manipulate and access tables from JavaScript.
You can use this to create custom interactive controls for your table without
the use of Shiny, or add cross-widget interactions beyond what Crosstalk provides.

Common use cases for the JavaScript API include:

- Export buttons to download table data to a CSV file
- Custom filter inputs located outside of the table
- Toggle buttons for row grouping or row expansion

#### Example: CSV download button

```{r}
library(htmltools)

htmltools::browsable(
  tagList(
    tags$button("Download as CSV", onclick = "Reactable.downloadDataCSV('cars-table')"),

    reactable(
      MASS::Cars93[, 1:5],
      defaultPageSize = 5,
      elementId = "cars-table"
    )
  )
)
```

## Using the JavaScript API

To use the JavaScript API, your table must first have a unique ID that distinguishes
it from other tables:

- For tables in static documents (e.g., R Markdown), add an ID through the
  `elementId` argument in `reactable()`:

  ```r
  reactable(MASS::Cars93, elementId = "cars-table")
  ```

- For tables in Shiny apps, the ID will be the Shiny output ID specified in
  `reactableOutput()`. For example, the table in this app uses `cars_table` as its ID:

  ```r
  library(shiny)

  ui <- fluidPage(
    reactableOutput("cars_table")
  )

  server <- function(input, output) {
    output$cars_table <- renderReactable({
      reactable(MASS::Cars93)
    })
  }

  shinyApp(ui, server)
  ```

Once your table has an ID, you can use any of the `Reactable` JavaScript functions
with that table ID. For example, to download data from the `cars-table` table to a
CSV file, the JavaScript code would look like this:

```js
// Download the "cars-table" data to a CSV file named 'cars.csv'
Reactable.downloadDataCSV('cars-table', 'cars.csv')
```

To try this out interactively, you can open your
[browser's developer tools](https://developer.mozilla.org/en-US/docs/Learn/Common_questions/What_are_browser_developer_tools)
and run this function in the
[JavaScript console](https://developer.mozilla.org/en-US/docs/Learn/Common_questions/What_are_browser_developer_tools#the_javascript_console).

### Creating custom interactive controls

Most users will likely want to use the JavaScript API through an interactive control,
such as a button, so they could decide when to download the table data. Using HTML,
you can create a `<button>` element with an `onclick` action that calls
the `Reactable` JavaScript function.

This example uses the `htmltools` package to render a CSV download button.
You can copy this code into an R console to view the output:

```{r, eval=FALSE}
library(htmltools)

htmltools::browsable(
  tagList(
    tags$button("Download as CSV", onclick = "Reactable.downloadDataCSV('cars-table', 'cars.csv')"),

    reactable(MASS::Cars93[, 1:5], elementId = "cars-table")
  )
)
```

::: {.callout-note}
**Note:** `htmltools::browsable()` is a convenient way to view the rendered HTML when
copying code into the console. It isn't required to render HTML in R Markdown documents
or Shiny apps.
:::

To reuse this button in other tables, you can also convert it into a function that
generates download buttons:

```{r, eval=FALSE}
library(htmltools)

csvDownloadButton <- function(tableId, label = "Download as CSV", filename = "data.csv") {
  htmltools::tags$button(
    label,
    onclick = sprintf("Reactable.downloadDataCSV('%s', '%s')", tableId, filename)
  )
}

htmltools::browsable(
  tagList(
    csvDownloadButton("cars-table", "Download as CSV", filename = "cars.csv"),

    reactable(MASS::Cars93[, 1:5], elementId = "cars-table")
  )
)
```

For more examples of custom controls that use the JavaScript API, check out the
[JavaScript API examples](examples.html#javascript-api).

## JavaScript API Reference

### `Reactable.downloadDataCSV()`

Downloads the table data to a CSV file. The downloaded file will be named `data.csv`
by default, but you can customize this using the optional `filename` argument.

The downloaded data will include any filters that have been applied, and
exclude any sorting or grouping. Hidden columns will also be included, but
this may be customizable in the future.

For further customization, you can use the `options` argument to only include specific columns,
exclude column headers, change the field separator, or change the decimal separator (new in v0.4.0).

```ts
Reactable.downloadDataCSV(
  tableId: string,
  filename = "data.csv",
  // New in v0.4.0
  options?: {
    columnIds: string[],
    headers: true,
    sep: ',',
    dec: '.'
  }
)
```

#### Examples

```js
// Download table data to a file named data.csv
Reactable.downloadDataCSV('cars-table')

// Download table data to a file named cars93.csv
Reactable.downloadDataCSV('cars-table', 'cars93.csv')

// Download table data to a tab-separated values file
Reactable.downloadDataCSV('cars-table', 'cars93.tsv', { sep: '\t' })

// Download table data with custom columns and headers excluded
Reactable.downloadDataCSV('cars-table', 'cars93.csv', {
  columnIds: ['Model', 'Type'],
  headers: false
})
```

### `Reactable.getDataCSV()`

::: {.callout}
New in v0.4.0
:::

Gets the table data as a CSV string. Same as [`Reactable.downloadDataCSV()`](#reactable-downloaddatacsv)
but does not download the data.

```ts
Reactable.getDataCSV(
  tableId: string,
  options?: {
    columnIds: string[],
    headers: true,
    sep: ',',
    dec: '.'
  }
)
```

#### Value

A string with the table data in CSV format.

#### Examples

```js
// Get table data as a CSV string
const csv = Reactable.getDataCSV('cars-table')
console.log(csv)

// Get table data as a tab-separted values string
const tsv = Reactable.getDataCSV('cars-table', { sep: '\t' })
console.log(tsv)
```

### `Reactable.setSearch()`

Sets the search value of a table. To clear the search, set the value to `undefined`.

```ts
Reactable.setSearch(
  tableId: string,
  value: any
)
```

#### Examples

```js
// Set the search value to "midsize"
Reactable.setSearch('cars-table', 'midsize')

// Clear the search value
Reactable.setSearch('cars-table', undefined)
```

### `Reactable.setFilter()`

Sets the filter value of a column. To clear the column filter, set the value
to `undefined`.

```ts
Reactable.setFilter(
  tableId: string,
  columnId: string,
  value: any
)
```

#### Examples

```js
// Set the filter value of the "Type" column to "midsize"
Reactable.setFilter('cars-table', 'Type', 'midsize')

// Clear the filter value of the "Type" column
Reactable.setFilter('cars-table', 'Type', undefined)
```

### `Reactable.setAllFilters()`

Sets all column filter values in the table. To clear the column filters,
set `filters` to an empty array, `[]`.

```ts
Reactable.setAllFilters(
  tableId: string,
  filters: Array<{ id: string, value: any }>
)
```

#### Examples

```js
// Set the column filters for the "Type" column
Reactable.setAllFilters('cars-table', [{ id: 'Type', value: 'midsize' }])

// Set the column filters for the "Type" and "Model" columns
Reactable.setAllFilters('cars-table', [
  { id: 'Type', value: 'midsize' },
  { id: 'Model', value: 'legend' }
])

// Clear all column filters
Reactable.setAllFilters([])
```

### `Reactable.toggleGroupBy()`

Toggles the `groupBy` state for a column between grouped and ungrouped.
To enable or disable grouping explicitly, set the optional `isGrouped` argument
to `true` or `false`.

```ts
Reactable.toggleGroupBy(
  tableId: string,
  columnId: string,
  isGrouped?: boolean
)
```

#### Examples

```js
// Toggle groupBy state for the "Type" column
Reactable.toggleGroupBy('cars-table', 'Type')

// Enable grouping for the "Type" column
Reactable.toggleGroupBy('cars-table', 'Type', true)

// Disable grouping for the "Type" column
Reactable.toggleGroupBy('cars-table', 'Type', false)
```

### `Reactable.setGroupBy()`

Sets the `groupBy` columns for the table. To clear the `groupBy` columns,
set `columnIds` to an empty array, `[]`.

```ts
Reactable.setGroupBy(
  tableId: string,
  columnIds: string[]
)
```

#### Examples

```js
// Set the groupBy columns to "Type" and "Manufacturer"
Reactable.setGroupBy('cars-table', ['Type', 'Manufacturer'])

// Clear the groupBy columns
Reactable.setGroupBy('cars-table', [])
```

### `Reactable.toggleAllRowsExpanded()`

Toggles the expanded state of all rows in the table between expanded and collapsed.
To expand or collapse rows explicitly, set the optional `isExpanded` argument to
`true` or `false`.

```ts
Reactable.toggleAllRowsExpanded(
  tableId: string,
  isExpanded?: boolean
)
```

#### Examples

```js
// Toggle expanded state for all rows
Reactable.toggleAllRowsExpanded('cars-table')

// Expand all rows
Reactable.toggleAllRowsExpanded('cars-table', true)

// Collapse all rows
Reactable.toggleAllRowsExpanded('cars-table', false)
```

### `Reactable.setMeta()`

::: {.callout}
New in v0.4.0
:::

Sets the custom metadata for the table. `meta` can either be an object with new
values, or a function that takes the previous metadata object and returns new values.

New values are merged into the current metadata, so only the values specified in
`meta` will be updated. To clear all metadata, set `meta` to `undefined`.

```ts
Reactable.setMeta(
  tableId: string,
  meta?: object | Function
)
```

#### Examples

```js
// Set metadata, updating the values for the `count`, `enabled`, and `formatter` properties
Reactable.setMeta('cars-table', { count: 123, enabled: true, formatter: value => '$' + value })

// Set metadata using a function that increments `count` and toggles an `enabled` boolean
Reactable.setMeta('cars-table', prevMeta => {
  return { count: prevMeta.count + 1, enabled: !prevMeta.enabled }
})

// Clear metadata
Reactable.setMeta('cars-table', undefined)
```

### `Reactable.toggleHideColumn()`

::: {.callout}
New in v0.4.0
:::

Toggles the hidden state for a column between hidden and shown.
To hide or show a column explicitly, set the optional `isHidden` argument to `true` or `false`.

```ts
Reactable.toggleHideColumn(
  tableId: string,
  columnId: string,
  isHidden?: boolean
)
```

#### Examples

```js
// Toggle hidden state for the "Type" column
Reactable.toggleHideColumn('cars-table', 'Type')

// Hide the "Type" column
Reactable.toggleHideColumn('cars-table', 'Type', true)

// Show the "Type" column
Reactable.toggleHideColumn('cars-table', 'Type', false)
```

### `Reactable.setHiddenColumns()`

::: {.callout}
New in v0.4.0
:::

Sets the hidden columns for the table. `columnIds` can either be an array of column IDs, or a
function that takes the previous hidden column IDs and returns new column IDs.

To clear the hidden columns, set `columnIds` to an empty array, `[]`.

```ts
Reactable.setHiddenColumns(
  tableId: string,
  columnIds: string[] | Function
)
```

#### Examples

```js
// Set the hidden columns to "Type" and "Manufacturer"
Reactable.setHiddenColumns('cars-table', ['Type', 'Manufacturer'])

// Set hidden columns using a function that adds "Type" to the existing hidden columns
Reactable.setMeta('cars-table', prevHiddenColumns => {
  return prevHiddenColumns.concat('Type')
})

// Clear the hidden columns
Reactable.setHiddenColumns('cars-table', [])
```

### `Reactable.setData()`

::: {.callout}
New in v0.4.0
:::

Updates the table data. `data` can either be in row or column format. In row format, `data` is an array
of row objects. In column format, `data` is an object containing arrays of column values. `data` should
have all the same columns as the original table data.

When updating data, the selected rows and current page will reset by default. All other state will persist,
including sorting, filtering, and grouping state.

```ts
Reactable.setData(
  tableId: string,
  data: Array<object> | object
)
```

#### Examples

```js
// Update data in row format
Reactable.setData('cars-table', [
  { Model: 'Legend', Type: 'Midsize' },
  { Model: 'Integra', Type: 'Small' }
])

// Update data in column format
Reactable.setData('cars-table', {
  Model: ['Legend', 'Integra'],
  Type: ['Midsize', 'Small']
})
```

### `Reactable.getState()`

Gets the current state of a table.

```ts
Reactable.getState(tableId: string)
```

#### Value {#state-properties}

An object with the following properties:

```{r, echo=FALSE, asis=TRUE}
stateProps <- dplyr::tribble(
  ~Property, ~Example, ~Description,
  "sorted", '[{ id: "Petal.Length", desc: true }, ...]', "columns being sorted in the table",
  "page", "2", "page index (zero-based)",
  "pageSize", "10", "page size",
  "pages", "5", "number of pages",
  "filters", '[{ id: "Species", value: "petal" }]', "column filter values",
  "searchValue", '"petal"', "table search value",
  "selected", '[0, 1, 4]', "selected row indices (zero-based)",
  "pageRows", '[{ Petal.Length: 1.7, Species: "setosa" }, ...]', "current row data on the page",
  "sortedData", '[{ Petal.Length: 1.7, Species: "setosa" }, ...]', "current row data in the table (after sorting, filtering, grouping)",
  "data", '[{ Petal.Length: 1.7, Species: "setosa" }, ...]', "original row data in the table",
  "meta", '{ custom: 123 }', tagList("custom table metadata from", tags$code("reactable()"), "(new in v0.4.0)"),
  "hiddenColumns", '["Petal.Length"]', "columns being hidden in the table"
)

propsTable(stateProps)
```

#### Examples

```js
Reactable.getState('cars-table')
// { page: 2, searchValue: 'petal', ... }
```

### `Reactable.onStateChange()`

::: {.callout}
New in v0.4.0
:::

Sets up a function that will be called whenever the table state changes.

`listenerFn` should be a function that takes the current table state, an object with
properties described at [state properties](#state-properties).

`Reactable.onStateChange()` can only be run when the table has finished rendering,
which isn't guaranteed to happen at page load time. To ensure that the table is ready
before running `Reactable.onStateChange()`, you can either use `htmlwidgets::onRender()`
for static widgets and Shiny outputs, or `htmlwidgets::onStaticRenderComplete()` for
static widgets only.

```ts
Reactable.onStateChange(
  tableId: string,
  listenerFn: Function
)
```

#### Value

`Reactable.onStateChange` returns a function that can be called to cancel the listener function.

#### Examples

```js
Reactable.onStateChange('cars-table', state => {
  console.log('selected rows:', state.selected)
})

// Cancel the listener function
const cancel = Reactable.onStateChange('cars-table', listenerFn)
cancel()
```

##### Static widgets

```{r, eval=FALSE}
library(htmltools)

data <- MASS::Cars93[, c("Manufacturer", "Model", "Type", "Price")]

browsable(
  tagList(
    "Table state:",
    tags$pre(id = "tbl-state", "{}"),
    reactable(
      data,
      searchable = TRUE,
      selection = "multiple",
      onClick = "select",
      elementId = "tbl"
    ),
    htmlwidgets::onStaticRenderComplete("
      Reactable.onStateChange('tbl', state => {
        const { selected, sorted, searchValue, pageIndex } = state
        document.getElementById('tbl-state').textContent = JSON.stringify({
          selected,
          sorted,
          searchValue,
          pageIndex
        })
      })
    ")
  )
)
```

##### Shiny outputs

```{r, eval=FALSE}
library(shiny)

data <- MASS::Cars93[, c("Manufacturer", "Model", "Type", "Price")]

ui <- fluidPage(
  reactableOutput("tbl"),
  verbatimTextOutput("tbl_state")
)

server <- function(input, output) {
  output$tbl <- renderReactable({
    tbl <- reactable(data, searchable = TRUE)

    htmlwidgets::onRender(tbl, "() => {
      Reactable.onStateChange('tbl', state => {
        const { sorted, searchValue } = state
        Shiny.setInputValue('tbl_state', { sorted, searchValue })
      })
    }")
  })

  output$tbl_state <- renderPrint({
    writeLines("Table state:\n")
    print(input$tbl_state)
  })
}

shinyApp(ui, server)
```

```{css echo=FALSE}
/* rmarkdown html documents */
.main-container {
  font-family: -apple-system, BlinkMacSystemFont, "Segoe UI", Helvetica, Arial, sans-serif;
}

.main-container blockquote {
  font-size: inherit;
}

h3 {
  font-weight: 600;
}
```

---
title: "Shiny Demo"
output: html_document
---

```{r echo=FALSE}
library(htmltools)

tags$div(class = "iframe-container", tags$iframe(src = "https://glin.shinyapps.io/reactable"))
```

```{css echo=FALSE}
.iframe-container {
  display: flex;
  flex-direction: column;
  margin-top: 1rem;
  height: 100%;
  overflow: hidden;
}

.iframe-container iframe {
  width: 100%;
  height: 100%;
  border: 0;
}

/* rmarkdown html documents */
.main-container {
  padding: 0;
  max-width: 100%;
  height: 100vh;
}

body {
  overflow-y: hidden;
}

#header {
  display: none;
}

/* pkgdown articles */
.template-article {
  width: 100%;
  height: 100%;
}

.row {
  height: 100%;
}

.row > main {
  max-width: 100%;
}

.page-header {
  display: none;
}
```

# static rendering

    Code
      cat(rendered$html)
    Output
      <div class="reactable html-widget html-fill-item-overflow-hidden html-fill-item" data-react-ssr id="stable-id-static-rendering" style="width:auto;height:auto;"><div class="Reactable ReactTable"><div class="rt-table" role="table"><div class="rt-thead" role="rowgroup" style="min-width:300px"><div class="rt-tr rt-tr-header" role="row" style="flex:1 0 auto;min-width:300px"><div class="rt-th rt-align-right" role="columnheader" style="position:relative;flex:100 0 auto;min-width:100px;width:100px" aria-label="Sort x" aria-sort="none" tabindex="0" data-sort-hint="ascending"><div class="rt-th-inner"><div class="rt-sort-header"><span class="rt-sort-left" aria-hidden="true"></span><div class="rt-text-content">x</div></div></div></div><div class="rt-th rt-align-left" role="columnheader" style="position:relative;flex:100 0 auto;min-width:100px;width:100px" aria-label="Sort y" aria-sort="none" tabindex="0" data-sort-hint="ascending"><div class="rt-th-inner"><div class="rt-sort-header"><div class="rt-text-content">y</div><span class="rt-sort-right" aria-hidden="true"></span></div></div></div><div class="rt-th rt-align-left" role="columnheader" style="position:relative;flex:100 0 auto;min-width:100px;width:100px" aria-label="Sort z" aria-sort="none" tabindex="0" data-sort-hint="ascending"><div class="rt-th-inner"><div class="rt-sort-header"><div class="rt-text-content">z</div><span class="rt-sort-right" aria-hidden="true"></span></div></div></div></div></div><div class="rt-tbody" role="rowgroup" style="min-width:300px"><div class="rt-tr-group"><div class="rt-tr" role="row" style="flex:1 0 auto;min-width:300px"><div class="rt-td rt-align-right" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">1</div></div><div class="rt-td rt-align-left" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">a</div></div><div class="rt-td rt-align-left" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">1,2,3</div></div></div></div><div class="rt-tr-group"><div class="rt-tr" role="row" style="flex:1 0 auto;min-width:300px"><div class="rt-td rt-align-right" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">2</div></div><div class="rt-td rt-align-left" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">column-y-cell</div></div><div class="rt-td rt-align-left" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">4,5,6</div></div></div></div><div class="rt-no-data" aria-live="assertive"></div></div></div></div></div>
      <script type="application/json" data-for="stable-id-static-rendering">{"x":{"tag":{"name":"Reactable","attribs":{"data":{"x":[1,2],"y":["a","column-y-cell"],"z":[[1,2,3],[4,5,6]]},"columns":[{"id":"x","name":"x","type":"numeric"},{"id":"y","name":"y","type":"character"},{"id":"z","name":"z","type":"AsIs"}],"elementId":"stable-id-static-rendering","dataKey":"f09fcb11dd011a5cefcb1725322e0cb4"},"children":[]},"class":"reactR_markup"},"evals":[],"jsHooks":[]}</script>

---

    Code
      cat(styleDep$head)
    Output
      <style data-emotion="reactable 1rdr7a4 1ub9dy9 d7nd0n">.reactable-1rdr7a4{color:blue;}.reactable-1ub9dy9{border-width:3px;}.reactable-d7nd0n{border-width:3px;}.rt-bordered .reactable-d7nd0n,.rt-outlined .reactable-d7nd0n{border-width:3px;}</style>

---

    Code
      cat(rendered$html)
    Output
      <div class="reactable html-widget html-fill-item-overflow-hidden html-fill-item" data-react-ssr id="stable-id-theme-critical-css" style="width:auto;height:auto;"><div class="Reactable ReactTable reactable-1rdr7a4"><div class="rt-table reactable-1ub9dy9" role="table"><div class="rt-thead" role="rowgroup" style="min-width:300px"><div class="rt-tr rt-tr-header" role="row" style="flex:1 0 auto;min-width:300px"><div class="rt-th rt-align-right reactable-d7nd0n" role="columnheader" style="position:relative;flex:100 0 auto;min-width:100px;width:100px" aria-label="Sort x" aria-sort="none" tabindex="0" data-sort-hint="ascending"><div class="rt-th-inner"><div class="rt-sort-header"><span class="rt-sort-left" aria-hidden="true"></span><div class="rt-text-content">x</div></div></div></div><div class="rt-th rt-align-left reactable-d7nd0n" role="columnheader" style="position:relative;flex:100 0 auto;min-width:100px;width:100px" aria-label="Sort y" aria-sort="none" tabindex="0" data-sort-hint="ascending"><div class="rt-th-inner"><div class="rt-sort-header"><div class="rt-text-content">y</div><span class="rt-sort-right" aria-hidden="true"></span></div></div></div><div class="rt-th rt-align-left reactable-d7nd0n" role="columnheader" style="position:relative;flex:100 0 auto;min-width:100px;width:100px" aria-label="Sort z" aria-sort="none" tabindex="0" data-sort-hint="ascending"><div class="rt-th-inner"><div class="rt-sort-header"><div class="rt-text-content">z</div><span class="rt-sort-right" aria-hidden="true"></span></div></div></div></div></div><div class="rt-tbody" role="rowgroup" style="min-width:300px"><div class="rt-tr-group"><div class="rt-tr" role="row" style="flex:1 0 auto;min-width:300px"><div class="rt-td rt-align-right reactable-1ub9dy9" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">1</div></div><div class="rt-td rt-align-left reactable-1ub9dy9" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">a</div></div><div class="rt-td rt-align-left reactable-1ub9dy9" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">1,2,3</div></div></div></div><div class="rt-tr-group"><div class="rt-tr" role="row" style="flex:1 0 auto;min-width:300px"><div class="rt-td rt-align-right reactable-1ub9dy9" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">2</div></div><div class="rt-td rt-align-left reactable-1ub9dy9" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">column-y-cell</div></div><div class="rt-td rt-align-left reactable-1ub9dy9" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">4,5,6</div></div></div></div><div class="rt-no-data" aria-live="assertive"></div></div></div></div></div>
      <script type="application/json" data-for="stable-id-theme-critical-css">{"x":{"tag":{"name":"Reactable","attribs":{"data":{"x":[1,2],"y":["a","column-y-cell"],"z":[[1,2,3],[4,5,6]]},"columns":[{"id":"x","name":"x","type":"numeric"},{"id":"y","name":"y","type":"character"},{"id":"z","name":"z","type":"AsIs"}],"theme":{"color":"blue","borderWidth":3},"elementId":"stable-id-theme-critical-css","dataKey":"f09fcb11dd011a5cefcb1725322e0cb4"},"children":[]},"class":"reactR_markup"},"evals":[],"jsHooks":[]}</script>

---

    Code
      cat(rendered$html)
    Output
      <div class="reactable html-widget html-fill-item-overflow-hidden html-fill-item" data-react-ssr id="stable-id-custom-js-evals" style="width:auto;height:auto;"><div class="Reactable ReactTable"><div class="rt-table" role="table"><div class="rt-thead" role="rowgroup" style="min-width:300px"><div class="rt-tr rt-tr-header" role="row" style="flex:1 0 auto;min-width:300px"><div class="rt-th rt-align-right" role="columnheader" style="position:relative;flex:100 0 auto;min-width:100px;width:100px" aria-label="Sort x" aria-sort="none" tabindex="0" data-sort-hint="ascending"><div class="rt-th-inner"><div class="rt-sort-header"><span class="rt-sort-left" aria-hidden="true"></span><div class="rt-text-content">x</div></div></div></div><div class="rt-th rt-align-left" role="columnheader" style="position:relative;flex:100 0 auto;min-width:100px;width:100px" aria-label="Sort y" aria-sort="none" tabindex="0" data-sort-hint="ascending"><div class="rt-th-inner"><div class="rt-sort-header"><div class="rt-text-content">y</div><span class="rt-sort-right" aria-hidden="true"></span></div></div></div><div class="rt-th rt-align-left" role="columnheader" style="position:relative;flex:100 0 auto;min-width:100px;width:100px" aria-label="Sort z" aria-sort="none" tabindex="0" data-sort-hint="ascending"><div class="rt-th-inner"><div class="rt-sort-header"><div class="rt-text-content">z</div><span class="rt-sort-right" aria-hidden="true"></span></div></div></div></div></div><div class="rt-tbody" role="rowgroup" style="min-width:300px"><div class="rt-tr-group"><div class="rt-tr" role="row" style="flex:1 0 auto;min-width:300px"><div class="rt-td rt-align-right" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">js-rendered_1_</div></div><div class="rt-td rt-align-left" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner"><b>a</b></div></div><div class="rt-td rt-align-left" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">1,2,3</div></div></div></div><div class="rt-tr-group"><div class="rt-tr" role="row" style="flex:1 0 auto;min-width:300px"><div class="rt-td rt-align-right" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">js-rendered_2_</div></div><div class="rt-td rt-align-left" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner"><b>column-y-cell</b></div></div><div class="rt-td rt-align-left" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">4,5,6</div></div></div></div><div class="rt-no-data" aria-live="assertive"></div></div></div></div></div>
      <script type="application/json" data-for="stable-id-custom-js-evals">{"x":{"tag":{"name":"Reactable","attribs":{"data":{"x":[1,2],"y":["a","column-y-cell"],"z":[[1,2,3],[4,5,6]]},"columns":[{"id":"x","name":"x","type":"numeric","cell":"cellInfo => `js-rendered_${cellInfo.value}_`"},{"id":"y","name":"y","type":"character","cell":[{"name":"b","attribs":{},"children":["a"]},{"name":"b","attribs":{},"children":["column-y-cell"]}]},{"id":"z","name":"z","type":"AsIs"}],"elementId":"stable-id-custom-js-evals","dataKey":"bd88f9f4f4eb8f75073fc3a22a74c1cd"},"children":[]},"class":"reactR_markup"},"evals":["tag.attribs.columns.0.cell"],"jsHooks":[]}</script>

---

    Code
      cat(rendered$html)
    Output
      <div class="reactable html-widget html-fill-item-overflow-hidden html-fill-item" data-react-ssr id="stable-id-external-React" style="width:auto;height:auto;"><div class="Reactable ReactTable"><div class="rt-table" role="table"><div class="rt-thead" role="rowgroup" style="min-width:300px"><div class="rt-tr rt-tr-header" role="row" style="flex:1 0 auto;min-width:300px"><div class="rt-th rt-align-right" role="columnheader" style="position:relative;flex:100 0 auto;min-width:100px;width:100px" aria-label="Sort x" aria-sort="none" tabindex="0" data-sort-hint="ascending"><div class="rt-th-inner"><div class="rt-sort-header"><span class="rt-sort-left" aria-hidden="true"></span><div class="rt-text-content">x</div></div></div></div><div class="rt-th rt-align-left" role="columnheader" style="position:relative;flex:100 0 auto;min-width:100px;width:100px" aria-label="Sort y" aria-sort="none" tabindex="0" data-sort-hint="ascending"><div class="rt-th-inner"><div class="rt-sort-header"><div class="rt-text-content">y</div><span class="rt-sort-right" aria-hidden="true"></span></div></div></div><div class="rt-th rt-align-left" role="columnheader" style="position:relative;flex:100 0 auto;min-width:100px;width:100px" aria-label="Sort z" aria-sort="none" tabindex="0" data-sort-hint="ascending"><div class="rt-th-inner"><div class="rt-sort-header"><div class="rt-text-content">z</div><span class="rt-sort-right" aria-hidden="true"></span></div></div></div></div></div><div class="rt-tbody" role="rowgroup" style="min-width:300px"><div class="rt-tr-group"><div class="rt-tr" role="row" style="flex:1 0 auto;min-width:300px"><div class="rt-td rt-align-right" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">1</div></div><div class="rt-td rt-align-left" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner"><b>a</b></div></div><div class="rt-td rt-align-left" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">1,2,3</div></div></div></div><div class="rt-tr-group"><div class="rt-tr" role="row" style="flex:1 0 auto;min-width:300px"><div class="rt-td rt-align-right" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">2</div></div><div class="rt-td rt-align-left" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner"><b>column-y-cell</b></div></div><div class="rt-td rt-align-left" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner">4,5,6</div></div></div></div><div class="rt-no-data" aria-live="assertive"></div></div></div></div></div>
      <script type="application/json" data-for="stable-id-external-React">{"x":{"tag":{"name":"Reactable","attribs":{"data":{"x":[1,2],"y":["a","column-y-cell"],"z":[[1,2,3],[4,5,6]]},"columns":[{"id":"x","name":"x","type":"numeric"},{"id":"y","name":"y","type":"character","cell":"cellInfo => React.createElement('b', null, cellInfo.value)"},{"id":"z","name":"z","type":"AsIs"}],"elementId":"stable-id-external-React","dataKey":"2f64b24b884b505b77491449ecc29b61"},"children":[]},"class":"reactR_markup"},"evals":["tag.attribs.columns.1.cell"],"jsHooks":[]}</script>

---

    "<div class=\"reactable html-widget html-fill-item-overflow-hidden html-fill-item\" data-react-ssr id=\"stable-id-row-details\" style=\"width:auto;height:auto;\"><div class=\"Reactable ReactTable\"><div class=\"rt-table\" role=\"table\"><div class=\"rt-thead\" role=\"rowgroup\" style=\"min-width:145px\"><div class=\"rt-tr rt-tr-header\" role=\"row\" style=\"flex:1 0 auto;min-width:145px\"><div class=\"rt-th rt-align-center\" role=\"columnheader\" style=\"position:relative;flex:0 0 auto;min-width:45px;width:45px;max-width:45px\"><div class=\"rt-th-inner\"><\/div><\/div><div class=\"rt-th rt-align-right\" role=\"columnheader\" style=\"position:relative;flex:100 0 auto;min-width:100px;width:100px\" aria-label=\"Sort x\" aria-sort=\"none\" tabindex=\"0\" data-sort-hint=\"ascending\"><div class=\"rt-th-inner\"><div class=\"rt-sort-header\"><span class=\"rt-sort-left\" aria-hidden=\"true\"><\/span><div class=\"rt-text-content\">x<\/div><\/div><\/div><\/div><\/div><\/div><div class=\"rt-tbody\" role=\"rowgroup\" style=\"min-width:145px\"><div class=\"rt-tr-group\"><div class=\"rt-tr\" role=\"row\" style=\"flex:1 0 auto;min-width:145px\"><div class=\"rt-td rt-align-center rt-td-expandable\" role=\"cell\" style=\"flex:0 0 auto;min-width:45px;width:45px;max-width:45px;text-overflow:clip;user-select:none\"><div class=\"rt-td-inner\"><button class=\"rt-expander-button\" aria-label=\"Toggle details\" aria-expanded=\"false\"><span class=\"rt-expander\" tabindex=\"-1\" aria-hidden=\"true\">​<\/span><\/button>​<\/div><\/div><div class=\"rt-td rt-align-right\" role=\"cell\" style=\"flex:100 0 auto;min-width:100px;width:100px\"><div class=\"rt-td-inner\">1<\/div><\/div><\/div><\/div><div class=\"rt-no-data\" aria-live=\"assertive\"><\/div><\/div><\/div><\/div><\/div>\n<script type=\"application/json\" data-for=\"stable-id-row-details\">{\"x\":{\"tag\":{\"name\":\"Reactable\",\"attribs\":{\"data\":{\"x\":[1]},\"columns\":[{\"id\":\".details\",\"name\":\"\",\"type\":null,\"sortable\":false,\"resizable\":false,\"filterable\":false,\"searchable\":false,\"width\":45,\"align\":\"center\",\"details\":[\"row-details\"]},{\"id\":\"x\",\"name\":\"x\",\"type\":\"numeric\"}],\"elementId\":\"stable-id-row-details\",\"dataKey\":\"cdcc049b22a5052f3797eb020abf2b4c\"},\"children\":[]},\"class\":\"reactR_markup\"},\"evals\":[],\"jsHooks\":[]}<\/script>"

---

    Code
      cat(rendered$html)
    Output
      <div class="reactable html-widget html-fill-item-overflow-hidden html-fill-item" data-react-ssr id="stable-id-html-widgets" style="width:auto;height:auto;"><div class="Reactable ReactTable"><div class="rt-table" role="table"><div class="rt-thead" role="rowgroup" style="min-width:100px"><div class="rt-tr rt-tr-header" role="row" style="flex:1 0 auto;min-width:100px"><div class="rt-th rt-align-right" role="columnheader" style="position:relative;flex:100 0 auto;min-width:100px;width:100px" aria-label="Sort x" aria-sort="none" tabindex="0" data-sort-hint="ascending"><div class="rt-th-inner"><div class="rt-sort-header"><span class="rt-sort-left" aria-hidden="true"></span><div class="rt-text-content">x</div></div></div></div></div></div><div class="rt-tbody" role="rowgroup" style="min-width:100px"><div class="rt-tr-group"><div class="rt-tr" role="row" style="flex:1 0 auto;min-width:100px"><div class="rt-td rt-align-right" role="cell" style="flex:100 0 auto;min-width:100px;width:100px"><div class="rt-td-inner"></div></div></div></div><div class="rt-no-data" aria-live="assertive"></div></div></div></div></div>
      <script type="application/json" data-for="stable-id-html-widgets">{"x":{"tag":{"name":"Reactable","attribs":{"data":{"x":[1]},"columns":[{"id":"x","name":"x","type":"numeric","cell":[{"name":"WidgetContainer","attribs":{"key":"e064ce22a125c7dc2d723f7491898939"},"children":[{"name":"Fragment","attribs":[],"children":[{"name":"span","attribs":{"id":"stable-id-sparkline","className":"sparkline html-widget "},"children":[]},{"name":"script","attribs":{"type":"application/json","data-for":"stable-id-sparkline"},"children":["{\"x\":{\"values\":[1,2],\"options\":{\"height\":20,\"width\":60},\"width\":60,\"height\":20},\"evals\":[],\"jsHooks\":[]}"]}]}]}]}],"elementId":"stable-id-html-widgets","dataKey":"f7d43b158943d129cc0ccd5b599b92e5"},"children":[]},"class":"reactR_markup"},"evals":[],"jsHooks":[]}</script>

---

    "<div class=\"reactable html-widget html-fill-item-overflow-hidden html-fill-item\" data-react-ssr id=\"stable-id-formatting-intl-polyfills\" style=\"width:auto;height:auto;\"><div class=\"Reactable ReactTable\"><div class=\"rt-table\" role=\"table\"><div class=\"rt-thead\" role=\"rowgroup\" style=\"min-width:900px\"><div class=\"rt-tr rt-tr-header\" role=\"row\" style=\"flex:1 0 auto;min-width:900px\"><div class=\"rt-th rt-align-left\" role=\"columnheader\" style=\"position:relative;flex:100 0 auto;min-width:100px;width:100px\" aria-label=\"Sort str\" aria-sort=\"none\" tabindex=\"0\" data-sort-hint=\"ascending\"><div class=\"rt-th-inner\"><div class=\"rt-sort-header\"><div class=\"rt-text-content\">str<\/div><span class=\"rt-sort-right\" aria-hidden=\"true\"><\/span><\/div><\/div><\/div><div class=\"rt-th rt-align-right\" role=\"columnheader\" style=\"position:relative;flex:100 0 auto;min-width:100px;width:100px\" aria-label=\"Sort pct\" aria-sort=\"none\" tabindex=\"0\" data-sort-hint=\"ascending\"><div class=\"rt-th-inner\"><div class=\"rt-sort-header\"><span class=\"rt-sort-left\" aria-hidden=\"true\"><\/span><div class=\"rt-text-content\">pct<\/div><\/div><\/div><\/div><div class=\"rt-th rt-align-right\" role=\"columnheader\" style=\"position:relative;flex:100 0 auto;min-width:100px;width:100px\" aria-label=\"Sort pct_digits\" aria-sort=\"none\" tabindex=\"0\" data-sort-hint=\"ascending\"><div class=\"rt-th-inner\"><div class=\"rt-sort-header\"><span class=\"rt-sort-left\" aria-hidden=\"true\"><\/span><div class=\"rt-text-content\">pct_digits<\/div><\/div><\/div><\/div><div class=\"rt-th rt-align-right\" role=\"columnheader\" style=\"position:relative;flex:100 0 auto;min-width:100px;width:100px\" aria-label=\"Sort currency_USD\" aria-sort=\"none\" tabindex=\"0\" data-sort-hint=\"ascending\"><div class=\"rt-th-inner\"><div class=\"rt-sort-header\"><span class=\"rt-sort-left\" aria-hidden=\"true\"><\/span><div class=\"rt-text-content\">currency_USD<\/div><\/div><\/div><\/div><div class=\"rt-th rt-align-right\" role=\"columnheader\" style=\"position:relative;flex:100 0 auto;min-width:100px;width:100px\" aria-label=\"Sort currency_EUR\" aria-sort=\"none\" tabindex=\"0\" data-sort-hint=\"ascending\"><div class=\"rt-th-inner\"><div class=\"rt-sort-header\"><span class=\"rt-sort-left\" aria-hidden=\"true\"><\/span><div class=\"rt-text-content\">currency_EUR<\/div><\/div><\/div><\/div><div class=\"rt-th rt-align-left\" role=\"columnheader\" style=\"position:relative;flex:100 0 auto;min-width:100px;width:100px\" aria-label=\"Sort date\" aria-sort=\"none\" tabindex=\"0\" data-sort-hint=\"ascending\"><div class=\"rt-th-inner\"><div class=\"rt-sort-header\"><div class=\"rt-text-content\">date<\/div><span class=\"rt-sort-right\" aria-hidden=\"true\"><\/span><\/div><\/div><\/div><div class=\"rt-th rt-align-left\" role=\"columnheader\" style=\"position:relative;flex:100 0 auto;min-width:100px;width:100px\" aria-label=\"Sort time\" aria-sort=\"none\" tabindex=\"0\" data-sort-hint=\"ascending\"><div class=\"rt-th-inner\"><div class=\"rt-sort-header\"><div class=\"rt-text-content\">time<\/div><span class=\"rt-sort-right\" aria-hidden=\"true\"><\/span><\/div><\/div><\/div><div class=\"rt-th rt-align-right\" role=\"columnheader\" style=\"position:relative;flex:100 0 auto;min-width:100px;width:100px\" aria-label=\"Sort num\" aria-sort=\"none\" tabindex=\"0\" data-sort-hint=\"ascending\"><div class=\"rt-th-inner\"><div class=\"rt-sort-header\"><span class=\"rt-sort-left\" aria-hidden=\"true\"><\/span><div class=\"rt-text-content\">num<\/div><\/div><\/div><\/div><div class=\"rt-th rt-align-right\" role=\"columnheader\" style=\"position:relative;flex:100 0 auto;min-width:100px;width:100px\" aria-label=\"Sort locale_hi_IN\" aria-sort=\"none\" tabindex=\"0\" data-sort-hint=\"ascending\"><div class=\"rt-th-inner\"><div class=\"rt-sort-header\"><span class=\"rt-sort-left\" aria-hidden=\"true\"><\/span><div class=\"rt-text-content\">locale_hi_IN<\/div><\/div><\/div><\/div><\/div><\/div><div class=\"rt-tbody\" role=\"rowgroup\" style=\"min-width:900px\"><div class=\"rt-tr-group\"><div class=\"rt-tr\" role=\"row\" style=\"flex:1 0 auto;min-width:900px\"><div class=\"rt-td rt-align-left\" role=\"cell\" style=\"flex:100 0 auto;min-width:100px;width:100px\"><div class=\"rt-td-inner\">pre_str_suffix<\/div><\/div><div class=\"rt-td rt-align-right\" role=\"cell\" style=\"flex:100 0 auto;min-width:100px;width:100px\"><div class=\"rt-td-inner\">75%<\/div><\/div><div class=\"rt-td rt-align-right\" role=\"cell\" style=\"flex:100 0 auto;min-width:100px;width:100px\"><div class=\"rt-td-inner\">52.9%<\/div><\/div><div class=\"rt-td rt-align-right\" role=\"cell\" style=\"flex:100 0 auto;min-width:100px;width:100px\"><div class=\"rt-td-inner\">$10.00<\/div><\/div><div class=\"rt-td rt-align-right\" role=\"cell\" style=\"flex:100 0 auto;min-width:100px;width:100px\"><div class=\"rt-td-inner\">€11.12<\/div><\/div><div class=\"rt-td rt-align-left\" role=\"cell\" style=\"flex:100 0 auto;min-width:100px;width:100px\"><div class=\"rt-td-inner\">_date_replaced_date_<\/div><\/div><div class=\"rt-td rt-align-left\" role=\"cell\" style=\"flex:100 0 auto;min-width:100px;width:100px\"><div class=\"rt-td-inner\">_time_replaced_time_<\/div><\/div><div class=\"rt-td rt-align-right\" role=\"cell\" style=\"flex:100 0 auto;min-width:100px;width:100px\"><div class=\"rt-td-inner\">1,234.1<\/div><\/div><div class=\"rt-td rt-align-right\" role=\"cell\" style=\"flex:100 0 auto;min-width:100px;width:100px\"><div class=\"rt-td-inner\">₹1,234,567.40<\/div><\/div><\/div><\/div><div class=\"rt-no-data\" aria-live=\"assertive\"><\/div><\/div><\/div><\/div><\/div>\n<script type=\"application/json\" data-for=\"stable-id-formatting-intl-polyfills\">{\"x\":{\"tag\":{\"name\":\"Reactable\",\"attribs\":{\"data\":{\"str\":[\"str\"],\"pct\":[0.75],\"pct_digits\":[0.529],\"currency_USD\":[10],\"currency_EUR\":[11.123],\"date\":[\"2019-05-06T03:22:15Z\"],\"time\":[\"2019-05-06T03:22:15Z\"],\"num\":[1234.1234],\"locale_hi_IN\":[1234567.4]},\"columns\":[{\"id\":\"str\",\"name\":\"str\",\"type\":\"character\",\"format\":{\"cell\":{\"prefix\":\"pre_\",\"suffix\":\"_suffix\"},\"aggregated\":{\"prefix\":\"pre_\",\"suffix\":\"_suffix\"}}},{\"id\":\"pct\",\"name\":\"pct\",\"type\":\"numeric\",\"format\":{\"cell\":{\"percent\":true},\"aggregated\":{\"percent\":true}}},{\"id\":\"pct_digits\",\"name\":\"pct_digits\",\"type\":\"numeric\",\"format\":{\"cell\":{\"percent\":true},\"aggregated\":{\"percent\":true}}},{\"id\":\"currency_USD\",\"name\":\"currency_USD\",\"type\":\"numeric\",\"format\":{\"cell\":{\"currency\":\"USD\"},\"aggregated\":{\"currency\":\"USD\"}}},{\"id\":\"currency_EUR\",\"name\":\"currency_EUR\",\"type\":\"numeric\",\"format\":{\"cell\":{\"currency\":\"EUR\"},\"aggregated\":{\"currency\":\"EUR\"}}},{\"id\":\"date\",\"name\":\"date\",\"type\":\"Date\",\"format\":{\"cell\":{\"prefix\":\"_date_\",\"suffix\":\"_date_\",\"datetime\":true},\"aggregated\":{\"prefix\":\"_date_\",\"suffix\":\"_date_\",\"datetime\":true}}},{\"id\":\"time\",\"name\":\"time\",\"type\":\"Date\",\"format\":{\"cell\":{\"prefix\":\"_time_\",\"suffix\":\"_time_\",\"time\":true},\"aggregated\":{\"prefix\":\"_time_\",\"suffix\":\"_time_\",\"time\":true}}},{\"id\":\"num\",\"name\":\"num\",\"type\":\"numeric\",\"format\":{\"cell\":{\"digits\":1,\"separators\":true},\"aggregated\":{\"digits\":1,\"separators\":true}}},{\"id\":\"locale_hi_IN\",\"name\":\"locale_hi_IN\",\"type\":\"numeric\",\"format\":{\"cell\":{\"separators\":true,\"currency\":\"INR\",\"locales\":\"hi-IN\"},\"aggregated\":{\"separators\":true,\"currency\":\"INR\",\"locales\":\"hi-IN\"}}}],\"elementId\":\"stable-id-formatting-intl-polyfills\",\"dataKey\":\"0519a440a28d31e079942abf20805ab1\"},\"children\":[]},\"class\":\"reactR_markup\"},\"evals\":[],\"jsHooks\":[]}<\/script>"

---

    Code
      cat(rendered$html)
    Output
      <div class="reactable html-widget html-fill-item-overflow-hidden html-fill-item" id="stable-id-CSR-fallback" style="width:auto;height:auto;"></div>
      <script type="application/json" data-for="stable-id-CSR-fallback">{"x":{"tag":{"name":"Reactable","attribs":{"data":{"x":[1,2],"y":["a","column-y-cell"],"z":[[1,2,3],[4,5,6]]},"columns":[{"id":"x","name":"x","type":"numeric","cell":"throw new Error('error rendering JS')"},{"id":"y","name":"y","type":"character"},{"id":"z","name":"z","type":"AsIs"}],"elementId":"stable-id-CSR-fallback","dataKey":"65bcb4103c068cbb97c7ef2f3a491dd2"},"children":[]},"class":"reactR_markup"},"evals":["tag.attribs.columns.0.cell"],"jsHooks":[]}</script>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reactable.R
\name{reactable}
\alias{reactable}
\title{Create an interactive data table}
\usage{
reactable(
  data,
  columns = NULL,
  columnGroups = NULL,
  rownames = NULL,
  groupBy = NULL,
  sortable = TRUE,
  resizable = FALSE,
  filterable = FALSE,
  searchable = FALSE,
  searchMethod = NULL,
  defaultColDef = NULL,
  defaultColGroup = NULL,
  defaultSortOrder = "asc",
  defaultSorted = NULL,
  pagination = TRUE,
  defaultPageSize = 10,
  showPageSizeOptions = FALSE,
  pageSizeOptions = c(10, 25, 50, 100),
  paginationType = "numbers",
  showPagination = NULL,
  showPageInfo = TRUE,
  minRows = 1,
  paginateSubRows = FALSE,
  details = NULL,
  defaultExpanded = FALSE,
  selection = NULL,
  defaultSelected = NULL,
  onClick = NULL,
  highlight = FALSE,
  outlined = FALSE,
  bordered = FALSE,
  borderless = FALSE,
  striped = FALSE,
  compact = FALSE,
  wrap = TRUE,
  showSortIcon = TRUE,
  showSortable = FALSE,
  class = NULL,
  style = NULL,
  rowClass = NULL,
  rowStyle = NULL,
  fullWidth = TRUE,
  width = NULL,
  height = NULL,
  theme = getOption("reactable.theme"),
  language = getOption("reactable.language"),
  meta = NULL,
  elementId = NULL,
  static = getOption("reactable.static", FALSE),
  selectionId = NULL
)
}
\arguments{
\item{data}{A data frame or matrix.

Can also be a \code{\link[crosstalk:SharedData]{crosstalk::SharedData}} object that wraps a data frame.}

\item{columns}{Named list of column definitions. See \code{\link[=colDef]{colDef()}}.}

\item{columnGroups}{List of column group definitions. See \code{\link[=colGroup]{colGroup()}}.}

\item{rownames}{Show row names? Defaults to \code{TRUE} if the data has row names.

To customize the row names column, use \code{".rownames"} as the column name.

Cells in the row names column are automatically marked up as row headers
for assistive technologies.}

\item{groupBy}{Character vector of column names to group by.

To aggregate data when rows are grouped, use the \code{aggregate} argument in \code{\link[=colDef]{colDef()}}.}

\item{sortable}{Enable sorting? Defaults to \code{TRUE}.}

\item{resizable}{Enable column resizing?}

\item{filterable}{Enable column filtering?}

\item{searchable}{Enable global table searching?}

\item{searchMethod}{Custom search method to use for global table searching.
A \code{\link[=JS]{JS()}} function that takes an array of row objects, an array of
column IDs, and the search value as arguments, and returns the filtered
array of row objects.}

\item{defaultColDef}{Default column definition used by every column. See \code{\link[=colDef]{colDef()}}.}

\item{defaultColGroup}{Default column group definition used by every column group.
See \code{\link[=colGroup]{colGroup()}}.}

\item{defaultSortOrder}{Default sort order. Either \code{"asc"} for ascending
order or \code{"desc"} for descending order. Defaults to \code{"asc"}.}

\item{defaultSorted}{Character vector of column names to sort by default.
Or to customize sort order, a named list with values of \code{"asc"} or \code{"desc"}.}

\item{pagination}{Enable pagination? Defaults to \code{TRUE}.}

\item{defaultPageSize}{Default page size for the table. Defaults to 10.}

\item{showPageSizeOptions}{Show page size options?}

\item{pageSizeOptions}{Page size options for the table. Defaults to 10, 25, 50, 100.}

\item{paginationType}{Pagination control to use. Either \code{"numbers"} for page
number buttons (the default), \code{"jump"} for a page jump, or \code{"simple"} to show
'Previous' and 'Next' buttons only.}

\item{showPagination}{Show pagination? Defaults to \code{TRUE} if the table has more
than one page.}

\item{showPageInfo}{Show page info? Defaults to \code{TRUE}.}

\item{minRows}{Minimum number of rows to show per page. Defaults to 1.}

\item{paginateSubRows}{When rows are grouped, paginate sub rows? Defaults to \code{FALSE}.}

\item{details}{Additional content to display when expanding a row. An R function
that takes the row index and column name as arguments, or a \code{\link[=JS]{JS()}} function
that takes a row info object as an argument. Can also be a \code{\link[=colDef]{colDef()}} to
customize the details expander column.}

\item{defaultExpanded}{Expand all rows by default?}

\item{selection}{Enable row selection? Either \code{"multiple"} or \code{"single"} for
multiple or single row selection.

To get the selected rows in Shiny, use \code{\link[=getReactableState]{getReactableState()}}.

To customize the selection column, use \code{".selection"} as the column name.}

\item{defaultSelected}{A numeric vector of default selected row indices.}

\item{onClick}{Action to take when clicking a cell. Either \code{"expand"} to expand
the row, \code{"select"} to select the row, or a \code{\link[=JS]{JS()}} function that takes a
row info object, column object, and table state object as arguments.}

\item{highlight}{Highlight table rows on hover?}

\item{outlined}{Add borders around the table?}

\item{bordered}{Add borders around the table and every cell?}

\item{borderless}{Remove inner borders from table?}

\item{striped}{Add zebra-striping to table rows?}

\item{compact}{Make tables more compact?}

\item{wrap}{Enable text wrapping? If \code{TRUE} (the default), long text will be
wrapped to multiple lines. If \code{FALSE}, text will be truncated to fit on one line.}

\item{showSortIcon}{Show a sort icon when sorting columns?}

\item{showSortable}{Show an indicator on sortable columns?}

\item{class}{Additional CSS classes to apply to the table.}

\item{style}{Inline styles to apply to the table. A named list or character string.

Note that if \code{style} is a named list, property names should be camelCased.}

\item{rowClass}{Additional CSS classes to apply to table rows. A character
string, a \code{\link[=JS]{JS()}} function that takes a row info object and table state object
as arguments, or an R function that takes a row index argument.}

\item{rowStyle}{Inline styles to apply to table rows. A named list, character
string, \code{\link[=JS]{JS()}} function that takes a row info object and table state object
as arguments, or an R function that takes a row index argument.

Note that if \code{rowStyle} is a named list, property names should be camelCased.
If \code{rowStyle} is a \code{\link[=JS]{JS()}} function, it should return a JavaScript object with
camelCased property names.}

\item{fullWidth}{Stretch the table to fill the full width of its container?
Defaults to \code{TRUE}.}

\item{width}{Width of the table in pixels. Defaults to \code{"auto"} for automatic sizing.

To set the width of a column, see \code{\link[=colDef]{colDef()}}.}

\item{height}{Height of the table in pixels. Defaults to \code{"auto"} for automatic sizing.}

\item{theme}{Theme options for the table, specified by
\code{\link[=reactableTheme]{reactableTheme()}}. Defaults to the global \code{reactable.theme} option.
Can also be a function that returns a \code{\link[=reactableTheme]{reactableTheme()}} or \code{NULL}.}

\item{language}{Language options for the table, specified by
\code{\link[=reactableLang]{reactableLang()}}. Defaults to the global \code{reactable.language} option.}

\item{meta}{Custom metadata to pass to JavaScript render functions or style functions.
A named list of values that can also be \code{\link[=JS]{JS()}} expressions or functions.
Custom metadata can be accessed using the \code{state.meta} property, and updated
using \code{updateReactable()} in Shiny or \code{Reactable.setMeta()} in the JavaScript API.}

\item{elementId}{Element ID for the widget.}

\item{static}{Render the table to static HTML? Defaults to the global
\code{reactable.static} option. Requires the V8 package, which is not installed
with reactable by default.

With static rendering, tables are pre-rendered to their initial HTML so they appear
immediately without any flash of content. Tables are then made interactive and
subsequently rendered by JavaScript as needed.

Static rendering is \strong{experimental}, and is not supported for tables
rendered via \code{\link[=reactableOutput]{reactableOutput()}} in Shiny.}

\item{selectionId}{\strong{Deprecated}. Use \code{\link[=getReactableState]{getReactableState()}} to get the selected rows
in Shiny.}
}
\value{
A \code{reactable} HTML widget that can be used in R Markdown documents
and Shiny applications, or viewed from an R console.
}
\description{
\code{reactable()} creates a data table from tabular data with sorting
and pagination by default. The data table is an HTML widget that can be
used in R Markdown documents and Shiny applications, or viewed from an
R console.
}
\note{
See the \href{https://glin.github.io/reactable/}{online documentation} for
additional details and examples.
}
\examples{
# Basic usage
reactable(iris)

# Grouping and aggregation
reactable(
  iris,
  groupBy = "Species",
  columns = list(
    Sepal.Length = colDef(aggregate = "count"),
    Sepal.Width = colDef(aggregate = "mean"),
    Petal.Length = colDef(aggregate = "sum"),
    Petal.Width = colDef(aggregate = "max")
  )
)

# Row details
reactable(iris, details = function(index) {
  htmltools::div(
    "Details for row: ", index,
    htmltools::tags$pre(paste(capture.output(iris[index, ]), collapse = "\n"))
  )
})

# Conditional styling
reactable(sleep, columns = list(
  extra = colDef(style = function(value) {
    if (value > 0) {
      color <- "green"
    } else if (value < 0) {
      color <- "red"
    } else {
      color <- "#777"
    }
    list(color = color, fontWeight = "bold")
  })
))

}
\seealso{
\itemize{
\item \code{\link[=renderReactable]{renderReactable()}} and \code{\link[=reactableOutput]{reactableOutput()}} for using reactable
in Shiny applications or interactive R Markdown documents.
\item \code{\link[=colDef]{colDef()}}, \code{\link[=colFormat]{colFormat()}}, and \code{\link[=colGroup]{colGroup()}} to customize columns.
\item \code{\link[=reactableTheme]{reactableTheme()}} and \code{\link[=reactableLang]{reactableLang()}} to customize the table.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/language.R
\name{reactableLang}
\alias{reactableLang}
\title{Language options}
\usage{
reactableLang(
  sortLabel = "Sort {name}",
  filterPlaceholder = "",
  filterLabel = "Filter {name}",
  searchPlaceholder = "Search",
  searchLabel = "Search",
  noData = "No rows found",
  pageNext = "Next",
  pagePrevious = "Previous",
  pageNumbers = "{page} of {pages}",
  pageInfo = "{rowStart}\u2013{rowEnd} of {rows} rows",
  pageSizeOptions = "Show {rows}",
  pageNextLabel = "Next page",
  pagePreviousLabel = "Previous page",
  pageNumberLabel = "Page {page}",
  pageJumpLabel = "Go to page",
  pageSizeOptionsLabel = "Rows per page",
  groupExpandLabel = "Toggle group",
  detailsExpandLabel = "Toggle details",
  selectAllRowsLabel = "Select all rows",
  selectAllSubRowsLabel = "Select all rows in group",
  selectRowLabel = "Select row",
  defaultGroupHeader = NULL,
  detailsCollapseLabel = NULL,
  deselectAllRowsLabel = NULL,
  deselectAllSubRowsLabel = NULL,
  deselectRowLabel = NULL
)
}
\arguments{
\item{sortLabel}{Accessible label for column sort buttons.
Takes a \code{{name}} parameter for the column name.}

\item{filterPlaceholder}{Placeholder for column filter inputs.}

\item{filterLabel}{Accessible label for column filter inputs.
Takes a \code{{name}} parameter for the column name.}

\item{searchPlaceholder}{Placeholder for the table search input.}

\item{searchLabel}{Accessible label for the table search input.}

\item{noData}{Placeholder text when the table has no data.}

\item{pageNext}{Text for the next page button.}

\item{pagePrevious}{Text for the previous page button.}

\item{pageNumbers}{Text for the page numbers info. Only used with the \code{"jump"} and
\code{"simple"} pagination types.
Takes the following parameters:
\itemize{
\item \code{{page}} for the current page
\item \code{{pages}} for the total number of pages
}}

\item{pageInfo}{Text for the page info.
Takes the following parameters:
\itemize{
\item \code{{rowStart}} for the starting row of the page
\item \code{{rowEnd}} for the ending row of the page
\item \code{{rows}} for the total number of rows
}}

\item{pageSizeOptions}{Text for the page size options input.
Takes a \code{{rows}} parameter for the page size options input.}

\item{pageNextLabel}{Accessible label for the next page button.}

\item{pagePreviousLabel}{Accessible label for the previous page button.}

\item{pageNumberLabel}{Accessible label for the page number buttons.
Only used with the the \code{"numbers"} pagination type.
Takes a \code{{page}} parameter for the page number.}

\item{pageJumpLabel}{Accessible label for the page jump input. Only used with
the \code{"jump"} pagination type.}

\item{pageSizeOptionsLabel}{Accessible label for the page size options input.}

\item{groupExpandLabel}{Accessible label for the row group expand button.}

\item{detailsExpandLabel}{Accessible label for the row details expand button.}

\item{selectAllRowsLabel}{Accessible label for the select all rows checkbox.}

\item{selectAllSubRowsLabel}{Accessible label for the select all sub rows checkbox.}

\item{selectRowLabel}{Accessible label for the select row checkbox.}

\item{defaultGroupHeader}{Deprecated and no longer used.}

\item{detailsCollapseLabel}{Deprecated and no longer used.}

\item{deselectAllRowsLabel}{Deprecated and no longer used.}

\item{deselectAllSubRowsLabel}{Deprecated and no longer used.}

\item{deselectRowLabel}{Deprecated and no longer used.}
}
\value{
A language options object that can be used to customize the language
strings in \code{reactable()}.
}
\description{
Use \code{reactableLang()} to customize the language strings in a table.
Language strings include both visible text and accessible labels that can
be read by assistive technology, such as screen readers.

To set the default language strings for all tables, use the global
\code{reactable.language} option.
}
\examples{
reactable(
  iris[1:30, ],
  searchable = TRUE,
  paginationType = "simple",
  language = reactableLang(
    searchPlaceholder = "Search...",
    noData = "No entries found",
    pageInfo = "{rowStart}\u2013{rowEnd} of {rows} entries",
    pagePrevious = "\u276e",
    pageNext = "\u276f",

    # Accessible labels for assistive technology, such as screen readers
    pagePreviousLabel = "Previous page",
    pageNextLabel = "Next page"
  )
)

# Set the default language for all tables
options(reactable.language = reactableLang(
  searchPlaceholder = "Search...",
  noData = "No entries found",
  pageInfo = "{rowStart} to {rowEnd} of {rows} entries"
))

reactable(iris[1:30, ], searchable = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/columns.R
\name{colDef}
\alias{colDef}
\title{Column definitions}
\usage{
colDef(
  name = NULL,
  aggregate = NULL,
  sortable = NULL,
  resizable = NULL,
  filterable = NULL,
  searchable = NULL,
  filterMethod = NULL,
  show = TRUE,
  defaultSortOrder = NULL,
  sortNALast = FALSE,
  format = NULL,
  cell = NULL,
  grouped = NULL,
  aggregated = NULL,
  header = NULL,
  footer = NULL,
  details = NULL,
  filterInput = NULL,
  html = FALSE,
  na = "",
  rowHeader = FALSE,
  minWidth = 100,
  maxWidth = NULL,
  width = NULL,
  align = NULL,
  vAlign = NULL,
  headerVAlign = NULL,
  sticky = NULL,
  class = NULL,
  style = NULL,
  headerClass = NULL,
  headerStyle = NULL,
  footerClass = NULL,
  footerStyle = NULL
)
}
\arguments{
\item{name}{Column header name.}

\item{aggregate}{Aggregate function to use when rows are grouped. The name
of a built-in aggregate function or a custom \code{\link[=JS]{JS()}} aggregate function.
Built-in aggregate functions are: \code{"mean"}, \code{"sum"}, \code{"max"}, \code{"min"},
\code{"median"}, \code{"count"}, \code{"unique"}, and \code{"frequency"}.

To enable row grouping, use the \code{groupBy} argument in \code{\link[=reactable]{reactable()}}.}

\item{sortable}{Enable sorting? Overrides the table option.}

\item{resizable}{Enable column resizing? Overrides the table option.}

\item{filterable}{Enable column filtering? Overrides the table option.}

\item{searchable}{Enable or disable global table searching for this column.
By default, global searching applies to all visible columns. Set this to
\code{FALSE} to exclude a visible column from searching, or \code{TRUE} to include a
hidden column in searching.}

\item{filterMethod}{Custom filter method to use for column filtering.
A \code{\link[=JS]{JS()}} function that takes an array of row objects, the column ID,
and the filter value as arguments, and returns the filtered array of
row objects.}

\item{show}{Show the column?

If \code{FALSE}, this column will be excluded from global table searching by
default. To include this hidden column in searching, set \code{searchable}
to \code{TRUE} in \code{\link[=colDef]{colDef()}}.}

\item{defaultSortOrder}{Default sort order. Either \code{"asc"} for ascending
order or \code{"desc"} for descending order. Overrides the table option.}

\item{sortNALast}{Always sort missing values (\link{NA} or \link{NaN}) last?}

\item{format}{Column formatting options. A \code{\link[=colFormat]{colFormat()}} object to
format all cells, or a named list of \code{\link[=colFormat]{colFormat()}} objects to format standard
cells (\code{"cell"}) and aggregated cells (\code{"aggregated"}) separately.}

\item{cell}{Custom cell renderer. An R function that takes the cell value,
row index, and column name as arguments, or a \code{\link[=JS]{JS()}} function that takes a
cell info object and table state object as arguments.}

\item{grouped}{Custom grouped cell renderer. A \code{\link[=JS]{JS()}} function that takes a
cell info object and table state object as arguments.}

\item{aggregated}{Custom aggregated cell renderer. A \code{\link[=JS]{JS()}} function that takes
a cell info object and table state object as arguments.}

\item{header}{Custom header renderer. An R function that takes the header value
and column name as arguments, or a \code{\link[=JS]{JS()}} function that takes a column
object and table state object as arguments.}

\item{footer}{Footer content or render function. Render functions can be an
R function that takes the column values and column name as arguments, or a
\code{\link[=JS]{JS()}} function that takes a column object and table state object as
arguments.}

\item{details}{Additional content to display when expanding a row. An R function
that takes the row index and column name as arguments, or a \code{\link[=JS]{JS()}} function
that takes a row info object and table state object as arguments.
Cannot be used on a \code{groupBy} column.}

\item{filterInput}{Custom filter input or render function. Render functions can
be an R function that takes the column values and column name as arguments,
or a \code{\link[=JS]{JS()}} function that takes a column object and table state object as
arguments.}

\item{html}{Render content as HTML? Raw HTML strings are escaped by default.}

\item{na}{String to display for missing values (i.e. \link{NA} or \link{NaN}).
By default, missing values are displayed as blank cells.}

\item{rowHeader}{Mark up cells in this column as row headers?

Set this to \code{TRUE} to help users navigate the table using assistive technologies.
When cells are marked up as row headers, assistive technologies will read them
aloud while navigating through cells in the table.

Cells in the row names column are automatically marked up as row headers.}

\item{minWidth}{Minimum width of the column in pixels. Defaults to 100.}

\item{maxWidth}{Maximum width of the column in pixels.}

\item{width}{Fixed width of the column in pixels. Overrides \code{minWidth} and \code{maxWidth}.}

\item{align}{Horizontal alignment of content in the column. One of
\code{"left"}, \code{"right"}, \code{"center"}. By default, all numbers are right-aligned,
while all other content is left-aligned.}

\item{vAlign}{Vertical alignment of content in data cells. One of \code{"top"}
(the default), \code{"center"}, \code{"bottom"}.}

\item{headerVAlign}{Vertical alignment of content in header cells. One of
\code{"top"} (the default), \code{"center"}, \code{"bottom"}.}

\item{sticky}{Make the column sticky when scrolling horizontally? Either
\code{"left"} or \code{"right"} to make the column stick to the left or right side.

If a sticky column is in a column group, all columns in the group will
automatically be made sticky, including the column group header.}

\item{class}{Additional CSS classes to apply to cells. Can also be an R function
that takes the cell value, row index, and column name as arguments, or a \code{\link[=JS]{JS()}}
function that takes a row info object, column object, and table state object
as arguments.

Note that R functions cannot apply classes to aggregated cells.}

\item{style}{Inline styles to apply to cells. A named list or character string.
Can also be an R function that takes the cell value and row index as arguments,
or a \code{\link[=JS]{JS()}} function that takes a row info object, column object, and
table state object as arguments.

Note that R functions cannot apply styles to aggregated cells.
If \code{style} is a named list, property names should be camelCased.}

\item{headerClass}{Additional CSS classes to apply to the header.}

\item{headerStyle}{Inline styles to apply to the header. A named list or
character string.

Note that if \code{headerStyle} is a named list, property names should be camelCased.}

\item{footerClass}{Additional CSS classes to apply to the footer.}

\item{footerStyle}{Inline styles to apply to the footer. A named list or
character string.

Note that if \code{footerStyle} is a named list, property names should be camelCased.}
}
\value{
A column definition object that can be used to customize columns
in \code{reactable()}.
}
\description{
Use \code{colDef()} to customize the columns in a table.
}
\examples{
reactable(
  iris,
  columns = list(
    Sepal.Length = colDef(name = "Sepal Length"),
    Sepal.Width = colDef(filterable = TRUE),
    Petal.Length = colDef(show = FALSE),
    Petal.Width = colDef(defaultSortOrder = "desc")
  )
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reactable.R
\name{reactable-shiny}
\alias{reactable-shiny}
\alias{reactableOutput}
\alias{renderReactable}
\title{Shiny bindings for reactable}
\usage{
reactableOutput(outputId, width = "auto", height = "auto", inline = FALSE)

renderReactable(expr, env = parent.frame(), quoted = FALSE)
}
\arguments{
\item{outputId}{Output variable to read from.}

\item{width, height}{A valid CSS unit (like \code{"100\%"}, \code{"400px"}, \code{"auto"})
or a number, which will be coerced to a string and have \code{"px"} appended.}

\item{inline}{Use an inline element for the table's container?}

\item{expr}{An expression that generates a \link{reactable} widget.}

\item{env}{The environment in which to evaluate \code{expr}.}

\item{quoted}{Is \code{expr} a quoted expression (with \code{\link[=quote]{quote()}})? This is useful
if you want to save an expression in a variable.}
}
\value{
\code{reactableOutput()} returns a \code{reactable} output element that can be
included in a Shiny UI.

\code{renderReactable()} returns a \code{reactable} render function that can be
assigned to a Shiny output slot.
}
\description{
Output and render functions for using reactable within Shiny
applications and interactive R Markdown documents.
}
\note{
See the \href{https://glin.github.io/reactable/articles/shiny-demo.html}{online demo}
for additional examples of using reactable in Shiny.
}
\examples{
# Run in an interactive R session
if (interactive()) {

library(shiny)
library(reactable)

ui <- fluidPage(
 titlePanel("reactable example"),
 reactableOutput("table")
)

server <- function(input, output, session) {
  output$table <- renderReactable({
   reactable(iris)
 })
}

shinyApp(ui, server)
}

}
\seealso{
\code{\link[=updateReactable]{updateReactable()}} for updating a reactable instance in Shiny.

\code{\link[=getReactableState]{getReactableState()}} for getting the state of a reactable instance in Shiny.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reactable.R
\name{knit_print.reactable}
\alias{knit_print.reactable}
\title{Print a reactable widget for knitr}
\usage{
knit_print.reactable(x, ...)
}
\arguments{
\item{x}{A \code{\link[=reactable]{reactable()}} instance.}

\item{...}{Additional arguments passed to the S3 method.}
}
\description{
This S3 method exists to enable \code{\link[=reactable]{reactable()}}'s \code{static} rendering option.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reactable.R
\name{widget_html.reactable}
\alias{widget_html.reactable}
\title{Called by HTMLWidgets to produce the widget's root element}
\usage{
widget_html.reactable(id, style, class, ...)
}
\arguments{
\item{id}{Element ID.}

\item{style}{Element style.}

\item{class}{Element class.}

\item{...}{Additional arguments.}
}
\description{
Called by HTMLWidgets to produce the widget's root element
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/theme.R
\name{reactableTheme}
\alias{reactableTheme}
\title{Theme options}
\usage{
reactableTheme(
  color = NULL,
  backgroundColor = NULL,
  borderColor = NULL,
  borderWidth = NULL,
  stripedColor = NULL,
  highlightColor = NULL,
  cellPadding = NULL,
  style = NULL,
  tableStyle = NULL,
  headerStyle = NULL,
  groupHeaderStyle = NULL,
  tableBodyStyle = NULL,
  rowGroupStyle = NULL,
  rowStyle = NULL,
  rowStripedStyle = NULL,
  rowHighlightStyle = NULL,
  rowSelectedStyle = NULL,
  cellStyle = NULL,
  footerStyle = NULL,
  inputStyle = NULL,
  filterInputStyle = NULL,
  searchInputStyle = NULL,
  selectStyle = NULL,
  paginationStyle = NULL,
  pageButtonStyle = NULL,
  pageButtonHoverStyle = NULL,
  pageButtonActiveStyle = NULL,
  pageButtonCurrentStyle = NULL
)
}
\arguments{
\item{color}{Default text color.}

\item{backgroundColor}{Default background color.}

\item{borderColor}{Default border color.}

\item{borderWidth}{Default border width.}

\item{stripedColor}{Default row stripe color.}

\item{highlightColor}{Default row highlight color.}

\item{cellPadding}{Default cell padding.}

\item{style}{Additional CSS for the table.}

\item{tableStyle}{Additional CSS for the table element (excludes
the pagination bar and search input).}

\item{headerStyle}{Additional CSS for header cells.}

\item{groupHeaderStyle}{Additional CSS for group header cells.}

\item{tableBodyStyle}{Additional CSS for the table body element.}

\item{rowGroupStyle}{Additional CSS for row groups.}

\item{rowStyle}{Additional CSS for rows.}

\item{rowStripedStyle}{Additional CSS for striped rows.}

\item{rowHighlightStyle}{Additional CSS for highlighted rows.}

\item{rowSelectedStyle}{Additional CSS for selected rows.}

\item{cellStyle}{Additional CSS for cells.}

\item{footerStyle}{Additional CSS for footer cells.}

\item{inputStyle}{Additional CSS for inputs.}

\item{filterInputStyle}{Additional CSS for filter inputs.}

\item{searchInputStyle}{Additional CSS for the search input.}

\item{selectStyle}{Additional CSS for table select controls.}

\item{paginationStyle}{Additional CSS for the pagination bar.}

\item{pageButtonStyle, pageButtonHoverStyle, pageButtonActiveStyle, pageButtonCurrentStyle}{Additional CSS for page buttons, page buttons with hover or active
states, and the current page button.}
}
\value{
A theme options object that can be used to customize the default
styling in \code{reactable()}.
}
\description{
Use \code{reactableTheme()} to customize the default styling of a table.
You can set theme variables to change the default styles, or add custom CSS
to specific elements of the table.

The \code{color} variables are specified as character strings of CSS color values.
The \code{width} and \code{padding} variables are specified as either character strings
of CSS width and padding values, or numeric pixel values. The \code{style} arguments
take custom CSS as named lists of camelCased properties.

To set the default theme for all tables, use the global \code{reactable.theme} option.
}
\details{
You can use nested CSS selectors in \code{style} arguments to target
the current element, using \code{&} as the selector, or other child elements
(just like in Sass). This is useful for adding pseudo-classes like \verb{&:hover},
or adding styles in a certain context like \verb{.outer-container &}.
}
\examples{
reactable(
  iris[1:30, ],
  searchable = TRUE,
  striped = TRUE,
  highlight = TRUE,
  bordered = TRUE,
  theme = reactableTheme(
    borderColor = "#dfe2e5",
    stripedColor = "#f6f8fa",
    highlightColor = "#f0f5f9",
    cellPadding = "8px 12px",
    style = list(
      fontFamily = "-apple-system, BlinkMacSystemFont, Segoe UI, Helvetica, Arial, sans-serif"
    ),
    searchInputStyle = list(width = "100\%")
  )
)

# Set the default theme for all tables
options(reactable.theme = reactableTheme(
  color = "hsl(233, 9\%, 87\%)",
  backgroundColor = "hsl(233, 9\%, 19\%)",
  borderColor = "hsl(233, 9\%, 22\%)",
  stripedColor = "hsl(233, 12\%, 22\%)",
  highlightColor = "hsl(233, 12\%, 24\%)",
  inputStyle = list(backgroundColor = "hsl(233, 9\%, 25\%)"),
  selectStyle = list(backgroundColor = "hsl(233, 9\%, 25\%)"),
  pageButtonHoverStyle = list(backgroundColor = "hsl(233, 9\%, 25\%)"),
  pageButtonActiveStyle = list(backgroundColor = "hsl(233, 9\%, 28\%)")
))

reactable(
  iris[1:30, ],
  filterable = TRUE,
  showPageSizeOptions = TRUE,
  striped = TRUE,
  highlight = TRUE,
  details = function(index) paste("Details for row", index)
)

# Use nested selectors to highlight headers when sorting
reactable(
  iris[1:30, ],
  columns = list(Sepal.Length = colDef(sortable = FALSE)),
  showSortable = TRUE,
  theme = reactableTheme(
    headerStyle = list(
      "&:hover[aria-sort]" = list(background = "hsl(0, 0\%, 96\%)"),
      "&[aria-sort='ascending'], &[aria-sort='descending']" = list(background = "hsl(0, 0\%, 96\%)"),
      borderColor = "#555"
    )
  )
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shiny.R
\name{updateReactable}
\alias{updateReactable}
\title{Update a reactable instance}
\usage{
updateReactable(
  outputId,
  data = NULL,
  selected = NULL,
  expanded = NULL,
  page = NULL,
  meta = NULL,
  session = NULL
)
}
\arguments{
\item{outputId}{The Shiny output ID of the \code{reactable} instance.}

\item{data}{Table data. A data frame or matrix.

\code{data} should have the same columns as the original table data.
When updating \code{data}, the selected rows, expanded rows, and current page
will reset unless explicitly specified. All other state will persist,
including sorting, filtering, and grouping state.}

\item{selected}{Selected rows. Either a numeric vector of row indices,
or \code{NA} to deselect all rows.}

\item{expanded}{Expanded rows. Either \code{TRUE} to expand all rows, or \code{FALSE}
to collapse all rows.}

\item{page}{The current page. A single, positive integer.}

\item{meta}{Custom table metadata. Either a named list with new values, or \code{NA}
to clear all metadata. New values are merged into the current metadata, so only
the values specified in \code{meta} will be updated.}

\item{session}{The Shiny session object. Defaults to the current Shiny session.}
}
\value{
None
}
\description{
\code{updateReactable()} updates a reactable instance within a Shiny application.
}
\examples{
# Run in an interactive R session
if (interactive()) {

library(shiny)
library(reactable)

data <- MASS::Cars93[, 1:7]

ui <- fluidPage(
  actionButton("select_btn", "Select rows"),
  actionButton("clear_btn", "Clear selection"),
  actionButton("expand_btn", "Expand rows"),
  actionButton("collapse_btn", "Collapse rows"),
  actionButton("page_btn", "Change page"),
  selectInput("filter_type", "Filter type", unique(data$Type), multiple = TRUE),
  reactableOutput("table")
)

server <- function(input, output) {
  output$table <- renderReactable({
    reactable(
      data,
      filterable = TRUE,
      searchable = TRUE,
      selection = "multiple",
      details = function(index) paste("Details for row:", index)
    )
  })

  observeEvent(input$select_btn, {
    # Select rows
    updateReactable("table", selected = c(1, 3, 5))
  })

  observeEvent(input$clear_btn, {
    # Clear row selection
    updateReactable("table", selected = NA)
  })

  observeEvent(input$expand_btn, {
    # Expand all rows
    updateReactable("table", expanded = TRUE)
  })

  observeEvent(input$collapse_btn, {
    # Collapse all rows
    updateReactable("table", expanded = FALSE)
  })

  observeEvent(input$page_btn, {
    # Change current page
    updateReactable("table", page = 3)
  })

  observe({
    # Filter data
    filtered <- if (length(input$filter_type) > 0) {
      data[data$Type \%in\% input$filter_type, ]
    } else {
      data
    }
    updateReactable("table", data = filtered)
  })
}

shinyApp(ui, server)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/columns.R
\name{colFormat}
\alias{colFormat}
\title{Column formatting options}
\usage{
colFormat(
  prefix = NULL,
  suffix = NULL,
  digits = NULL,
  separators = FALSE,
  percent = FALSE,
  currency = NULL,
  datetime = FALSE,
  date = FALSE,
  time = FALSE,
  hour12 = NULL,
  locales = NULL
)
}
\arguments{
\item{prefix}{Prefix string.}

\item{suffix}{Suffix string.}

\item{digits}{Number of decimal digits to use for numbers.}

\item{separators}{Whether to use grouping separators for numbers, such as
thousands separators or thousand/lakh/crore separators. The format is
locale-dependent.}

\item{percent}{Format number as a percentage? The format is locale-dependent.}

\item{currency}{Currency format. An ISO 4217 currency code such as \code{"USD"}
for the US dollar, \code{"EUR"} for the euro, or \code{"CNY"} for the Chinese RMB.
The format is locale-dependent.}

\item{datetime}{Format as a locale-dependent date-time?}

\item{date}{Format as a locale-dependent date?}

\item{time}{Format as a locale-dependent time?}

\item{hour12}{Whether to use 12-hour time (\code{TRUE}) or 24-hour time (\code{FALSE}).
The default time convention is locale-dependent.}

\item{locales}{Locales to use for number, date, time, and currency formatting.
A character vector of BCP 47 language tags, such as \code{"en-US"} for English
(United States), \code{"hi"} for Hindi, or \code{"sv-SE"} for Swedish (Sweden).
Defaults to the locale of the user's browser.

Multiple locales may be specified to provide a fallback language in case
a locale is unsupported. When multiple locales are specified, the first
supported locale will be used.

See a list of \href{https://learn.microsoft.com/en-us/openspecs/office_standards/ms-oe376/6c085406-a698-4e12-9d4d-c3b0ee3dbc4a}{common BCP 47 language tags}
for reference.}
}
\value{
A column format object that can be used to customize data formatting
in \code{colDef()}.
}
\description{
Use \code{colFormat()} to add data formatting to a column.
}
\examples{
data <- data.frame(
  price_USD = c(123456.56, 132, 5650.12),
  price_INR = c(350, 23208.552, 1773156.4),
  number_FR = c(123456.56, 132, 5650.12),
  temp = c(22, NA, 31),
  percent = c(0.9525556, 0.5, 0.112),
  date = as.Date(c("2019-01-02", "2019-03-15", "2019-09-22"))
)

reactable(data, columns = list(
  price_USD = colDef(format = colFormat(prefix = "$", separators = TRUE, digits = 2)),
  price_INR = colDef(format = colFormat(currency = "INR", separators = TRUE, locales = "hi-IN")),
  number_FR = colDef(format = colFormat(locales = "fr-FR")),
  temp = colDef(format = colFormat(suffix = " \u00b0C")),
  percent = colDef(format = colFormat(percent = TRUE, digits = 1)),
  date = colDef(format = colFormat(date = TRUE, locales = "en-GB"))
))

# Date formatting
datetimes <- as.POSIXct(c("2019-01-02 3:22:15", "2019-03-15 09:15:55", "2019-09-22 14:20:00"))
data <- data.frame(
  datetime = datetimes,
  date = datetimes,
  time = datetimes,
  time_24h = datetimes,
  datetime_pt_BR = datetimes
)

reactable(data, columns = list(
  datetime = colDef(format = colFormat(datetime = TRUE)),
  date = colDef(format = colFormat(date = TRUE)),
  time = colDef(format = colFormat(time = TRUE)),
  time_24h = colDef(format = colFormat(time = TRUE, hour12 = FALSE)),
  datetime_pt_BR = colDef(format = colFormat(datetime = TRUE, locales = "pt-BR"))
))

# Currency formatting
data <- data.frame(
  USD = c(12.12, 2141.213, 0.42, 1.55, 34414),
  EUR = c(10.68, 1884.27, 0.37, 1.36, 30284.32),
  INR = c(861.07, 152122.48, 29.84, 110, 2444942.63),
  JPY = c(1280, 226144, 44.36, 164, 3634634.61),
  MAD = c(115.78, 20453.94, 4.01, 15, 328739.73)
)

reactable(data, columns = list(
  USD = colDef(
    format = colFormat(currency = "USD", separators = TRUE, locales = "en-US")
  ),
  EUR = colDef(
    format = colFormat(currency = "EUR", separators = TRUE, locales = "de-DE")
  ),
  INR = colDef(
    format = colFormat(currency = "INR", separators = TRUE, locales = "hi-IN")
  ),
  JPY = colDef(
    format = colFormat(currency = "JPY", separators = TRUE, locales = "ja-JP")
  ),
  MAD = colDef(
    format = colFormat(currency = "MAD", separators = TRUE, locales = "ar-MA")
  )
))

# Formatting aggregated cells
data <- data.frame(
  States = state.name,
  Region = state.region,
  Area = state.area
)

reactable(
  data,
  groupBy = "Region",
  columns = list(
    States = colDef(
      aggregate = "count",
      format = list(
        aggregated = colFormat(suffix = " states")
      )
    ),
    Area = colDef(
      aggregate = "sum",
      format = colFormat(suffix = " mi\u00b2", separators = TRUE)
    )
  )
)

}
\seealso{
Custom cell rendering in \code{\link[=colDef]{colDef()}} to customize data formatting
beyond what the built-in formatters provide.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reactable.R
\name{as.tags.reactable}
\alias{as.tags.reactable}
\title{Convert a reactable widget to HTML tags}
\usage{
\method{as.tags}{reactable}(x, standalone = FALSE, ...)
}
\arguments{
\item{x}{a \code{\link[=reactable]{reactable()}} instance.}

\item{standalone}{Logical value indicating whether the widget is being
rendered in a standalone context.}

\item{...}{Additional arguments passed to the S3 method.}
}
\description{
This S3 method exists to enable \code{\link[=reactable]{reactable()}}'s \code{static} rendering option.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/columns.R
\name{colGroup}
\alias{colGroup}
\title{Column group definitions}
\usage{
colGroup(
  name = NULL,
  columns = NULL,
  header = NULL,
  html = FALSE,
  align = NULL,
  headerVAlign = NULL,
  sticky = NULL,
  headerClass = NULL,
  headerStyle = NULL
)
}
\arguments{
\item{name}{Column group header name.}

\item{columns}{Character vector of column names in the group.}

\item{header}{Custom header renderer. An R function that takes the header value
as an argument, or a \code{\link[=JS]{JS()}} function that takes a column object and
table state object as arguments.}

\item{html}{Render header content as HTML? Raw HTML strings are escaped by default.}

\item{align}{Horizontal alignment of content in the column group header. One of
\code{"left"}, \code{"right"}, \code{"center"} (the default).}

\item{headerVAlign}{Vertical alignment of content in the column group header. One of
\code{"top"} (the default), \code{"center"}, \code{"bottom"}.}

\item{sticky}{Make the column group sticky when scrolling horizontally? Either
\code{"left"} or \code{"right"} to make the column group stick to the left or right side.

If a column group is sticky, all columns in the group will automatically
be made sticky.}

\item{headerClass}{Additional CSS classes to apply to the header.}

\item{headerStyle}{Inline styles to apply to the header. A named list or
character string.

Note that if \code{headerStyle} is a named list, property names should be camelCased.}
}
\value{
A column group definition object that can be used to create column
groups in \code{reactable()}.
}
\description{
Use \code{colGroup()} to create column groups in a table.
}
\examples{
reactable(
  iris,
  columns = list(
    Sepal.Length = colDef(name = "Length"),
    Sepal.Width = colDef(name = "Width"),
    Petal.Length = colDef(name = "Length"),
    Petal.Width = colDef(name = "Width")
  ),
  columnGroups = list(
    colGroup(name = "Sepal", columns = c("Sepal.Length", "Sepal.Width")),
    colGroup(name = "Petal", columns = c("Petal.Length", "Petal.Width"))
  )
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shiny.R
\name{getReactableState}
\alias{getReactableState}
\title{Get the state of a reactable instance}
\usage{
getReactableState(outputId, name = NULL, session = NULL)
}
\arguments{
\item{outputId}{The Shiny output ID of the \code{reactable} instance.}

\item{name}{Character vector of state value(s) to get. Values must be one of \code{"page"},
\code{"pageSize"}, \code{"pages"}, \code{sorted}, or \code{"selected"}. If unspecified, all values will
be returned.}

\item{session}{The Shiny session object. Defaults to the current Shiny session.}
}
\value{
If \code{name} is specified, one of the following values:
\itemize{
\item \code{page}: the current page
\item \code{pageSize}: the page size
\item \code{pages}: the number of pages
\item \code{sorted}: the sorted columns - a named list of columns with values of \code{"asc"} for
ascending order or \code{"desc"} for descending order, or \code{NULL} if no columns are sorted
\item \code{selected}: the selected rows - a numeric vector of row indices, or \code{NULL} if no rows are selected
}

If \code{name} contains more than one value, \code{getReactableState()} returns a named list of
the specified values.

If \code{name} is unspecified, \code{getReactableState()} returns a named list containing all values.

If the table has not been rendered yet, \code{getReactableState()} returns \code{NULL}.
}
\description{
\code{getReactableState()} gets the state of a reactable instance within a Shiny application.
}
\examples{
# Run in an interactive R session
if (interactive()) {

library(shiny)
library(reactable)
library(htmltools)

ui <- fluidPage(
  actionButton("prev_page_btn", "Previous page"),
  actionButton("next_page_btn", "Next page"),
  reactableOutput("table"),
  verbatimTextOutput("table_state"),
  uiOutput("selected_row_details")
)

server <- function(input, output) {
  output$table <- renderReactable({
    reactable(
      MASS::Cars93[, 1:5],
      showPageSizeOptions = TRUE,
      selection = "multiple",
      onClick = "select"
    )
  })

  output$table_state <- renderPrint({
    state <- req(getReactableState("table"))
    print(state)
  })

  observeEvent(input$prev_page_btn, {
    # Change to the previous page
    page <- getReactableState("table", "page")
    if (page > 1) {
      updateReactable("table", page = page - 1)
    }
  })

  observeEvent(input$next_page_btn, {
    # Change to the next page
    state <- getReactableState("table")
    if (state$page < state$pages) {
      updateReactable("table", page = state$page + 1)
    }
  })

  output$selected_row_details <- renderUI({
    selected <- getReactableState("table", "selected")
    req(selected)
    details <- MASS::Cars93[selected, -c(1:5)]
    tagList(
      h2("Selected row details"),
      tags$pre(
        paste(capture.output(print(details, width = 1200)), collapse = "\n")
      )
    )
  })
}

shinyApp(ui, server)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reactable.R
\docType{package}
\name{reactable-package}
\alias{reactable-package}
\alias{_PACKAGE}
\title{reactable: Interactive Data Tables for R}
\description{
Interactive data tables for R, based on the 'React Table' JavaScript library. Provides an HTML widget that can be used in 'R Markdown' or 'Quarto' documents, 'Shiny' applications, or viewed from an R console.
}
\details{
See the \href{https://glin.github.io/reactable/}{online documentation} for
examples and an extensive usage guide.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://glin.github.io/reactable/}
  \item \url{https://github.com/glin/reactable}
  \item Report bugs at \url{https://github.com/glin/reactable/issues}
}

}
\author{
\strong{Maintainer}: Greg Lin \email{glin@glin.io}

Other contributors:
\itemize{
  \item Tanner Linsley (React Table library) [contributor, copyright holder]
  \item  Emotion team and other contributors (Emotion library) [contributor, copyright holder]
  \item Kent Russell (reactR package) [contributor, copyright holder]
  \item Ramnath Vaidyanathan (htmlwidgets package) [contributor, copyright holder]
  \item Joe Cheng (htmlwidgets package) [contributor, copyright holder]
  \item JJ Allaire (htmlwidgets package) [contributor, copyright holder]
  \item Yihui Xie (htmlwidgets package) [contributor, copyright holder]
  \item Kenton Russell (htmlwidgets package) [contributor, copyright holder]
  \item  Facebook, Inc. and its affiliates (React library) [contributor, copyright holder]
  \item  FormatJS (FormatJS libraries) [contributor, copyright holder]
  \item  Feross Aboukhadijeh, and other contributors (buffer library) [contributor, copyright holder]
  \item Roman Shtylman (process library) [contributor, copyright holder]
  \item James Halliday (stream-browserify library) [contributor, copyright holder]
  \item  Posit Software, PBC [funder, copyright holder]
}

}
\keyword{internal}
